# CSIRTMU Case Database

Database for case management. Used for managing cases, workflows and tasks.

All data is stored in a PostgreSQL database.

Links to other services (e.g. case-observable, task-victim, ...) are managed by Case Federation.

## Content

- [CSIRTMU Case Database](#csirtmu-case-database)
  - [Data Model](#data-model)
  - [Configuration](#configuration)
  - [Dependencies](#dependencies)
  - [Endpoints](#endpoints)
  - [Local development](#local-development)
  - [Docker](#docker)
  - [Helm](#helm)

---

## Data Model

![](images/data_model_noattr.png)

A more detailed data model can be found in the _images_ directory.

---

## Configuration

All configuration files are in _config_ directory or _tests/config_ if running locally.

[Configuration Management Library](https://gitlab.orion.lan/orion/backend-python/libraries/csirtmu-dataclass-config-loader)

### Environment Variables

- **CASE_DATABASE_CONFIG_PATH**: "Path to service configuration yaml file"
- **JWT_AUTH_BACKEND_CONFIG_PATH**: "Path to authorization configuration yaml file"
- 
### Authorization

- Authorization is configured using the [authorization.polar](https://docs.osohq.com/) file in the _config_ directory.

- All users (even unauthenticated) have the **is:default** role, which grants the user all permissions.

---

## Dependencies

- Authorization
  => [Service Gateway](https://gitlab.orion.lan/orion/backend-python/applications/csirtmu-service-gateway/)
- PostgreSQL instance
- Redis instance
- S3 storage (e.g. MinIO)

---

## Endpoints

The application listens by default on port 8000.

- **/**: Graph(i)QL endpoint
- **/graphql**: Graph(i)QL endpoint
- **/healthcheck/liveness**: Liveness healthcheck
- **/healthcheck/readiness**: Readiness healthcheck
- **/metrics**: Prometheus metrics endpoint
- **/template/{template_type:str}/{template_id:str}/{template_name:str}**: Template endpoint
  - **template_type**: template type ("workflow", "case")
  - **template_id**: template id
  - **template_name**: template name
- **/file/{file_id:int}/{file_name:str}**: File endpoint
  - **file_id**: file id
  - **file_name**: file name
- **/\_auth_backend/version**: Authorization backend version endpoint (for use with Service Gateway)

---

## Local Development

Configuration for local development.

- Required dependencies:
  - Python 3.9+
  - pip
  - [Poetry](https://python-poetry.org/docs/master/#installing-with-the-official-installer)
```console
$ poetry config repositories.ORION https://gitlab.orion.lan/api/v4/projects/2485/packages/pypi/simple
$ poetry config http-basic.ORION token <GITLAB_TOKEN>
$ poetry install --with dev
$ export CASE_DATABASE_CONFIG_PATH=./tests/config/config-test.yaml
$ export JWT_AUTH_BACKEND_CONFIG_PATH=./tests/config/config_jwt_auth.yaml
```

### Server

* Start the server locally.

  ```console
  $ poetry run uvicorn --port 8000 --host 127.0.0.1 --factory csirtmu.case_database.app:app
  ```

### Tests

* Run tests.
  ```console
  $ poetry run tox
  ```

---

## Docker
* To be able to build and run the image, follow the instructions:
  * Build and run Docker image:
    ```console
    $ docker build -t csirtmu-case-database --build-arg POETRY_HTTP_BASIC_ORION_USERNAME=token --build-arg POETRY_HTTP_BASIC_ORION_PASSWORD=<GITLAB_TOKEN> .`
    $ docker run -d -p 127.0.0.1:8000:8000 -n csirtmu-case-database-container csirtmu-case-database
    ```
* Build args **POETRY_HTTP_BASIC_ORION_USERNAME** and **POETRY_HTTP_BASIC_ORION_PASSWORD** are necessary
to download Python packages from private registries.


---

## Helm

Install to Kubernetes cluster using helm.
  * Add *orion* helm repository:
    ```console
    $ helm repo add --username orion --password ${PACKAGE_REGISTRY_PULL} gitlab-orion https://gitlab.orion.lan/api/v4/projects/2485/packages/helm/stable
    ```
  * Download chart dependencies:
    ```console
    $ cd csirtmu-case-database-chart
    $ helm dependency update
    ```
  * Create *csirtmu-registry-credentials* secret with credentials to a container registry:
    ```console
    $ kubectl create secret docker-registry csirtmu-registry-credentials --docker-server=registry.gitlab.orion.lan:443 --docker-username=orion --docker-password=<password>
    ```
  * To deploy Helm chart into kubernetes cluster run:
    ```console
    $ helm upgrade --install case-database .
    ```
---

---

## Workflow/stage/task - states & resolving

### Workflow
* `IN_PROGRESS` - initial state when new case is opened, which persists until all stages are in their final state.
* `DONE` - final state that can be achieved when all stages are in their final state and at least one stage is not `REJECTED`.
* `REJECTED` - final state that can be achieved when all stages are in `REJECTED` state. This can be done by rejecting
whole workflow, which will automatically reject all stages.
### Stage
* `IN_PROGRESS` - initial state for a stage that can typically (see warning) be worked on. This state persists 
until all tasks of the stage are not in their final state.
  - (! warning: Sometimes `PENDING` stage can automatically change its state to `IN_PROGRESS` due to skipping some tasks.
  However, the stage cannot actually be worked on because it does not contain any `OPEN` tasks, only `PENDING`/`SKIPPED` tasks.)
* `PENDING` - initial state for a stage that contains only `PENDING` tasks. It cannot be worked on, but can be 
rejected.
* `DONE` - final state that can be achieved when all tasks are in their final state and at least one task is not `REJECTED`.
* `REJECTED` - final state that can be achieved when all tasks are in `REJECTED` state. This can be done by rejecting
whole stage, which will automatically reject all tasks.
### Task
* `OPEN` - default state for a task that can be worked on. It can be either resolved or rejected. The option will affect
the `PENDING` successor.
  - (! warning: watch out for `OPEN` task flagged as `AGGREGATOR` - see [flags](#flags))
* `PENDING` - default state for a task that depends on some predecessor that is not yet in its final state.
* `RUNNING` - this state can be seen when the task has `AUTOMATED` flag and is indicating that pipeline is not yet completed.
* `RESOLVED` - final state that can be manually set when all mandatory steps of task are done. Resolving the task will change
`PENDING` successor to `OPEN` if all other successor's predecessors are `RESOLVED`/`SKIPPED`, but at least one is `RESOLVED`.
* `REJECTED` - final state that can be manually set when it is required. Rejecting the task will change `PENDING` successor
to `REJECTED` regardless of the state of the other successor's predecessors.
* `SKIPPED` - final state that was set automatically by pipeline or by resolving decision predecessor as false.
Skipping the task will change `PENDING` successor to `SKIPPED` if all other successor's predecessors are `SKIPPED`.

    #### Flags
    * `AUTOMATED` - indicates a task that is doing some action without any interference. However, it requires 
    manual trigger of the pipeline.
    * `AUTOTRIGGER` - extends `AUTOMATED` flag and will automatically trigger the pipeline when task is opened.
    * `AGGREGATOR` - indicates a task that is blocked by several subtasks that need to be solved first, because the
    output of the task depends on the output of its subtasks. 
    * `DECISION` - indicates a task whose output is either True or False (Decision, Confirmation, etc.)

Example flowchart that shows how the states change throughout resolving tasks.
![](images/workflow_states.png)

Example flowchart that shows branching on decision task.
![](images/flowchart_decision_task.png)

Example flowchart that shows states on task blocked by subtasks.
![](images/flowchart_blocked_task.png)