from dataclasses import dataclass
from os import getenv
from pathlib import Path
from typing import Optional

from csirtmu.dataclass_config_loader import Omit
from csirtmu.dataclass_config_loader.load import load_dataclass_config
from sqlalchemy import create_engine
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine
from sqlalchemy.future import Engine
from sqlalchemy.pool import QueuePool

BASE_DIR = Path(__file__).parent.parent.parent


@dataclass
class LocalFileStorage:
    file_directory: str = '/tmp'


@dataclass
class Redis:
    host: str
    port: int
    username: Optional[str] = None
    password: Optional[str] = None


@dataclass
class MinioFileStorage:
    host: str
    bucket: str
    access_key: str
    secret_key: str
    port: int


@dataclass
class UserProfileIntrospection:
    host: str
    port: int
    secure: bool = False
    enabled: bool = False


@dataclass
class Database:
    host: str
    port: int
    username: str
    password: str
    database: str
    engine: Omit[Engine] = None
    async_engine: Omit[AsyncEngine] = None


@dataclass
class Config:
    database: Database
    file_storage_type: str
    external_url: str
    local_file_storage: LocalFileStorage
    minio_file_storage: MinioFileStorage
    redis: Redis
    user_profile_introspection: UserProfileIntrospection
    debug: Optional[bool] = False


def _load_config(path: Path) -> Config:
    config = load_dataclass_config(path, Config, "CASE_DATABASE")

    config.database.engine = create_engine(
        f'postgresql+psycopg2://{config.database.username}:{config.database.password}@{config.database.host}:{config.database.port}/{config.database.database}',
        future=True, echo=config.debug, poolclass=QueuePool)

    config.database.async_engine = create_async_engine(
        f'postgresql+asyncpg://{config.database.username}:{config.database.password}@{config.database.host}:{config.database.port}/{config.database.database}',
        future=True, echo=config.debug, poolclass=QueuePool)

    if config.file_storage_type not in ['local', 'minio']:
        raise ValueError(f'Invalid file storage type {config.file_storage_type}! Valid are: minio, local')

    return config


CONFIG = _load_config(Path(getenv('CASE_DATABASE_CONFIG_PATH', (BASE_DIR / 'config/config.yaml').resolve().__str__())))
