import json
from typing import List, Optional

import strawberry
from csirtmu.strawberry_dmt.permissions import CAN_CREATE_BULK, CAN_CREATE_ONE, CAN_QUERY_ONE, CAN_UPDATE_ONE
from csirtmu.strawberry_dmt.strawberry_types import CreateResult
from csirtmu.strawberry_dmt.locale.locale import Languages
from starlette.datastructures import UploadFile
from strawberry.file_uploads import Upload
from strawberry.types import Info

from csirtmu.case_database import models
from csirtmu.case_database.enums import TaskStateEnum
from csirtmu.case_database.gql.enums import KnownInputOutputTypes
from csirtmu.case_database.gql.inputs import CaseCreate, FileCreate, FileCreateComment, FileRefCreate, TaskCreate, \
    TaskCreateBulk, TaskCreateParentless, TaskCreateParentlessBulk, TemplateCreate
from csirtmu.case_database.gql.resolvers_description import DESCRIPTIONS
from csirtmu.case_database.gql.schema_auto_generated import CaseDBUser, mutations, queries, registry
from csirtmu.case_database.gql.strawberry_types import MutationResult, MutationResultField
from csirtmu.case_database.models import EmailMessageInputOutput, FileInputOutput, JSONInputOutput
from csirtmu.case_database.permission_classes import is_authorized
from csirtmu.case_database.services.service_layer import ServiceLayer
from csirtmu.case_database.utils import get_file_location


@strawberry.type
class Query(queries):
    @strawberry.field(permission_classes=[is_authorized(models.CaseDBUser, CAN_QUERY_ONE)],
                      description=DESCRIPTIONS["get_current_user"])
    async def get_current_user(self, info: Info) -> CaseDBUser.get_strawberry_info().entity:
        return info.context['request'].user


@strawberry.type
class Mutation(mutations):
    @strawberry.field(permission_classes=[is_authorized(models.Workflow, CAN_CREATE_ONE)],
                      description=DESCRIPTIONS["create_workflow_from_template"])
    async def create_workflow_from_template(self, info: Info, workflow_template_id: strawberry.ID,
                                            case_id: strawberry.ID,
                                            workflow_name: Optional[str] = None) -> MutationResult:
        x = info.context['async_session_factory']()
        async with info.context['async_session_factory']() as session:
            if not (case := await ServiceLayer.get_case(session, case_id)):
                raise ValueError("Case with this ID does not exist!")

            workflow = await ServiceLayer.get_workflow_from_template(session, workflow_template_id,
                                                                     case.case_identifier)
            if workflow_name:
                workflow.name = workflow_name
            workflow.case = case
            await session.commit()
            await session.refresh(workflow)
            return MutationResult(id=workflow.id)

    @strawberry.field(permission_classes=[is_authorized(models.Task, CAN_CREATE_ONE)],
                      description=DESCRIPTIONS["create_task_from_template"])
    async def create_task_from_template(self, info: Info, task_template_id: strawberry.ID,
                                        stage_id: strawberry.ID, task_name: Optional[str] = None) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            if not (stage := await ServiceLayer.get_stage(session, stage_id)):
                raise ValueError("Stage with this ID does not exist!")

            task = await ServiceLayer.get_task_from_template(session, task_template_id, stage.case_identifier)
            if task_name:
                task.name = task_name

            task.stage = stage
            await session.commit()
            await session.refresh(task)
            return MutationResult(id=task.id)

    @strawberry.field(permission_classes=[is_authorized(models.Case, CAN_CREATE_ONE)],
                      description=DESCRIPTIONS["create_case"])
    async def create_case(self, info: Info, data: CaseCreate) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            try:
                if data.workflow_templates:
                    workflow_templates = data.workflow_templates
                else:
                    workflow_templates = []
                case = await ServiceLayer.create_case(session, data)
                await session.flush([case])
                for workflow_template in workflow_templates:
                    workflow = await ServiceLayer.get_workflow_from_template(session, workflow_template.id,
                                                                             case.case_identifier)
                    workflow.case = case
                    if workflow_template.name:
                        workflow.name = workflow_template.name
                await session.commit()
                await session.refresh(case)
                return MutationResult(id=case.id)
            except Exception as ex:
                await session.rollback()
                raise ex

    @strawberry.field(permission_classes=[is_authorized(models.Case, CAN_CREATE_ONE)],
                      description=DESCRIPTIONS["create_case_bulk"])
    async def create_case_bulk(self, info: Info, data_list: List[CaseCreate]) -> List[CreateResult]:
        cases = []
        async with info.context['async_session_factory']() as session:
            for data in data_list:
                try:
                    if data.workflow_templates:
                        workflow_templates = data.workflow_templates
                    else:
                        workflow_templates = []
                    case = await ServiceLayer.create_case(session, data)
                    await session.flush([case])
                    for workflow_template in workflow_templates:
                        workflow = await ServiceLayer.get_workflow_from_template(session, workflow_template.id,
                                                                                 case.case_identifier)
                        workflow.case = case
                        if workflow_template.name:
                            workflow.name = workflow_template.name
                    await session.commit()
                    await session.refresh(case)
                    cases.append(CreateResult(id=case.id))
                except Exception as ex:
                    await session.rollback()
                    raise ex
            return cases

    @strawberry.field(permission_classes=[is_authorized(models.Stage, CAN_UPDATE_ONE)],
                      description=DESCRIPTIONS["reject_stage"])
    async def reject_stage(self, info: Info, stage_id: strawberry.ID) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            if not await ServiceLayer.reject_stage(session, stage_id):
                raise ValueError('Could not resolve stage!')
            await session.commit()
            return MutationResult(id=stage_id)

    @strawberry.field(permission_classes=[is_authorized(models.Workflow, CAN_UPDATE_ONE)],
                      description=DESCRIPTIONS["reject_workflow"])
    async def reject_workflow(self, info: Info, workflow_id: strawberry.ID) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            if not await ServiceLayer.reject_workflow(session, workflow_id):
                raise ValueError('Could not reject workflow!')
            await session.commit()
            return MutationResult(id=workflow_id)

    @strawberry.field(permission_classes=[is_authorized(models.Case, CAN_UPDATE_ONE)],
                      description=DESCRIPTIONS["reject_case"])
    async def reject_case(self, info: Info, case_id: strawberry.ID) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            if not await ServiceLayer.reject_case(session, case_id):
                raise ValueError('Could not reject case!')
            await session.commit()
            return MutationResult(id=case_id)

    @strawberry.field(permission_classes=[is_authorized(models.Case, CAN_UPDATE_ONE)],
                      description=DESCRIPTIONS["resolve_case"])
    async def resolve_case(self, info: Info, case_id: strawberry.ID) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            if not await ServiceLayer.resolve_case(session, case_id):
                raise ValueError('Could not resolve case!')
            await session.commit()
            return MutationResult(id=case_id)

    @strawberry.field(permission_classes=[is_authorized(models.Task, CAN_UPDATE_ONE)],
                      description=DESCRIPTIONS["resolve_task"])
    async def resolve_task(self, info: Info, task_id: strawberry.ID) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            if not await ServiceLayer.resolve_task(session, task_id):
                raise ValueError("Could not resolve task!")
            await session.commit()
            return MutationResult(id=task_id)

    @strawberry.field(permission_classes=[is_authorized(models.Task, CAN_UPDATE_ONE)],
                      description=DESCRIPTIONS["resolve_task_with_decision"])
    async def resolve_task_with_decision(self, info: Info, task_id: strawberry.ID, decision: bool) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            if not await ServiceLayer.resolve_task(session, task_id, decision=decision):
                raise ValueError("Could not resolve task!")
            await session.commit()
            return MutationResult(id=task_id)

    @strawberry.field(permission_classes=[is_authorized(models.Task, CAN_UPDATE_ONE)],
                      description=DESCRIPTIONS["reject_task"])
    async def reject_task(self, info: Info, task_id: strawberry.ID) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            if not await ServiceLayer.reject_task(session, task_id):
                raise ValueError("Could not reject task!")
            await session.commit()
            return MutationResult(id=task_id)

    @strawberry.field(permission_classes=[is_authorized(models.Task, CAN_UPDATE_ONE)],
                      description=DESCRIPTIONS["skip_task"])
    async def skip_task(self, info: Info, task_id: strawberry.ID) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            if not await ServiceLayer.skip_task(session, task_id):
                raise ValueError(f"Could not skip task {task_id}")
            await session.commit()
            return MutationResult(id=task_id)

    # File manipulation
    @strawberry.field(permission_classes=[is_authorized(models.CaseFileItem, CAN_CREATE_ONE)],
                      description=DESCRIPTIONS["create_file"])
    async def create_file(self, info: Info, data: FileCreate) -> Optional[MutationResult]:
        async with info.context['async_session_factory']() as session:
            case_file = await ServiceLayer.create_file(session, data, requires_parent=True)
            await session.commit()
            await session.refresh(case_file)
            return MutationResult(id=case_file.id)

    @strawberry.field(permission_classes=[is_authorized(models.CaseFileItem, CAN_CREATE_ONE)],
                      description=DESCRIPTIONS["create_file_from_comment_in_case"])
    async def create_file_from_comment_in_case(self, info: Info, case_id: strawberry.ID,
                                               data: FileCreateComment) -> Optional[MutationResultField]:
        async with info.context['async_session_factory']() as session:
            file_create_data = FileCreate(file=data.file, display_name=data.display_name, extension=data.extension,
                                          name=data.name, created_by=data.created_by)
            case_file = await ServiceLayer.create_file_from_comment_on_case(session, file_create_data, case_id=case_id)
            await session.commit()
            await session.refresh(case_file)
            return MutationResultField(field=case_file.src)

    @strawberry.field(permission_classes=[is_authorized(models.CaseFileItem, CAN_CREATE_ONE)],
                      description=DESCRIPTIONS["create_file_from_comment_in_task"])
    async def create_file_from_comment_in_task(self, info: Info, task_id: strawberry.ID,
                                               data: FileCreateComment) -> Optional[MutationResultField]:
        async with info.context['async_session_factory']() as session:
            file_create_data = FileCreate(file=data.file, display_name=data.display_name, extension=data.extension,
                                          name=data.name, created_by=data.created_by)
            case_file = await ServiceLayer.create_file_from_comment_on_task(session, file_create_data, task_id=task_id)
            await session.commit()
            await session.refresh(case_file)
            return MutationResultField(field=case_file.src)

    @strawberry.field(permission_classes=[is_authorized(models.CaseFileItem, CAN_CREATE_ONE)],
                      description=DESCRIPTIONS["create_file_ref"])
    async def create_file_ref(self, info: Info, data: FileRefCreate) -> Optional[MutationResult]:
        async with info.context['async_session_factory']() as session:
            case_file_ref = await ServiceLayer.create_file_ref(session, data)
            await session.commit()
            await session.refresh(case_file_ref)
            return MutationResult(id=case_file_ref.id)

    @strawberry.field(permission_classes=[is_authorized(models.CaseFileItem, CAN_CREATE_BULK)],
                      description=DESCRIPTIONS["create_file_bulk"])
    async def create_file_bulk(self, info: Info, data_list: List[FileCreate]) -> Optional[List[MutationResult]]:
        async with info.context['async_session_factory']() as session:
            files = []
            for file_info in data_list:
                case_file = await ServiceLayer.create_file(session, file_info, requires_parent=True)
                files.append(case_file)
            await session.commit()

            results = []
            for result in files:
                await session.refresh(result)
                results.append(MutationResult(id=result.id))
            return results

    @strawberry.field(permission_classes=[is_authorized(models.Task, CAN_CREATE_ONE)],
                      description=DESCRIPTIONS["create_task_parentless"])
    async def create_task_parentless(self, info: Info, task: TaskCreateParentless,
                                     input_file: Optional[Upload] = None) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            task.input_file = input_file
            task_obj = await ServiceLayer.create_task(session, task)
            await session.commit()
            await session.refresh(task_obj)
            return MutationResult(id=task_obj.id)

    @strawberry.field(permission_classes=[is_authorized(models.Task, CAN_CREATE_BULK)],
                      description=DESCRIPTIONS["create_task_parentless_bulk"])
    async def create_task_parentless_bulk(self, info: Info, data_list: List[TaskCreateParentlessBulk]) -> \
            Optional[List[MutationResult]]:
        async with info.context['async_session_factory']() as session:
            tasks = []
            for data in data_list:
                data.task.input_file = data.input_file
                task_obj = await ServiceLayer.create_task(session, data.task)
                tasks.append(task_obj)
            await session.commit()

            results = []
            for result in tasks:
                await session.refresh(result)
                results.append(MutationResult(id=result.id))
            return results

    @strawberry.field(permission_classes=[is_authorized(models.Task, CAN_CREATE_ONE)],
                      description=DESCRIPTIONS["create_task"])
    async def create_task(self, info: Info, task: TaskCreate,
                          input_file: Optional[Upload] = None) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            task.input_file = input_file
            task_obj = await ServiceLayer.create_task(session, task)
            await session.commit()
            await session.refresh(task_obj)
            return MutationResult(id=task_obj.id)

    @strawberry.field(permission_classes=[is_authorized(models.Task, CAN_CREATE_BULK)],
                      description=DESCRIPTIONS["create_task_bulk"])
    async def create_task_bulk(self, info: Info, data_list: List[TaskCreateBulk]) -> Optional[
        List[MutationResult]]:
        async with info.context['async_session_factory']() as session:
            tasks = []
            for data in data_list:
                data.task.input_file = data.input_file
                task_obj = await ServiceLayer.create_task(session, data.task)
                tasks.append(task_obj)
            await session.commit()

            results = []
            for result in tasks:
                await session.refresh(result)
                results.append(MutationResult(id=result.id))
            return results

    @strawberry.field(permission_classes=[is_authorized(models.Task, CAN_UPDATE_ONE)],
                      description=DESCRIPTIONS["create_task_input"])
    async def create_task_input(self, info: Info, id: strawberry.ID, input_type: KnownInputOutputTypes,
                                input_format: Optional[str] = None, description: Optional[str] = None) -> Optional[
        MutationResult]:
        async with info.context['async_session_factory']() as session:
            task = await ServiceLayer.get_task(session, task_id=id)
            if not task:
                raise ValueError(f"Task with id {id} does not exist!")
            if task.input is not None:
                raise ValueError("Task already has input!")
            if task.output is not None and task.output.value is not None:
                raise ValueError("Task already has set output value, can't change input!")
            input_obj = await ServiceLayer.prepare_input_output_object(session, input_type, None, input_format,
                                                                       description)
            task.input = input_obj
            await session.commit()
            return MutationResult(id=id)

    @strawberry.field(permission_classes=[is_authorized(models.Task, CAN_UPDATE_ONE)],
                      description=DESCRIPTIONS["create_task_output"])
    async def create_task_output(self, info: Info, id: strawberry.ID, output_type: KnownInputOutputTypes,
                                 output_format: Optional[str] = None, description: Optional[str] = None) -> Optional[
        MutationResult]:
        async with info.context['async_session_factory']() as session:
            task = await ServiceLayer.get_task(session, task_id=id)
            if not task:
                raise ValueError(f"Task with id {id} does not exist!")
            if task.output is not None:
                raise ValueError("Task already has output!")
            output_obj = await ServiceLayer.prepare_input_output_object(session, output_type, None, output_format,
                                                                        description)
            task.output = output_obj
            await session.commit()
            return MutationResult(id=id)

    @strawberry.field(permission_classes=[is_authorized(models.Task, CAN_UPDATE_ONE)],
                      description=DESCRIPTIONS["set_task_input"])
    async def set_task_input(self, info: Info, id: strawberry.ID, input: str) -> Optional[MutationResult]:
        async with info.context['async_session_factory']() as session:
            task = await ServiceLayer.get_task(session, task_id=id)

            if not task.input:
                raise ValueError("Task does not have specified input type!")
            if not input:
                raise ValueError(f"Missing value!")
            if task.output is not None and task.output.value is not None:
                raise ValueError("Output is already set, can't change input!")
            task.input.value = input

            if isinstance(task.input, JSONInputOutput):
                if task.output is not None and \
                        isinstance(task.output, JSONInputOutput):
                    task.output.json_initial_value = task.input.value

            await session.commit()
            return MutationResult(id=id)

    @strawberry.field(permission_classes=[is_authorized(models.Task, CAN_UPDATE_ONE)],
                      description=DESCRIPTIONS["set_task_file_input"])
    async def set_task_file_input(self, info: Info, id: strawberry.ID, file_input: Upload) -> Optional[MutationResult]:
        with info.context["async_session_factory"]() as session:
            task = await ServiceLayer.get_task(session, id)
            if not task:
                raise ValueError(f"Task with ID {id} does not exist!")
            if not isinstance(task.input, (FileInputOutput, EmailMessageInputOutput)):
                raise ValueError(f"Task does not expect file as input!")
            if task.output is not None and task.output.value is not None:
                raise ValueError("Output is already set, can't change input!")

            create_file_data = FileCreate(file=file_input, name=file_input.filename)
            new_file = await ServiceLayer.create_file(session, create_file_data)
            task.input.value = get_file_location(new_file.id, file_input.filename)
            await session.commit()
            return MutationResult(id=id)

    @strawberry.field(permission_classes=[is_authorized(models.Task, CAN_UPDATE_ONE)],
                      description=DESCRIPTIONS["set_task_output"])
    async def set_task_output(self, info: Info, id: strawberry.ID, output: str) -> Optional[MutationResult]:
        async with info.context['async_session_factory']() as session:
            task = await ServiceLayer.get_task(session, task_id=id)

            if not task.output:
                raise ValueError("Task does not have specified output type!")
            else:
                if not output:
                    raise ValueError(f"Missing value!")
                task.output.value = output
            if task.state == TaskStateEnum.RUNNING:
                task.state = TaskStateEnum.OPEN
            await session.commit()
            return MutationResult(id=id)

    @strawberry.field(permission_classes=[is_authorized(models.Task, CAN_UPDATE_ONE)],
                      description=DESCRIPTIONS["set_task_file_output"])
    async def set_task_file_output(self, info: Info, id: strawberry.ID, file_output: Upload) -> Optional[
        MutationResult]:
        with info.context["async_session_factory"]() as session:
            task = await ServiceLayer.get_task(session, id)

            if not task.output:
                raise ValueError("Task does not have specified output type!")

            if not isinstance(task.output, (FileInputOutput, EmailMessageInputOutput)):
                raise ValueError(f"Task does not expect file output!")

            create_file_data = FileCreate(file=file_output, name=file_output.filename)
            new_file = await ServiceLayer.create_file(session, create_file_data)
            task.output.value = get_file_location(new_file.id, file_output.filename)
            if task.state == TaskStateEnum.RUNNING:
                task.state = TaskStateEnum.OPEN
            await session.commit()
            return MutationResult(id=id)

    @strawberry.field(permission_classes=[is_authorized(models.Task, CAN_UPDATE_ONE)],
                      description=DESCRIPTIONS["update_output_from_blocking_tasks"])
    async def update_output_from_blocking_tasks(self, info: Info, id: strawberry.ID) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            await ServiceLayer.update_blocked_task_output(session, id)
            await session.commit()
            return MutationResult(id=id)

    @strawberry.field(permission_classes=[is_authorized(models.Task, CAN_UPDATE_ONE)],
                      description=DESCRIPTIONS["trigger_task"])
    async def trigger_task(self, task_id: strawberry.ID, info: Info) -> str:
        async with info.context['async_session_factory']() as session:
            try:
                await ServiceLayer.trigger_task(session, task_id)
                await session.commit()
            except ValueError as ex:
                return str(ex)
            return "published"

    @strawberry.field(permission_classes=[is_authorized(models.WorkflowTemplate, CAN_CREATE_ONE)],
                      description=DESCRIPTIONS["upload_workflow_template"])
    async def upload_workflow_template(self, info: Info, file: Upload) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            if isinstance(file, UploadFile):
                file_content_raw = (await file.read()).decode()
            else:
                file_content_raw = bytes(file).decode()
            file_content = json.loads(file_content_raw)

            data = TemplateCreate(name=file_content.get("name"),
                                  description=file_content.get("description"),
                                  template=file_content_raw)
            workflow_template = await ServiceLayer.create_workflow_template(session, data)

            await session.commit()
            await session.refresh(workflow_template)
            return MutationResult(id=workflow_template.id)

    @strawberry.field(permission_classes=[is_authorized(models.WorkflowTemplate, CAN_CREATE_ONE)],
                      description=DESCRIPTIONS["create_workflow_template"])
    async def create_workflow_template(self, info: Info, data: TemplateCreate) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            try:
                workflow_template = await ServiceLayer.create_workflow_template(session, data)
                await session.commit()
                await session.refresh(workflow_template)
                return MutationResult(id=workflow_template.id)
            except Exception as ex:
                await session.rollback()
                raise ex

    @strawberry.field(permission_classes=[is_authorized(models.WorkflowTemplate, CAN_CREATE_BULK)],
                      description=DESCRIPTIONS["create_workflow_template_bulk"])
    async def create_workflow_template_bulk(self, info: Info, data_list: List[TemplateCreate]) -> Optional[
        List[MutationResult]]:
        async with info.context['async_session_factory']() as session:
            templates = []
            for data in data_list:
                template = await ServiceLayer.create_workflow_template(session, data)
                templates.append(template)
            await session.commit()

            results = []
            for result in templates:
                await session.refresh(result)
                results.append(MutationResult(id=result.id))
            return results

    @strawberry.field(permission_classes=[is_authorized(models.WorkflowTemplate, CAN_CREATE_ONE)],
                      description=DESCRIPTIONS["create_template_from_workflow"])
    async def create_template_from_workflow(self, info: Info, workflow_id: strawberry.ID) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            try:
                workflow_template = await ServiceLayer.create_template_from_workflow(session, workflow_id)
                await session.commit()
                await session.refresh(workflow_template)
                return MutationResult(id=workflow_template.id)
            except Exception as ex:
                await session.rollback()
                raise ex

    @strawberry.field(permission_classes=[is_authorized(models.TaskTemplate, CAN_CREATE_ONE)],
                      description=DESCRIPTIONS["upload_task_template"])
    async def upload_task_template(self, info: Info, file: Upload) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            if isinstance(file, UploadFile):
                file_content_raw = (await file.read()).decode()
            else:
                file_content_raw = bytes(file).decode()
            file_content = json.loads(file_content_raw)

            data = TemplateCreate(name=file_content.get("name"),
                                  description=file_content.get("description"),
                                  template=file_content_raw)
            task_template = await ServiceLayer.create_task_template(session, data)

            await session.commit()
            await session.refresh(task_template)
            return MutationResult(id=task_template.id)

    @strawberry.field(permission_classes=[is_authorized(models.TaskTemplate, CAN_CREATE_ONE)],
                      description=DESCRIPTIONS["create_task_template"])
    async def create_task_template(self, info: Info, data: TemplateCreate) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            try:
                task_template = await ServiceLayer.create_task_template(session, data)
                await session.commit()
                await session.refresh(task_template)
                return MutationResult(id=task_template.id)
            except Exception as ex:
                await session.rollback()
                raise ex

    @strawberry.field(permission_classes=[is_authorized(models.TaskTemplate, CAN_CREATE_BULK)],
                      description=DESCRIPTIONS["create_task_template_bulk"])
    async def create_task_template_bulk(self, info: Info, data_list: List[TemplateCreate]) -> Optional[
        List[MutationResult]]:
        async with info.context['async_session_factory']() as session:
            templates = []
            for data in data_list:
                template = await ServiceLayer.create_task_template(session, data)
                templates.append(template)
            await session.commit()

            results = []
            for result in templates:
                await session.refresh(result)
                results.append(MutationResult(id=result.id))
            return results

    @strawberry.field(permission_classes=[is_authorized(models.TaskTemplate, CAN_CREATE_ONE)],
                      description=DESCRIPTIONS["create_template_from_task"])
    async def create_template_from_task(self, info: Info, task_id: strawberry.ID) -> MutationResult:
        async with info.context['async_session_factory']() as session:
            try:
                task_template = await ServiceLayer.create_template_from_task(session, task_id)
                await session.commit()
                await session.refresh(task_template)
                return MutationResult(id=task_template.id)
            except Exception as ex:
                await session.rollback()
                raise ex


extended_types = []
schema: strawberry.federation.Schema = registry.schema(Query, Mutation, types=extended_types, docs_language=Languages.CS)
