from csirtmu.strawberry_dmt import mappings
from csirtmu.strawberry_dmt.field_hints import Relationship
from csirtmu.strawberry_dmt.model_type import model_type, ModelTypeBase, model_type_union, ModelTypeUnion, \
    model_type_interface, ModelTypeInterface
from csirtmu.strawberry_dmt.schema_registry import SchemaRegistry
from sqlalchemy import BigInteger

from csirtmu.case_database import models
from csirtmu.case_database.gql.strawberry_types import LargeJsInt, LargeJsIntFieldFilter

mappings.VALUE_FIELDS_TYPE_MAP[BigInteger] = LargeJsInt
mappings.INPUT_FIELDS_TYPE_MAP[BigInteger] = LargeJsInt
mappings.FILTER_FIELDS_TYPE_MAP[BigInteger] = LargeJsIntFieldFilter


@model_type(models.CaseDBUser)
class CaseDBUser(ModelTypeBase):
    roles = Relationship("CaseDBRole")
    created_cases = Relationship("Case")
    created_case_items = Relationship("CaseFileItem")
    cases = Relationship("Case")
    workflows = Relationship("Workflow")
    tasks = Relationship("Task")
    comments = Relationship("Comment")
    stages = Relationship("Stage")
    created_workflows = Relationship("Workflow")
    created_tasks = Relationship("Task")
    created_stages = Relationship("Stage")

    class Meta:
        exclude_filter_fields = ['global_identifier']
        exclude_query_fields = ['global_identifier']
        exclude_create_fields = ['global_identifier']
        exclude_update_fields = ['global_identifier']


@model_type(models.CaseDBRole)
class CaseDBRole(ModelTypeBase):
    granted_to = Relationship("CaseDBUser")


# External ticket

@model_type(models.CaseManagementExternalReference)
class CaseManagementExternalReference(ModelTypeBase):
    related_cases = Relationship("Case")
    related_tasks = Relationship("Task")

    class Meta:
        exclude_create_fields = ['last_updated', 'created', 'unique_identifier']
        exclude_update_fields = ['last_updated', 'created', 'unique_identifier']


@model_type(models.Activity)
class BaseActivity(ModelTypeBase):
    pass


@model_type_interface(models.Activity)
class Activity(ModelTypeInterface):
    pass


@model_type(models.Comment, interfaces=[Activity])
class Comment(ModelTypeBase):
    author = Relationship("CaseDBUser")
    case = Relationship("Case")
    task = Relationship("Task")
    reply_to = Relationship("Comment")
    replies = Relationship("Comment")

    class Meta:
        exclude_create_fields = ['last_updated', 'created']
        exclude_update_fields = ['last_updated', 'created']


@model_type(models.CaseRelationship)
class CaseRelationship(ModelTypeBase):
    source = Relationship("Case")
    target = Relationship("Case")


@model_type(models.Case)
class Case(ModelTypeBase):
    created_by = Relationship("CaseDBUser")
    assignees = Relationship("CaseDBUser")
    events = Relationship("Event")
    external_references = Relationship("CaseManagementExternalReference")
    case_file_root = Relationship("Directory")
    source_relationships = Relationship("CaseRelationship")
    target_relationships = Relationship("CaseRelationship")
    labels = Relationship("Label")
    workflows = Relationship("Workflow")
    comments = Relationship("Comment")

    class Meta:
        exclude_create_fields = ['last_updated', 'created', 'case_file_root', 'case_identifier',
                                 'source_relationships', 'target_relationships', 'weight', 'external_references',
                                 'events', 'comments', 'workflows']
        exclude_update_fields = ['last_updated', 'created', 'case_file_root', 'case_identifier']


@model_type(models.CaseFileItem)
class CaseFileItemBase(ModelTypeBase):
    parent = Relationship("Directory")
    created_by = Relationship("CaseDBUser")

    class Meta:
        exclude_create_fields = ['created']
        exclude_update_fields = ['created']


@model_type(models.File)
class File(ModelTypeBase):
    parent = Relationship("Directory")
    created_by = Relationship("CaseDBUser")

    class Meta:
        exclude_create_fields = ['created']
        exclude_update_fields = ['created']


@model_type(models.Directory)
class Directory(ModelTypeBase):
    case = Relationship("Case")
    parent = Relationship("Directory")
    created_by = Relationship("CaseDBUser")
    children = Relationship("CaseFileItem")

    class Meta:
        exclude_create_fields = ['created', 'case']
        exclude_update_fields = ['created', 'case']


@model_type_union(CaseFileItemBase, [File, Directory])
class CaseFileItem(ModelTypeUnion):
    pass


# Event
@model_type_interface(models.Event)
class Event(ModelTypeInterface):
    related_cases = Relationship("Case")
    related_tasks = Relationship("Task")

    class Meta:
        exclude_create_fields = ['unique_identifier']
        exclude_update_fields = ['unique_identifier']


@model_type(models.Event, interfaces=[Event])
class BaseEvent(ModelTypeBase):
    class Meta:
        exclude_create_fields = ['unique_identifier']
        exclude_update_fields = ['unique_identifier']


@model_type(models.StorageEvent, interfaces=[Event])
class StorageEvent(ModelTypeBase):
    class Meta:
        exclude_create_fields = ['unique_identifier']
        exclude_update_fields = ['unique_identifier']


@model_type(models.FlowEvent, interfaces=[Event])
class FlowEvent(ModelTypeBase):
    flow_data = Relationship("FlowEventData")

    class Meta:
        exclude_create_fields = ['unique_identifier']
        exclude_update_fields = ['unique_identifier']


@model_type(models.FlowEventData)
class FlowEventData(ModelTypeBase):
    flow_event = Relationship(FlowEvent)


# Label

@model_type(models.Label)
class Label(ModelTypeBase):
    cases = Relationship("Case")
    tasks = Relationship("Task")


# Workflow
@model_type(models.WorkflowTemplate)
class WorkflowTemplate(ModelTypeBase):
    pass


@model_type(models.Workflow)
class Workflow(ModelTypeBase):
    stages = Relationship("Stage")
    assignees = Relationship("CaseDBUser")
    case = Relationship("Case")
    created_by = Relationship("CaseDBUser")

    class Meta:
        exclude_create_fields = ['state', 'progress', 'case_identifier']
        exclude_update_fields = ['state', 'progress', 'case_identifier']


@model_type(models.Stage)
class Stage(ModelTypeBase):
    workflow = Relationship("Workflow")
    tasks = Relationship("Task")
    assignees = Relationship("CaseDBUser")
    created_by = Relationship("CaseDBUser")

    class Meta:
        exclude_create_fields = ['state', 'progress', 'case_identifier']
        exclude_update_fields = ['state', 'progress', 'case_identifier']


@model_type(models.TaskRelationship)
class TaskRelationship(ModelTypeBase):
    source = Relationship("Task")
    target = Relationship("Task")


@model_type(models.TaskReference)
class TaskReference(ModelTypeBase):
    source = Relationship("Task")
    target = Relationship("Task")


@model_type(models.TaskTemplate)
class TaskTemplate(ModelTypeBase):
    pass


@model_type(models.Task)
class Task(ModelTypeBase):
    stage = Relationship("Stage")
    assignees = Relationship("CaseDBUser")
    comments = Relationship("Comment")
    source_relationships = Relationship("TaskRelationship")
    target_relationships = Relationship("TaskRelationship")
    refers_to = Relationship("TaskReference")
    referenced_by = Relationship("TaskReference")

    events = Relationship("Event")
    external_references = Relationship("CaseManagementExternalReference")
    created_by = Relationship("CaseDBUser")

    input = Relationship("TaskInputOutput")
    output = Relationship("TaskInputOutput")

    labels = Relationship("Label")

    class Meta:
        exclude_create_fields = ['case_identifier']
        exclude_update_fields = ['case_identifier']

@model_type_interface(models.TaskInputOutput)
class TaskInputOutput(ModelTypeInterface):
    input_of = Relationship("Task")
    output_of = Relationship("Task")


@model_type(models.TaskInputOutput, interfaces=[TaskInputOutput])
class BaseTaskInputOutput(ModelTypeBase):
    pass


@model_type(models.StringInputOutput, interfaces=[TaskInputOutput])
class StringInputOutput(ModelTypeBase):
    pass


@model_type(models.URLInputOutput, interfaces=[TaskInputOutput])
class URLInputOutput(ModelTypeBase):
    pass


@model_type(models.IPInputOutput, interfaces=[TaskInputOutput])
class IPInputOutput(ModelTypeBase):
    pass


@model_type(models.DomainInputOutput, interfaces=[TaskInputOutput])
class DomainInputOutput(ModelTypeBase):
    pass


@model_type(models.EmailAddressInputOutput, interfaces=[TaskInputOutput])
class EmailAddressInputOutput(ModelTypeBase):
    pass


@model_type(models.EmailMessageInputOutput, interfaces=[TaskInputOutput])
class EmailMessageInputOutput(ModelTypeBase):
    pass


@model_type(models.HashInputOutput, interfaces=[TaskInputOutput])
class HashInputOutput(ModelTypeBase):
    pass


@model_type(models.BoolInputOutput, interfaces=[TaskInputOutput])
class BoolInputOutput(ModelTypeBase):
    pass


@model_type(models.DateInputOutput, interfaces=[TaskInputOutput])
class DateInputOutput(ModelTypeBase):
    pass


@model_type(models.DateTimeInputOutput, interfaces=[TaskInputOutput])
class DateTimeInputOutput(ModelTypeBase):
    pass


@model_type(models.JSONInputOutput, interfaces=[TaskInputOutput])
class JSONInputOutput(ModelTypeBase):
    pass


@model_type(models.FileRefInputOutput, interfaces=[TaskInputOutput])
class FileRefInputOutput(ModelTypeBase):
    pass


@model_type(models.FileInputOutput, interfaces=[TaskInputOutput])
class FileInputOutput(ModelTypeBase):
    pass


registry = SchemaRegistry([
    CaseDBUser,
    CaseDBRole,
    Label,
    StorageEvent,
    Event,
    Case,
    CaseFileItemBase,
    CaseManagementExternalReference,
    CaseFileItem,
    File,
    Directory,
    CaseRelationship,
    Workflow,
    Task,
    Stage,
    TaskRelationship,
    TaskReference,
    WorkflowTemplate,
    TaskTemplate,
    Activity,
    BaseActivity,
    Comment,
    TaskInputOutput,
    BaseTaskInputOutput,
    StringInputOutput,
    BoolInputOutput,
    URLInputOutput,
    DomainInputOutput,
    IPInputOutput,
    DateInputOutput,
    DateTimeInputOutput,
    JSONInputOutput,
    HashInputOutput,
    FileInputOutput,
    FileRefInputOutput,
    EmailAddressInputOutput,
    EmailMessageInputOutput,
    FlowEvent,
    FlowEventData
])

queries = registry.queries()
mutations = registry.mutations()
