import random
import string

from csirtmu.strawberry_dmt.field_hints import Relationship
from csirtmu.strawberry_dmt.sqla_base import AutoBase, AutoPolymorphicBase, pk_name_for, Base, \
    table_name_for, pk_m2m_name_for
from sqlalchemy import Column, ForeignKey, Integer, Table
from sqlalchemy import String, DateTime, Enum
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import Session
from sqlalchemy.orm import backref, relationship
from sqlalchemy.sql.functions import func

from csirtmu.case_database.config import CONFIG
from csirtmu.case_database.enums import CasePriorityEnum, CaseRelationshipTypeEnum, CaseStateEnum, CaseTypeEnum
from csirtmu.case_database.models.mixins import OrderingWeightMixin
from csirtmu.case_database.utils import get_file_location

ALPHABET = string.ascii_lowercase + string.digits


# helpers
def generate_case_identifier():
    case_identifier = "".join(random.choices(ALPHABET, k=8))
    with Session(bind=CONFIG.database.engine, expire_on_commit=False) as session:
        while session.query(Case).filter_by(case_identifier=case_identifier).first() is not None:
            case_identifier = random.choices(ALPHABET, k=8)
    return case_identifier


def copy_case_identifier(model):
    def copy_with_context(context):
        if (case_identifier := context.get_current_parameters()["case_identifier"]) is not None:
            return case_identifier
        if (model_id := context.get_current_parameters()[f"{table_name_for(model)}_id"]) is not None:
            with Session(bind=CONFIG.database.engine, expire_on_commit=False) as session:
                case_identifier = session.query(model.case_identifier).filter_by(id=model_id).first()
                return case_identifier[0]
    return copy_with_context


# Association tables
__case_events_m2m__ = Table(
    f'{table_name_for("Case")}__m2m__{table_name_for("Event")}',
    Base.metadata,
    Column(pk_m2m_name_for('Case'), Integer, ForeignKey('case.id', ondelete='CASCADE'), primary_key=True),
    Column(pk_m2m_name_for('Event'), Integer,
           ForeignKey('event.id', ondelete='CASCADE'), primary_key=True)
)

__case_labels_m2m__ = Table(
    f'{table_name_for("Case")}__m2m__{table_name_for("Label")}',
    Base.metadata,
    Column(pk_m2m_name_for('Case'), Integer, ForeignKey('case.id', ondelete='CASCADE'), primary_key=True),
    Column(pk_m2m_name_for('Label'), Integer,
           ForeignKey('label.id', ondelete='CASCADE'), primary_key=True)
)

__case_assignees_m2m__ = Table(
    f'{table_name_for("Case")}__m2m__{table_name_for("CaseDBUser")}',
    Base.metadata,
    Column(pk_m2m_name_for('Case'), Integer, ForeignKey('case.id', ondelete='CASCADE'), primary_key=True),
    Column(pk_m2m_name_for('CaseDBUser'), Integer,
           ForeignKey('case_db_user.id', ondelete='CASCADE'), primary_key=True)
)

__case_external_references_m2m__ = Table(
    f'{table_name_for("Case")}__m2m__{table_name_for("CaseManagementExternalReference")}',
    Base.metadata,
    Column(pk_m2m_name_for('Case'), Integer, ForeignKey('case.id', ondelete='CASCADE'), primary_key=True),
    Column(pk_m2m_name_for('CaseManagementExternalReference'), Integer,
           ForeignKey('case_management_external_reference.id', ondelete='CASCADE'), primary_key=True)
)


class Case(OrderingWeightMixin, AutoBase):
    name = Column(String, nullable=False)
    case_identifier = Column(String(8), unique=True, nullable=False, default=generate_case_identifier)
    description = Column(String)
    state = Column(Enum(CaseStateEnum), default=CaseStateEnum.OPEN)
    priority = Column(Enum(CasePriorityEnum), nullable=False)
    type = Column(Enum(CaseTypeEnum), nullable=False)
    created = Column(DateTime(timezone=True), server_default=func.now())
    created_by_id = Column(Integer, ForeignKey(pk_name_for('CaseDBUser'), ondelete='CASCADE'), index=True)
    created_by = relationship(
        'CaseDBUser', backref='created_cases', foreign_keys=[created_by_id])
    last_updated = Column(
        DateTime(timezone=True), onupdate=func.current_timestamp(), server_default=func.now())
    due_date = Column(DateTime(timezone=True))
    source_relationships = Relationship("CaseRelationship")
    target_relationships = Relationship("CaseRelationship")
    labels = relationship('Label', secondary=__case_labels_m2m__, backref='cases')
    workflows = relationship('Workflow', back_populates='case')
    comments = relationship('Comment', back_populates='case')
    assignees = relationship('CaseDBUser',
                             secondary=__case_assignees_m2m__,
                             backref='cases',
                             )

    events = relationship('Event', secondary=__case_events_m2m__, backref='related_cases')
    external_references = relationship('CaseManagementExternalReference', secondary=__case_external_references_m2m__,
                                       backref='related_cases')


class CaseRelationship(AutoBase):
    source_id = Column(Integer, ForeignKey(pk_name_for(Case), ondelete='CASCADE'), nullable=False, index=True)
    source = relationship(
        Case, backref="source_relationships", foreign_keys=[source_id])
    target_id = Column(Integer, ForeignKey(pk_name_for(Case), ondelete='CASCADE'), nullable=False, index=True)
    target = relationship(
        Case, backref="target_relationships", foreign_keys=[target_id])
    type = Column(Enum(CaseRelationshipTypeEnum), nullable=False)


class CaseFileItem(AutoPolymorphicBase(on_column='type')):
    name = Column(String, nullable=False)
    created = Column(DateTime(timezone=True), server_default=func.now())
    created_by_id = Column(Integer, ForeignKey(pk_name_for("CaseDBUser"), ondelete='CASCADE'))
    created_by = relationship("CaseDBUser", backref="created_case_items")
    parent_id = Column(Integer,
                       ForeignKey(pk_name_for("Directory"), use_alter=True, ondelete="CASCADE", name="fk_cfi_parent"),
                       nullable=True)
    parent = relationship("Directory",
                          backref=backref("children", cascade="all, delete, delete-orphan", passive_deletes=True),
                          foreign_keys=[parent_id])
    type = Column(String(20))


class File(CaseFileItem):
    id = Column(Integer, ForeignKey(pk_name_for(CaseFileItem), ondelete='CASCADE'), primary_key=True)
    external_src = Column(String)
    display_name = Column(String)
    extension = Column(String)

    @hybrid_property
    def src(self) -> str:
        if self.external_src:
            return self.external_src
        return get_file_location(self.id, self.name)


class Directory(CaseFileItem):
    id = Column(Integer, ForeignKey(pk_name_for(CaseFileItem), ondelete="CASCADE"), primary_key=True)
    case_id = Column(Integer,
                     ForeignKey(pk_name_for("Case"), use_alter=True, ondelete="CASCADE", name="fk_root_directory"),
                     nullable=True)
    case = relationship("Case", backref=backref("case_file_root", uselist=False), foreign_keys=[case_id])
