import ipaddress
import json
import re
from typing import Optional

import jsonschema
from csirtmu.strawberry_dmt.sqla_base import AutoSingleTablePolymorphicBase, pk_name_for
from sqlalchemy import Boolean, Column, Date, DateTime, ForeignKey, Integer, String
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import backref, relationship
from validators import domain as validate_domain, email as validate_email_address, url as validate_url
from validators.utils import ValidationFailure


class TaskInputOutput(AutoSingleTablePolymorphicBase('type')):
    type = Column(String(50))
    description = Column(String)
    input_of_id = Column(Integer,
                         ForeignKey(pk_name_for("Task"), use_alter=True, ondelete="CASCADE", name="fk_task_input"),
                         index=True)
    input_of = relationship("Task", backref=backref("input", uselist=False), uselist=False, foreign_keys=[input_of_id])
    output_of_id = Column(Integer,
                          ForeignKey(pk_name_for('Task'), use_alter=True, ondelete="CASCADE", name="fk_task_output"),
                          index=True)
    output_of = relationship("Task", backref=backref("output", uselist=False), uselist=False,
                             foreign_keys=[output_of_id])

    # naive copy constructor working only on values
    def copy(self):
        return self.__class__(value=self.value, format_field=self.format_field)

    @property
    def value(self):
        return None

    @value.setter
    def value(self, value):
        pass

    @property
    def format_field(self):
        return None

    @format_field.setter
    def format_field(self, value):
        pass

    @hybrid_property
    def stringified_value(self) -> Optional[str]:
        pass

    def __init__(self, value=None, format_field=None, description=None):
        if format_field is not None:
            self.format_field = format_field
        if value is not None:
            self.value = self.validate(value)
        self.description = description

    def validate(self, value):
        return value


class StringInputOutput(TaskInputOutput):
    string_value = Column(String)
    string_regex = Column(String)

    @property
    def value(self):
        return self.string_value

    @value.setter
    def value(self, value):
        self.string_value = self.validate(value)

    @property
    def format_field(self):
        return self.string_regex

    @format_field.setter
    def format_field(self, value):
        try:
            re.compile(value)
            self.string_regex = value
        except:
            raise ValueError(f"{value} is not valid regex pattern!")

    @hybrid_property
    def stringified_value(self) -> Optional[str]:
        return self.string_value

    def validate(self, value):
        if self.format_field:
            if not re.match(self.format_field, value):
                raise ValueError(f"{value} does not satisfy pattern {self.format_field}!")
        return value


class JSONInputOutput(TaskInputOutput):
    json_value = Column(String)
    json_schema = Column(String)
    json_initial_value = Column(String)
    json_formly_schema = Column(String)

    @property
    def value(self):
        return self.json_value

    @value.setter
    def value(self, value):
        self.json_value = self.validate(value)

    @property
    def format_field(self):
        if self.json_schema is None:
            return None
        return json.loads(self.json_schema)

    @format_field.setter
    def format_field(self, value):
        try:
            if isinstance(value, str):
                data = json.loads(value)
            elif isinstance(value, dict):
                data = value
            else:
                raise ValueError('Unexpected value!')
            self.json_schema = json.dumps(data)
        except:
            raise ValueError("JSON schema must be valid JSON!")

    def validate(self, value):
        try:
            data = value
            if isinstance(value, str):
                data = json.loads(value)
            elif isinstance(value, dict):
                data = value
            value = json.dumps(data)
        except Exception as e:
            raise ValueError("Not JSON!")

        if schema := self.format_field:
            jsonschema.validate(data, schema)
        return value

    @hybrid_property
    def stringified_value(self) -> Optional[str]:
        return self.json_value


class BoolInputOutput(TaskInputOutput):
    boolean_value = Column(Boolean)

    @property
    def value(self):
        return self.boolean_value

    @value.setter
    def value(self, value):
        self.boolean_value = self.validate(value)

    def validate(self, value):
        if value in ["false", "False", "F", False]:
            return False
        elif value in ["true", "True", "T", True]:
            return True
        else:
            raise ValueError("Not boolean value!")

    @hybrid_property
    def stringified_value(self) -> Optional[str]:
        if self.boolean_value is not None:
            return str(self.boolean_value)
        return None


class HashInputOutput(TaskInputOutput):
    hash_value = Column(String)

    @property
    def value(self):
        return self.hash_value

    @value.setter
    def value(self, value):
        self.hash_value = self.validate(value)

    @hybrid_property
    def stringified_value(self) -> Optional[str]:
        return self.hash_value


class DomainInputOutput(TaskInputOutput):
    domain_value = Column(String)

    @property
    def value(self):
        return self.domain_value

    @value.setter
    def value(self, value):
        self.domain_value = self.validate(value)

    def validate(self, value):
        if isinstance(validate_domain(value), ValidationFailure):
            raise ValueError(f"{value} is not a valid domain name!")
        return value

    @hybrid_property
    def stringified_value(self) -> Optional[str]:
        return self.domain_value


class URLInputOutput(TaskInputOutput):
    url_value = Column(String)

    @property
    def value(self):
        return self.url_value

    @value.setter
    def value(self, value):
        self.url_value = self.validate(value)

    def validate(self, value):
        if isinstance(validate_url(value), ValidationFailure):
            raise ValueError(f"{value} is not a valid URL!")
        return value

    @hybrid_property
    def stringified_value(self) -> Optional[str]:
        return self.url_value


class IPInputOutput(TaskInputOutput):
    ip_value = Column(String)

    @property
    def value(self):
        return self.ip_value

    @value.setter
    def value(self, value):
        self.ip_value = self.validate(value)

    def validate(self, value):
        try:
            ipaddress.ip_address(value)
        except ValueError:
            raise ValueError(f"{value} is not a valid IP address!")
        return value

    @hybrid_property
    def stringified_value(self) -> Optional[str]:
        return self.ip_value


class DateInputOutput(TaskInputOutput):
    date_value = Column(Date)

    @property
    def value(self):
        return self.date_value

    @value.setter
    def value(self, value):
        self.date_value = self.validate(value)

    @hybrid_property
    def stringified_value(self) -> Optional[str]:
        return self.date_value


class DateTimeInputOutput(TaskInputOutput):
    datetime_value = Column(DateTime)

    @property
    def value(self):
        return self.datetime_value

    @value.setter
    def value(self, value):
        self.datetime_value = self.validate(value)

    @hybrid_property
    def stringified_value(self) -> Optional[str]:
        return self.datetime_value


class EmailAddressInputOutput(TaskInputOutput):
    email_address_value = Column(String)

    @property
    def value(self):
        return self.email_address_value

    @value.setter
    def value(self, value):
        self.email_address_value = self.validate(value)

    def validate(self, value):
        if isinstance(validate_email_address(value), ValidationFailure):
            raise ValueError(f"{value} is not a valid Email Address!")
        return value

    @hybrid_property
    def stringified_value(self) -> Optional[str]:
        return self.email_address_value


class EmailMessageInputOutput(TaskInputOutput):
    email_message_value = Column(String)

    @property
    def value(self):
        return self.email_message_value

    @value.setter
    def value(self, value):
        self.email_message_value = self.validate(value)

    @hybrid_property
    def stringified_value(self) -> Optional[str]:
        return self.email_message_value


class FileRefInputOutput(TaskInputOutput):
    file_ref_value = Column(String)

    @property
    def value(self):
        return self.file_ref_value

    @value.setter
    def value(self, value):
        self.file_ref_value = self.validate(value)

    @hybrid_property
    def stringified_value(self) -> Optional[str]:
        return self.file_ref_value


class FileInputOutput(TaskInputOutput):
    file_value = Column(String)

    @property
    def value(self):
        return self.file_value

    @value.setter
    def value(self, value):
        self.file_value = self.validate(value)

    @hybrid_property
    def stringified_value(self) -> Optional[str]:
        return self.file_value
