import asyncio
from functools import cached_property
from typing import List

from csirtmu.starlette_jwt_auth.models import MappedUser
from csirtmu.strawberry_dmt.sqla_base import AutoBase, m2m_for
from gql.transport.exceptions import TransportError
from sqlalchemy import Column, String
from sqlalchemy.orm import Session, relationship

# Association tables
__user_roles_m2m__ = m2m_for('CaseDBUser', 'CaseDBRole')

from csirtmu.case_database.config import CONFIG
from csirtmu.case_database.services.icons_service import IconsService
from csirtmu.case_database.user_profile_client import DEFAULT_USERNAME, UserProfileIntrospectionClient

USER_INTROSPECTION_CLIENT = UserProfileIntrospectionClient()


class UserMeta(type(AutoBase), type(MappedUser)):
    pass


class CaseDBUser(AutoBase, MappedUser, metaclass=UserMeta):
    name = Column(String, nullable=False)
    first_name = Column(String, nullable=True)
    last_name = Column(String, nullable=True)
    email = Column(String, nullable=True)
    icon_src = Column(String)
    roles = relationship('CaseDBRole', secondary=__user_roles_m2m__,
                         back_populates='granted_to')
    global_identifier = Column(String, nullable=False, unique=True)

    @property
    def identity(self) -> str:
        return self.id

    @property
    def is_authenticated(self) -> bool:
        return True

    @property
    def display_name(self) -> str:
        return self.name

    @cached_property
    def capabilities(self):
        return [r.name for r in self.roles]

    @staticmethod
    def create_or_get(access_token: str, user_id: str, capabilities: List[str]) -> 'MappedUser':
        with Session(bind=CONFIG.database.engine, expire_on_commit=False) as s:
            user = s.query(CaseDBUser).filter_by(global_identifier=user_id).first()
            roles = s.query(CaseDBRole).where(CaseDBRole.name.in_(capabilities)).all()

            missing_role_names = list(set(capabilities) - set(r.name for r in roles))
            if missing_role_names:
                roles = roles + [CaseDBRole(name=name) for name in missing_role_names]
                s.add_all(roles)
                s.commit()

            if not user:
                icon_service = IconsService()

                user = CaseDBUser(name=DEFAULT_USERNAME, global_identifier=user_id, roles=roles,
                                  icon_src=icon_service.get_icon(user_id))
                s.add(user)
            elif user.roles != roles:
                user.roles = roles
                s.add(user)

            if user.name == DEFAULT_USERNAME:
                try:
                    user_introspection = USER_INTROSPECTION_CLIENT\
                        .get_user_introspection_for_current_user(user_token=access_token)
                except TransportError:
                    user_introspection = None
                if user_introspection:
                    user.name = user_introspection.username
                    user.first_name = user_introspection.first_name
                    user.last_name = user_introspection.last_name
                    user.email = user_introspection.email
                    s.add(user)
            s.commit()
            return user


class CaseDBRole(AutoBase):
    name = Column(String, nullable=False)
    granted_to = relationship(
        'CaseDBUser', secondary=__user_roles_m2m__, back_populates="roles")
