from csirtmu.strawberry_dmt.sqla_base import AutoBase, Base, pk_m2m_name_for, pk_name_for, \
    table_name_for
from sqlalchemy import Column, DateTime, ForeignKey, Integer, String, Table
from sqlalchemy.orm import relationship
from sqlalchemy.sql.functions import func
from sqlalchemy.sql.sqltypes import Boolean, Enum, Float

from csirtmu.case_database.enums import StageStateEnum, TaskReferenceTypeEnum, \
    TaskRelationshipTypeEnum, TaskStateEnum, WorkflowStateEnum
from csirtmu.case_database.models.case import Case, copy_case_identifier
from csirtmu.case_database.models.mixins import OrderingWeightMixin

__workflow_assignees_m2m__ = Table(
    f'{table_name_for("Workflow")}__m2m__{table_name_for("CaseDBUser")}',
    Base.metadata,
    Column(pk_m2m_name_for('Workflow'), Integer, ForeignKey('workflow.id', ondelete='CASCADE'), primary_key=True),
    Column(pk_m2m_name_for('CaseDBUser'), Integer,
           ForeignKey('case_db_user.id', ondelete='CASCADE'), primary_key=True)
)

__stage_assignees_m2m__ = Table(
    f'{table_name_for("Stage")}__m2m__{table_name_for("CaseDBUser")}',
    Base.metadata,
    Column(pk_m2m_name_for('Stage'), Integer, ForeignKey('stage.id', ondelete='CASCADE'), primary_key=True),
    Column(pk_m2m_name_for('CaseDBUser'), Integer,
           ForeignKey('case_db_user.id', ondelete='CASCADE'), primary_key=True)
)

__task_assignees_m2m__ = Table(
    f'{table_name_for("Task")}__m2m__{table_name_for("CaseDBUser")}',
    Base.metadata,
    Column(pk_m2m_name_for('Task'), Integer, ForeignKey('task.id', ondelete='CASCADE'), primary_key=True),
    Column(pk_m2m_name_for('CaseDBUser'), Integer,
           ForeignKey('case_db_user.id', ondelete='CASCADE'), primary_key=True)
)

__task_events_m2m__ = Table(
    f'{table_name_for("Task")}__m2m__{table_name_for("Event")}',
    Base.metadata,
    Column(pk_m2m_name_for('Task'), Integer, ForeignKey('task.id', ondelete='CASCADE'), primary_key=True),
    Column(pk_m2m_name_for('Event'), Integer,
           ForeignKey('event.id', ondelete='CASCADE'), primary_key=True)
)

__task_external_references_m2m__ = Table(
    f'{table_name_for("Task")}__m2m__{table_name_for("CaseManagementExternalReference")}',
    Base.metadata,
    Column(pk_m2m_name_for('Task'), Integer, ForeignKey('task.id', ondelete='CASCADE'), primary_key=True),
    Column(pk_m2m_name_for('CaseManagementExternalReference'), Integer,
           ForeignKey('case_management_external_reference.id', ondelete='CASCADE'), primary_key=True)
)

__task_labels_m2m__ = Table(
    f'{table_name_for("Task")}__m2m__{table_name_for("Label")}',
    Base.metadata,
    Column(pk_m2m_name_for('Task'), Integer, ForeignKey('task.id', ondelete='CASCADE'), primary_key=True),
    Column(pk_m2m_name_for('Label'), Integer,
           ForeignKey('label.id', ondelete='CASCADE'), primary_key=True)
)


class Workflow(OrderingWeightMixin, AutoBase):
    name = Column(String, nullable=False)
    case_identifier = Column(String(8), default=copy_case_identifier(Case))
    description = Column(String, nullable=False)
    created = Column(DateTime(timezone=True), server_default=func.now())
    last_updated = Column(
        DateTime(timezone=True), onupdate=func.current_timestamp(), server_default=func.now())
    due_date = Column(DateTime(timezone=True))
    assignees = relationship("CaseDBUser", secondary=__workflow_assignees_m2m__, backref="workflows")
    case_id = Column(Integer, ForeignKey(pk_name_for("Case"), ondelete='CASCADE'), index=True, nullable=False)
    case = relationship("Case", back_populates="workflows")

    created_by_id = Column(Integer, ForeignKey(pk_name_for('CaseDBUser'), ondelete='CASCADE'), index=True)
    created_by = relationship(
        'CaseDBUser', backref='created_workflows', foreign_keys=[created_by_id])

    progress = Column(Float, default=0.0, nullable=False)
    state = Column(Enum(WorkflowStateEnum), default=WorkflowStateEnum.IN_PROGRESS, nullable=False)


class Stage(OrderingWeightMixin, AutoBase):
    name = Column(String, nullable=False)
    case_identifier = Column(String(8), default=copy_case_identifier(Workflow))
    description = Column(String, nullable=False)
    created = Column(DateTime(timezone=True), server_default=func.now())
    last_updated = Column(
        DateTime(timezone=True), onupdate=func.current_timestamp(), server_default=func.now())
    due_date = Column(DateTime(timezone=True))
    workflow_id = Column(Integer, ForeignKey(pk_name_for("Workflow"), ondelete='CASCADE'), index=True, nullable=False)
    workflow = relationship("Workflow", backref="stages", foreign_keys=[workflow_id])
    tasks = relationship("Task", back_populates="stage")
    assignees = relationship("CaseDBUser", secondary=__stage_assignees_m2m__, backref="stages")

    created_by_id = Column(Integer, ForeignKey(pk_name_for('CaseDBUser'), ondelete='CASCADE'), index=True)
    created_by = relationship(
        'CaseDBUser', backref='created_stages', foreign_keys=[created_by_id])

    state = Column(Enum(StageStateEnum), default=StageStateEnum.PENDING, nullable=False, index=True)
    progress = Column(Float, default=0.0, nullable=False)


class Task(OrderingWeightMixin, AutoBase):
    name = Column(String, nullable=False)
    case_identifier = Column(String(8), default=copy_case_identifier(Stage))
    description = Column(String, nullable=False)
    created = Column(DateTime(timezone=True), server_default=func.now())
    last_updated = Column(
        DateTime(timezone=True), onupdate=func.current_timestamp(), server_default=func.now())
    due_date = Column(DateTime(timezone=True))
    labels = relationship('Label', secondary=__task_labels_m2m__, backref='tasks')
    state = Column(Enum(TaskStateEnum), default=TaskStateEnum.OPEN, index=True)
    stage_id = Column(Integer, ForeignKey(pk_name_for("Stage"), ondelete='CASCADE'), index=True, nullable=False)
    stage = relationship("Stage", back_populates="tasks")
    comments = relationship('Comment', back_populates='task')
    created_by_id = Column(Integer, ForeignKey(pk_name_for("CaseDBUser"), ondelete='CASCADE'), index=True)
    created_by = relationship(
        'CaseDBUser', backref='created_tasks', foreign_keys=[created_by_id])
    assignees = relationship("CaseDBUser", secondary=__task_assignees_m2m__, backref="tasks")
    events = relationship('Event', secondary=__task_events_m2m__, backref='related_tasks')
    external_references = relationship('CaseManagementExternalReference', secondary=__task_external_references_m2m__,
                                       backref='related_tasks')
    # flags
    is_automated = Column(Boolean, default=False)
    autotrigger = Column(Boolean, default=False)
    processing_queue = Column(String)
    aggregates_results = Column(Boolean, default=False)
    is_decision = Column(Boolean, default=False)
    decision = Column(Boolean)
    output_is_joined = Column(Boolean, default=False)


class TaskReference(AutoBase):
    source_id = Column(Integer,
                       ForeignKey(pk_name_for("Task"), use_alter=True, ondelete="CASCADE", name="fk_source_reference"),
                       nullable=False)
    source = relationship("Task", backref="refers_to", foreign_keys=[source_id])
    target_id = Column(Integer,
                       ForeignKey(pk_name_for("Task"), use_alter=True, ondelete="CASCADE", name="fk_target_reference"),
                       nullable=False)
    target = relationship("Task", backref="referenced_by", foreign_keys=[target_id])
    relationship_type = Column(Enum(TaskReferenceTypeEnum), nullable=False)


class TaskRelationship(AutoBase):
    source_id = Column(Integer, ForeignKey(pk_name_for("Task"), use_alter=True, ondelete="CASCADE",
                                           name="fk_source_relationship"), nullable=False, index=True)
    source = relationship("Task", backref="source_relationships", foreign_keys=[source_id])
    target_id = Column(Integer, ForeignKey(pk_name_for("Task"), use_alter=True, ondelete="CASCADE",
                                           name="fk_target_relationship"), nullable=False, index=True)
    target = relationship("Task", backref="target_relationships", foreign_keys=[target_id])
    relationship_type = Column(Enum(TaskRelationshipTypeEnum), nullable=False)
