from abc import ABC
from io import BytesIO
from pathlib import Path

from minio import Minio

from csirtmu.case_database.config import CONFIG


class FileService(ABC):
    def store_file(self, file_name: str, file: bytes, file_id: int = None):
        return f"{CONFIG.external_url}/file/{file_id}/{file_name}"

    def get_file(self, file_link: str):
        pass


class LocalFileService(FileService):
    def __init__(self, file_path: str):
        self.file_path = file_path

    def store_file(self, file_name: str, file: bytes, file_id: int = None):
        directory_location = Path(self.file_path)
        full_file_name = f"{file_id}_{file_name}"

        Path(directory_location).mkdir(parents=True, exist_ok=True)
        file_path = str((directory_location / full_file_name).resolve())

        with open(file_path, "wb") as f:
            f.write(file)

        return super().store_file(file_name, None, file_id)

    def get_file(self, file_link: str, file_id: int = None):
        file_path = str((Path(self.file_path) / f"{file_id}_{file_link}"))
        with open(file_path, "rb") as stored_file:
            content = stored_file.read()
            return content


class S3FileService(FileService):
    def __init__(self, host: str, bucket: str = "casedb_files", acces_key: str = None, secret_key: str = None):
        self.host = host
        self.bucket = bucket
        self.client = Minio(self.host, access_key=acces_key, secret_key=secret_key, secure=False)

    def store_file(self, file_name: str, file: bytes, file_id: int = None):
        if not self.client.bucket_exists(self.bucket):
            self.client.make_bucket(self.bucket)

        path = f"{file_id}_{file_name}"
        with BytesIO(file) as memory_file:
            self.client.put_object(self.bucket, path, memory_file, len(file))

        return super().store_file(file_name, None, file_id)

    def get_file(self, file_link: str, file_id: int = None):
        file = self.client.get_object(self.bucket, f"{file_id}_{file_link}")
        file_content = file.data
        return file_content


class FileHandler:
    @staticmethod
    def get_file_handler() -> "FileHandler":
        if CONFIG.file_storage_type == "minio":
            return FileHandler(S3FileService(f"{CONFIG.minio_file_storage.host}:{CONFIG.minio_file_storage.port}",
                                             CONFIG.minio_file_storage.bucket,
                                             acces_key=CONFIG.minio_file_storage.access_key,
                                             secret_key=CONFIG.minio_file_storage.secret_key))
        elif CONFIG.file_storage_type == "local":
            return FileHandler(LocalFileService(CONFIG.local_file_storage.file_directory))

        return None

    def __init__(self, file_service: "FileService"):
        self.file_service = file_service

    def store_file(self, file_name: str, file: bytes, file_id: int = None):
        return self.file_service.store_file(file_name, file, file_id)

    def get_file(self, file_link: str, file_id: int = None):
        return self.file_service.get_file(file_link, file_id)
