import json
import string
from itertools import chain
from pathlib import Path
from typing import Optional, Union

import strawberry
from jsonschema.validators import validate
from redis import Redis
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import selectinload
from starlette.datastructures import UploadFile

from csirtmu.case_database.config import CONFIG
from csirtmu.case_database.enums import CaseStateEnum, StageStateEnum, TaskRelationshipTypeEnum, TaskStateEnum, \
    WorkflowStateEnum
from csirtmu.case_database.gql.enums import KnownInputOutputTypes
from csirtmu.case_database.gql.inputs import CaseCreate, FileCreate, FileRefCreate, TaskCreate, TaskCreateParentless, \
    TemplateCreate
from csirtmu.case_database.models import Case, CaseDBUser, Directory, File, JSONInputOutput, Label, Stage, Task, \
    TaskInputOutput, TaskRelationship, TaskTemplate, Workflow, WorkflowTemplate
from csirtmu.case_database.services.file_service import FileHandler
from csirtmu.case_database.services.mappings import input_output_map
from csirtmu.case_database.services.task_generator_service import TaskGenerator
from csirtmu.case_database.services.template_generator_service import TemplateGenerator
from csirtmu.case_database.services.workflow_generator_service import WorkflowGenerator
from csirtmu.case_database.utils import get_file_location, has_multiple_predecessors, has_resolved_predecessors, \
    has_skipped_predecessors

JSON_SCHEMAS_DIR = Path(__file__).resolve().parent / 'json_schemas'
ALPHABET = string.ascii_lowercase + string.digits


class ServiceLayer:
    @staticmethod
    async def get_user(session: AsyncSession, user_id) -> CaseDBUser:
        if user_id is None:
            return None
        return (await session.execute(select(CaseDBUser).filter_by(id=int(user_id)))).scalar()

    @staticmethod
    async def get_label(session: AsyncSession, label_id) -> Label:
        if label_id is None:
            return None
        return (await session.execute(select(Label).filter_by(id=int(label_id)))).scalar()

    @staticmethod
    async def get_case(session: AsyncSession, case_id: strawberry.ID) -> Case:
        options = selectinload(Case.workflows)
        if case_id is None:
            return None
        return (await session.execute(select(Case).filter_by(id=int(case_id)).options(options))).scalar()

    @staticmethod
    async def get_task(session: AsyncSession, task_id: strawberry.ID) -> Task:
        if task_id is None:
            return None
        options = [
            selectinload(Task.source_relationships).selectinload(TaskRelationship.target),
            selectinload(Task.input),
            selectinload(Task.output),
            selectinload(Task.stage),
        ]
        expr = select(Task).filter_by(id=int(task_id))
        for opt in options:
            expr = expr.options(opt)
        return (await session.execute(expr)).scalar()

    @staticmethod
    async def get_stage(session: AsyncSession, stage_id: strawberry.ID) -> Stage:
        options = selectinload(Stage.tasks)
        if stage_id is None:
            return None
        return (await session.execute(select(Stage).filter_by(id=int(stage_id)).options(options))).scalar()

    @staticmethod
    async def get_workflow(session: AsyncSession, workflow_id: strawberry.ID) -> Workflow:
        options = selectinload(Workflow.stages)
        if workflow_id is None:
            return None
        return (await session.execute(select(Workflow).filter_by(id=int(workflow_id)).options(options))).scalar()

    @staticmethod
    async def get_directory(session: AsyncSession, directory_id: strawberry.ID) -> Optional[Directory]:
        if directory_id is None:
            return None
        return (await session.execute(select(Directory).filter_by(id=int(directory_id)))).scalar()

    @staticmethod
    async def skip_task(session: AsyncSession, task_id: strawberry.ID) -> bool:
        options = [selectinload(Task.source_relationships).selectinload(TaskRelationship.target)
                   .selectinload(Task.target_relationships).selectinload(TaskRelationship.source)]
        expr = select(Task).filter_by(id=int(task_id))
        for opt in options:
            expr = expr.options(opt)
        task = (await session.execute(expr)).scalar()
        if not task:
            raise ValueError(f"Task with id {task_id} does not exist!")
        for relationship in task.source_relationships:
            if has_multiple_predecessors(relationship.target):
                if has_skipped_predecessors(relationship.target, task):
                    await ServiceLayer.skip_task(session, relationship.target.id)
                elif has_resolved_predecessors(relationship.target, task):
                    await ServiceLayer.open_task(session, relationship.target.id)
            else:
                await ServiceLayer.skip_task(session, relationship.target.id)
        task.state = TaskStateEnum.SKIPPED
        return True

    @staticmethod
    async def resolve_case(session: AsyncSession, case_id: strawberry.ID) -> bool:
        case = await ServiceLayer.get_case(session, case_id)
        if not case:
            raise ValueError(f"Case with id {case_id} does not exist!")
        for workflow in case.workflows:
            if workflow.state not in [WorkflowStateEnum.DONE, WorkflowStateEnum.REJECTED]:
                raise ValueError('Not all workflows are completed!')
        case.state = CaseStateEnum.RESOLVED
        return True

    @staticmethod
    async def reject_case(session: AsyncSession, case_id: strawberry.ID) -> bool:
        case = await ServiceLayer.get_case(session, case_id)
        if not case:
            raise ValueError(f"Case with id {case_id} does not exist!")
        for workflow in case.workflows:
            await ServiceLayer.reject_workflow(session, workflow.id)
        case.state = CaseStateEnum.REJECTED
        return True

    @staticmethod
    async def reject_workflow(session: AsyncSession, workflow_id: strawberry.ID) -> bool:
        workflow = await ServiceLayer.get_workflow(session, workflow_id)
        if not workflow:
            raise ValueError(f"Workflow with id {workflow_id} does not exist!")
        for stage in workflow.stages:
            await ServiceLayer.reject_stage(session, stage.id)
        workflow.state = WorkflowStateEnum.REJECTED
        return True

    @staticmethod
    async def reject_stage(session: AsyncSession, stage_id: strawberry.ID) -> bool:
        stage = await ServiceLayer.get_stage(session, stage_id)
        if not stage:
            raise ValueError(f"Stage with id {stage_id} does not exist!")
        for task in stage.tasks:
            await ServiceLayer.reject_task(session, task.id)
        stage.state = StageStateEnum.REJECTED
        return True

    @staticmethod
    async def reject_task(session: AsyncSession, task_id: strawberry.ID) -> bool:
        task = await ServiceLayer.get_task(session, task_id)
        if not task:
            raise ValueError(f"Task with id {task_id} does not exist!")

        #collect all tasks to reject
        tasks_to_reject = set([int(task_id)])
        top_q = select(TaskRelationship.source_id, TaskRelationship.target_id).filter(TaskRelationship.source_id == int(task_id)).cte(name='task_children', recursive=True)
        bottom_q = select(TaskRelationship.source_id, TaskRelationship.target_id).join(top_q, top_q.c.target_id == TaskRelationship.source_id)
        q = select(top_q.union(bottom_q))
        r = (await session.execute(q)).all()
        tasks_to_reject.update(chain(*r))
        # reject tasks
        tasks = list((await session.execute(select(Task).filter(Task.id.in_(tasks_to_reject)))).scalars())
        for task in tasks:
            task.state = TaskStateEnum.REJECTED
        return True

    @staticmethod
    async def trigger_task(session: AsyncSession, task_id):
        options = [selectinload(Task.input),
                   selectinload(Task.stage).selectinload(Stage.workflow).selectinload(Workflow.case)]
        expr = select(Task).filter_by(id=int(task_id))
        for opt in options:
            expr = expr.options(opt)
        task = (await session.execute(expr)).scalar()

        if task is None:
            raise ValueError("Could not find such task")

        serialized_task = {
            "case_id": task.stage.workflow.case.id,
            "task_id": task_id
        }

        if task.input:
            serialized_task["input"] = task.input.stringified_value

        if task.processing_queue is None:
            raise ValueError("There is not defined process")

        connection = Redis(CONFIG.redis.host, CONFIG.redis.port, db=0, username=CONFIG.redis.username,
                           password=CONFIG.redis.password)
        connection.rpush(task.processing_queue, json.dumps(serialized_task))
        task.state = TaskStateEnum.RUNNING

    @staticmethod
    async def open_task(session: AsyncSession, task_id) -> bool:
        task = await ServiceLayer.get_task(session, task_id)
        if not task:
            raise ValueError(f"Task with id {task_id} does not exist!")

        for relationship in task.source_relationships:
            if relationship.relationship_type == TaskRelationshipTypeEnum.BLOCKED_BY:
                await ServiceLayer.open_task(session, relationship.target.id)

        task.state = TaskStateEnum.OPEN
        if task.is_automated and task.autotrigger and task.input:
            try:
                await ServiceLayer.trigger_task(session, task_id)
            except Exception as e:
                pass
        return True

    @staticmethod
    async def reactivate_task(session: AsyncSession, task_id):
        options = [selectinload(Task.output),
                   selectinload(Task.source_relationships).selectinload(TaskRelationship.target),
                   selectinload(Task.target_relationships).selectinload(TaskRelationship.source),
                   selectinload(Task.stage).selectinload(Stage.workflow).selectinload(Workflow.case)]
        expr = select(Task).filter_by(id=int(task_id))
        for opt in options:
            expr = expr.options(opt)
        task = (await session.execute(expr)).scalar()
        # no task found
        if not task:
            raise ValueError(f"Task with id {task_id} does not exist!")

        # task is either in active state or rejected, we do not do anything
        if task.state == [TaskStateEnum.REJECTED, TaskStateEnum.OPEN, TaskStateEnum.RUNNING, TaskStateEnum.PENDING]:
            return True

        # if the case was resolved as well we should reopen it
        if task.stage.workflow.case.state == CaseStateEnum.RESOLVED:
            task.stage.workflow.case.state = CaseStateEnum.OPEN
        # we do not care about number of predecessors, if all of them are skipped or resolved, we should open this task, pending otherwise
        if has_resolved_predecessors(task, task):
            task.state = TaskStateEnum.OPEN
        else:
            task.state = TaskStateEnum.PENDING

        for relationship in task.source_relationships:
            # we do not touch blocking tasks, if they were previously skipped, new blocker does not change their handling
            # we must reactivate normal successors, since new blocker CAN change their handling
            if relationship.relationship_type != TaskRelationshipTypeEnum.BLOCKED_BY:
                await ServiceLayer.reactivate_task(session, relationship.target.id)
        return True

    @staticmethod
    async def resolve_task(session: AsyncSession, task_id: strawberry.ID, decision: Optional[bool] = None) -> bool:
        options = [selectinload(Task.output),
                   selectinload(Task.source_relationships).selectinload(TaskRelationship.target)
                   .selectinload(Task.target_relationships).selectinload(TaskRelationship.source),
                   selectinload(Task.source_relationships).selectinload(TaskRelationship.target)
                   .selectinload(Task.input),
                   selectinload(Task.target_relationships).selectinload(TaskRelationship.source)
                   .selectinload(Task.source_relationships).selectinload(TaskRelationship.target)]
        expr = select(Task).filter_by(id=int(task_id))
        for opt in options:
            expr = expr.options(opt)
        task = (await session.execute(expr)).scalar()
        if not task:
            raise ValueError(f"Task with id {task_id} does not exist!")

        if task.output and task.output.value is None:
            raise ValueError(f"Task output value is required but it is not set!")

        if task.is_decision is True:
            if decision is None:
                raise ValueError(f"Task requires decision to be resolved!")
            task.decision = decision
            for relationship in task.source_relationships:
                relationship_type = relationship.relationship_type
                if relationship.target.state in [TaskStateEnum.PENDING, TaskStateEnum.OPEN]:
                    if relationship_type in [TaskRelationshipTypeEnum.ON_TRUE, TaskRelationshipTypeEnum.ON_FALSE]:
                        if (relationship_type == TaskRelationshipTypeEnum.ON_TRUE and task.decision) or \
                                (relationship_type == TaskRelationshipTypeEnum.ON_FALSE and not task.decision):
                            if has_multiple_predecessors(relationship.target):
                                if has_resolved_predecessors(relationship.target, task):
                                    await ServiceLayer.open_task(session, relationship.target.id)
                            else:
                                if task.output and type(relationship.target.input) == type(
                                        task.output) and relationship.target.input.value is None:
                                    relationship.target.input.value = task.output.value
                                await ServiceLayer.open_task(session, relationship.target.id)
                        else:
                            await ServiceLayer.skip_task(session, relationship.target.id)

        if task.aggregates_results and not task.output_is_joined:
            raise ValueError("Blocking tasks outputs were not joined!")

        for relationship in task.source_relationships:
            if relationship.target.state in [TaskStateEnum.PENDING, TaskStateEnum.OPEN, TaskStateEnum.RUNNING]:
                if relationship.relationship_type == TaskRelationshipTypeEnum.BLOCKED_BY:
                    raise ValueError("Not all blocking tasks were resolved!")
                if relationship.relationship_type == TaskRelationshipTypeEnum.ON_COMPLETION:
                    if has_multiple_predecessors(relationship.target):
                        if has_resolved_predecessors(relationship.target, task):
                            await ServiceLayer.open_task(session, relationship.target.id)
                    else:
                        if task.output and \
                                type(relationship.target.input) == type(task.output):
                            if relationship.target.input.value is None:
                                try:
                                    relationship.target.input.value = task.output.value
                                except:
                                    pass
                            if isinstance(relationship.target.input,
                                          JSONInputOutput) and relationship.target.input.json_initial_value is None:
                                relationship.target.input.json_initial_value = task.output.value
                        await ServiceLayer.open_task(session, relationship.target.id)
        task.state = TaskStateEnum.RESOLVED
        return True

    @staticmethod
    async def create_case(session: AsyncSession, case_input: CaseCreate) -> Case:
        case = Case(name=case_input.name,
                    priority=case_input.priority,
                    type=case_input.type,
                    case_file_root=Directory(name="case_root")
                    )
        if case_input.description:
            case.description = case_input.description

        if case_input.state:
            case.state = case_input.state

        if case_input.created_by:
            case.created_by = await ServiceLayer.get_user(session, case_input.created_by)

        if case_input.due_date:
            case.due_date = case_input.due_date

        if case_input.labels:
            case_input.labels = map(int, case_input.labels)
            case.labels = list((await session.execute(select(Label).filter(Label.id.in_(case_input.labels)))).scalars())
        if case_input.assignees:
            case_input.assignees = map(int, case_input.assignees)
            case.assignees = list(
                (await session.execute(select(CaseDBUser).filter(CaseDBUser.id.in_(case_input.assignees)))).scalars())
        session.add(case)
        return case

    @staticmethod
    async def get_workflow_from_template(session: AsyncSession, template_id: strawberry.ID,
                                         case_identifier=None) -> Workflow:
        generator = WorkflowGenerator()
        workflow_template = (
            await session.execute(select(WorkflowTemplate.template).filter_by(id=int(template_id)))).one().template
        workflow, relationships = generator.parse_workflow_template(template=workflow_template,
                                                                    case_identifier=case_identifier)
        session.add(workflow)
        session.add_all(relationships)
        return workflow

    @staticmethod
    async def get_task_from_template(session: AsyncSession, template_id: strawberry.ID, case_identifier=None) -> Task:
        generator = TaskGenerator()
        task_template = (
            await session.execute(select(TaskTemplate.template).filter_by(id=int(template_id)))).one().template
        task = generator.parse_task_template(task_template=task_template, case_identifier=case_identifier)
        session.add(task)
        return task

    @staticmethod
    async def prepare_input_output_object(session: AsyncSession, io_type, io_value=None, io_format=None,
                                          description=None) -> TaskInputOutput:
        io_object = input_output_map.get(io_type, TaskInputOutput)(value=io_value, format_field=io_format,
                                                                   description=description)
        session.add(io_object)
        return io_object

    @staticmethod
    async def create_task(session: AsyncSession, task_input: Union[TaskCreate, TaskCreateParentless]) -> Task:
        if not (stage := await ServiceLayer.get_stage(session, task_input.stage)) and task_input.stage:
            raise ValueError(f"Stage {task_input.stage} does not exist")

        if getattr(task_input, "predecessors"):
            predecessors = task_input.predecessors
            if len(predecessors) == 1:
                if not (predecessor := await ServiceLayer.get_task(session, predecessors[0].predecessor)):
                    raise ValueError("Unknown predecessor!")
                if stage is None:
                    stage = predecessor.stage
            else:
                if stage is None:
                    raise ValueError(
                        "Couldn't resolve stage automatically from predecessors. Manual specification is required.")

        task = Task(
            name=task_input.name,
            description=task_input.description,
            stage=stage,
            state=task_input.state,
            is_automated=task_input.is_automated,
            aggregates_results=task_input.aggregates_results,
            due_date=task_input.due_date,
            processing_queue=task_input.processing_queue,
            is_decision=task_input.is_decision,
            autotrigger=task_input.autotrigger
        )

        # handling inputs
        if task_input.input_type:
            if task_input.input_type in [KnownInputOutputTypes.FILE, KnownInputOutputTypes.EMAIL_MESSAGE]:
                if task_input.input_file:
                    file_create_info = FileCreate(file=task_input.input_file)
                    if not (file := await ServiceLayer.create_file(session, file_create_info)):
                        raise ValueError("Could not save file!")
                    task_input.input = get_file_location(file.id, file.name)

            task.input = await ServiceLayer.prepare_input_output_object(session, task_input.input_type,
                                                                        task_input.input, task_input.input_format,
                                                                        task_input.input_description)

        # handling outputs
        ## if task is aggregator, enforce json output
        if task.aggregates_results:
            task_input.output_type = KnownInputOutputTypes.JSON

        if task_input.output_type:
            task.output = await ServiceLayer.prepare_input_output_object(session, task_input.output_type,
                                                                         task_input.output, task_input.output_format,
                                                                         task_input.output_description)

        if task_input.created_by:
            task.created_by = await ServiceLayer.get_user(session, task_input.created_by)

        relationships = []
        if getattr(task_input, "predecessors"):
            predecessors = task_input.predecessors
            predecessors_states = []
            for predecessor_input in predecessors:
                if not (predecessor := await ServiceLayer.get_task(session, predecessor_input.predecessor)):
                    raise ValueError("Unknown predecessor!")
                if predecessor_input.predecessor_condition == TaskRelationshipTypeEnum.BLOCKED_BY:
                    if not predecessor.aggregates_results:
                        raise ValueError("Predecessor is not aggregator, can't create blocking task!")
                    predecessor.output_is_joined = False
                relationships.append(TaskRelationship(source=predecessor, target=task,
                                                      relationship_type=predecessor_input.predecessor_condition))

                if len(predecessors) == 1:
                    if predecessor.state == TaskStateEnum.RESOLVED:
                        if predecessor.output and \
                                type(task.input) == type(predecessor.output):
                            if task.input.value is None:
                                try:
                                    task.input.value = predecessor.output.value
                                except:
                                    pass
                            if isinstance(type, JSONInputOutput) and task.input.json_initial_value is None:
                                task.input.json_initial_value = task.output.value

                predecessors_states.append(predecessor.state)

                if predecessor.state in [TaskStateEnum.OPEN, TaskStateEnum.RUNNING, TaskStateEnum.RESOLVED,
                                         TaskStateEnum.SKIPPED] and predecessor_input.predecessor_condition == TaskRelationshipTypeEnum.BLOCKED_BY:
                    task.state = TaskStateEnum.PENDING
                    if predecessor.state in [TaskStateEnum.RESOLVED, TaskStateEnum.SKIPPED]:
                        await ServiceLayer.reactivate_task(session, predecessor.id)
                    if predecessor.state == TaskStateEnum.OPEN:
                        task.state = TaskStateEnum.OPEN

            if all(state in [TaskStateEnum.RESOLVED, TaskStateEnum.SKIPPED] for state in predecessors_states):
                if not all(state in [TaskStateEnum.SKIPPED] for state in predecessors_states):
                    task.state = TaskStateEnum.OPEN

        else:
            task.state = TaskStateEnum.OPEN

        # input output special cases handling
        if isinstance(task.input, JSONInputOutput) and \
                isinstance(task.output, JSONInputOutput) and \
                task.input.value is not None:
            task.output.json_initial_value = task.input.value

        session.add(task)
        session.add_all(relationships)
        return task

    @staticmethod
    async def _create_template(session: AsyncSession, template_input: TemplateCreate, template_class, schema_path):
        with open(schema_path) as f:
            json_schema = json.loads(f.read())
        validate(instance=json.loads(template_input.template), schema=json_schema)

        template = template_class(name=template_input.name,
                                  description=template_input.description,
                                  template=template_input.template)
        session.add(template)
        return template

    @staticmethod
    async def create_task_template(session: AsyncSession, task_template_input: TemplateCreate) -> TaskTemplate:
        schema_path = JSON_SCHEMAS_DIR / 'task_template_schema.json'
        return await ServiceLayer._create_template(session, task_template_input, TaskTemplate, schema_path)

    @staticmethod
    async def create_workflow_template(session: AsyncSession,
                                       workflow_template_input: TemplateCreate) -> WorkflowTemplate:
        schema_path = JSON_SCHEMAS_DIR / 'workflow_template_schema.json'
        return await ServiceLayer._create_template(session, workflow_template_input, WorkflowTemplate, schema_path)

    @staticmethod
    async def create_file_from_comment_on_case(session: AsyncSession, file_info: FileCreate, case_id: strawberry.ID):
        case = (await session.execute(
            select(Case).filter_by(id=int(case_id)).options(selectinload(Case.case_file_root)))).scalar()
        if case is None:
            raise ValueError(f"Case with id {case_id} does not exist!")
        if case.case_file_root is None:
            raise ValueError(f"Fatal error: case with id {case_id} does not have case root directory!")
        return await ServiceLayer.create_file_from_comment(session, file_info, case.case_file_root.id)

    @staticmethod
    async def create_file_from_comment_on_task(session: AsyncSession, file_info: FileCreate, task_id: strawberry.ID):
        task = (await session.execute(
            select(Task).filter_by(id=int(task_id)).options(
                selectinload(Task.stage).
                selectinload(Stage.workflow).
                selectinload(Workflow.case).
                selectinload(Case.case_file_root)
            ))).scalar()
        if task is None:
            raise ValueError(f"Task with id {task_id} does not exist!")
        if (root_dir := task.stage.workflow.case.case_file_root) is None:
            raise ValueError(
                f"Fatal error: case with id {task.stage.workflow.case.id} does not have case root directory!")
        return await ServiceLayer.create_file_from_comment(session, file_info, root_dir.id)

    @staticmethod
    async def create_file_from_comment(session: AsyncSession, file_info: FileCreate, case_root_dir: strawberry.ID):
        root_dir = await ServiceLayer.get_directory(session, case_root_dir)
        if not root_dir:
            raise ValueError("Strange error: case does not have case root directory!")
        comment_dir_query = select(Directory).filter_by(parent_id=int(root_dir.id), name='.comment_files')
        if (comments_dir := (await session.execute(comment_dir_query)).scalars().first()) is None:
            comments_dir = Directory(parent=root_dir, name=".comment_files")
            session.add(comments_dir)
            await session.flush()
        file_info.parent = comments_dir.id
        return await ServiceLayer.create_file(session, file_info, requires_parent=True)

    @staticmethod
    async def create_file(session: AsyncSession, file_info: FileCreate, requires_parent=False):
        # check content
        if isinstance(file_info.file, UploadFile):
            content = await file_info.file.read()
            file_info.name = file_info.file.filename
        else:
            content = str.encode(file_info.file)

        if not file_info.display_name:
            index = file_info.name.find(".")
            if index > 0:
                file_info.display_name = file_info.name[0:file_info.name.find(".")]
            else:
                file_info.display_name = file_info.name[0:index]

        if not file_info.extension:
            index = file_info.name.find(".") + 1
            if index > 0:
                file_info.extension = file_info.name[file_info.name.find(".") + 1:]
            else:
                file_info.extension = "txt"

        if not file_info.name:
            file_info.name = f"{file_info.display_name}.{file_info.extension}"

        # prepare file skeleton
        case_file = File(
            name=file_info.name,
            display_name=file_info.display_name,
            extension=file_info.extension
        )

        directory = await ServiceLayer.get_directory(session, file_info.parent)
        if not directory and requires_parent:
            raise ValueError("Parent not found while required!")
        if directory:
            case_file.parent = directory

        session.add(case_file)

        # to get ID
        await session.flush()
        await session.refresh(case_file)

        file_handler = FileHandler.get_file_handler()
        file_handler.store_file(case_file.name, content, case_file.id)

        return case_file

    @staticmethod
    async def create_file_ref(session: AsyncSession, ref_info: FileRefCreate):
        directory = await ServiceLayer.get_directory(session, ref_info.parent)
        if not directory:
            raise ValueError("Parent does not exist!")

        case_file_ref = File(
            name=ref_info.source,
            display_name=ref_info.display_name,
            parent=directory
        )
        session.add(case_file_ref)
        return case_file_ref

    @staticmethod
    async def _generate_workflow_template(session: AsyncSession, workflow_id: strawberry.ID) -> TemplateCreate:
        generator = TemplateGenerator()
        options = [
            selectinload(Workflow.stages).selectinload(Stage.tasks)
            .selectinload(Task.source_relationships).selectinload(TaskRelationship.target),
            selectinload(Workflow.stages).selectinload(Stage.tasks).selectinload(Task.input),
            selectinload(Workflow.stages).selectinload(Stage.tasks).selectinload(Task.output),
        ]
        expr = select(Workflow).filter_by(id=int(workflow_id))
        for opt in options:
            expr = expr.options(opt)
        workflow = (await session.execute(expr)).scalar()
        template = generator.parse_workflow(workflow)

        return TemplateCreate(name=workflow.name,
                              description=workflow.description,
                              template=template.to_json())

    @staticmethod
    async def create_template_from_workflow(session: AsyncSession, workflow_id: strawberry.ID):
        workflow_template = await ServiceLayer._generate_workflow_template(session, workflow_id)
        return await ServiceLayer.create_workflow_template(session, workflow_template)

    @staticmethod
    async def _generate_task_template(session: AsyncSession, task_id: strawberry.ID) -> TemplateCreate:
        generator = TemplateGenerator()
        task = await ServiceLayer.get_task(session, task_id)
        template = generator.parse_task(task)

        return TemplateCreate(name=task.name,
                              description=task.description,
                              template=template.to_json())

    @staticmethod
    async def create_template_from_task(session: AsyncSession, task_id: strawberry.ID):
        task_template = await ServiceLayer._generate_task_template(session, task_id)
        return await ServiceLayer.create_task_template(session, task_template)

    @staticmethod
    async def update_blocked_task_output(session: AsyncSession, task_id: strawberry.ID):
        options = [selectinload(Task.output),
                   selectinload(Task.source_relationships).selectinload(TaskRelationship.target)
                   .selectinload(Task.output)]
        expr = select(Task).filter_by(id=int(task_id))
        for opt in options:
            expr = expr.options(opt)
        task = (await session.execute(expr)).scalar()
        output = None
        if task.aggregates_results is False:
            raise ValueError("This task does not aggregate results from blocking tasks!")
        # aggregates result must be json for aggregation to work
        if isinstance(task.output, JSONInputOutput):
            output = []
            for relationship in task.source_relationships:
                if relationship.relationship_type == TaskRelationshipTypeEnum.BLOCKED_BY:
                    if relationship.target.state in [TaskStateEnum.PENDING, TaskStateEnum.RUNNING,
                                                     TaskStateEnum.OPEN]:
                        raise ValueError("Not all blocking tasks finished!")
                    if (relationship.target.is_decision is False or (
                            relationship.target.is_decision is True and relationship.target.decision is True)):
                        if isinstance(relationship.target.output, JSONInputOutput):
                            output.append(json.loads(relationship.target.output.value))
                        else:
                            output.append({"value": str(relationship.target.output.stringified_value)})

        task.output.value = output
        task.output_is_joined = True
        return True
