from dataclasses import dataclass
from typing import List, Optional, Tuple, Dict

from dataclasses_json import dataclass_json

from csirtmu.case_database.enums import TaskRelationshipTypeEnum, TaskStateEnum
from csirtmu.case_database.models import Workflow, Stage, TaskRelationship
from csirtmu.case_database.services.task_generator_service import TaskGenerator, TaskTemplateBase


@dataclass_json
@dataclass
class RelationshipRecord:
    source_id: int
    target_ids: List[int]


@dataclass_json
@dataclass
class TaskTemplate(TaskTemplateBase):
    id: int = 0
    on_completion: Optional[List[int]] = None
    on_false: Optional[List[int]] = None
    on_true: Optional[List[int]] = None
    blocked_by: Optional[List[int]] = None


@dataclass_json
@dataclass
class StageTemplate:
    name: str
    description: str
    tasks: List[TaskTemplate]


@dataclass_json
@dataclass
class WorkflowTemplate:
    name: str
    description: str
    stages: List[StageTemplate]


class WorkflowGenerator:
    def __init__(self):
        self.cache = {}
        self.relationships: Dict[str, List[RelationshipRecord]] = {"on_completion": [], "on_true": [], "on_false": [],
                                                                   "blocked_by": []}

    def __evaluate_relationships(self):
        relationships = []
        for rel_type, rel_list in self.relationships.items():
            for rel_record in rel_list:
                source = self.cache[rel_record.source_id]
                for target_id in rel_record.target_ids:
                    if (target := self.cache[target_id]) is None:
                        raise ValueError(f"Invalid template. Task with ID={target_id} does not exists!")
                    target.state = TaskStateEnum.PENDING
                    relationships.append(TaskRelationship(
                        source=source,
                        target=target,
                        relationship_type=TaskRelationshipTypeEnum(rel_type)
                    )
                    )
        return relationships

    def __parse_task_template(self, task_template: TaskTemplate, case_identifier=None):
        task = TaskGenerator().parse_task_template(task_template, case_identifier=case_identifier)

        if task_template.on_completion:
            self.relationships["on_completion"].append(
                RelationshipRecord(task_template.id, task_template.on_completion))

        if task_template.on_false:
            self.relationships["on_false"].append(RelationshipRecord(task_template.id, task_template.on_false))

        if task_template.on_true:
            self.relationships["on_true"].append(RelationshipRecord(task_template.id, task_template.on_true))
            
        if task_template.blocked_by:
            self.relationships["blocked_by"].append(RelationshipRecord(task_template.id, task_template.blocked_by))

        if self.cache.get(task_template.id):
            raise ValueError(f"Invalid template. Task with the same ID={task_template.id} already exists!")
        self.cache[task_template.id] = task
        return task

    def __parse_stage_template(self, stage_template: StageTemplate, case_identifier=None):
        tasks = []
        for task_template in stage_template.tasks:
            tasks.append(self.__parse_task_template(task_template, case_identifier))
        return Stage(name=stage_template.name, description=stage_template.description, tasks=tasks,
                     case_identifier=case_identifier)

    def parse_workflow_template(self, template, case_identifier=None) -> Tuple[Workflow, List[TaskRelationship]]:
        workflow_template = WorkflowTemplate.from_json(template)
        stages = []
        for index, stage_template in enumerate(workflow_template.stages):
            stage = self.__parse_stage_template(stage_template, case_identifier)
            stages.append(stage)

        relationships = self.__evaluate_relationships()
        self.cache = {}
        self.relationships = {}

        return Workflow(
            name=workflow_template.name, description=workflow_template.description, stages=stages,
            case_identifier=case_identifier), relationships
