from dataclasses import dataclass
from typing import Optional

from gql import Client, gql
from gql.dsl import DSLSchema, DSLQuery, dsl_gql
from gql.transport.requests import RequestsHTTPTransport

from csirtmu.case_database.config import CONFIG

DEFAULT_USERNAME = 'UNKNOWN'


@dataclass
class UserIntrospection:
    user_id: str
    username: str
    first_name: str
    last_name: str
    email: str


class UserProfileIntrospectionClient:
    def __init__(self):
        self._enabled = CONFIG.user_profile_introspection.enabled
        self._url = 'https' if CONFIG.user_profile_introspection.secure else 'http' + '://' \
                    + CONFIG.user_profile_introspection.host + ':' + str(CONFIG.user_profile_introspection.port)
        self._schema = None

    def _fetch_schema(self, client: Client):
        query = gql("""query IntrospectionQuery {
                      __schema {
                          types {
                              name
                          }
                      }
                  } 
              """)
        client.execute_sync(query)
        return client.schema

    def get_user_introspection_for_current_user(self, user_token: str) -> Optional[UserIntrospection]:
        if not self._enabled:
            return None

        client = Client(
            transport=RequestsHTTPTransport(url=self._url, headers={'Authorization': f"Token {user_token}"}),
            fetch_schema_from_transport=True
        )

        if not self._schema:
            self._schema = DSLSchema(self._fetch_schema(client))

        query = DSLQuery(query=self._schema.Query.getTokenIntrospection()
                         .select(self._schema.User.id)
                         .select(self._schema.User.username)
                         .select(self._schema.User.email)
                         .select(self._schema.User.firstName)
                         .select(self._schema.User.lastName)
                         )
        result = client.execute_sync(dsl_gql(query))
        user = result.get('query')
        if user:
            return UserIntrospection(
                user_id=user.get('id'),
                username=user.get('username'),
                first_name=user.get('firstName'),
                last_name=user.get('lastName'),
                email=user.get('email'))
        return None
