import asyncio
import os
import random
import time
from datetime import datetime, timedelta
from pathlib import Path
from typing import Generator

import alembic.config
import pytest
import pytest_asyncio
from authlib.jose import jwt
from csirtmu.starlette_jwt_auth.config import CONFIG as AUTH_CONFIG
from csirtmu.strawberry_dmt.authorization import CapableUser
from csirtmu.strawberry_dmt.sqla_base import Base
from gql.dsl import DSLSchema
from graphql import build_schema
from sqlalchemy import text
from sqlalchemy.orm import close_all_sessions
from sqlalchemy_oso import SQLAlchemyOso
from starlette.testclient import TestClient

from csirtmu.case_database import app
from csirtmu.case_database.config import CONFIG
from csirtmu.case_database.models import CaseDBUser
from csirtmu.case_database.views import AppState
from tests.shared_queries import SharedQueries

BASE_DIR = Path(__file__).resolve().parent
ASSET_DIR = BASE_DIR / 'assets'


@pytest.fixture(scope="session")
def strict_oso():
    oso = SQLAlchemyOso(Base)
    oso.register_class(Base)
    oso.register_class(CapableUser, name='User')
    oso.register_class(CaseDBUser, name='CustomUser')
    oso.load_files([(BASE_DIR / 'config/authorization-test.polar').resolve().__str__()])
    return oso


@pytest_asyncio.fixture(autouse=True, scope='function')
async def migrate_tables():
    try:
        with CONFIG.database.engine.connect() as conn:
            conn.execute(text('DROP SCHEMA IF EXISTS "public" CASCADE;'))
            conn.commit()
    except Exception:
        pass
    try:
        cwd = os.getcwd()
        os.chdir(str(BASE_DIR.parent))
        print("Creating tables")
        with CONFIG.database.engine.connect() as conn:
            conn.execute(text('CREATE SCHEMA "public";'))
            conn.commit()
        alembic.config.main(['upgrade', 'head'])
        os.chdir(cwd)
    finally:
        yield
        close_all_sessions()
        with CONFIG.database.engine.connect() as conn:
            conn.execute(text('DROP SCHEMA IF EXISTS "public" CASCADE;'))
            conn.commit()
            conn.execute(text('CREATE SCHEMA "public";'))
            conn.commit()


@pytest_asyncio.fixture(scope='function', autouse=True)
async def cleanup_engines():
    yield
    await CONFIG.database.async_engine.dispose()
    CONFIG.database.engine.dispose()


@pytest_asyncio.fixture
async def client(strict_oso):
    with TestClient(app.create_app(strict_oso, AppState())) as c:
        yield c


@pytest_asyncio.fixture
async def schema() -> DSLSchema:
    with open(ASSET_DIR / 'schema/schema.graphql') as f:
        schema = f.read()
    return DSLSchema(build_schema(schema))


@pytest_asyncio.fixture
async def token():
    return jwt.encode(header={'alg': 'HS512'},
                      payload={
                          'iss': AUTH_CONFIG.gateway.token_issuer,
                          'aud': AUTH_CONFIG.gateway.token_audience,
                          'exp': datetime.now() + timedelta(days=120),
                          'iat': datetime.now(),
                          'jti': random.randint(0, 10000),
                          'scope': 'eduperson_entitlement',
                          'username': 'Pepa',
                          'capabilities': ["urn:geant:orion.lan:res:all:role:member#identity.orion.lan"]
                      },
                      key=AUTH_CONFIG.gateway.token_key).decode("utf-8")


@pytest_asyncio.fixture(scope='function')
async def case_workflow_stage_task_fixture(client, schema, token):
    start = time.time()
    result = {'case': {}}
    create_case_query = SharedQueries.create_case(schema)
    response = client.post('/graphql', json=dict(query=create_case_query,
                                                 variables=dict(name="Case1",
                                                                description="Case1",
                                                                type="INCIDENT",
                                                                priority="MEDIUM",
                                                                )
                                                 ), headers={'Authorization': f'Token {token}'})
    case = response.json()
    result['case']['id'] = case['data']['query']['id']
    result['case']['workflows'] = []

    for i in range(3):
        create_workflow_query = SharedQueries.create_workflow(schema)
        response = client.post('/graphql', json=dict(query=create_workflow_query,
                                                     variables=dict(
                                                         name=f'Workflow{i} for case {result["case"]["id"]}',
                                                         description=f'Workflow{i} for case {result["case"]["id"]}',
                                                         case=result["case"]["id"]
                                                     )), headers={'Authorization': f'Token {token}'})
        workflow = response.json()
        result['case']['workflows'].append({'id': workflow['data']['query']['id']})

    for workflow in result['case']['workflows']:
        workflow['stages'] = []
        for i in range(2):
            create_stage_query = SharedQueries.create_stage(schema)
            response = client.post('/graphql', json=dict(query=create_stage_query,
                                                         variables=dict(
                                                             name=f'Stage - {workflow["id"]}',
                                                             description=f'Stage - {workflow["id"]}',
                                                             workflow=workflow["id"]
                                                         )),
                                   headers={'Authorization': f'Token {token}'})
            data = response.json()
            workflow['stages'].append({'id': data['data']['query']['id'], 'tasks': []})

        for stage in workflow['stages']:
            create_task_query = SharedQueries.create_task(schema)
            create_child_task_query = SharedQueries.create_task_child(schema)
            for i in range(3):
                response = client.post('/graphql', json=dict(query=create_task_query,
                                                             variables=dict(
                                                                 name=f'Task - {stage["id"]}',
                                                                 description=f'Task - {stage["id"]}',
                                                                 stage=stage["id"]
                                                             )),
                                       headers={'Authorization': f'Token {token}'})
                data = response.json()
                stage['tasks'].append({'id': data['data']['query']['id'], 'children': []})

            for task in stage['tasks']:
                response = client.post('/graphql', json=dict(query=create_child_task_query,
                                                             variables=dict(
                                                                 name=f'Child task of {task["id"]} - {stage["id"]}',
                                                                 description=f'Child task of {task["id"]} - {stage["id"]}',
                                                                 stage=stage["id"],
                                                                 predecessors=[dict(predecessor=task['id'])]
                                                             )),
                                       headers={'Authorization': f'Token {token}'})
                data = response.json()
                task['children'].append(data['data']['query']['id'])

    print(f'case_workflow_stage_task_fixture took {time.time() - start}')
    return result


@pytest.fixture(scope="session", autouse=True)
def event_loop() -> Generator:
    loop = asyncio.get_event_loop_policy().new_event_loop()
    asyncio.set_event_loop(loop)
    yield loop
    loop.close()
