import json

from gql.dsl import DSLVariableDefinitions, DSLQuery, dsl_gql, DSLMutation
from graphql import print_ast


class SharedQueries:
    @staticmethod
    def create_case(schema):
        var = DSLVariableDefinitions()
        query = DSLMutation(query=schema.Mutation.createCase.args(data=dict(
            name=var.name,
            description=var.description,
            type=var.type,
            priority=var.priority,
            workflowTemplates=var.templates
        )).select(
            schema.MutationResult.id
        ))
        query.variable_definitions = var
        return print_ast(dsl_gql(query))

    @staticmethod
    def create_case_bulk(schema):
        var = DSLVariableDefinitions()
        query = DSLMutation(query=schema.Mutation.createCaseBulk.args(dataList=var.data_list).select(
            schema.CreateResult.id
        ))
        query.variable_definitions = var
        return print_ast(dsl_gql(query))

    @staticmethod
    def create_workflow(schema):
        var = DSLVariableDefinitions()
        query = DSLMutation(query=schema.Mutation.createWorkflow.args(data=dict(
            description=var.description,
            name=var.name,
            case=var.case
        )).select(
            schema.CreateResult.id
        ))
        query.variable_definitions = var
        return print_ast(dsl_gql(query))

    @staticmethod
    def upload_workflow(schema):
        var = DSLVariableDefinitions()
        query = DSLMutation(query=schema.Mutation.uploadWorkflowTemplate.args(file=var.file).select(
            schema.MutationResult.id
        )
        )
        query.variable_definitions = var
        upload_workflow_template = print_ast(dsl_gql(query))

        map = json.dumps({"file": ["variables.file"]})
        operations = json.dumps({
            "query": upload_workflow_template,
            "variables": {"file": None}
        })

        return dict(operations=operations, map=map)

    @staticmethod
    def upload_task(schema):
        var = DSLVariableDefinitions()
        query = DSLMutation(query=schema.Mutation.uploadTaskTemplate.args(file=var.file).select(
            schema.MutationResult.id
        )
        )
        query.variable_definitions = var
        upload_task_template = print_ast(dsl_gql(query))

        map = json.dumps({"file": ["variables.file"]})
        operations = json.dumps({
            "query": upload_task_template,
            "variables": {"file": None}
        })

        return dict(operations=operations, map=map)

    @staticmethod
    def create_stage(schema):
        var = DSLVariableDefinitions()
        query = DSLMutation(query=schema.Mutation.createStage.args(data=dict(
            name=var.name,
            description=var.description,
            workflow=var.workflow
        )).select(
            schema.CreateResult.id
        ))
        query.variable_definitions = var
        return print_ast(dsl_gql(query))

    @staticmethod
    def create_task(schema):
        var = DSLVariableDefinitions()
        query = DSLMutation(query=schema.Mutation.createTaskParentless.args(task=dict(
            description=var.description,
            name=var.name,
            stage=var.stage,
            inputType=var.input_type,
            inputFormat=var.input_format,
            input=var.input,
            outputType=var.output_type,
            output=var.output,
            aggregatesResults=var.aggregate_result,
            isAutomated=var.is_automated,
            autotrigger=var.autotrigger,
            processingQueue=var.processing_queue,
            isDecision=var.is_decision
        )).select(
            schema.MutationResult.id
        ))
        query.variable_definitions = var
        return print_ast(dsl_gql(query))

    @staticmethod
    def create_task_child(schema):
        var = DSLVariableDefinitions()
        query = DSLMutation(query=schema.Mutation.createTask.args(task=dict(
            predecessors=var.predecessors,
            description=var.description,
            name=var.name,
            stage=var.stage,
            inputType=var.input_type,
            inputFormat=var.input_format,
            input=var.input,
            outputType=var.output_type,
            output=var.output,
            aggregatesResults=var.aggregate_result,
            isAutomated=var.is_automated,
            autotrigger=var.autotrigger,
            processingQueue=var.processing_queue,
            isDecision=var.is_decision
        )).select(
            schema.MutationResult.id
        ))
        query.variable_definitions = var
        return print_ast(dsl_gql(query))

    @staticmethod
    def create_task_bulk(schema):
        var = DSLVariableDefinitions()
        query = DSLMutation(query=schema.Mutation.createTaskBulk.args(dataList=var.data_list).select(
            schema.MutationResult.id
        ))
        query.variable_definitions = var
        return print_ast(dsl_gql(query))

    @staticmethod
    def create_task_parentless_bulk(schema):
        var = DSLVariableDefinitions()
        query = DSLMutation(query=schema.Mutation.createTaskParentlessBulk.args(dataList=var.data_list).select(
            schema.MutationResult.id
        ))
        query.variable_definitions = var
        return print_ast(dsl_gql(query))

    @staticmethod
    def get_cases(schema):
        query = DSLQuery(query=schema.Query.getCaseList.select(
            schema.CasePaginated.data.select(schema.Case.id,
                                             schema.Case.workflows.select(
                                                 schema.WorkflowNestedList.data.select(
                                                     schema.Workflow.id)))))
        return print_ast(dsl_gql(query))

    @staticmethod
    def resolve_task(schema):
        var = DSLVariableDefinitions()
        query = DSLMutation(
            mutation=schema.Mutation.resolveTask.args(taskId=var.task_id).select(
                schema.MutationResult.id
            )
        )
        query.variable_definitions = var
        return print_ast(dsl_gql(query))

    @staticmethod
    def resolve_decision_task(schema):
        var = DSLVariableDefinitions()
        query = DSLMutation(
            mutation=schema.Mutation.resolveTaskWithDecision.args(taskId=var.task_id, decision=var.decision).select(
                schema.MutationResult.id
            )
        )
        query.variable_definitions = var
        return print_ast(dsl_gql(query))

    @staticmethod
    def get_stage_progress(schema):
        var = DSLVariableDefinitions()
        query = DSLQuery(
            query=schema.Query.getStage.args(id=var.stage_id).select(
                schema.Stage.id,
                schema.Stage.progress,
                schema.Stage.state,
                schema.Stage.tasks.select(
                    schema.TaskNestedList.data.select(
                        schema.Task.id,
                        schema.Task.name,
                        schema.Task.state
                    )
                )
            )
        )
        query.variable_definitions = var

        return print_ast(dsl_gql(query))

    @staticmethod
    def get_workflow_progress(schema):
        var = DSLVariableDefinitions()
        query = DSLQuery(
            query=schema.Query.getWorkflow.args(id=var.workflow_id).select(
                schema.Workflow.id,
                schema.Workflow.state,
                schema.Workflow.progress,
                schema.Workflow.stages.select(
                    schema.StageNestedList.data.select(
                        schema.Stage.id,
                        schema.Stage.progress,
                        schema.Stage.state,
                        schema.Stage.tasks.select(
                            schema.TaskNestedList.data.select(
                                schema.Task.id,
                                schema.Task.name,
                                schema.Task.state
                            )
                        )
                    )
                )
            )
        )
        query.variable_definitions = var

        return print_ast(dsl_gql(query))

    @staticmethod
    def get_task(schema):
        var = DSLVariableDefinitions()
        query = DSLQuery(query=schema.Query.getTask.args(id=var.id).select(
            schema.Task.id,
            schema.Task.state,
            schema.Task.sourceRelationships.select(
                schema.TaskRelationshipNestedList.data.select(
                    schema.TaskRelationship.target.select(
                        schema.Task.id,
                        schema.Task.state
                    )
                )
            ),
            schema.Task.targetRelationships.select(
                schema.TaskRelationshipNestedList.data.select(
                    schema.TaskRelationship.source.select(
                        schema.Task.id,
                        schema.Task.state
                    )
                )
            )
        ))
        query.variable_definitions = var

        return print_ast(dsl_gql(query))

    @staticmethod
    def create_task_relationship(schema):
        var = DSLVariableDefinitions()
        query = DSLMutation(query=schema.Mutation.createTaskRelationship.args(data=var.data).select(
            schema.CreateResult.id
        ))
        query.variable_definitions = var
        return print_ast(dsl_gql(query))

    @staticmethod
    def get_workflow_template(schema):
        var = DSLVariableDefinitions()
        query = DSLQuery(query=schema.Query.getWorkflowTemplate.args(id=var.id).select(
            schema.WorkflowTemplate.template))
        query.variable_definitions = var
        return print_ast(dsl_gql(query))
