import pytest as pytest
from gql.dsl import DSLMutation, DSLQuery, dsl_gql
from graphql import print_ast

from csirtmu.case_database.enums import CasePriorityEnum, CaseStateEnum, CaseTypeEnum, WorkflowStateEnum
from tests.conftest import ASSET_DIR
from tests.shared_queries import SharedQueries


class TestCase:
    @pytest.mark.asyncio
    async def test_simple_query(self, client, token, schema):
        query = print_ast(dsl_gql(DSLQuery(query=schema.Query.getCaseList.select(
            schema.CasePaginated.totalCount,
            schema.CasePaginated.data.select(
                schema.Case.id,
                schema.Case.name
            )
        ))))
        response = client.post('/graphql', json=dict(query=query),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200

        result = response.json()
        assert result['data']['query']['totalCount'] == 0
        assert len(result['data']['query']['data']) == 0

    @pytest.mark.asyncio
    async def test_create_case(self, client, token, schema):
        create_case = SharedQueries.create_case(schema)

        response = client.post('/graphql', json=dict(query=create_case,
                                                     variables=dict(name="Case1",
                                                                    description="Case1",
                                                                    type="INCIDENT",
                                                                    priority="MEDIUM")
                                                     ), headers={'Authorization': f'Token {token}'})
        assert response.status_code == 200
        result = response.json()
        assert "errors" not in result
        case_id = result['data']["query"]["id"]

        get_case = print_ast(dsl_gql(DSLQuery(query=schema.Query.getCase.args(id=case_id).select(
            schema.Case.name,
            schema.Case.description,
            schema.Case.state,
            schema.Case.type,
            schema.Case.priority,
            schema.Case.caseFileRoot.select(
                schema.Directory.id
            )
        ))))

        response = client.post('/graphql', json=dict(query=get_case),
                               headers={'Authorization': f'Token {token}'})
        assert response.status_code == 200
        result = response.json()
        assert "errors" not in result
        result = result['data']['query']
        assert result['name'] == 'Case1'
        assert result['description'] == 'Case1'
        assert result['type'].lower() == CaseTypeEnum.INCIDENT.value.lower()
        assert result['priority'].lower() == CasePriorityEnum.MEDIUM.value.lower()
        assert result['state'].lower() == CaseStateEnum.OPEN.value.lower()
        assert result['caseFileRoot'] is not None and str(result['caseFileRoot']['id']).isnumeric()

    @pytest.mark.asyncio
    async def test_create_case_bulk(self, client, token, schema):
        create_cases = SharedQueries.create_case_bulk(schema)

        response = client.post('/graphql', json=dict(query=create_cases,
                                                     variables=dict(data_list=[
                                                         dict(name="CaseFirst", description="C1", type="INCIDENT",
                                                              priority="MEDIUM"),
                                                         dict(name="CaseSecond", description="C2", type="ALERT",
                                                              priority="HIGH")])),
                               headers={'Authorization': f'Token {token}'})
        assert response.status_code == 200
        result = response.json()
        assert "errors" not in result

        id_list = response.json()['data']['query']
        ids = set(result.get("id") for result in id_list)
        c_root = None
        for case_id in ids:
            get_case = print_ast(dsl_gql(DSLQuery(query=schema.Query.getCase.args(id=case_id).select(
                schema.Case.name,
                schema.Case.description,
                schema.Case.state,
                schema.Case.type,
                schema.Case.priority,
                schema.Case.caseFileRoot.select(
                    schema.Directory.id
                )
            ))))

            response = client.post('/graphql', json=dict(query=get_case),
                                   headers={'Authorization': f'Token {token}'})
            assert response.status_code == 200
            result = response.json()
            assert "errors" not in result
            result = result['data']['query']
            assert result['name'] in ['CaseFirst', 'CaseSecond']
            if result['name'] == 'CaseFirst':
                d = 'C1'
                c_type = CaseTypeEnum.INCIDENT.value.lower()
                c_priority = CasePriorityEnum.MEDIUM.value.lower()
            else:
                d = 'C2'
                c_type = CaseTypeEnum.ALERT.value.lower()
                c_priority = CasePriorityEnum.HIGH.value.lower()
            assert result['description'] == d
            assert result['type'].lower() == c_type
            assert result['priority'].lower() == c_priority
            assert result['caseFileRoot'] is not None and str(result['caseFileRoot']['id']).isnumeric()
            if c_root is None:
                c_root = int(result['caseFileRoot']['id'])
            else:
                assert c_root != int(result['caseFileRoot']['id'])

    @pytest.mark.asyncio
    async def test_create_case_with_template(self, client, token, schema):
        with open(ASSET_DIR / 'templates/generic_template.json') as f:
            template = f.read()

        workflow_template = print_ast(dsl_gql(DSLMutation(
            query=schema.Mutation.createWorkflowTemplate.args(data=dict(name="Generic", description="Generic template",
                                                                        template=template)).select(
                schema.MutationResult.id
            )
        )))

        response = client.post('/graphql', json=dict(query=workflow_template),
                               headers=dict(Authorization=f'Token {token}'))

        workflow_template_id = response.json()['data']['query']['id']
        workflow_template_input = {"id": workflow_template_id}

        create_case = SharedQueries.create_case(schema)

        response = client.post('/graphql', json=dict(query=create_case,
                                                     variables=dict(name="Case1",
                                                                    description="Case1",
                                                                    type="INCIDENT",
                                                                    priority="MEDIUM",
                                                                    templates=[workflow_template_input])
                                                     ), headers={'Authorization': f'Token {token}'})
        assert response.status_code == 200
        result = response.json()
        assert "errors" not in result
        assert 'id' in result['data']["query"]

        get_case = print_ast(dsl_gql(DSLQuery(query=schema.Query.getCase.args(id=result['data']["query"]['id']).select(
            schema.Case.workflows.select(
                schema.WorkflowNestedList.totalCount,
                schema.WorkflowNestedList.data.select(
                    schema.Workflow.stages.select(
                        schema.StageNestedList.totalCount,
                        schema.StageNestedList.data.select(
                            schema.Stage.tasks.select(
                                schema.TaskNestedList.totalCount
                            )
                        )
                    )
                )
            )
        ))))

        response = client.post('/graphql', json=dict(query=get_case),
                               headers={'Authorization': f'Token {token}'})
        assert response.status_code == 200
        result = response.json()
        assert "errors" not in result
        result = result['data']['query']
        assert result['workflows']['totalCount'] == 1
        assert result['workflows']['data'][0]['stages']['totalCount'] == 2

    @pytest.mark.asyncio
    async def test_create_case_bulk_with_template(self, client, token, schema):
        with open(ASSET_DIR / 'templates/generic_template.json') as f:
            template = f.read()

        workflow_template = print_ast(dsl_gql(DSLMutation(
            query=schema.Mutation.createWorkflowTemplate.args(data=dict(name="Generic", description="Generic template",
                                                                        template=template)).select(
                schema.MutationResult.id
            )
        )))

        response = client.post('/graphql', json=dict(query=workflow_template),
                               headers=dict(Authorization=f'Token {token}'))

        workflow_template_id = response.json()['data']['query']['id']
        workflow_template_input = {"id": workflow_template_id}

        create_cases = SharedQueries.create_case_bulk(schema)

        response = client.post('/graphql', json=dict(query=create_cases,
                                                     variables=dict(data_list=[
                                                         dict(name="CaseFirst", description="C1", type="INCIDENT",
                                                              priority="MEDIUM",
                                                              workflowTemplates=[workflow_template_input]),
                                                         dict(name="CaseSecond", description="C2", type="ALERT",
                                                              priority="HIGH",
                                                              workflowTemplates=[workflow_template_input])]))
                               , headers={'Authorization': f'Token {token}'})
        assert response.status_code == 200
        result = response.json()
        assert "errors" not in result
        id_list = response.json()['data']['query']
        ids = set(result.get("id") for result in id_list)
        c_root = None

        for case_id in ids:
            get_case = print_ast(dsl_gql(DSLQuery(query=schema.Query.getCase.args(id=case_id).select(
                schema.Case.name,
                schema.Case.description,
                schema.Case.state,
                schema.Case.type,
                schema.Case.priority,
                schema.Case.caseFileRoot.select(
                    schema.Directory.id
                ),
                schema.Case.workflows.select(
                    schema.WorkflowNestedList.totalCount,
                    schema.WorkflowNestedList.data.select(
                        schema.Workflow.stages.select(
                            schema.StageNestedList.totalCount,
                            schema.StageNestedList.data.select(
                                schema.Stage.tasks.select(
                                    schema.TaskNestedList.totalCount
                                )
                            )
                        )
                    )
                )
            ))))

            response = client.post('/graphql', json=dict(query=get_case),
                                   headers={'Authorization': f'Token {token}'})
            assert response.status_code == 200
            result = response.json()
            assert "errors" not in result
            result = result['data']['query']
            assert result['name'] in ['CaseFirst', 'CaseSecond']
            if result['name'] == 'CaseFirst':
                d = 'C1'
                c_type = CaseTypeEnum.INCIDENT.value.lower()
                c_priority = CasePriorityEnum.MEDIUM.value.lower()
            else:
                d = 'C2'
                c_type = CaseTypeEnum.ALERT.value.lower()
                c_priority = CasePriorityEnum.HIGH.value.lower()
            assert result['description'] == d
            assert result['type'].lower() == c_type
            assert result['priority'].lower() == c_priority
            assert result['caseFileRoot'] is not None and str(result['caseFileRoot']['id']).isnumeric()
            if c_root is None:
                c_root = int(result['caseFileRoot']['id'])
            else:
                assert c_root != int(result['caseFileRoot']['id'])
            assert result['workflows']['totalCount'] == 1
            assert result['workflows']['data'][0]['stages']['totalCount'] == 2

    @pytest.mark.asyncio
    async def test_case_reject(self, client, case_workflow_stage_task_fixture, token, schema):
        get_case = print_ast(
            dsl_gql(
                DSLQuery(query=schema.Query.getCase.args(id=case_workflow_stage_task_fixture['case']['id']).select(
                    schema.Case.state,
                    schema.Case.workflows.select(
                        schema.WorkflowNestedList.data.select(
                            schema.Workflow.state,
                        )
                    )
                ))))

        response = client.post('/graphql', json=dict(query=get_case), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == CaseStateEnum.OPEN.value.lower()
        assert result['workflows']['data'][0]['state'].lower() == WorkflowStateEnum.IN_PROGRESS.value.lower()

        reject_case = print_ast(
            dsl_gql(
                DSLMutation(query=schema.Mutation.rejectCase.args(caseId=case_workflow_stage_task_fixture['case']['id'])
                    .select(
                    schema.MutationResult.id
                ))))

        response = client.post('/graphql', json=dict(query=reject_case), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200

        response = client.post('/graphql', json=dict(query=get_case), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == CaseStateEnum.REJECTED.value.lower()
        assert result['workflows']['data'][0]['state'].lower() == WorkflowStateEnum.REJECTED.value.lower()

    @pytest.mark.asyncio
    async def test_case_not_finished_resolve(self, client, case_workflow_stage_task_fixture, token, schema):
        get_case = print_ast(
            dsl_gql(
                DSLQuery(query=schema.Query.getCase.args(id=case_workflow_stage_task_fixture['case']['id']).select(
                    schema.Case.state,
                    schema.Case.workflows.select(
                        schema.WorkflowNestedList.data.select(
                            schema.Workflow.state,
                        )
                    )
                ))))

        resolve_case = print_ast(dsl_gql(DSLMutation(query=schema.Mutation.resolve_case.args(caseId=1).select(
            schema.MutationResult.id
        ))))

        response = client.post('/graphql', json=dict(query=get_case), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == CaseStateEnum.OPEN.value.lower()
        assert result['workflows']['data'][0]['state'].lower() == WorkflowStateEnum.IN_PROGRESS.value.lower()

        response = client.post('/graphql', json=dict(query=resolve_case), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert 'errors' in response.json()
        assert 'Not all workflows are completed!' in response.json()['errors'][0]['message']

        response = client.post('/graphql', json=dict(query=get_case), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == CaseStateEnum.OPEN.value.lower()
        assert result['workflows']['data'][0]['state'].lower() == WorkflowStateEnum.IN_PROGRESS.value.lower()

    @pytest.mark.asyncio
    async def test_case_resolve(self, client, case_workflow_stage_task_fixture, token, schema):
        get_case_workflows_query = print_ast(
            dsl_gql(
                DSLQuery(query=schema.Query.getCase
                    .args(id=case_workflow_stage_task_fixture['case']['id']).select(
                    schema.Case.state,
                    schema.Case.workflows.select(
                        schema.WorkflowNestedList.data.select(
                            schema.Workflow.state,
                        )
                    )
                ))))

        resolve_task = SharedQueries.resolve_task(schema)
        response = client.post('/graphql', json=dict(query=get_case_workflows_query),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == CaseStateEnum.OPEN.value.lower()
        assert result['workflows']['data'][0]['state'].lower() == WorkflowStateEnum.IN_PROGRESS.value.lower()

        # DEBUG

        for workflow in case_workflow_stage_task_fixture['case']['workflows']:
            for stage in workflow['stages']:
                for task in stage['tasks']:
                    client.post('/graphql', json=dict(query=resolve_task, variables=dict(task_id=task['id'])),
                                headers=dict(Authorization=f'Token {token}'))
                    for child_task_id in task['children']:
                        client.post('/graphql', json=dict(query=resolve_task, variables=dict(task_id=child_task_id)),
                                    headers=dict(Authorization=f'Token {token}'))
        resolve_case = print_ast(
            dsl_gql(
                DSLMutation(query=schema.Mutation.resolve_case
                            .args(caseId=case_workflow_stage_task_fixture['case']['id'])
                            .select(schema.MutationResult.id)
                            )))

        response = client.post('/graphql', json=dict(query=resolve_case), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert "errors" not in response.json()

        response = client.post('/graphql', json=dict(query=get_case_workflows_query),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == CaseStateEnum.RESOLVED.value.lower()
        assert result['workflows']['data'][0]['state'].lower() == WorkflowStateEnum.DONE.value.lower()
