import pytest as pytest
from gql.dsl import DSLMutation, dsl_gql
from graphql import print_ast

from csirtmu.case_database.enums import StageStateEnum, TaskStateEnum
from tests.shared_queries import SharedQueries


class TestStage:
    @pytest.mark.asyncio
    async def test_stage_progress(self, client, case_workflow_stage_task_fixture, token, schema):
        get_stage = SharedQueries.get_stage_progress(schema)
        resolve_task = SharedQueries.resolve_task(schema)

        stage = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]

        response = client.post('/graphql', json=dict(query=get_stage, variables={"stage_id": stage['id']}),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['progress'] == 0.0

        completed_tasks = 0
        for task in stage['tasks']:
            client.post('/graphql', json=dict(query=resolve_task, variables=dict(task_id=task['id'])),
                        headers=dict(Authorization=f'Token {token}'))
            for child_task_id in task['children']:
                client.post('/graphql',
                            json=dict(query=resolve_task, variables=dict(task_id=child_task_id)),
                            headers=dict(Authorization=f'Token {token}'))
            completed_tasks += 1
            # Get progress
            response = client.post('/graphql', json=dict(query=get_stage, variables={"stage_id": stage['id']}),
                                   headers=dict(Authorization=f'Token {token}'))
            assert response.status_code == 200
            result = response.json()['data']['query']
            expected_progress = round(completed_tasks / len(stage['tasks']) * 100, 2)
            assert round(float(result['progress']), 2) == expected_progress

    @pytest.mark.asyncio
    async def test_stage_state(self, client, case_workflow_stage_task_fixture, token, schema):
        get_stage = SharedQueries.get_stage_progress(schema)
        resolve_task = SharedQueries.resolve_task(schema)
        stage = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]

        response = client.post('/graphql', json=dict(query=get_stage, variables={"stage_id": stage['id']}),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == StageStateEnum.IN_PROGRESS.value.lower()

        # Resolve first task to put Stage into IN_PROGRESS state
        client.post('/graphql', json=dict(query=resolve_task, variables={"task_id": stage['tasks'][0]['id']}),
                    headers=dict(Authorization=f'Token {token}'))
        for child_task_id in stage['tasks'][0]['children']:
            client.post('/graphql',
                        json=dict(query=resolve_task, variables=dict(task_id=child_task_id)),
                        headers=dict(Authorization=f'Token {token}'))

        response = client.post('/graphql', json=dict(query=get_stage, variables={"stage_id": stage['id']}),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == StageStateEnum.IN_PROGRESS.value.lower()

        # Resolve all remaining tasks
        for task in stage['tasks'][1:]:
            client.post('/graphql', json=dict(query=resolve_task, variables=dict(task_id=task['id'])),
                        headers=dict(Authorization=f'Token {token}'))
            for child_task_id in task['children']:
                client.post('/graphql',
                            json=dict(query=resolve_task, variables=dict(task_id=child_task_id)),
                            headers=dict(Authorization=f'Token {token}'))

        # All tasks are resolved, Stage state should be DONE
        response = client.post('/graphql', json=dict(query=get_stage, variables={"stage_id": stage['id']}),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == StageStateEnum.DONE.value.lower()

    @pytest.mark.asyncio
    async def test_stage_reject(self, client, case_workflow_stage_task_fixture, token, schema):
        stage = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]

        get_stage = SharedQueries.get_stage_progress(schema)
        reject_stage = print_ast(dsl_gql(DSLMutation(query=schema.Mutation.rejectStage.args(stageId=stage['id']).select(
            schema.MutationResult.id
        ))))

        response = client.post('/graphql', json=dict(query=get_stage, variables={"stage_id": stage['id']}),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == StageStateEnum.IN_PROGRESS.value.lower()

        client.post('/graphql', json=dict(query=reject_stage), headers=dict(Authorization=f'Token {token}'))

        response = client.post('/graphql', json=dict(query=get_stage, variables={"stage_id": stage['id']}),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == StageStateEnum.REJECTED.value.lower()

        for task in result['tasks']['data']:
            assert task['state'].lower() == TaskStateEnum.REJECTED.value.lower()
