import json

import pytest as pytest
from deepdiff import DeepDiff
from gql.dsl import DSLMutation, DSLQuery, dsl_gql, DSLInlineFragment
from graphql import print_ast
from redis import Redis

from csirtmu.case_database.enums import TaskStateEnum
from tests.conftest import ASSET_DIR
from tests.shared_queries import SharedQueries


class TestTask:
    @pytest.mark.asyncio
    async def test_task_template_create(self, client, token, schema):
        with open(ASSET_DIR / 'templates/task_template.json') as f:
            template = f.read()

        task_template = print_ast(dsl_gql(DSLMutation(
            query=schema.Mutation.createTaskTemplate.args(data=dict(name="Task template", description="Task template",
                                                                    template=template)).select(
                schema.MutationResult.id
            )
        )))

        response = client.post('/graphql', json=dict(query=task_template),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()
        assert "errors" not in result

    @pytest.mark.asyncio
    async def test_upload_task_template(self, client, token, schema):
        data = SharedQueries.upload_task(schema)

        with open(ASSET_DIR / 'templates/task_template.json', "rb") as file:
            response = client.post('/graphql',
                                   data=data,
                                   files={"file": file},
                                   headers=dict(Authorization=f'Token {token}'))

        assert response.status_code == 200
        result = response.json()
        assert "errors" not in result

    @pytest.mark.asyncio
    async def test_upload_invalid_task_template(self, client, token, schema):
        data = SharedQueries.upload_task(schema)

        with open(ASSET_DIR / 'templates/invalid_task_template_1.json', "rb") as file:
            response = client.post('/graphql',
                                   data=data,
                                   files={"file": file},
                                   headers=dict(Authorization=f'Token {token}'))

        assert response.status_code == 200
        result = response.json()
        assert "errors" in result

    @pytest.mark.asyncio
    async def test_skip_task(self, client, case_workflow_stage_task_fixture, token, schema):
        task_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['tasks'][0]['id']
        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.id,
            schema.Task.state,
            schema.Task.sourceRelationships.select(
                schema.TaskRelationshipNestedList.data.select(
                    schema.TaskRelationship.target.select(
                        schema.Task.id,
                        schema.Task.state
                    )
                )
            )
        ))))

        skip_task = print_ast(dsl_gql(query=DSLMutation(schema.Mutation.skipTask(taskId=task_id).select(
            schema.MutationResult.id
        ))))

        response = client.post('/graphql',
                               json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == TaskStateEnum.OPEN.value.lower()

        response = client.post('/graphql', json=dict(query=skip_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200

        response = client.post('/graphql',
                               json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == TaskStateEnum.SKIPPED.value.lower()

        for r in result['sourceRelationships']['data']:
            assert r['target']['state'].lower() == TaskStateEnum.SKIPPED.value.lower()

    @pytest.mark.asyncio
    async def test_reject_task(self, client, case_workflow_stage_task_fixture, token, schema):
        task_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['tasks'][0]['id']
        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.id,
            schema.Task.state,
            schema.Task.sourceRelationships.select(
                schema.TaskRelationshipNestedList.data.select(
                    schema.TaskRelationship.target.select(
                        schema.Task.id,
                        schema.Task.state
                    )
                )
            )
        ))))

        reject_task = print_ast(dsl_gql(DSLMutation(query=schema.Mutation.rejectTask(taskId=task_id).select(
            schema.MutationResult.id
        ))))

        response = client.post('/graphql',
                               json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == TaskStateEnum.OPEN.value.lower()

        response = client.post('/graphql', json=dict(query=reject_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200

        response = client.post('/graphql',
                               json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == TaskStateEnum.REJECTED.value.lower()

        for r in result['sourceRelationships']['data']:
            assert r['target']['state'].lower() == TaskStateEnum.REJECTED.value.lower()

    @pytest.mark.asyncio
    async def test_set_input(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="Generic task",
                                                                    input_type="URL", stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert str(response.json()['data']['query']['id']).isnumeric()

        task_id = response.json()['data']['query']['id']

        set_input = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.setTaskInput(id=task_id, input="https://google.com").select(
                schema.MutationResult.id
            ))))

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.input.select(
                schema.TaskInputOutput.stringifiedValue
            )
        ))))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert response.json()['data']['query']['input']['stringifiedValue'] is None

        response = client.post('/graphql', json=dict(query=set_input), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert 'errors' not in response.json()

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert response.json()['data']['query']['input']['stringifiedValue'] == 'https://google.com'

    @pytest.mark.asyncio
    async def test_set_output(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="Generic task",
                                                                    output_type="URL", stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert str(response.json()['data']['query']['id']).isnumeric()

        task_id = response.json()['data']['query']['id']

        set_output = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.setTaskOutput(id=task_id, output="https://google.com").select(
                schema.MutationResult.id
            ))))

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.output.select(
                schema.TaskInputOutput.stringifiedValue
            )
        ))))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert response.json()['data']['query']['output']['stringifiedValue'] is None

        response = client.post('/graphql', json=dict(query=set_output), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert 'errors' not in response.json()

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert response.json()['data']['query']['output']['stringifiedValue'] == 'https://google.com'

    @pytest.mark.asyncio
    async def test_set_invalid_output(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="Generic task",
                                                                    output_type="URL", stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert str(response.json()['data']['query']['id']).isnumeric()

        task_id = response.json()['data']['query']['id']
        set_output = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.setTaskOutput(id=task_id, output="invalid").select(
                schema.MutationResult.id
            ))))

        response = client.post('/graphql', json=dict(query=set_output), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert 'errors' in response.json()

    @pytest.mark.asyncio
    async def test_set_invalid_input(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="Generic task",
                                                                    input_type="URL", stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert str(response.json()['data']['query']['id']).isnumeric()

        task_id = response.json()['data']['query']['id']
        set_input = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.setTaskInput(id=task_id, input="invalid").select(
                schema.MutationResult.id
            ))))

        response = client.post('/graphql', json=dict(query=set_input), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert 'errors' in response.json()

    @pytest.mark.asyncio
    async def test_create_task_input(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="Generic task",
                                                                    stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert str(response.json()['data']['query']['id']).isnumeric()
        task_id = response.json()['data']['query']['id']

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.input.select(
                schema.TaskInputOutput.id,
                schema.TaskInputOutput.type
            )
        ))))

        create_input = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.createTaskInput(id=task_id, inputType="URL").select(
                schema.MutationResult.id
            ))))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['input'] is None

        response = client.post('/graphql', json=dict(query=create_input), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['input']['type'].lower() == 'url_input_output'
        assert str(result['input']['id']).isnumeric()

    @pytest.mark.asyncio
    async def test_create_task_output(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="Generic task",
                                                                    stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert str(response.json()['data']['query']['id']).isnumeric()
        task_id = response.json()['data']['query']['id']

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.output.select(
                schema.TaskInputOutput.id,
                schema.TaskInputOutput.type
            )
        ))))

        create_output = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.createTaskOutput(id=task_id, outputType="URL").select(
                schema.MutationResult.id
            ))))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['output'] is None

        response = client.post('/graphql', json=dict(query=create_output), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['output']['type'].lower() == 'url_input_output'
        assert str(result['output']['id']).isnumeric()

    @pytest.mark.asyncio
    async def test_create_task_input_with_pattern(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="Generic task",
                                                                    stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert str(response.json()['data']['query']['id']).isnumeric()
        task_id = response.json()['data']['query']['id']

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.input.select(
                schema.TaskInputOutput.id,
                schema.TaskInputOutput.type,
                DSLInlineFragment().on(schema.StringInputOutput).select(schema.StringInputOutput.stringRegex)
            )
        ))))

        create_input = print_ast(
            dsl_gql(DSLMutation(
                query=schema.Mutation.createTaskInput(id=task_id, inputType="STRING", inputFormat="abc").select(
                    schema.MutationResult.id
                ))))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['input'] is None

        response = client.post('/graphql', json=dict(query=create_input), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['input']['type'].lower() == 'string_input_output'
        assert result['input']['stringRegex'] == "abc"
        assert str(result['input']['id']).isnumeric()

    @pytest.mark.asyncio
    async def test_create_task_output_with_pattern(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="Generic task",
                                                                    stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert str(response.json()['data']['query']['id']).isnumeric()
        task_id = response.json()['data']['query']['id']

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.output.select(
                schema.TaskInputOutput.id,
                schema.TaskInputOutput.type,
                DSLInlineFragment().on(schema.StringInputOutput).select(schema.StringInputOutput.stringRegex)
            )
        ))))

        create_output = print_ast(
            dsl_gql(DSLMutation(
                query=schema.Mutation.createTaskOutput(id=task_id, outputType="STRING", outputFormat="abc").select(
                    schema.MutationResult.id
                ))))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['output'] is None

        response = client.post('/graphql', json=dict(query=create_output), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['output']['type'].lower() == 'string_input_output'
        assert result['output']['stringRegex'] == "abc"
        assert str(result['output']['id']).isnumeric()

    @pytest.mark.asyncio
    async def test_create_task_input_that_already_exists(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="Generic task",
                                                                    input_type="URL", stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert str(response.json()['data']['query']['id']).isnumeric()
        task_id = response.json()['data']['query']['id']

        create_input = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.createTaskInput(id=task_id, inputType="URL").select(
                schema.MutationResult.id
            ))))

        response = client.post('/graphql', json=dict(query=create_input), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert 'errors' in response.json()

    @pytest.mark.asyncio
    async def test_create_task_output_that_already_exists(self, client, case_workflow_stage_task_fixture, token,
                                                          schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="Generic task",
                                                                    output_type="URL", stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert str(response.json()['data']['query']['id']).isnumeric()

        task_id = response.json()['data']['query']['id']

        create_output = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.createTaskOutput(id=task_id, outputType="URL").select(
                schema.MutationResult.id
            ))))

        response = client.post('/graphql', json=dict(query=create_output), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert 'errors' in response.json()

    @pytest.mark.asyncio
    async def test_create_json_initial_value_on_create(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="Generic task",
                                                                    input_type="JSON",
                                                                    input="{\"key\": \"value\"}",
                                                                    output_type="JSON",
                                                                    stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert str(response.json()['data']['query']['id']).isnumeric()
        task_id = response.json()['data']['query']['id']

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.id,
            schema.Task.input.select(
                DSLInlineFragment().on(schema.JSONInputOutput).select(schema.JSONInputOutput.jsonValue)
            ),
            schema.Task.output.select(
                DSLInlineFragment().on(schema.JSONInputOutput).select(
                    schema.JSONInputOutput.jsonInitialValue)
            )
        ))))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()
        assert result['data']['query']['output']['jsonInitialValue'] == result['data']['query']['input'][
            'jsonValue']

    @pytest.mark.asyncio
    async def test_task_set_json_value_check_initial_value(self, client, case_workflow_stage_task_fixture, token,
                                                           schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="Generic task",
                                                                    input_type="JSON",
                                                                    output_type="JSON",
                                                                    stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert str(response.json()['data']['query']['id']).isnumeric()
        task_id = response.json()['data']['query']['id']

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.id,
            schema.Task.input.select(
                DSLInlineFragment().on(schema.JSONInputOutput).select(schema.JSONInputOutput.jsonValue)
            ),
            schema.Task.output.select(
                DSLInlineFragment().on(schema.JSONInputOutput).select(
                    schema.JSONInputOutput.jsonInitialValue)
            )
        ))))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()
        assert result['data']['query']['output']['jsonInitialValue'] is None
        assert result['data']['query']['input']['jsonValue'] is None

        task_input = "{\"key\": \"value\"}"

        set_task_input = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.setTaskInput.args(id=task_id, input=task_input).select(
                schema.MutationResult.id
            ))))

        client.post('/graphql', json=dict(query=set_task_input), headers=dict(Authorization=f'Token {token}'))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()
        assert result['data']['query']['input']['jsonValue'] == task_input
        assert result['data']['query']['output']['jsonInitialValue'] == result['data']['query']['input'][
            'jsonValue']

    @pytest.mark.asyncio
    async def test_task_string_field_invalid_pattern(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="GenericTask",
                                                                    input_type="STRING", input_format="a\dc",
                                                                    stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert str(response.json()['data']['query']['id']).isnumeric()
        task_id = response.json()['data']['query']['id']

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.id,
            schema.Task.input.select(
                DSLInlineFragment().on(schema.StringInputOutput).select(schema.StringInputOutput.stringValue)
            ),
        ))))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()
        assert result['data']['query']['input']['stringValue'] is None

        set_task_invalid_input = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.setTaskInput.args(id=task_id, input="abc").select(
                schema.MutationResult.id
            ))))

        response = client.post('/graphql', json=dict(query=set_task_invalid_input),
                               headers=dict(Authorization=f'Token {token}'))
        assert 'errors' in response.json()

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        result = response.json()
        assert result['data']['query']['input']['stringValue'] is None

    @pytest.mark.asyncio
    async def test_task_string_field_valid_pattern(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="GenericTask",
                                                                    input_type="STRING", input_format="a\dc",
                                                                    stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert str(response.json()['data']['query']['id']).isnumeric()
        task_id = response.json()['data']['query']['id']

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.id,
            schema.Task.input.select(
                DSLInlineFragment().on(schema.StringInputOutput).select(schema.StringInputOutput.stringValue)
            ),
        ))))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))

        result = response.json()
        assert result['data']['query']['input']['stringValue'] is None

        set_task_valid_input = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.setTaskInput.args(id=task_id, input="a3c").select(
                schema.MutationResult.id
            ))))

        client.post('/graphql', json=dict(query=set_task_valid_input), headers=dict(Authorization=f'Token {token}'))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()
        assert result['data']['query']['input']['stringValue'] == "a3c"

    @pytest.mark.asyncio
    async def test_task_json_field_invalid_schema(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="GenericTask",
                                                                    input_type="JSON",
                                                                    input_format="{\"type\": \"object\", \"properties\": {\"name\": {\"type\": \"string\"}}}",
                                                                    stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert str(response.json()['data']['query']['id']).isnumeric()
        task_id = response.json()['data']['query']['id']

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.id,
            schema.Task.input.select(
                DSLInlineFragment().on(schema.JSONInputOutput).select(schema.JSONInputOutput.jsonValue)
            ),
        ))))

        invalid_json = "{\"name\": 3}"

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()
        assert result['data']['query']['input']['jsonValue'] is None

        set_task_invalid_input = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.setTaskInput.args(id=task_id, input=invalid_json).select(
                schema.MutationResult.id
            ))))

        response = client.post('/graphql', json=dict(query=set_task_invalid_input),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert 'errors' in response.json()

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()
        assert result['data']['query']['input']['jsonValue'] is None

    @pytest.mark.asyncio
    async def test_task_json_field_valid_schema(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="GenericTask",
                                                                    input_type="JSON",
                                                                    input_format="{\"type\": \"object\", \"properties\": {\"name\": {\"type\": \"string\"}}}",
                                                                    stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        task_id = response.json()['data']['query']['id']

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.id,
            schema.Task.input.select(
                DSLInlineFragment().on(schema.JSONInputOutput).select(schema.JSONInputOutput.jsonValue)
            ),
        ))))

        invalid_json = "{\"name\": \"michal\"}"

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        result = response.json()
        assert result['data']['query']['input']['jsonValue'] is None

        set_task_invalid_input = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.setTaskInput.args(id=task_id, input=invalid_json).select(
                schema.MutationResult.id
            ))))

        response = client.post('/graphql', json=dict(query=set_task_invalid_input),
                               headers=dict(Authorization=f'Token {token}'))
        assert 'errors' not in response.json()

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        result = response.json()
        assert result['data']['query']['input']['jsonValue'] == "{\"name\": \"michal\"}"

    @pytest.mark.asyncio
    async def test_resolve_without_output_task(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="Generic task",
                                                                    output_type="URL", stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        task_id = response.json()['data']['query']['id']

        resolve_task = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.resolveTask(taskId=task_id).select(
                schema.MutationResult.id
            ))))

        response = client.post('/graphql', json=dict(query=resolve_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert 'errors' in response.json()

    @pytest.mark.asyncio
    async def test_resolve_valid_task(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="Generic task",
                                                                    output_type="URL", output="https://google.com",
                                                                    stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        task_id = response.json()['data']['query']['id']

        resolve_task = print_ast(dsl_gql(DSLMutation(query=schema.Mutation.resolveTask(taskId=task_id).select(
            schema.MutationResult.id
        ))))

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.output.select(
                schema.TaskInputOutput.stringifiedValue
            ),
            schema.Task.state
        ))))

        response = client.post('/graphql', json=dict(query=resolve_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert 'errors' not in response.json()

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert response.json()['data']['query']['output']['stringifiedValue'] == 'https://google.com'
        assert response.json()['data']['query']['state'].lower() == TaskStateEnum.RESOLVED.value.lower()

    @pytest.mark.asyncio
    async def test_resolve_decision_task_on_true(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        create_child = SharedQueries.create_task_child(schema)

        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="Generic task",
                                                                    stage=stage_id, output_type="BOOL",
                                                                    is_decision=True)),
                               headers=dict(Authorization=f'Token {token}'))
        task_id = response.json()['data']['query']['id']

        response = client.post('/graphql', json=dict(query=create_child,
                                                     variables=dict(name="True 1", description="Child on true task",
                                                                    predecessors=[dict(predecessor=task_id,
                                                                                       predecessorCondition="ON_TRUE")])),
                               headers=dict(Authorization=f'Token {token}'))
        true_child_id = response.json()['data']['query']['id']

        response = client.post('/graphql', json=dict(query=create_child,
                                                     variables=dict(name="False 1", description="Child on false task",
                                                                    predecessors=[dict(predecessor=task_id,
                                                                                       predecessorCondition="ON_FALSE")])),
                               headers=dict(Authorization=f'Token {token}'))
        false_child_id = response.json()['data']['query']['id']

        set_task_output = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.setTaskOutput.args(id=task_id, output="True").select(
                schema.MutationResult.id
            ))))

        resolve_task = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.resolveTaskWithDecision(taskId=task_id, decision=True).select(
                schema.MutationResult.id
            ))))

        client.post('/graphql', json=dict(query=set_task_output), headers=dict(Authorization=f'Token {token}'))
        client.post('/graphql', json=dict(query=resolve_task), headers=dict(Authorization=f'Token {token}'))

        # parent
        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.state
        ))))
        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.json()['data']['query']['state'].lower() == TaskStateEnum.RESOLVED.value.lower()

        # on true
        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=true_child_id).select(
            schema.Task.state
        ))))
        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.json()['data']['query']['state'].lower() == TaskStateEnum.OPEN.value.lower()

        # on false
        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=false_child_id).select(
            schema.Task.state
        ))))
        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.json()['data']['query']['state'].lower() == TaskStateEnum.SKIPPED.value.lower()

    @pytest.mark.asyncio
    async def test_resolve_decision_task_on_false(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        create_child = SharedQueries.create_task_child(schema)

        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="Generic task",
                                                                    stage=stage_id, is_decision=True)),
                               headers=dict(Authorization=f'Token {token}'))
        task_id = response.json()['data']['query']['id']

        response = client.post('/graphql', json=dict(query=create_child,
                                                     variables=dict(name="True 1", description="Child on true task",
                                                                    predecessors=[dict(predecessor=task_id,
                                                                                       predecessorCondition="ON_TRUE")])),
                               headers=dict(Authorization=f'Token {token}'))
        true_child_id = response.json()['data']['query']['id']

        response = client.post('/graphql', json=dict(query=create_child,
                                                     variables=dict(name="False 1", description="Child on false task",
                                                                    predecessors=[dict(predecessor=task_id,
                                                                                       predecessorCondition="ON_FALSE")])),
                               headers=dict(Authorization=f'Token {token}'))
        false_child_id = response.json()['data']['query']['id']

        resolve_task = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.resolveTaskWithDecision(taskId=task_id, decision=False).select(
                schema.MutationResult.id
            ))))

        response = client.post('/graphql', json=dict(query=resolve_task), headers=dict(Authorization=f'Token {token}'))

        # parent
        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.state
        ))))
        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.json()['data']['query']['state'].lower() == TaskStateEnum.RESOLVED.value.lower()

        # on true
        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=true_child_id).select(
            schema.Task.state
        ))))
        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.json()['data']['query']['state'].lower() == TaskStateEnum.SKIPPED.value.lower()

        # on false
        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=false_child_id).select(
            schema.Task.state
        ))))
        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.json()['data']['query']['state'].lower() == TaskStateEnum.OPEN.value.lower()

    @pytest.mark.asyncio
    async def test_resolve_task_copy_output_to_input(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        create_child = SharedQueries.create_task_child(schema)
        io_type = "STRING"
        value = "test string"

        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="T1",
                                                                    output_type=io_type, output=value, stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        parent_id = response.json()['data']['query']['id']

        response = client.post('/graphql', json=dict(query=create_child,
                                                     variables=dict(name="Task 2", description="T2",
                                                                    input_type=io_type,
                                                                    predecessors=[dict(predecessor=parent_id)])),
                               headers=dict(Authorization=f'Token {token}'))
        child_id = response.json()['data']['query']['id']

        resolve_task = print_ast(dsl_gql(DSLMutation(query=schema.Mutation.resolveTask(taskId=parent_id).select(
            schema.MutationResult.id
        ))))

        client.post('/graphql', json=dict(query=resolve_task), headers=dict(Authorization=f'Token {token}'))

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=child_id).select(
            schema.Task.state,
            schema.Task.input.select(
                schema.TaskInputOutput.stringifiedValue
            ),
        ))))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))

        assert response.json()['data']['query']['input']['stringifiedValue'] == value

    @pytest.mark.asyncio
    async def test_resolve_task_copy_output_to_input_with_pattern(self, client, case_workflow_stage_task_fixture, token,
                                                                  schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        create_child = SharedQueries.create_task_child(schema)
        io_type = "STRING"
        value = "test string"

        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="T1",
                                                                    output_type=io_type, output=value,
                                                                    output_format=".*", stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        parent_id = response.json()['data']['query']['id']

        response = client.post('/graphql', json=dict(query=create_child,
                                                     variables=dict(name="Task 2", description="T2",
                                                                    input_type=io_type,
                                                                    predecessors=[dict(predecessor=parent_id)],
                                                                    input_format="test.*")),
                               headers=dict(Authorization=f'Token {token}'))
        child_id = response.json()['data']['query']['id']

        resolve_task = print_ast(dsl_gql(DSLMutation(query=schema.Mutation.resolveTask(taskId=parent_id).select(
            schema.MutationResult.id
        ))))

        client.post('/graphql', json=dict(query=resolve_task), headers=dict(Authorization=f'Token {token}'))

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=child_id).select(
            schema.Task.state,
            schema.Task.input.select(
                schema.TaskInputOutput.stringifiedValue
            ),
        ))))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))

        assert response.json()['data']['query']['input']['stringifiedValue'] == value

    @pytest.mark.asyncio
    async def test_resolve_task_do_not_copy_output_to_input_due_to_pattern(self, client,
                                                                           case_workflow_stage_task_fixture, token,
                                                                           schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        create_child = SharedQueries.create_task_child(schema)
        io_type = "STRING"
        value = "test string"

        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="T1",
                                                                    output_type=io_type, output=value, stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        parent_id = response.json()['data']['query']['id']

        response = client.post('/graphql', json=dict(query=create_child,
                                                     variables=dict(name="Task 2", description="T2",
                                                                    input_type=io_type,
                                                                    predecessors=[dict(predecessor=parent_id)],
                                                                    input_format="abc")),
                               headers=dict(Authorization=f'Token {token}'))
        child_id = response.json()['data']['query']['id']

        resolve_task = print_ast(dsl_gql(DSLMutation(query=schema.Mutation.resolveTask(taskId=parent_id).select(
            schema.MutationResult.id
        ))))

        client.post('/graphql', json=dict(query=resolve_task), headers=dict(Authorization=f'Token {token}'))

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=child_id).select(
            schema.Task.state,
            schema.Task.input.select(
                schema.TaskInputOutput.stringifiedValue
            ),
        ))))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))

        assert response.json()['data']['query']['input']['stringifiedValue'] is None

    @pytest.mark.asyncio
    async def test_resolve_task_do_not_copy_output_to_input_cause_already_exists(self, client,
                                                                                 case_workflow_stage_task_fixture,
                                                                                 token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        create_child = SharedQueries.create_task_child(schema)
        value = "test string"

        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="T1",
                                                                    output_type="STRING", output=value,
                                                                    stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        parent_id = response.json()['data']['query']['id']

        response = client.post('/graphql', json=dict(query=create_child,
                                                     variables=dict(name="Task 2", description="T2",
                                                                    input_type="STRING",
                                                                    predecessors=[dict(predecessor=parent_id)],
                                                                    input="Coolio")),
                               headers=dict(Authorization=f'Token {token}'))
        child_id = response.json()['data']['query']['id']

        resolve_task = print_ast(dsl_gql(DSLMutation(query=schema.Mutation.resolveTask(taskId=parent_id).select(
            schema.MutationResult.id
        ))))

        client.post('/graphql', json=dict(query=resolve_task), headers=dict(Authorization=f'Token {token}'))

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=child_id).select(
            schema.Task.state,
            schema.Task.input.select(
                schema.TaskInputOutput.stringifiedValue
            ),
        ))))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))

        assert response.json()['data']['query']['input']['stringifiedValue'] == "Coolio"

    @pytest.mark.asyncio
    async def test_resolve_task_do_not_copy_output_to_input_due_to_mismatch(self, client,
                                                                            case_workflow_stage_task_fixture, token,
                                                                            schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        create_child = SharedQueries.create_task_child(schema)
        value = "test string"

        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="T1",
                                                                    output_type="STRING", output=value,
                                                                    stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        parent_id = response.json()['data']['query']['id']

        response = client.post('/graphql', json=dict(query=create_child,
                                                     variables=dict(name="Task 2", description="T2",
                                                                    input_type="JSON",
                                                                    predecessors=[dict(predecessor=parent_id)])),
                               headers=dict(Authorization=f'Token {token}'))
        child_id = response.json()['data']['query']['id']

        resolve_task = print_ast(dsl_gql(DSLMutation(query=schema.Mutation.resolveTask(taskId=parent_id).select(
            schema.MutationResult.id
        ))))

        client.post('/graphql', json=dict(query=resolve_task), headers=dict(Authorization=f'Token {token}'))

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=child_id).select(
            schema.Task.state,
            schema.Task.input.select(
                schema.TaskInputOutput.stringifiedValue
            ),
        ))))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))

        assert response.json()['data']['query']['input']['stringifiedValue'] is None

    @pytest.mark.asyncio
    async def test_create_task_bulk(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        create_generic_childs = SharedQueries.create_task_bulk(schema)
        create_decision_childs = SharedQueries.create_task_bulk(schema)

        io_type = "STRING"
        value = "test string"

        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1", description="T1",
                                                                    output_type=io_type, output=value, stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        parent_id = response.json()['data']['query']['id']

        response = client.post('/graphql', json=dict(query=create_generic_childs,
                                                     variables=dict(data_list=[
                                                         dict(task=dict(name="Task 1.1", description="T1.1",
                                                                        inputType=io_type,
                                                                        predecessors=[dict(predecessor=parent_id)])),
                                                         dict(task=dict(name="Task 1.2", description="T1.2",
                                                                        inputType=io_type,
                                                                        predecessors=[dict(predecessor=parent_id)]))])),
                               headers=dict(Authorization=f'Token {token}'))

        child_id_list = response.json()['data']['query']
        child_ids = set(result.get("id") for result in child_id_list)

        response = client.post('/graphql', json=dict(query=create_decision_childs,
                                                     variables=dict(data_list=[
                                                         dict(task=dict(isDecision=True, name="Task 1.3",
                                                                        description="T1.3",
                                                                        inputType=io_type,
                                                                        predecessors=[dict(predecessor=parent_id)])),
                                                         dict(task=dict(isDecision=True, name="Task 1.4",
                                                                        description="T1.4",
                                                                        inputType=io_type,
                                                                        predecessors=[dict(predecessor=parent_id)]))])),
                               headers=dict(Authorization=f'Token {token}'))

        child_id_list = response.json()['data']['query']
        child_ids.update(set(result.get("id") for result in child_id_list))

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=parent_id).select(
            schema.Task.sourceRelationships.select(
                schema.TaskRelationshipNestedList.data.select(
                    schema.TaskRelationship.target.select(
                        schema.Task.id
                    )
                )
            )
        ))))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']['sourceRelationships']['data']
        target_ids = set(int(r.get("target").get('id')) for r in result)

        assert child_ids == target_ids

    @pytest.mark.asyncio
    async def test_create_task_parentless_bulk(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_generic_parentless = SharedQueries.create_task_parentless_bulk(schema)
        create_decision_parentless = SharedQueries.create_task_parentless_bulk(schema)

        io_type = "STRING"
        value = "test string"

        response = client.post('/graphql', json=dict(query=create_generic_parentless,
                                                     variables=dict(data_list=[
                                                         dict(task=dict(name="Task 1", description="T1",
                                                                        inputType=io_type, stage=stage_id)),
                                                         dict(task=dict(name="Task 2", description="T2",
                                                                        inputType=io_type, stage=stage_id))])),
                               headers=dict(Authorization=f'Token {token}'))

        id_list = response.json()['data']['query']
        ids = set(result.get("id") for result in id_list)

        response = client.post('/graphql', json=dict(query=create_decision_parentless,
                                                     variables=dict(data_list=[
                                                         dict(task=dict(isDecision=True,name="Task 3", description="T3",
                                                                        inputType=io_type, stage=stage_id)),
                                                         dict(task=dict(isDecision=True,name="Task 4", description="T4",
                                                                        inputType=io_type, stage=stage_id))])),
                               headers=dict(Authorization=f'Token {token}'))

        id_list = response.json()['data']['query']
        ids.update(set(result.get("id") for result in id_list))

        for id in ids:
            get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=id).select(
                schema.Task.sourceRelationships.select(
                    schema.TaskRelationshipNestedList.data.select(
                        schema.TaskRelationship.target.select(
                            schema.Task.id
                        )
                    )
                )
            ))))

            response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
            assert response.status_code == 200
            result = response.json()['data']['query']['sourceRelationships']['data']
            assert result == []

    @pytest.mark.asyncio
    async def test_create_task_from_template(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        with open(ASSET_DIR / 'templates/task_template.json') as f:
            template = f.read()
        json_template = json.loads(template)
        task_template = print_ast(dsl_gql(DSLMutation(
            query=schema.Mutation.createTaskTemplate.args(data=dict(name="Task template", description="Task template",
                                                                    template=template)).select(
                schema.MutationResult.id
            )
        )))
        response = client.post('/graphql', json=dict(query=task_template),
                               headers=dict(Authorization=f'Token {token}'))
        template_id = response.json()['data']['query']['id']

        create_task = print_ast(dsl_gql(DSLMutation(
            query=schema.Mutation.createTaskFromTemplate.args(taskTemplateId=template_id,
                                                              stageId=stage_id).select(
                schema.MutationResult.id
            )
        )))
        response = client.post('/graphql', json=dict(query=create_task),
                               headers={'Authorization': f'Token {token}'})
        task_id = response.json()['data']['query']['id']

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.name,
            schema.Task.isDecision,
            schema.Task.input.select(
                schema.TaskInputOutput.type
            ),
            schema.Task.output.select(
                schema.TaskInputOutput.type
            )
        ))))

        response = client.post('/graphql', json=dict(query=get_task),
                               headers={'Authorization': f'Token {token}'})
        assert response.status_code == 200
        result = response.json()
        assert "errors" not in result
        result = result['data']['query']
        assert result['name'] == json_template['name']
        assert result['isDecision'] is True
        assert result['input']['type'] == f"{json_template['input_type']}_input_output"
        assert result['output']['type'] == f"{json_template['output_type']}_input_output"

    @pytest.mark.asyncio
    async def test_create_template_from_task(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_tasks = SharedQueries.create_task_parentless_bulk(schema)
        response = client.post('/graphql', json=dict(query=create_tasks,
                                                     variables=dict(data_list=[
                                                         dict(task=dict(isDecision=True, name="Task #1",
                                                                        description="Task #1 description",
                                                                        stage=stage_id)),
                                                         dict(task=dict(isDecision=True, name="Task #2",
                                                                        description="Task #2 description",
                                                                        stage=stage_id))])
                                                     ), headers={'Authorization': f'Token {token}'})

        task_ids = [result.get("id") for result in response.json()['data']['query']]

        create_template = print_ast(dsl_gql(DSLMutation(
            query=schema.Mutation.createTemplateFromTask.args(taskId=task_ids[0]).select(
                schema.MutationResult.id
            )
        )))
        response = client.post('/graphql', json=dict(query=create_template),
                               headers={'Authorization': f'Token {token}'})

        assert response.status_code == 200
        assert "errors" not in response.json()

    @pytest.mark.asyncio
    async def test_task_blocked_by_with_aggregate_decisions(self, client, case_workflow_stage_task_fixture, token,
                                                            schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        create_child = SharedQueries.create_task_child(schema)

        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1",
                                                                    description="Decision task",
                                                                    aggregate_result=True,
                                                                    output_type='JSON',
                                                                    is_decision=True,
                                                                    stage=stage_id)),
                               headers=dict(Authorization=f'Token {token}'))
        task_id = response.json()['data']['query']['id']

        response = client.post('/graphql', json=dict(query=create_child,
                                                     variables=dict(name="T 1", description="Child 1",
                                                                    predecessors=[dict(predecessor=task_id,
                                                                                       predecessorCondition="BLOCKED_BY")],
                                                                    output_type="JSON",
                                                                    is_decision=True)),
                               headers=dict(Authorization=f'Token {token}'))
        child_1 = response.json()['data']['query']['id']

        response = client.post('/graphql', json=dict(query=create_child,
                                                     variables=dict(name="T 2", description="Child 2",
                                                                    predecessors=[dict(predecessor=task_id,
                                                                                       predecessorCondition="BLOCKED_BY")],
                                                                    output_type="JSON",
                                                                    is_decision=True)),
                               headers=dict(Authorization=f'Token {token}'))
        child_2 = response.json()['data']['query']['id']

        # parent
        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.state
        ))))
        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.json()['data']['query']['state'].lower() == TaskStateEnum.OPEN.value.lower()

        # child 1
        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=child_1).select(
            schema.Task.state
        ))))
        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.json()['data']['query']['state'].lower() == TaskStateEnum.OPEN.value.lower()

        set_output = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.setTaskOutput(id=child_1, output="{\"a\": \"1\"}").select(
                schema.MutationResult.id
            ))))

        resolve_task = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.resolveTaskWithDecision(taskId=child_1, decision=True).select(
                schema.MutationResult.id
            ))))

        response = client.post('/graphql', json=dict(query=set_output), headers=dict(Authorization=f'Token {token}'))
        response = client.post('/graphql', json=dict(query=resolve_task), headers=dict(Authorization=f'Token {token}'))

        # child 2
        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=child_2).select(
            schema.Task.state
        ))))
        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.json()['data']['query']['state'].lower() == TaskStateEnum.OPEN.value.lower()

        set_output = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.setTaskOutput(id=child_2, output="{\"b\": \"2\"}").select(
                schema.MutationResult.id
            ))))

        resolve_task = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.resolveTaskWithDecision(taskId=child_2, decision=False).select(
                schema.MutationResult.id
            ))))
        response = client.post('/graphql', json=dict(query=set_output), headers=dict(Authorization=f'Token {token}'))
        response = client.post('/graphql', json=dict(query=resolve_task), headers=dict(Authorization=f'Token {token}'))

        # check parent output

        update_output = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.updateOutputFromBlockingTasks(id=task_id).select(
                schema.MutationResult.id
            ))))

        response = client.post('/graphql', json=dict(query=update_output), headers=dict(Authorization=f'Token {token}'))

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.output.select(
                schema.TaskInputOutput.stringifiedValue
            )
        ))))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert DeepDiff(json.loads(response.json()['data']['query']['output']['stringifiedValue']), json.loads(
            json.dumps([{"a": "1"}])), ignore_order=True) == {}

    @pytest.mark.asyncio
    async def test_task_blocked_by_with_aggregate(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        create_child = SharedQueries.create_task_child(schema)

        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1",
                                                                    description="Decision task",
                                                                    aggregate_result=True,
                                                                    output_type='JSON',
                                                                    stage=stage_id, is_decision=True)),
                               headers=dict(Authorization=f'Token {token}'))
        task_id = response.json()['data']['query']['id']

        response = client.post('/graphql', json=dict(query=create_child,
                                                     variables=dict(name="T 1", description="Child 1",
                                                                    predecessors=[dict(predecessor=task_id,
                                                                                       predecessorCondition="BLOCKED_BY", )],
                                                                    output_type="JSON")),
                               headers=dict(Authorization=f'Token {token}'))
        child_1 = response.json()['data']['query']['id']

        response = client.post('/graphql', json=dict(query=create_child,
                                                     variables=dict(name="T 2", description="Child 2",
                                                                    predecessors=[dict(predecessor=task_id,
                                                                                       predecessorCondition="BLOCKED_BY", )],
                                                                    output_type="JSON")),
                               headers=dict(Authorization=f'Token {token}'))
        child_2 = response.json()['data']['query']['id']

        # parent
        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.state
        ))))
        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.json()['data']['query']['state'].lower() == TaskStateEnum.OPEN.value.lower()

        # child 1
        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=child_1).select(
            schema.Task.state
        ))))
        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.json()['data']['query']['state'].lower() == TaskStateEnum.OPEN.value.lower()

        set_output = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.setTaskOutput(id=child_1, output="{\"a\": \"1\"}").select(
                schema.MutationResult.id
            ))))

        resolve_task = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.resolveTask(taskId=child_1).select(
                schema.MutationResult.id
            ))))

        response = client.post('/graphql', json=dict(query=set_output), headers=dict(Authorization=f'Token {token}'))
        response = client.post('/graphql', json=dict(query=resolve_task), headers=dict(Authorization=f'Token {token}'))

        # child 2
        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=child_2).select(
            schema.Task.state
        ))))
        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.json()['data']['query']['state'].lower() == TaskStateEnum.OPEN.value.lower()

        set_output = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.setTaskOutput(id=child_2, output="{\"b\": \"2\"}").select(
                schema.MutationResult.id
            ))))

        resolve_task = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.resolveTask(taskId=child_2).select(
                schema.MutationResult.id
            ))))
        response = client.post('/graphql', json=dict(query=set_output), headers=dict(Authorization=f'Token {token}'))
        response = client.post('/graphql', json=dict(query=resolve_task), headers=dict(Authorization=f'Token {token}'))

        # check parent output

        update_output = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.updateOutputFromBlockingTasks(id=task_id).select(
                schema.MutationResult.id
            ))))

        response = client.post('/graphql', json=dict(query=update_output), headers=dict(Authorization=f'Token {token}'))

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.output.select(
                schema.TaskInputOutput.stringifiedValue
            )
        ))))

        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert DeepDiff(json.loads(response.json()['data']['query']['output']['stringifiedValue']),
                        json.loads(json.dumps([{"a": "1"}, {"b": "2"}])), ignore_order=True) == {}

    @pytest.mark.asyncio
    async def test_autotrigger_task(self, client, case_workflow_stage_task_fixture, token, schema, monkeypatch):
        # FAILED
        def redis_mockreturn(*args, **kwargs):
            return 1

        monkeypatch.setattr(Redis, "rpush", redis_mockreturn)

        task_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['tasks'][0]['id']
        automated_children_ids = []

        create_child_task_query = SharedQueries.create_task_child(schema)

        # Create 2 automated child task, with True/False autotrigger flag and assert their state is PENDING
        for autotrigger in [True, False]:
            response = client.post('/graphql', json=dict(query=create_child_task_query,
                                                         variables=dict(
                                                             name=f'Child task of {task_id}',
                                                             description=f'Child task of {task_id}',
                                                             predecessors=[dict(predecessor=task_id)],
                                                             is_automated=True,
                                                             input="https://google.com",
                                                             input_type="URL",
                                                             processing_queue="processing_queue",
                                                             autotrigger=autotrigger
                                                         )),
                                   headers={'Authorization': f'Token {token}'})

            assert response.status_code == 200
            result = response.json()
            assert "errors" not in result
            child_id = result['data']['query']['id']
            automated_children_ids.append(child_id)

            get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=child_id).select(
                schema.Task.state
            ))))
            response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
            assert response.status_code == 200
            result = response.json()
            assert "errors" not in result
            assert result['data']['query']['state'].lower() == TaskStateEnum.PENDING.value.lower()

        # Resolve parent task, which should OPEN both child tasks.
        # First one should be automatically triggered which means its state should be RUNNING.
        resolve_task_query = SharedQueries.resolve_task(schema)
        response = client.post('/graphql', json=dict(query=resolve_task_query, variables=dict(task_id=task_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        assert "errors" not in response.json()

        for child_task_id in automated_children_ids:
            get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=child_task_id).select(
                schema.Task.state
            ))))
            response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
            assert response.status_code == 200
            result = response.json()
            assert "errors" not in result

            if child_task_id == automated_children_ids[0]:
                assert result['data']['query']['state'].lower() == TaskStateEnum.RUNNING.value.lower()
            elif child_task_id == automated_children_ids[1]:
                assert result['data']['query']['state'].lower() == TaskStateEnum.OPEN.value.lower()

    @pytest.mark.asyncio
    async def test_skip_multipredecessor_task(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_tasks = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['tasks']
        task_ids = [task['id'] for task in stage_tasks]
        third_task_id = task_ids[2]
        get_task = SharedQueries.get_task(schema)

        response = client.post('/graphql', json=dict(query=get_task, variables=dict(id=third_task_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        child_task_id = result['sourceRelationships']['data'][0]['target']['id']

        for task_id in task_ids:
            if task_id != third_task_id:
                create_relationships = print_ast(dsl_gql(
                    DSLMutation(
                        query=schema.Mutation.createTaskRelationship.args(data=dict(relationshipType="ON_COMPLETION",
                                                                                    source=task_id,
                                                                                    target=child_task_id)).select(
                            schema.CreateResult.id
                        ))))
                client.post('/graphql', json=dict(query=create_relationships),
                            headers=dict(Authorization=f'Token {token}'))

        for task_id in task_ids:
            response = client.post('/graphql',
                                   json=dict(query=get_task, variables=dict(id=child_task_id)),
                                   headers=dict(Authorization=f'Token {token}'))
            assert response.status_code == 200
            result = response.json()['data']['query']
            assert result['state'].lower() == TaskStateEnum.PENDING.value.lower()

            skip_task = print_ast(dsl_gql(query=DSLMutation(schema.Mutation.skipTask(taskId=task_id).select(
                schema.MutationResult.id
            ))))
            response = client.post('/graphql', json=dict(query=skip_task),
                                   headers=dict(Authorization=f'Token {token}'))
            assert response.status_code == 200

            response = client.post('/graphql',
                                   json=dict(query=get_task, variables=dict(id=task_id)),
                                   headers=dict(Authorization=f'Token {token}'))
            assert response.status_code == 200
            result = response.json()['data']['query']
            assert result['state'].lower() == TaskStateEnum.SKIPPED.value.lower()

        response = client.post('/graphql',
                               json=dict(query=get_task, variables=dict(id=child_task_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == TaskStateEnum.SKIPPED.value.lower()

    @pytest.mark.asyncio
    async def test_open_multipredecessor_task(self, client, case_workflow_stage_task_fixture, token, schema):
        stage_tasks = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['tasks']
        task_ids = [task['id'] for task in stage_tasks]
        third_task_id = task_ids[2]
        get_task = SharedQueries.get_task(schema)

        response = client.post('/graphql', json=dict(query=get_task, variables=dict(id=third_task_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        child_task_id = result['sourceRelationships']['data'][0]['target']['id']

        for task_id in task_ids:
            if task_id != third_task_id:
                create_relationships = print_ast(dsl_gql(
                    DSLMutation(
                        query=schema.Mutation.createTaskRelationship.args(data=dict(relationshipType="ON_COMPLETION",
                                                                                    source=task_id,
                                                                                    target=child_task_id)).select(
                            schema.CreateResult.id
                        ))))
                client.post('/graphql', json=dict(query=create_relationships),
                            headers=dict(Authorization=f'Token {token}'))

        for task_id in task_ids:
            response = client.post('/graphql',
                                   json=dict(query=get_task, variables=dict(id=child_task_id)),
                                   headers=dict(Authorization=f'Token {token}'))
            assert response.status_code == 200
            result = response.json()['data']['query']
            assert result['state'].lower() == TaskStateEnum.PENDING.value.lower()

            resolve_task = print_ast(dsl_gql(query=DSLMutation(schema.Mutation.resolveTask(taskId=task_id).select(
                schema.MutationResult.id
            ))))
            response = client.post('/graphql', json=dict(query=resolve_task),
                                   headers=dict(Authorization=f'Token {token}'))
            assert response.status_code == 200

            response = client.post('/graphql',
                                   json=dict(query=get_task, variables=dict(id=task_id)),
                                   headers=dict(Authorization=f'Token {token}'))
            assert response.status_code == 200
            result = response.json()['data']['query']
            assert result['state'].lower() == TaskStateEnum.RESOLVED.value.lower()

        response = client.post('/graphql',
                               json=dict(query=get_task, variables=dict(id=child_task_id)),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == TaskStateEnum.OPEN.value.lower()

    @pytest.mark.asyncio
    async def test_create_multipredecessor_task(self, client, case_workflow_stage_task_fixture, token, schema):
        create_child_task_query = SharedQueries.create_task_child(schema)
        get_task = SharedQueries.get_task(schema)
        resolve_task = SharedQueries.resolve_task(schema)

        for i in range(2):
            stage = case_workflow_stage_task_fixture['case']['workflows'][i]['stages'][0]
            stage_tasks = stage['tasks']
            task_ids = [task['id'] for task in stage_tasks]
            expected_state = None

            # Create task where predecessors have opened state
            if i == 0:
                expected_state = TaskStateEnum.PENDING.value.lower()

            # Create task where predecessors have resolved state
            if i == 1:
                expected_state = TaskStateEnum.OPEN.value.lower()

                for task_id in task_ids:
                    client.post('/graphql', json=dict(query=resolve_task, variables=dict(task_id=task_id)),
                                headers=dict(Authorization=f'Token {token}'))

            response = client.post('/graphql', json=dict(query=create_child_task_query,
                                                         variables=dict(
                                                             name=f'Child task of {task_ids}',
                                                             description=f'Child task of {task_ids}',
                                                             stage=stage["id"],
                                                             predecessors=[dict(predecessor=id) for id in task_ids]
                                                         )),
                                   headers={'Authorization': f'Token {token}'})
            assert response.status_code == 200
            data = response.json()
            multipredecessor_task_id = data['data']['query']['id']

            response = client.post('/graphql', json=dict(query=get_task, variables=dict(id=multipredecessor_task_id)),
                                   headers=dict(Authorization=f'Token {token}'))
            assert response.status_code == 200
            result = response.json()['data']['query']
            assert result['state'].lower() == expected_state

    @pytest.mark.asyncio
    async def test_create_multipredecessor_task_with_no_stage(self, client, case_workflow_stage_task_fixture, token,
                                                              schema):
        stage = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]
        stage_tasks = stage['tasks']
        task_ids = [task['id'] for task in stage_tasks]

        create_child_task_query = SharedQueries.create_task_child(schema)

        response = client.post('/graphql', json=dict(query=create_child_task_query,
                                                     variables=dict(
                                                         name=f'Child task of {task_ids}',
                                                         description=f'Child task of {task_ids}',
                                                         predecessors=[dict(predecessor=id) for id in task_ids]
                                                     )),
                               headers={'Authorization': f'Token {token}'})
        assert response.status_code == 200
        result = response.json()
        assert "errors" in result

    @pytest.mark.asyncio
    async def test_output_is_joined_flag(self, client, case_workflow_stage_task_fixture, token,
                                         schema):
        stage_id = case_workflow_stage_task_fixture['case']['workflows'][0]['stages'][0]['id']
        create_task = SharedQueries.create_task(schema)
        create_child = SharedQueries.create_task_child(schema)

        response = client.post('/graphql', json=dict(query=create_task,
                                                     variables=dict(name="Task 1",
                                                                    description="Decision task",
                                                                    aggregate_result=True,
                                                                    output_type='JSON',
                                                                    stage=stage_id, is_decision=True)),
                               headers=dict(Authorization=f'Token {token}'))
        task_id = response.json()['data']['query']['id']

        response = client.post('/graphql', json=dict(query=create_child,
                                                     variables=dict(name="T 1", description="Child 1",
                                                                    predecessors=[dict(predecessor=task_id,
                                                                                       predecessorCondition="BLOCKED_BY", )],
                                                                    output_type="JSON")),
                               headers=dict(Authorization=f'Token {token}'))
        child_1 = response.json()['data']['query']['id']

        # parent
        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.outputIsJoined
        ))))
        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.json()['data']['query']['outputIsJoined'] is False

        # child 1
        set_output = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.setTaskOutput(id=child_1, output="{\"a\": \"1\"}").select(
                schema.MutationResult.id
            ))))

        resolve_task = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.resolveTask(taskId=child_1).select(
                schema.MutationResult.id
            ))))

        response = client.post('/graphql', json=dict(query=set_output),
                               headers=dict(Authorization=f'Token {token}'))
        response = client.post('/graphql', json=dict(query=resolve_task),
                               headers=dict(Authorization=f'Token {token}'))

        # check parent output flag
        update_output = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.updateOutputFromBlockingTasks(id=task_id).select(
                schema.MutationResult.id
            ))))

        response = client.post('/graphql', json=dict(query=update_output),
                               headers=dict(Authorization=f'Token {token}'))

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.outputIsJoined
        ))))
        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.json()['data']['query']['outputIsJoined'] is True

        # add new child
        response = client.post('/graphql', json=dict(query=create_child,
                                                     variables=dict(name="T 2", description="Child 2",
                                                                    predecessors=[dict(predecessor=task_id,
                                                                                       predecessorCondition="BLOCKED_BY", )],
                                                                    output_type="JSON")),
                               headers=dict(Authorization=f'Token {token}'))
        child_2 = response.json()['data']['query']['id']

        # check parent output flag
        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.outputIsJoined
        ))))
        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.json()['data']['query']['outputIsJoined'] is False

        # child 2
        set_output = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.setTaskOutput(id=child_2, output="{\"b\": \"2\"}").select(
                schema.MutationResult.id
            ))))

        resolve_task = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.resolveTask(taskId=child_2).select(
                schema.MutationResult.id
            ))))
        response = client.post('/graphql', json=dict(query=set_output),
                               headers=dict(Authorization=f'Token {token}'))
        response = client.post('/graphql', json=dict(query=resolve_task),
                               headers=dict(Authorization=f'Token {token}'))

        # check parent output flag
        update_output = print_ast(
            dsl_gql(DSLMutation(query=schema.Mutation.updateOutputFromBlockingTasks(id=task_id).select(
                schema.MutationResult.id
            ))))

        response = client.post('/graphql', json=dict(query=update_output),
                               headers=dict(Authorization=f'Token {token}'))

        get_task = print_ast(dsl_gql(DSLQuery(query=schema.Query.getTask.args(id=task_id).select(
            schema.Task.outputIsJoined
        ))))
        response = client.post('/graphql', json=dict(query=get_task), headers=dict(Authorization=f'Token {token}'))
        assert response.json()['data']['query']['outputIsJoined'] is True
