import json

import pytest as pytest
from gql.dsl import DSLMutation, dsl_gql, DSLInlineFragment, DSLQuery
from graphql import print_ast

from csirtmu.case_database.enums import StageStateEnum, TaskStateEnum, WorkflowStateEnum, TaskRelationshipTypeEnum
from tests.conftest import ASSET_DIR
from tests.shared_queries import SharedQueries


class TestWorkflow:
    @pytest.mark.asyncio
    async def test_workflow_template_create(self, client, token, schema):
        with open(ASSET_DIR / 'templates/generic_template.json') as f:
            template = f.read()

        workflow_template = print_ast(dsl_gql(DSLMutation(
            query=schema.Mutation.createWorkflowTemplate.args(data=dict(name="Generic", description="Generic template",
                                                                        template=template)).select(
                schema.MutationResult.id
            )
        )))

        response = client.post('/graphql', json=dict(query=workflow_template),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()
        assert "errors" not in result

    @pytest.mark.asyncio
    async def test_upload_workflow_template(self, client, token, schema):
        data = SharedQueries.upload_workflow(schema)

        with open(ASSET_DIR / 'templates/generic_template.json', "rb") as file:
            response = client.post('/graphql',
                                   data=data,
                                   files={"file": file},
                                   headers=dict(Authorization=f'Token {token}'))

        assert response.status_code == 200
        result = response.json()
        assert "errors" not in result

    @pytest.mark.asyncio
    async def test_upload_invalid_workflow_template(self, client, token, schema):
        data = SharedQueries.upload_workflow(schema)

        with open(ASSET_DIR / 'templates/invalid_workflow_template_3.json', "rb") as file:
            response = client.post('/graphql',
                                   data=data,
                                   files={"file": file},
                                   headers=dict(Authorization=f'Token {token}'))

        assert response.status_code == 200
        result = response.json()
        assert "errors" in result

    @pytest.mark.asyncio
    async def test_workflow_progress(self, client, case_workflow_stage_task_fixture, token, schema):
        workflow_query = SharedQueries.get_workflow_progress(schema)
        resolve_task = SharedQueries.resolve_task(schema)
        workflow = case_workflow_stage_task_fixture['case']['workflows'][0]
        response = client.post('/graphql', json=dict(query=workflow_query, variables=dict(workflow_id=workflow['id'])),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['progress'] == 0.0

        for task in workflow['stages'][0]['tasks']:
            response = client.post('/graphql', json=dict(query=resolve_task,
                                                         variables=dict(task_id=task['id'])),
                                   headers=dict(Authorization=f'Token {token}'))
            assert response.status_code == 200
            for child_task_id in task['children']:
                client.post('/graphql',
                            json=dict(query=resolve_task, variables=dict(task_id=child_task_id)),
                            headers=dict(Authorization=f'Token {token}'))

        response = client.post('/graphql', json=dict(query=workflow_query, variables=dict(workflow_id=workflow['id'])),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['progress'] == 50.0

    @pytest.mark.asyncio
    async def test_workflow_state(self, client, case_workflow_stage_task_fixture, token, schema):
        workflow_query = SharedQueries.get_workflow_progress(schema)
        resolve_task = SharedQueries.resolve_task(schema)
        workflow = case_workflow_stage_task_fixture['case']['workflows'][0]

        response = client.post('/graphql', json=dict(query=workflow_query, variables=dict(workflow_id=workflow['id'])),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == WorkflowStateEnum.IN_PROGRESS.value.lower()
        for stage in workflow['stages']:
            for task in stage['tasks']:
                client.post('/graphql', json=dict(query=resolve_task, variables=dict(task_id=task['id'])),
                            headers=dict(Authorization=f'Token {token}'))
                for child_task_id in task['children']:
                    client.post('/graphql',
                                json=dict(query=resolve_task, variables=dict(task_id=child_task_id)),
                                headers=dict(Authorization=f'Token {token}'))

        response = client.post('/graphql', json=dict(query=workflow_query, variables=dict(workflow_id=workflow['id'])),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == WorkflowStateEnum.DONE.value.lower()

    @pytest.mark.asyncio
    async def test_workflow_reject(self, client, case_workflow_stage_task_fixture, token, schema):
        workflow_query = SharedQueries.get_workflow_progress(schema)
        workflow = case_workflow_stage_task_fixture['case']['workflows'][0]

        response = client.post('/graphql', json=dict(query=workflow_query, variables=dict(workflow_id=workflow['id'])),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['state'].lower() == WorkflowStateEnum.IN_PROGRESS.value.lower()

        reject_workflow = print_ast(dsl_gql(DSLMutation(query=schema.Mutation.rejectWorkflow(workflowId=workflow['id'])
        .select(
            schema.MutationResult.id
        ))))

        client.post('/graphql', json=dict(query=reject_workflow), headers=dict(Authorization=f'Token {token}'))

        response = client.post('/graphql', json=dict(query=workflow_query, variables=dict(workflow_id=workflow['id'])),
                               headers=dict(Authorization=f'Token {token}'))
        assert response.status_code == 200
        result = response.json()['data']['query']
        assert result['progress'] == 100.0
        assert result['state'].lower() == WorkflowStateEnum.REJECTED.value.lower()

        for stage in result['stages']['data']:
            assert stage['state'].lower() == StageStateEnum.REJECTED.value.lower()
            assert stage['progress'] == 100.0
            for task in stage['tasks']['data']:
                assert task['state'].lower() == TaskStateEnum.REJECTED.value.lower()

    @pytest.mark.asyncio
    async def test_create_workflow_from_template(self, client, token, schema):
        with open(ASSET_DIR / 'templates/minimalistic_template.json') as f:
            template = f.read()
        json_template = json.loads(template)
        workflow_template = print_ast(dsl_gql(DSLMutation(
            query=schema.Mutation.createWorkflowTemplate.args(data=dict(name="Generic", description="Generic template",
                                                                        template=template)).select(
                schema.MutationResult.id
            )
        )))
        response = client.post('/graphql', json=dict(query=workflow_template),
                               headers=dict(Authorization=f'Token {token}'))
        template_id = response.json()['data']['query']['id']

        create_case = SharedQueries.create_case(schema)
        response = client.post('/graphql', json=dict(query=create_case,
                                                     variables=dict(name="Case1",
                                                                    description="Case1",
                                                                    type="INCIDENT",
                                                                    priority="MEDIUM")
                                                     ), headers={'Authorization': f'Token {token}'})
        case_id = response.json()['data']['query']['id']

        create_workflow = print_ast(dsl_gql(DSLMutation(
            query=schema.Mutation.createWorkflowFromTemplate.args(workflowTemplateId=template_id,
                                                                  caseId=case_id).select(
                schema.MutationResult.id
            )
        )))
        response = client.post('/graphql', json=dict(query=create_workflow),
                               headers={'Authorization': f'Token {token}'})
        workflow_id = response.json()['data']['query']['id']

        get_workflow = print_ast(dsl_gql(DSLQuery(query=schema.Query.getWorkflow.args(id=workflow_id).select(
            schema.Workflow.name,
            schema.Workflow.stages.select(
                schema.StageNestedList.totalCount,
                schema.StageNestedList.data.select(
                    schema.Stage.name,
                    schema.Stage.tasks.select(
                        schema.TaskNestedList.totalCount,
                        schema.TaskNestedList.data.select(
                            schema.Task.name,
                            schema.Task.input.select(
                                schema.TaskInputOutput.type,
                                DSLInlineFragment().on(schema.StringInputOutput).select(
                                    schema.StringInputOutput.stringRegex
                                )
                            )
                        )
                    )
                )
            )
        ))))
        response = client.post('/graphql', json=dict(query=get_workflow),
                               headers={'Authorization': f'Token {token}'})
        assert response.status_code == 200
        result = response.json()
        assert "errors" not in result
        result = result['data']['query']
        assert result['name'] == json_template['name']
        assert result['stages']['totalCount'] == 1
        stage = result['stages']['data'][0]
        assert stage['name'] == json_template['stages'][0]['name']
        assert stage['tasks']['totalCount'] == 1
        task = stage['tasks']['data'][0]
        assert task['name'] == json_template['stages'][0]['tasks'][0]['name']
        assert task['input']['type'] == f"{json_template['stages'][0]['tasks'][0]['input_type']}_input_output"
        assert task['input']['stringRegex'] == json_template['stages'][0]['tasks'][0]['input_format']

    @pytest.mark.asyncio
    async def test_create_workflow_from_invalid_template_1(self, client, token, schema):
        data = SharedQueries.upload_workflow(schema)

        with open(ASSET_DIR / 'templates/invalid_workflow_template_1.json', "rb") as file:
            response = client.post('/graphql',
                                   data=data,
                                   files={"file": file},
                                   headers=dict(Authorization=f'Token {token}'))

        assert response.status_code == 200
        template_id = response.json()['data']['query']['id']

        create_case = SharedQueries.create_case(schema)
        response = client.post('/graphql', json=dict(query=create_case,
                                                     variables=dict(name="Case1",
                                                                    description="Case1",
                                                                    type="INCIDENT",
                                                                    priority="MEDIUM")
                                                     ), headers={'Authorization': f'Token {token}'})
        case_id = response.json()['data']['query']['id']

        create_workflow = print_ast(dsl_gql(DSLMutation(
            query=schema.Mutation.createWorkflowFromTemplate.args(workflowTemplateId=template_id,
                                                                  caseId=case_id).select(
                schema.MutationResult.id
            )
        )))
        response = client.post('/graphql', json=dict(query=create_workflow),
                               headers={'Authorization': f'Token {token}'})

        assert response.status_code == 200
        assert "errors" in response.json()

    @pytest.mark.asyncio
    async def test_create_workflow_from_invalid_template_2(self, client, token, schema):
        data = SharedQueries.upload_workflow(schema)

        with open(ASSET_DIR / 'templates/invalid_workflow_template_2.json', "rb") as file:
            response = client.post('/graphql',
                                   data=data,
                                   files={"file": file},
                                   headers=dict(Authorization=f'Token {token}'))

        assert response.status_code == 200
        template_id = response.json()['data']['query']['id']

        create_case = SharedQueries.create_case(schema)
        response = client.post('/graphql', json=dict(query=create_case,
                                                     variables=dict(name="Case1",
                                                                    description="Case1",
                                                                    type="INCIDENT",
                                                                    priority="MEDIUM")
                                                     ), headers={'Authorization': f'Token {token}'})
        case_id = response.json()['data']['query']['id']

        create_workflow = print_ast(dsl_gql(DSLMutation(
            query=schema.Mutation.createWorkflowFromTemplate.args(workflowTemplateId=template_id,
                                                                  caseId=case_id).select(
                schema.MutationResult.id
            )
        )))
        response = client.post('/graphql', json=dict(query=create_workflow),
                               headers={'Authorization': f'Token {token}'})

        assert response.status_code == 200
        assert "errors" in response.json()

    @pytest.mark.asyncio
    async def test_create_template_from_workflow(self, client, token, schema):
        create_case = SharedQueries.create_case(schema)

        response = client.post('/graphql', json=dict(query=create_case,
                                                     variables=dict(name="Case1",
                                                                    description="Case1",
                                                                    type="INCIDENT",
                                                                    priority="MEDIUM")
                                                     ), headers={'Authorization': f'Token {token}'})
        result = response.json()
        case_id = result['data']["query"]["id"]
        create_workflow = SharedQueries.create_workflow(schema)
        response = client.post('/graphql', json=dict(query=create_workflow,
                                                     variables=dict(name="Workflow #1",
                                                                    description="Description of workflow...",
                                                                    case=case_id)
                                                     ), headers={'Authorization': f'Token {token}'})

        workflow_id = response.json()['data']['query']['id']

        create_stages = print_ast(dsl_gql(DSLMutation(
            query=schema.Mutation.createStageBulk.args(
                dataList=[dict(name="Stage #1", description="Stage #1 description", workflow=workflow_id),
                          dict(name="Stage #2", description="Stage #2 description", workflow=workflow_id)]).select(
                schema.CreateResult.id
            )
        )))
        response = client.post('/graphql', json=dict(query=create_stages),
                               headers={'Authorization': f'Token {token}'})

        stage_ids = [result.get("id") for result in response.json()['data']['query']]

        create_tasks = SharedQueries.create_task_parentless_bulk(schema)
        response = client.post('/graphql', json=dict(query=create_tasks,
                                                     variables=dict(data_list=[
                                                         dict(isDecision=True, name="Task #1", description="Task #1 description",
                                                              stage=stage_ids[0]),
                                                         dict(isDecision=True, name="Task #2", description="Task #2 description",
                                                              stage=stage_ids[1])])
                                                     ), headers={'Authorization': f'Token {token}'})

        create_template = print_ast(dsl_gql(DSLMutation(
            query=schema.Mutation.createTemplateFromWorkflow.args(workflowId=workflow_id).select(
                schema.MutationResult.id
            )
        )))
        response = client.post('/graphql', json=dict(query=create_template),
                               headers={'Authorization': f'Token {token}'})

        assert response.status_code == 200
        assert "errors" not in response.json()

    @pytest.mark.asyncio
    async def test_create_template_and_validate_relationships(self, client, token, schema):
        create_case = SharedQueries.create_case(schema)

        response = client.post('/graphql', json=dict(query=create_case,
                                                     variables=dict(name="Case1",
                                                                    description="Case1",
                                                                    type="INCIDENT",
                                                                    priority="MEDIUM")
                                                     ), headers={'Authorization': f'Token {token}'})
        result = response.json()
        case_id = result['data']["query"]["id"]

        create_workflow = SharedQueries.create_workflow(schema)
        response = client.post('/graphql', json=dict(query=create_workflow,
                                                     variables=dict(name="Workflow #1",
                                                                    description="Description of workflow...",
                                                                    case=case_id)
                                                     ), headers={'Authorization': f'Token {token}'})

        workflow_id = response.json()['data']['query']['id']

        create_stages = print_ast(dsl_gql(DSLMutation(
            query=schema.Mutation.createStageBulk.args(
                dataList=[dict(name="Stage #1", description="Stage #1 description", workflow=workflow_id),
                          dict(name="Stage #2", description="Stage #2 description", workflow=workflow_id)]).select(
                schema.CreateResult.id
            )
        )))
        response = client.post('/graphql', json=dict(query=create_stages),
                               headers={'Authorization': f'Token {token}'})

        stage_ids = [result.get("id") for result in response.json()['data']['query']]

        create_tasks = SharedQueries.create_task_parentless_bulk(schema)
        response = client.post('/graphql', json=dict(query=create_tasks,
                                                     variables=dict(data_list=[
                                                         dict(task=dict(isDecision=True, name="Task #1", description="Task #1 description",
                                                              stage=stage_ids[0])),
                                                         dict(task=dict(isDecision=True,name="Task #2", description="Task #2 description",
                                                              stage=stage_ids[0])),
                                                         dict(task=dict(isDecision=True,name="Task #3", description="Task #3 description",
                                                              stage=stage_ids[0])),
                                                         dict(task=dict(isDecision=True,name="Task #4", description="Task #4 description",
                                                              stage=stage_ids[1])),
                                                     ])
                                                     ), headers={'Authorization': f'Token {token}'})

        task_ids = [result.get("id") for result in response.json()['data']['query']]

        create_relationship = SharedQueries.create_task_relationship(schema)
        response = client.post('/graphql', json=dict(query=create_relationship,
                                                     variables=dict(data=dict(relationshipType="BLOCKED_BY",
                                                                    source=task_ids[1], target=task_ids[2]))
                                                     ), headers={'Authorization': f'Token {token}'})

        create_template = print_ast(dsl_gql(DSLMutation(
            query=schema.Mutation.createTemplateFromWorkflow.args(workflowId=workflow_id).select(
                schema.MutationResult.id
            )
        )))
        response = client.post('/graphql', json=dict(query=create_template),
                               headers={'Authorization': f'Token {token}'})

        assert response.status_code == 200
        assert "errors" not in response.json()

        template_id = response.json()['data']['query']['id']

        get_workflow_template = SharedQueries.get_workflow_template(schema)
        response = client.post('/graphql', json=dict(query=get_workflow_template,
                                                     variables=dict(id=template_id)
                                                     ), headers={'Authorization': f'Token {token}'})

        response_template = response.json()['data']['query']['template']
        assert f'"blocked_by": [{task_ids[2]}]' in response_template
