import os

from csirtmu.starlette_jwt_auth.models import MappedUser
from csirtmu.starlette_jwt_auth.routes import auth_backend_version_route
from csirtmu.starlette_jwt_auth.token_authentication_backend import TokenAuthenticationBackend, on_auth_error
from csirtmu.strawberry_dmt.authorization import CapableUser
from csirtmu.strawberry_dmt.custom_view import GraphQLCustomASGIApp
from csirtmu.strawberry_dmt.sqla_base import Base
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy_oso import SQLAlchemyOso
from starlette.applications import Starlette
from starlette.authentication import UnauthenticatedUser
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.middleware.cors import CORSMiddleware

from csirtmu.integration_test_service.config import BASE_DIR, CONFIG
from csirtmu.integration_test_service.schema import schema
from csirtmu.integration_test_service.utils import PayloadBouncer
from csirtmu.integration_test_service.views import create_payload_bouncer, healthcheck_liveness, healthcheck_readiness


def get_oso():
    oso = SQLAlchemyOso(Base)
    oso.register_class(Base)
    oso.register_class(CapableUser)
    oso.register_class(UnauthenticatedUser)
    oso.register_class(MappedUser, name='User')
    oso.register_class(PayloadBouncer)
    oso.load_files([os.getenv('TEST_SERVICE_OSO', (BASE_DIR / 'config/authorization.polar').resolve().__str__())])
    return oso


def create_app(engine: Engine, debug: bool = False):
    session_maker = sessionmaker(engine, future=True)
    app_ = Starlette(debug=True, routes=[auth_backend_version_route])
    app_.add_middleware(
        CORSMiddleware, allow_headers=["*"], allow_origins=["*"], allow_methods=["*"]
    )

    oso = get_oso()
    app_.add_middleware(AuthenticationMiddleware, backend=TokenAuthenticationBackend(), on_error=on_auth_error)

    graphql_app = GraphQLCustomASGIApp(
        schema, session_factory=session_maker, debug=debug, oso=oso)

    paths = ["/", "/graphql"]
    for path in paths:
        app_.add_route(path, graphql_app)
        app_.add_websocket_route(path, graphql_app)

    app_.add_route(path='/healthcheck/liveness', route=healthcheck_liveness, methods=['GET'])
    app_.add_route(path='/healthcheck/readiness', route=healthcheck_readiness, methods=['GET'])

    app_.add_route('/payload-bouncer', create_payload_bouncer(oso, authorized=False),
                   methods=['GET', 'PUT', 'POST', 'DELETE'])
    app_.add_route('/payload-bouncer-authorized', create_payload_bouncer(oso), methods=['GET', 'PUT', 'POST', 'DELETE'])
    return app_


def app():  # pragma: no cover
    Base.metadata.drop_all(CONFIG.database.engine)
    Base.metadata.create_all(CONFIG.database.engine)

    return create_app(CONFIG.database.engine, debug=True)
