import asyncio
import logging
import sys
from os import getenv
from typing import Optional, Union

import aiohttp
import requests
import uvloop
from aiohttp import ClientTimeout, DummyCookieJar
from packaging.specifiers import SpecifierSet
from packaging.version import Version
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session, sessionmaker
from starlette import status
from starlette.applications import Starlette
from starlette.middleware.cors import CORSMiddleware
from starlette.middleware.gzip import GZipMiddleware
from starlette.middleware.trustedhost import TrustedHostMiddleware
from starlette.requests import Request
from starlette.responses import Response
from starlette_prometheus import PrometheusMiddleware, metrics
from tenacity import RetryError, after_log, retry, retry_if_exception_type, stop_after_attempt, \
    wait_exponential, wait_random

from csirtmu.service_gateway.auth_backend.api_token_authentication_backend import (
    APITokenAuthenticationBackend,
)
from csirtmu.service_gateway.auth_backend.api_token_model import APITokenModel, Base
from csirtmu.service_gateway.auth_backend.auth_utils import (
    get_token_created_and_expiration_datetime,
)
from csirtmu.service_gateway.auth_backend.cached_user_model import JWTUser
from csirtmu.service_gateway.auth_backend.multi_backend_authentication_middleware import (  # noqa: E501
    MultiBackendAuthenticationMiddleware,
)
from csirtmu.service_gateway.auth_backend.oidc_authentication_backend import (
    OIDCAuthenticationBackend,
)
from csirtmu.service_gateway.config import BASE_DIR, CONFIG, GatewayEndpoint
from csirtmu.service_gateway.utils import get_version
from csirtmu.service_gateway.versions import SUPPORTED_VERSIONS
from csirtmu.service_gateway.views import aiohttp_proxy_request, create_healthcheck_liveness, \
    create_healthcheck_readiness, get_gateway_version, on_auth_error
from csirtmu.service_gateway.views_admin import (
    create_api_token,
    delete_api_token,
    get_api_token_capabilities,
    get_api_token_list,
)

logger = logging.getLogger(__package__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(logging.Formatter("%(asctime)s [%(levelname)s]: %(message)s"))
logger.addHandler(handler)


class ServiceGateway:
    debug: bool
    app: Starlette

    def __init__(self) -> None:
        self.debug = CONFIG.debug
        self._client_session: Optional[aiohttp.ClientSession] = None
        self.app = Starlette(debug=self.debug)
        self.healthy = True
        self._add_middleware_stack()
        self._add_routes()
        loop = asyncio.get_event_loop()
        self._load_task = loop.create_task(self._load_default_values_to_db())
        self.api_token_backend: APITokenAuthenticationBackend
        self.oidc_backend: OIDCAuthenticationBackend
        if not loop.is_running():
            loop.run_until_complete(asyncio.wait([self._load_task]))

    @property
    def client_session(self) -> aiohttp.ClientSession:
        if self._client_session is None:
            self._client_session = aiohttp.ClientSession(cookie_jar=DummyCookieJar(),
                                                         connector=aiohttp.TCPConnector(limit=60,
                                                                                        enable_cleanup_closed=True),
                                                         timeout=ClientTimeout(total=60))
        return self._client_session

    def __del__(self):
        if hasattr(self, '_load_task'):
            self._load_task.cancel()
        try:
            event_loop = asyncio.get_running_loop()
            event_loop.create_task(self.client_session.close())
        except RuntimeError:
            asyncio.new_event_loop().run_until_complete(self.client_session.close())

    def _add_middleware_stack(self) -> None:
        self.app.add_middleware(GZipMiddleware, minimum_size=500)
        # Auth
        session_factory = sessionmaker(CONFIG.database.engine, expire_on_commit=False, class_=AsyncSession)
        self.api_token_backend = APITokenAuthenticationBackend(config=CONFIG, session_factory=session_factory)
        self.oidc_backend = OIDCAuthenticationBackend(config=CONFIG)
        self.app.add_middleware(
            MultiBackendAuthenticationMiddleware,
            backends=[self.api_token_backend, self.oidc_backend],
            on_error=on_auth_error,
        )
        # TrustedHost and CORS
        self.app.add_middleware(
            TrustedHostMiddleware, allowed_hosts=CONFIG.allowed_hosts
        )
        if CONFIG.cors.enabled:
            self.app.add_middleware(
                CORSMiddleware,
                allow_origins=CONFIG.cors.allow_origins,
                allow_methods=CONFIG.cors.allow_methods,
                allow_headers=CONFIG.cors.allow_headers,
                allow_credentials=CONFIG.cors.allow_credentials,
                expose_headers=CONFIG.cors.expose_headers,
                max_age=CONFIG.cors.max_age,
            )
        self.app.add_middleware(PrometheusMiddleware)

    @staticmethod
    def _check_backend_version(gateway_path, _endpoint: Union[str, GatewayEndpoint]):
        endpoint = f'\'{gateway_path}\' => \'{_endpoint}\''
        try:
            version_base_path = _endpoint if isinstance(_endpoint, str) else _endpoint.version_path
            resp = requests.get(f'{version_base_path}/_auth_backend/version', timeout=3)
            if resp.status_code != status.HTTP_200_OK:
                logger.warning(f'Unable to check authentication backend version for {endpoint}, '
                               f'requesting \'{version_base_path}/_auth_backend/version\'')
            else:
                version = resp.json()
                backend = f'\'{version["kind"]}@{version["version"]}\''
                spec = SpecifierSet(SUPPORTED_VERSIONS[get_version()][version['kind']])
                if not spec.contains(Version(version['version'])):
                    logger.error(f'Backend version check failed for {endpoint}. '
                                 f'Required found {backend} but required {spec}')
                    raise RuntimeError(f'Auth backend endpoint: {endpoint} is not compatible with this version of'
                                       f' Service Gateway')
                else:
                    logger.info(f'Found compatible backend for {endpoint} - {backend}')
        except:
            logger.warning(f'Unable to contact endpoint {endpoint}')

    def _add_routes(self) -> None:
        for gateway_path, _endpoint in CONFIG.routes.items():
            self.app.add_route(
                path=gateway_path + '{rest_of_path:path}',
                route=self._get_proxy_view,
                methods=['GET', 'POST', 'PUT', 'DELETE', 'OPTIONS'],
            )
            self._check_backend_version(gateway_path, _endpoint)

        self.app.add_route(path='/_gateway/version', route=get_gateway_version, methods=['GET'])
        self.app.add_route(path='/_gateway/healthcheck/liveness', route=create_healthcheck_liveness(self),
                           methods=['GET'])
        self.app.add_route(path='/_gateway/healthcheck/readiness', route=create_healthcheck_readiness(self),
                           methods=['GET'])
        self.app.add_route(path='/_gateway/metrics', route=metrics)
        self.app.add_route(path='/_gateway/api-token', route=get_api_token_list, methods=['GET'])
        self.app.add_route(path='/_gateway/api-token', route=create_api_token, methods=['POST'])
        self.app.add_route(path='/_gateway/api-token', route=delete_api_token, methods=['DELETE'])
        self.app.add_route(path='/_gateway/api-token-capabilities', route=get_api_token_capabilities, methods=['GET'])

    async def _get_proxy_view(self, request: Request) -> Response:
        return await aiohttp_proxy_request(request, self.client_session)

    async def _load_default_values_to_db(self) -> None:
        @retry(stop=stop_after_attempt(10),
               wait=wait_exponential(multiplier=1, min=2, max=10) + wait_random(0, 2),
               retry=retry_if_exception_type(OperationalError),
               retry_error_callback=lambda _: exit(1),
               after=after_log(logger, logging.INFO))
        async def load():
            self.create_tables()
            if CONFIG.api_token_config.load_on_init:
                with Session(CONFIG.database.engine_sync) as s:
                    token = (
                        s.query(APITokenModel)
                        .filter_by(token=CONFIG.api_token_config.load_on_init)
                        .first()
                    )
                    if not token:
                        created, expires = get_token_created_and_expiration_datetime()
                        token = APITokenModel(
                            token=CONFIG.api_token_config.load_on_init,
                            capabilities=CONFIG.api_token_config.token_capabilities,
                            created_by='SERVICE_GATEWAY|INITIAL_TOKEN',
                            created=created,
                            expires=expires,
                        )
                        s.add(token)
                        s.commit()
                logger.info("Database setup completed")

        try:
            await load()
        except RetryError:
            self.healthy = False

    @staticmethod
    def create_tables() -> None:
        Base.metadata.create_all(CONFIG.database.engine_sync)

    def get_app(self) -> Starlette:
        return self.app


def app() -> Starlette:  # pragma: no cover
    # Switch Asyncio to uvloop
    uvloop.install()
    # Load Oso
    CONFIG.oso.register_class(JWTUser)
    CONFIG.oso.load_files(
        [getenv('SERVICE_GATEWAY_OSO', BASE_DIR / 'config/authorization.polar')]
    )
    setattr(JWTUser, 'api_roles', CONFIG.api_roles)  # noqa: B010

    return ServiceGateway().get_app()
