import uuid
from dataclasses import asdict
from datetime import datetime, timedelta
from typing import Optional

import aioredis
from aioredis import Redis
from authlib.jose import jwt
from orjson import orjson
from sqlalchemy import select
from sqlalchemy.orm import sessionmaker
from starlette.authentication import (
    AuthCredentials,
    AuthenticationBackend,
    AuthenticationError,
    BaseUser,
)
from starlette.requests import HTTPConnection, Request

from csirtmu.service_gateway.auth_backend.api_token_model import APITokenModel
from csirtmu.service_gateway.auth_backend.cached_user_model import JWTUser
from csirtmu.service_gateway.config import CONFIG, GatewayConfig


class APITokenAuthenticationBackend(AuthenticationBackend):
    cache: Redis

    def __init__(self, config: GatewayConfig, session_factory: sessionmaker):
        self.config = config
        self.cache = self.config.cache.client
        self.session_factory = session_factory

    async def authenticate(
            self, request: Request
    ) -> Optional[tuple[AuthCredentials, BaseUser]]:
        if not (api_token := self._extract_token(request)):
            return None

        async with self.session_factory() as s:
            result = await s.execute(
                select(APITokenModel).where(APITokenModel.token == api_token)
            )

        if not (token_record := result.scalar()):
            return None

        if token_record.expires <= datetime.utcnow():
            raise AuthenticationError('Token is expired!')

        cached_entry: Optional[bytes] = await self.cache.get(token_record.token)
        if cached_entry and cached_entry != b'{}':
            cached_token_container: JWTUser = JWTUser(**orjson.loads(cached_entry))
            return AuthCredentials(['authenticated']), cached_token_container

        expiration = datetime.now() + timedelta(
            seconds=self.config.proxy.token_cache_expiration_seconds
        )
        token_container: JWTUser = self._create_token_container(
            token_record, int(expiration.timestamp())
        )

        await self.cache.setex(
            api_token,
            self.config.proxy.token_cache_expiration_seconds,
            orjson.dumps(asdict(token_container)),
        )

        return AuthCredentials(['authenticated']), token_container

    def _extract_token(self, request: HTTPConnection) -> Optional[str]:
        if api_token := request.headers.get(self.config.api_token_config.header_name):
            prefix = self.config.api_token_config.token_prefix + ' '
            if api_token.startswith(prefix):
                return api_token[len(prefix):]

    def _create_token_container(
            self, token: APITokenModel, exp: int
    ) -> Optional[JWTUser]:
        iss, _ = token.created_by.split('|')

        token_payload = {
            'exp': exp,
            'iat': datetime.now().timestamp(),
            'iss': self.config.proxy.token_issuer,
            'aud': self.config.proxy.token_audience,
            'jti': uuid.uuid4().__str__(),
            'username': f'{self.config.proxy.token_issuer}|{CONFIG.api_token_config.token_username}',
            'scope': CONFIG.api_token_config.token_scopes,
            'capabilities': token.capabilities,
        }

        jwt_token = self._generate_jwt_token(token_payload)

        return JWTUser(
            username=token.created_by,
            jwt_token=jwt_token,
            access_token=token.token,
            issuer_name=iss,
            capabilities=token.capabilities,
        )

    def _generate_jwt_token(self, payload: dict) -> Optional[str]:
        return jwt.encode(
            {'alg': 'HS512'}, payload, key=self.config.proxy.token_key
        ).decode('utf-8')
