from functools import partial
from http.client import HTTPConnection
from typing import Callable, Optional, List

from starlette.authentication import (
    AuthCredentials,
    AuthenticationBackend,
    AuthenticationError,
)
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.responses import Response
from starlette.types import ASGIApp, Receive, Scope, Send

from csirtmu.service_gateway.auth_backend.cached_user_model import JWTUser


class MultiBackendAuthenticationMiddleware(AuthenticationMiddleware):
    def __init__(
            self,
            app: ASGIApp,
            backends: List[AuthenticationBackend],
            on_error: Callable[[HTTPConnection, AuthenticationError], Response] = None,
    ) -> None:
        if len(backends) < 1:
            raise AttributeError('At least one backend required')
        self.backends = backends
        self._app = app
        super().__init__(app, backends[0], on_error)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        super_call = super().__call__

        async def callback(
                backends: List[AuthenticationBackend],
                scope_: Scope,
                receive_: Receive,
                send_: Send,
        ) -> Optional[tuple[AuthCredentials, JWTUser]]:
            if (
                    'auth' in scope_
                    and not scope_['auth'].scopes
                    and not scope_['user'].is_authenticated
            ):
                backend = next(backends, None)
                if not backend:
                    return await self._app(scope_, receive_, send_)
                self.backend = backend
                return await super_call(scope, receive, send)
            return await self._app(scope_, receive_, send_)

        self.app = partial(callback, (b for b in self.backends))
        return await super().__call__(scope, receive, send)
