import uuid
from dataclasses import asdict
from datetime import datetime
from typing import Dict, List, Optional, Union

import aioredis
import orjson
from aioredis import Redis
from authlib.jose import JWTClaims, jwt
from authlib.jose.errors import DecodeError, ExpiredTokenError
from authlib.oidc.core import UserInfo
from csirtmu.oidc.utils.oidc_issuer import OpenIDConnectIssuer
from csirtmu.oidc.utils.utils import extract_access_token, extract_auth_header
from requests import HTTPError
from requests.exceptions import InvalidHeader
from starlette.authentication import (
    AuthCredentials,
    AuthenticationBackend,
    AuthenticationError,
)
from starlette.requests import HTTPConnection

from csirtmu.service_gateway.auth_backend.cached_user_model import JWTUser
from csirtmu.service_gateway.config import GatewayConfig
from csirtmu.service_gateway.utils import get_oidc_issuer_from_dataclass


class OIDCAuthenticationBackend(AuthenticationBackend):
    cache: Redis
    config: GatewayConfig
    # If oidc_provider header is missing in request, default to this provider
    default_issuer: Optional[OpenIDConnectIssuer]
    # Used to match provider with name specified in oidc_provider header [name, issuer]
    issuers: Dict[str, OpenIDConnectIssuer]
    cache_key_prefix_separator: str = '|'

    def __init__(self, config: GatewayConfig):
        self.config = config
        self.default_oidc_issuer_name, self.oidc_issuers \
            = get_oidc_issuer_from_dataclass(self.config.oidc_config.issuers)
        self.cache = aioredis.from_url(self.config.cache.url, db=self.config.cache.db,
                                       username=self.config.cache.username, password=self.config.cache.password)

    async def authenticate(
            self, request: HTTPConnection
    ) -> Optional[tuple[AuthCredentials, JWTUser]]:
        access_token = self._extract_token(request)

        if access_token is None:
            return None

        if issuer_name := request.headers.get(
                self.config.oidc_config.oidc_provider_header_name
        ):
            try:
                issuer = self.oidc_issuers[issuer_name]
            except KeyError:
                raise AuthenticationError(
                    'Requested issuer is not registered on the backend'
                )
        else:
            # If issuer header is missing, use default issuer
            if self.default_oidc_issuer_name is not None:
                issuer = self.oidc_issuers[self.default_oidc_issuer_name]
                issuer_name = self.default_oidc_issuer_name
            else:
                raise AuthenticationError(
                    'Missing issuer header and default is not configured'
                )

        cache_key = issuer_name + self.cache_key_prefix_separator + access_token
        cache_entry: Optional[bytes] = await self.cache.get(cache_key)
        if cache_entry and cache_entry != b'{}':
            cached_user: JWTUser = JWTUser(**orjson.loads(cache_entry))
            # User saved in cache, forward
            return AuthCredentials(['authenticated']), cached_user

        try:
            verified_access_token = await issuer.decode_and_verify_token(access_token)
        except DecodeError:
            return None
        except ValueError:
            raise AuthenticationError('Invalid access token')
        except ExpiredTokenError:
            raise AuthenticationError('Access token expired')

        try:
            userinfo = await issuer.get_userinfo(access_token)
        except HTTPError:
            raise AuthenticationError('Invalid access token')

        user: JWTUser = self._create_user(
            access_token, verified_access_token, userinfo, issuer_name=issuer_name
        )

        cache_ttl = int(verified_access_token['exp'] - datetime.now().timestamp())
        await self.cache.setex(cache_key, cache_ttl, orjson.dumps(asdict(user)))

        return AuthCredentials(['authenticated']), user

    def _extract_token(self, request: HTTPConnection) -> Optional[str]:
        auth_header = extract_auth_header(
            request, self.config.oidc_config.access_token_header_name
        )

        if auth_header is None:
            return None

        try:
            access_token = extract_access_token(
                auth_header, self.config.oidc_config.access_token_prefix
            )
            if access_token is None:
                return None
        except InvalidHeader as ex:
            raise AuthenticationError(ex)

        return access_token

    def _extract_capabilities_from_watched_claims(self, userinfo: UserInfo) -> List[Union[str, int, float, bool]]:
        watched_claims_found = set(self.config.oidc_config.watched_claims).intersection(userinfo.keys())

        capabilities = []
        for claim in watched_claims_found:
            # Claim values should be JSON (dict / list) or scalar values
            mapped_value = userinfo.get(claim)
            if isinstance(mapped_value, list):
                capabilities += mapped_value
            else:
                capabilities.append(mapped_value)
        return [orjson.dumps(cap).decode() if isinstance(cap, dict) else cap for cap in capabilities]

    def _create_user(
            self,
            raw_access: str,
            access_token: JWTClaims,
            userinfo: UserInfo,
            issuer_name: str,
    ) -> Optional[JWTUser]:
        username = access_token.get('iss') + '|' + userinfo.get('sub')

        capabilities = self._extract_capabilities_from_watched_claims(userinfo)
        token_payload = {
            'username': username,
            'scope': access_token['scope'],
            'jti': access_token.get('jti', uuid.uuid4()),
            'capabilities': capabilities,
        }
        jwt_token = self._generate_jwt_token(token_payload, access_token.get('exp', -1))
        return JWTUser(
            username=username,
            jwt_token=jwt_token,
            access_token=raw_access,
            issuer_name=issuer_name,
            capabilities=capabilities,
        )

    def _generate_jwt_token(self, payload: dict, expire_at: int) -> Optional[str]:
        metadata = {
            'exp': expire_at,
            'iat': datetime.now().timestamp(),
            'iss': self.config.proxy.token_issuer,
            'aud': self.config.proxy.token_audience,
        }
        return jwt.encode(
            {'alg': 'HS512'}, metadata | payload, key=self.config.proxy.token_key
        ).decode('utf-8')
