from dataclasses import dataclass, field
from os import getenv
from pathlib import Path
from typing import Dict, List, Optional, Union

import aioredis
from csirtmu.dataclass_config_loader import load_dataclass_config, Omit
from oso import Oso
from sqlalchemy import engine
from sqlalchemy.engine import Engine, create_engine
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy_oso import SQLAlchemyOso

from csirtmu.service_gateway.auth_backend.api_token_model import Base as ApiTokenBase

BASE_DIR = Path(__file__).parent.parent.parent


@dataclass
class OIDCIssuer:
    name: Optional[str]
    url: Optional[str]
    default_kid: Optional[str]
    default: Optional[bool] = False


@dataclass
class OIDCConfig:
    issuers: Optional[List[OIDCIssuer]]
    scopes: Optional[List[str]]
    oidc_provider_header_name: Optional[str]
    access_token_header_name: Optional[str]
    access_token_prefix: Optional[str]
    watched_claims: Optional[List[str]]


@dataclass
class Database:
    host: Optional[str]
    port: Optional[int]
    username: Optional[str]
    password: Optional[str]
    database: Optional[str]
    engine: Omit['Engine']
    engine_sync: Omit['Engine']


@dataclass
class Proxy:
    token_issuer: Optional[str]
    token_header_name: Optional[str]
    token_prefix: Optional[str]
    token_audience: Optional[str]
    token_key: Optional[str]
    token_cache_expiration_seconds: Optional[int]


@dataclass
class APIRoles:
    api_claim: Optional[str]  # Claim containing TOKEN ADMIN / SUPERADMIN capabilities
    admin_capability: Optional[str]


@dataclass
class ApiTokenConfig:
    # Specifies how token authorization header should look
    # header_name: token_prefix some_token
    # Authorization: Token ekcFi4w48h-jUk1ra696BBU9UdZ4BvmYUCW2do1BoHY
    header_name: Optional[str]
    token_prefix: Optional[str]
    # You can specify custom claim and capabilities
    # claim: List[capabilities] will be given to API tokens
    token_claim: Optional[str]
    token_capabilities: Optional[List[str]]

    expiration_seconds: Optional[int]
    # Token that will be loaded to DB when before GW starts
    load_on_init: Optional[str]
    # Length of generated tokens
    token_length: Optional[int] = 32
    # Required claims for our OIDC auth middleware
    token_username: Optional[str] = 'api-token'
    token_scopes: Optional[List[str]] = field(default_factory=lambda: ['api_token'])


@dataclass
class CORS:
    enabled: Optional[bool] = True
    allow_origins: Optional[List[str]] = field(default_factory=lambda: [])
    allow_origins_regex: Optional[List[str]] = field(default_factory=lambda: [])
    allow_methods: Optional[List[str]] = field(
        default_factory=lambda: ['GET', 'POST', 'OPTIONS', 'HEAD']
    )
    allow_headers: Optional[List[str]] = field(default_factory=lambda: [])
    expose_headers: Optional[List[str]] = field(default_factory=lambda: [])
    allow_credentials: Optional[bool] = False
    max_age: Optional[int] = 600  # Browser CORS response cache time in seconds


@dataclass
class GatewayEndpoint:
    path: Optional[str]
    version_path: Optional[str]

    def __str__(self):
        return self.path


@dataclass
class Cache:
    url: Optional[str]
    username: Optional[str]
    password: Optional[str]
    db: Optional[int]
    client: Omit[aioredis.Redis]


@dataclass
class GatewayConfig:
    database: Database
    cache: Cache
    oidc_config: OIDCConfig
    api_roles: APIRoles
    api_token_config: ApiTokenConfig
    proxy: Proxy
    allowed_hosts: Optional[List[str]]
    cors: CORS
    routes: Optional[Dict[str, Union[str, GatewayEndpoint]]] = field(default_factory=lambda: {})
    debug: Optional[bool] = False
    oso: Optional['Oso'] = field(default_factory=lambda: SQLAlchemyOso(ApiTokenBase))


def _load_config(path: Path) -> GatewayConfig:
    config = load_dataclass_config(path, GatewayConfig, 'SERVICE_GATEWAY')

    config.database.engine = create_async_engine(
        engine.URL.create(
            drivername='postgresql+asyncpg',
            username=config.database.username,
            password=config.database.password,
            host=config.database.host,
            port=config.database.port,
            database=config.database.database,
        ),
        future=True,
    )
    config.database.engine_sync = create_engine(
        engine.URL.create(
            drivername='postgresql+psycopg2',
            username=config.database.username,
            password=config.database.password,
            host=config.database.host,
            port=config.database.port,
            database=config.database.database,
        )
    )

    config.cache.client = aioredis.from_url(config.cache.url, db=config.cache.db, username=config.cache.username,
                                            password=config.cache.password)

    return config


CONFIG = _load_config(
    Path(
        getenv(
            'SERVICE_GATEWAY_CONFIG_PATH',
            (BASE_DIR / 'config/config.yaml').resolve().__str__(),
        )
    )
)
