from functools import cache
from pathlib import Path

import aiohttp
import tomli
from aiohttp import ClientConnectionError
from starlette import status
from starlette.authentication import requires
from starlette.datastructures import MutableHeaders
from starlette.requests import Request
from starlette.responses import JSONResponse, Response

from csirtmu.service_gateway.config import CONFIG


async def get_gateway_version(request: Request) -> Response:
    @cache
    def get_version():
        with open(Path(__file__).parent / "../../pyproject.toml", "rb") as f:
            pyproject = tomli.load(f)
        return pyproject['tool']['poetry']['version']

    return Response(get_version())


OK = "✌ healthy 👍"
NOK = "🪦 healthyn't 🐁"


def create_healthcheck_liveness(gateway: 'ServiceGateway'):
    async def healthcheck_liveness(request: Request) -> Response:
        return JSONResponse(OK if gateway.healthy else NOK, status_code=status.HTTP_200_OK)

    return healthcheck_liveness


def create_healthcheck_readiness(gateway: 'ServiceGateway'):
    async def healthcheck_readiness(request: Request) -> Response:
        status_data = {"gateway": OK if gateway.healthy else NOK, "database": OK, "cache": OK}
        try:
            CONFIG.database.engine_sync.connect().close()
        except Exception:
            status_data["database"] = NOK
        try:
            await CONFIG.cache.client.get('')
        except Exception:
            status_data['cache'] = NOK

        if NOK in status_data.values():
            return JSONResponse(status_data, status_code=status.HTTP_503_SERVICE_UNAVAILABLE)
        return JSONResponse(status_data, status_code=status.HTTP_200_OK)

    return healthcheck_readiness


@requires('authenticated', status_code=401)
async def aiohttp_proxy_request(
        request: Request, client_session: aiohttp.ClientSession
) -> Response:
    # Extract proxy path from request path
    path = request.url.path[:request.url.path.rfind(request.path_params['rest_of_path'])]
    url = f'{CONFIG.routes[path]}{request.path_params["rest_of_path"]}'

    # Drop the 'Host' header since we don't know it
    # Aiohttp won't follow redirects if host header is set
    mutable_headers = MutableHeaders(raw=request.headers.raw)
    mutable_headers.__delitem__('host')

    mutable_headers.__setitem__(
        'Authorization', CONFIG.proxy.token_prefix + ' ' + request.user.jwt_token
    )
    try:
        async with client_session.request(
                method=request.method.lower(),
                url=url,
                data=await request.body(),  # TODO: this raises ClientDisconnect with many users sending requests,
                                            #       try streaming it?
                params=request.query_params,
                headers=mutable_headers,
                cookies=request.cookies,
        ) as resp:
            resp_headers = resp.headers.copy()
            resp_headers.pop('Transfer-Encoding', None)
            resp_headers.pop('Content-Encoding', None)
            resp_headers.pop('Content-Length', None)
            response = Response(
                content=await resp.content.read(),
                status_code=resp.status,
                headers=dict(resp_headers),
            )

        return response
    except ClientConnectionError:  # pragma: no cover
        return Response(status_code=404)


def on_auth_error(request: Request, exc: Exception) -> JSONResponse:  # pragma: no cover
    return JSONResponse({'error': str(exc)}, status_code=401)
