from json import JSONDecodeError
from secrets import token_urlsafe

from sqlalchemy.future import select
from sqlalchemy_oso import authorized_sessionmaker
from starlette.authentication import requires
from starlette.requests import Request
from starlette.responses import JSONResponse, Response

from csirtmu.service_gateway.auth_backend.api_token_model import APITokenModel
from csirtmu.service_gateway.auth_backend.auth_utils import (
    get_token_created_and_expiration_datetime,
)
from csirtmu.service_gateway.config import CONFIG


@requires('authenticated', status_code=401)
async def get_api_token_list(request: Request) -> Response:
    session = authorized_sessionmaker(
        bind=CONFIG.database.engine_sync,
        expire_on_commit=False,
        get_oso=lambda: CONFIG.oso,
        get_user=lambda: request.user,
        get_checked_permissions=lambda: {APITokenModel: 'any'},
    )

    with session() as s:
        result = s.execute(select(APITokenModel))
        api_token_list = result.scalars()
        result = []
        for token in api_token_list:
            result.append(
                {
                    'token': token.token,
                    'capabilities': token.capabilities,
                    'created_by': token.created_by,
                    'created': token.created.isoformat(),
                    'expires': token.expires.isoformat(),
                }
            )
    return JSONResponse(result)


@requires('authenticated', status_code=401)
async def get_api_token_capabilities(request: Request) -> Response:
    return JSONResponse({'capabilities': CONFIG.api_token_config.token_capabilities})


@requires('authenticated', status_code=401)
async def create_api_token(request: Request) -> Response:
    try:
        data = await request.json()
        capabilities = data.get('capabilities')
        expiration = data.get('expiration-seconds', None)
    except (JSONDecodeError, ValueError):
        return JSONResponse(
            {'error': 'Missing JSON attributes, required: capabilities, optional: expiration!'}, status_code=400
        )

    session = authorized_sessionmaker(
        bind=CONFIG.database.engine_sync,
        expire_on_commit=False,
        get_oso=lambda: CONFIG.oso,
        get_user=lambda: request.user,
        get_checked_permissions=lambda: {APITokenModel: 'any'},
    )

    created, expires = get_token_created_and_expiration_datetime(expiration)
    with session() as s:
        token = APITokenModel(
            token=token_urlsafe(32),
            capabilities=capabilities,
            created_by=request.user.username,
            created=created,
            expires=expires,
        )
        s.add(token)
        s.commit()
    return JSONResponse({'token': token.token}, status_code=201)


@requires('authenticated', status_code=401)
async def delete_api_token(request: Request) -> Response:
    try:
        data = await request.json()
        request_token = data['token']
    except (JSONDecodeError, ValueError):
        return JSONResponse(
            {'error': 'Missing JSON attribute \"token\"!'}, status_code=400
        )

    session = authorized_sessionmaker(
        bind=CONFIG.database.engine_sync,
        expire_on_commit=False,
        get_oso=lambda: CONFIG.oso,
        get_user=lambda: request.user,
        get_checked_permissions=lambda: {APITokenModel: 'any'},
    )

    with session() as s:
        token = s.query(APITokenModel).filter_by(token=request_token).first()
        if token:
            s.delete(token)
            s.commit()
            return JSONResponse({'status': 'Token deleted!'}, status_code=200)
        return JSONResponse(
            {'error': f'Token {request_token} not found!'}, status_code=404
        )
