import dataclasses
import logging
from os import getenv
from typing import List
from urllib.parse import urlparse

import httpx
import pytest
import pytest_asyncio
import uvloop
from csirtmu.auth_keycloak_fixture_loader.config import CONFIG as KEYCLOAK_CONFIG
from csirtmu.auth_keycloak_fixture_loader.load_keycloak_fixtures import _load
from csirtmu.auth_keycloak_test_utils.test_client import KeycloakTestClient
from csirtmu.auth_keycloak_test_utils.utils import (
    get_auth_endpoint_for_realm,
    get_tokens_for_user,
)
from httpx import AsyncClient
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import sessionmaker

from csirtmu.service_gateway.app import ServiceGateway
from csirtmu.service_gateway.auth_backend.api_token_authentication_backend import (
    APITokenAuthenticationBackend,
)
from csirtmu.service_gateway.auth_backend.api_token_model import Base
from csirtmu.service_gateway.auth_backend.cached_user_model import JWTUser
from csirtmu.service_gateway.auth_backend.multi_backend_authentication_middleware import (
    MultiBackendAuthenticationMiddleware,
)
from csirtmu.service_gateway.auth_backend.oidc_authentication_backend import (
    OIDCAuthenticationBackend,
)
from csirtmu.service_gateway.config import BASE_DIR, CONFIG
from csirtmu.service_gateway.views import on_auth_error

logger = logging.getLogger(__name__)


# needs to be run before app config is loaded
def pytest_sessionstart(session):
    Base.metadata.drop_all(CONFIG.database.engine_sync)
    Base.metadata.create_all(CONFIG.database.engine_sync)
    try:
        _load()
    except Exception as ex:
        logger.warning(f'An exception was encountered when configuring keycloak, dumping {ex}')


@pytest.fixture(scope='session')
def event_loop() -> uvloop.Loop:
    loop = uvloop.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope='session')
def keycloak_client():
    keycloak_url = urlparse(CONFIG.oidc_config.issuers[0].url)
    return KeycloakTestClient(
        host=keycloak_url.hostname,
        port=keycloak_url.port,
        master_user=KEYCLOAK_CONFIG.keycloak.master_user,
        master_pass=KEYCLOAK_CONFIG.keycloak.master_password,
    )


@pytest.fixture(scope='session')
def endpoint(keycloak_client):
    return keycloak_client.endpoint


@pytest.fixture(scope='session')
def user_with_permissions():
    for user in KEYCLOAK_CONFIG.users:
        if user.authorized:
            return user


@pytest.fixture(scope='session')
def user_without_permissions():
    for user in KEYCLOAK_CONFIG.users:
        if not user.authorized:
            return user


@pytest_asyncio.fixture(scope='session')
async def gateway() -> ServiceGateway:
    gateway = ServiceGateway()
    test_app = gateway.get_app()

    api_token_backend = APITokenAuthenticationBackend(config=CONFIG,
                                                      session_factory=sessionmaker(bind=CONFIG.database.engine,
                                                                                   expire_on_commit=False,
                                                                                   class_=AsyncSession))
    oidc_backend = OIDCAuthenticationBackend(config=CONFIG)

    # Security middleware
    test_app.add_middleware(
        MultiBackendAuthenticationMiddleware,
        backends=[api_token_backend, oidc_backend],
        on_error=on_auth_error,
    )

    CONFIG.oso.register_class(JWTUser)
    CONFIG.oso.load_files(
        [getenv('SERVICE_GATEWAY_OSO', BASE_DIR / 'config/authorization.polar')]
    )
    setattr(JWTUser, 'api_roles', CONFIG.api_roles)

    yield test_app
    await gateway.client_session.close()
    await gateway.oidc_backend.cache.flushall()


# Use if you want to invalidate tokens in test function
@pytest.fixture(scope='function')
def test_scoped_tokens(endpoint, user_with_permissions):
    return get_tokens_for_user(
        realm_auth_endpoint=get_auth_endpoint_for_realm(
            endpoint=endpoint, realm=CONFIG.oidc_config.issuers[0].name
        ),
        username=user_with_permissions.username,
        password=user_with_permissions.password,
        client_id='test-client',
    )


@pytest.fixture(scope='session')
def endpoint(keycloak_client):
    return keycloak_client.endpoint


# Tuple[access_token, refresh_token]
@pytest.fixture(scope='module')
def oidc_tokens(endpoint, user_with_permissions) -> tuple[str, str]:
    return get_tokens_for_user(
        realm_auth_endpoint=get_auth_endpoint_for_realm(
            endpoint=endpoint, realm=CONFIG.oidc_config.issuers[0].name
        ),
        username=user_with_permissions.username,
        password=user_with_permissions.password,
        client_id='test-client',
    )


@pytest.fixture(scope='module')
def oidc_tokens_unauthenticated_user(
        endpoint, user_without_permissions
) -> tuple[str, str]:
    return get_tokens_for_user(
        realm_auth_endpoint=get_auth_endpoint_for_realm(
            endpoint=endpoint, realm=KEYCLOAK_CONFIG.oidc.issuers[0].name
        ),
        username=user_without_permissions.username,
        password=user_without_permissions.password,
        client_id='test-client',
    )


@pytest.fixture(scope='module')
def alternative_realm_oidc_tokens(endpoint, user_with_permissions) -> tuple[str, str]:
    return get_tokens_for_user(
        realm_auth_endpoint=get_auth_endpoint_for_realm(
            endpoint=endpoint, realm=CONFIG.oidc_config.issuers[1].name
        ),
        username=user_with_permissions.username,
        password=user_with_permissions.password,
        client_id='test-client',
    )


@pytest.fixture(scope='function')
def mock_auth_backend_config(mocker):
    mock_gateway_config = dataclasses.make_dataclass('MockGatewayConfig', {})
    mock_config = dataclasses.make_dataclass(
        'MockConfig',
        {
            'debug': bool,
            'watched_claims': List[str],
            'gateway': mock_gateway_config,
            'jwt_username_separator': str,
        },
    )
    mocker.patch('csirtmu.starlette_jwt_auth.config.CONFIG', return_value=mock_config)
    yield mock_config, mock_gateway_config


@pytest.fixture(scope='session')
def endpoint(keycloak_client):
    return keycloak_client.endpoint


# Use if you want to invalidate tokens in test function
@pytest.fixture
def test_scoped_tokens(endpoint, user_with_permissions):
    return get_tokens_for_user(
        realm_auth_endpoint=get_auth_endpoint_for_realm(
            endpoint=endpoint, realm=CONFIG.oidc_config.issuers[0].name
        ),
        username=user_with_permissions.username,
        password=user_with_permissions.password,
        client_id='test-client',
    )


@pytest_asyncio.fixture(scope='function')
async def api_token(gateway, oidc_tokens):
    headers = {
        'Authorization': 'Bearer ' + oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            0
        ].name,
    }

    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response: httpx.Response = await ac.post('/_gateway/api-token',
                                                 json={'capabilities': CONFIG.api_token_config.token_capabilities})

    data = response.json()
    return data['token']
