from datetime import datetime, timedelta

import deepdiff
import httpx
import pytest
from authlib.jose import jwt
from httpx import AsyncClient

from csirtmu.service_gateway.config import CONFIG


@pytest.mark.asyncio
async def test_create_admin_token(gateway, oidc_tokens):
    headers = {
        'Authorization': 'Bearer ' + oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            0
        ].name,
    }

    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response: httpx.Response = await ac.post('/_gateway/api-token',
                                                 json={'capabilities': CONFIG.api_token_config.token_capabilities})

    response = response.json()
    assert 'token' in response


@pytest.mark.asyncio
async def test_user_can_delete_token(gateway, oidc_tokens, api_token):
    headers = {
        'Authorization': 'Bearer ' + oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            0
        ].name,
    }

    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response: httpx.Response = await ac.request(
            url='/_gateway/api-token', json={'token': api_token}, method='DELETE'
        )
    data = response.json()
    assert 'status' in data
    assert response.status_code == 200


@pytest.mark.asyncio
async def test_user_cannot_delete_token(
        gateway, alternative_realm_oidc_tokens, api_token
):
    headers = {
        'Authorization': 'Bearer ' + alternative_realm_oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            0
        ].name,
    }

    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response: httpx.Response = await ac.request(
            url='/_gateway/api-token', json={'token': api_token}, method='DELETE'
        )
    data = response.json()
    assert 'error' in data


@pytest.mark.asyncio
async def test_user_cannot_create_token(gateway, alternative_realm_oidc_tokens):
    headers = {
        'Authorization': 'Bearer ' + alternative_realm_oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            0
        ].name,
    }

    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response: httpx.Response = await ac.post(url='/_gateway/api-token')
    data = response.json()
    assert 'error' in data


@pytest.mark.asyncio
async def test_delete_missing_token(gateway, oidc_tokens):
    headers = {
        'Authorization': 'Bearer ' + oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            0
        ].name,
    }

    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response: httpx.Response = await ac.request(
            url='/_gateway/api-token', json={'token': 'not-found'}, method='DELETE'
        )
    data = response.json()
    assert 'error' in data


@pytest.mark.asyncio
async def test_delete_missing_token_in_request(gateway, oidc_tokens):
    headers = {
        'Authorization': 'Bearer ' + oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            0
        ].name,
    }

    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response: httpx.Response = await ac.request(url='/_gateway/api-token', method='DELETE')
    data = response.json()
    assert 'error' in data


@pytest.mark.asyncio
async def test_default_token_is_loaded(gateway, oidc_tokens):
    headers = {
        'Authorization': 'Bearer ' + oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            0
        ].name,
    }

    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response = await ac.get('/_gateway/api-token')

    data = response.json()
    assert len([token['token'] for token in data if token['token'] == CONFIG.api_token_config.load_on_init]) == 1


@pytest.mark.asyncio
async def test_list_capabilities(gateway, oidc_tokens):
    headers = {
        'Authorization': 'Bearer ' + oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            0
        ].name,
    }
    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response = await ac.get('/_gateway/api-token-capabilities')

    data = response.json()
    assert deepdiff.diff.DeepDiff(data['capabilities'], CONFIG.api_token_config.token_capabilities) == {}


@pytest.mark.asyncio
async def test_create_token_validate_requested_capabilities(gateway, oidc_tokens):
    headers = {
        'Authorization': 'Bearer ' + oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            0
        ].name,
    }

    assert len(CONFIG.api_token_config.token_capabilities) > 1  # Test would be pointless otherwise

    requested_capability = CONFIG.api_token_config.token_capabilities[0]
    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response: httpx.Response = await ac.post('/_gateway/api-token',
                                                 json={'capabilities': [requested_capability]})

    data = response.json()
    token = data['token']

    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers={CONFIG.api_token_config.header_name: f'{CONFIG.api_token_config.token_prefix} {token}'},
            follow_redirects=True,
    ) as ac:
        response = await ac.get('/test/payload-bouncer')
    data = response.json()
    jwt_token = data['headers']['authorization'].split(' ')[1]

    decoded_token = jwt.decode(jwt_token, key=CONFIG.proxy.token_key)
    assert requested_capability in decoded_token['capabilities']


@pytest.mark.asyncio
async def test_create_token_expiration_bigger_than_maximal(gateway, oidc_tokens):
    headers = {
        'Authorization': 'Bearer ' + oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            0
        ].name,
    }

    limit_exceeding_expiration_offset = CONFIG.api_token_config.expiration_seconds + 4096
    request_datetime = datetime.utcnow()
    expiration = request_datetime + timedelta(seconds=limit_exceeding_expiration_offset)
    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response: httpx.Response = await ac.post('/_gateway/api-token',
                                                 json={'capabilities': CONFIG.api_token_config.token_capabilities,
                                                       'expiration': expiration.isoformat()})

    data = response.json()
    token = data['token']

    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response: httpx.Response = await ac.get('/_gateway/api-token')

    data = response.json()

    token_api_record = None
    for token_record in data:
        if token_record['token'] == token:
            token_api_record = token_record
    assert token_api_record is not None

    assert request_datetime < datetime.fromisoformat(token_api_record['expires']) < expiration
