import datetime
from secrets import token_urlsafe

import httpx
import pytest
from httpx import AsyncClient
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import sessionmaker

from csirtmu.service_gateway.auth_backend.api_token_model import APITokenModel
from csirtmu.service_gateway.auth_backend.auth_utils import register_api_token
from csirtmu.service_gateway.config import CONFIG


@pytest.mark.asyncio
async def test_missing_token_header(gateway):
    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response = await ac.get('/test')

    assert response.status_code == 401


@pytest.mark.asyncio
async def test_valid_token_in_header(gateway, api_token):
    headers = {
        CONFIG.api_token_config.header_name: (
            f'{CONFIG.api_token_config.token_prefix} {api_token}'
        )
    }
    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response = await ac.get('/test')

    assert response.status_code == 200


@pytest.mark.asyncio
async def test_token_not_found(gateway):
    headers = {
        CONFIG.api_token_config.header_name: (
            f'{CONFIG.api_token_config.token_prefix} very-deterministic-and-public-token'
        )
    }
    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response: httpx.Response = await ac.get('/test')
    assert response.status_code == 401


@pytest.mark.asyncio
async def test_no_matching_backend(gateway):
    headers = {
        'invalid-header-name': (
            f'{CONFIG.api_token_config.token_prefix} very-deterministic-and-public-token'
        )
    }
    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response: httpx.Response = await ac.get('/test')
    assert response.status_code == 401


@pytest.mark.asyncio
async def test_token_cache(gateway):
    test_token = 'very-test-very-specific-token'
    headers = {
        CONFIG.api_token_config.header_name: (
            f'{CONFIG.api_token_config.token_prefix} {test_token}'
        )
    }
    await register_api_token(
        api_token=test_token,
        capabilities=[],
        created=datetime.datetime.now(),
        expires=datetime.datetime.now() + datetime.timedelta(hours=1),
        created_by='test_iss|test_user',
    )

    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response = await ac.get('/test')
    assert response.status_code == 200

    async_session = sessionmaker(
        CONFIG.database.engine, expire_on_commit=False, class_=AsyncSession
    )
    async with async_session() as s:
        async with s.begin():
            stmt = select(APITokenModel).filter(APITokenModel.token == test_token)
            result = await s.execute(stmt)
            token_model = result.scalars().one()
            await s.delete(token_model)
        await s.commit()

    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response: httpx.Response = await ac.get('/test')
    assert response.status_code == 401  # DB is checked first


@pytest.mark.asyncio
async def test_expired_token(gateway):
    test_token = token_urlsafe(32)
    headers = {
        CONFIG.api_token_config.header_name: CONFIG.api_token_config.token_prefix
                                             + ' '
                                             + test_token
    }
    created = datetime.datetime.utcnow()
    expires = created - datetime.timedelta(days=1)

    await register_api_token(
        api_token=test_token,
        capabilities=[],
        created=created,
        expires=expires,
        created_by='test_iss|test_user',
    )

    async with AsyncClient(
            app=gateway,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response = await ac.get('/test')
    assert response.status_code == 401
