import httpx
import pytest
import pytest_asyncio
from httpx import AsyncClient
from starlette.applications import Starlette
from starlette.middleware.authentication import AuthenticationMiddleware

from csirtmu.service_gateway.app import on_auth_error
from tests.fake_authentication_backend import FakeAuthenticationBackend


@pytest_asyncio.fixture
async def app_no_backend() -> Starlette:
    from csirtmu.service_gateway.app import ServiceGateway

    gateway = ServiceGateway()

    # Push user middleware to stack
    gateway.get_app().user_middleware.clear()
    # Clear stack
    gateway.get_app().middleware_stack = None

    # Add mock auth middleware to empty stack
    gateway.get_app().add_middleware(
        AuthenticationMiddleware,
        backend=FakeAuthenticationBackend(),
        on_error=on_auth_error,
    )

    yield gateway.get_app()
    await gateway.client_session.close()


@pytest.mark.asyncio
async def test_simple_server(app_no_backend):
    headers = {'Authorization': 'anything goes'}

    async with AsyncClient(
            app=app_no_backend,
            base_url='http://testserver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response: httpx.Response = await ac.get('/test/payload-bouncer')
    assert response.status_code == 200


@pytest.mark.asyncio
async def test_simple_server_with_data(app_no_backend):
    headers = {'Authorization': 'anything goes'}
    data = {'sample': 'json', 'v': -5}

    async with AsyncClient(
            app=app_no_backend,
            base_url='http://testerver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response: httpx.Response = await ac.post('/test/payload-bouncer', json=data)
    assert response.status_code == 200


@pytest.mark.asyncio
async def test_simple_server_with_empty_body(app_no_backend):
    headers = {'Authorization': 'anything goes'}

    async with AsyncClient(
            app=app_no_backend,
            base_url='http://testerver',
            headers=headers,
            follow_redirects=True,
    ) as ac:
        response: httpx.Response = await ac.post('/test/payload-bouncer', json={})
    assert response.status_code == 200
