import httpx
import pytest
import requests
from aiohttp import ClientResponseError
from authlib.jose import jwt
from starlette import status

from csirtmu.service_gateway.auth_backend.multi_backend_authentication_middleware import (
    MultiBackendAuthenticationMiddleware,
)
from csirtmu.service_gateway.auth_backend.oidc_authentication_backend import (
    OIDCAuthenticationBackend,
)
from csirtmu.service_gateway.config import CONFIG, OIDCIssuer
from csirtmu.service_gateway.utils import get_oidc_issuer_from_dataclass
from deepdiff import DeepDiff
from httpx import AsyncClient
from starlette.datastructures import Headers
from starlette.requests import Request
from tests.test_settings import MUNI_MOCK_MAPPED_CLAIMS
from csirtmu.service_gateway.app import ServiceGateway


@pytest.mark.asyncio
async def test_example_app_backend_version():
    ServiceGateway._check_backend_version('/testserver', CONFIG.routes['/test'])


# noinspection DuplicatedCode
@pytest.mark.asyncio
async def test_valid_token_provider_name(gateway, oidc_tokens):
    headers = {
        'Authorization': 'Bearer ' + oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            0
        ].name,
    }

    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get('/test', headers=headers)

    assert response.status_code == 200


# noinspection DuplicatedCode
@pytest.mark.asyncio
async def test_valid_token_alternative_provider(gateway, alternative_realm_oidc_tokens):
    headers = {
        'Authorization': 'Bearer ' + alternative_realm_oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            1
        ].name,
    }
    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get('/test', headers=headers)

    assert response.status_code == 200


@pytest.mark.asyncio
async def test_valid_token_unknown_provider(gateway, oidc_tokens):
    headers = {
        'Authorization': 'Bearer ' + oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: 'what-is-this',
    }
    async with AsyncClient(
            app=gateway, base_url='http://testserver', headers=headers
    ) as ac:
        response: httpx.Response = await ac.get('/test', headers=headers)

    assert response.status_code == 401


@pytest.mark.asyncio
async def test_valid_token_use_default_provider(gateway, test_scoped_tokens):
    headers = {'Authorization': 'Bearer ' + test_scoped_tokens[0]}
    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get('/test', headers=headers)

    assert response.status_code == 200


@pytest.mark.asyncio
async def test_kid_in_token_header(gateway, oidc_tokens):
    headers = {
        'Authorization': 'Bearer ' + oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            0
        ].name,
    }
    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get('/test', headers=headers)

    assert response.status_code == 200


@pytest.mark.asyncio
async def test_request_invalid_access_token(gateway, oidc_tokens):
    headers = {
        'Authorization': 'Bearer ' + 'INVALID_TOKEN',
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            1
        ].name,
    }
    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get('/test', headers=headers)

    assert response.status_code == 401


@pytest.mark.asyncio
async def test_request_without_headers(gateway, oidc_tokens):
    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get('/test')

    assert response.status_code == 401


@pytest.mark.asyncio
async def test_request_invalid_access_token_format(gateway, oidc_tokens):
    headers = {'Authorization': 'INVALID_FORMAT ' + 'INVALID_TOKEN'}
    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get('/test', headers=headers)

    assert response.status_code == 401


@pytest.mark.asyncio
async def test_unknown_route(gateway, oidc_tokens):
    headers = {'Authorization': 'Bearer ' + oidc_tokens[0]}
    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get('/unknown/', headers=headers)

    assert response.status_code == 404


@pytest.mark.asyncio
async def test_request_with_query_params(gateway, oidc_tokens):
    query_params = {'foo': 'oof', 'bar': '3'}

    headers = {'Authorization': 'Bearer ' + oidc_tokens[0]}
    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get(
            '/test/payload-bouncer-authorized', params=query_params, headers=headers
        )

    assert response.status_code == 200

    json_response = response.json()
    assert DeepDiff(json_response['args'], query_params) == {}


@pytest.mark.asyncio
async def test_request_unauthorized(gateway):
    query_params = {'foo': 'oof', 'bar': '3'}

    # TODO: add users without capabilities
    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get('/test/payload-bouncer-authorized', params=query_params)

    assert response.status_code == status.HTTP_401_UNAUTHORIZED


@pytest.mark.asyncio
async def test_request_with_query_params_in_url(gateway, oidc_tokens):
    headers = {'Authorization': 'Bearer ' + oidc_tokens[0]}
    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get(
            '/test/payload-bouncer-authorized?foo=oof&bar=3', headers=headers
        )

    assert response.status_code == 200

    json_response = response.json()
    should_respond_with = {'foo': 'oof', 'bar': '3'}
    assert DeepDiff(json_response['args'], should_respond_with) == {}


@pytest.mark.asyncio
async def test_big_response(gateway, oidc_tokens):
    headers = {'Authorization': 'Bearer ' + oidc_tokens[0]}
    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get('/muni/', headers=headers)

    assert response.status_code == 200


@pytest.mark.asyncio
async def test_complex_path(gateway, oidc_tokens):
    headers = {'Authorization': 'Bearer ' + oidc_tokens[0]}
    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get(
            '/muni/studenti/studijni-oddeleni-mu', headers=headers
        )

    assert response.status_code == 200


@pytest.mark.asyncio
async def test_expired_token(gateway, endpoint, test_scoped_tokens):
    oidc_response = requests.post(
        endpoint
        + f'/realms/{CONFIG.oidc_config.issuers[0].name}'
          '/protocol/openid-connect/logout',
        headers={'Authorization': f'Bearer {test_scoped_tokens[0]}'},
        data={'client_id': 'test-client', 'refresh_token': test_scoped_tokens[1]},
    )
    assert oidc_response.status_code == 204

    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        with pytest.raises(ClientResponseError):
            await ac.get(
                '/test', headers={'Authorization': f'Bearer {test_scoped_tokens[0]}'}
            )


@pytest.mark.asyncio
async def test_backend_missing_token_keyword(gateway, oidc_tokens):
    headers = {'Authorization': 'missing_keyword'}
    request = Request({'type': 'http', 'headers': Headers(headers).raw})
    auth_middleware = [
        middleware
        for middleware in gateway.user_middleware
        if middleware.cls is MultiBackendAuthenticationMiddleware
    ][0]
    backend: OIDCAuthenticationBackend = [
        backend
        for backend in auth_middleware.options['backends']
        if isinstance(backend, OIDCAuthenticationBackend)
    ][0]
    assert await backend.authenticate(request) is None


@pytest.mark.asyncio
async def test_backend_missing_invalid_token(gateway, endpoint, test_scoped_tokens):
    oidc_response = requests.post(endpoint + f'/realms/{CONFIG.oidc_config.issuers[0].name}'
                                             f'/protocol/openid-connect/logout',
                                  headers={'Authorization': f'Bearer {test_scoped_tokens[0]}'},
                                  data={
                                      'client_id': 'test-client',
                                      'refresh_token': test_scoped_tokens[1]},
                                  )
    assert oidc_response.status_code == 204

    headers = {'Authorization': f'Bearer {test_scoped_tokens[0]}'}
    request = Request({'type': 'http', 'headers': Headers(headers).raw})

    auth_middleware = [
        middleware
        for middleware in gateway.user_middleware
        if middleware.cls is MultiBackendAuthenticationMiddleware
    ][0]
    backend: OIDCAuthenticationBackend = [
        backend
        for backend in auth_middleware.options['backends']
        if isinstance(backend, OIDCAuthenticationBackend)
    ][0]

    with pytest.raises(ClientResponseError):
        await backend.authenticate(request)


@pytest.mark.asyncio
async def test_auth_header_present_in_proxy_request(gateway, endpoint, oidc_tokens):
    headers = {
        'Authorization': 'Bearer ' + oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            0
        ].name,
    }
    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get(
            '/test/payload-bouncer-authorized', headers=headers
        )

    assert response.status_code != 400

    response: dict = response.json()
    assert 'authorization' in response['headers']


@pytest.mark.asyncio
async def test_claims_present_in_proxy_request(gateway, endpoint, test_scoped_tokens):
    headers = {
        'Authorization': 'Bearer ' + test_scoped_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            0
        ].name,
    }
    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get(
            '/test/payload-bouncer-authorized', headers=headers
        )

    response: dict = response.json()
    assert 'headers' in response
    # TODO: currently uses postman api, use local/CI service instead?
    # This endpoints responds with request details, including original request headers
    requests_headers = response['headers']
    jwt_token: str = requests_headers['authorization']
    assert jwt_token.startswith('Token ')

    stripped_token = jwt_token.replace(CONFIG.proxy.token_prefix + ' ', '')
    decoded_token = jwt.decode(
        stripped_token,
        key=CONFIG.proxy.token_key,
    )

    token_capabilities = decoded_token.get('capabilities')
    for capability in MUNI_MOCK_MAPPED_CLAIMS:
        assert capability in token_capabilities


# Must be in line with example config
def test_issuer_dataclass_to_object():
    issuer_dataclasses = []
    for issuer in CONFIG.oidc_config.issuers:
        default = issuer.default
        issuer_dataclasses.append(
            OIDCIssuer(
                name=issuer.name,
                url=issuer.url,
                default_kid=issuer.default,
                default=default,
            )
        )

    default_issuer, issuers = get_oidc_issuer_from_dataclass(issuer_dataclasses)
    assert default_issuer is not None
    assert default_issuer == CONFIG.oidc_config.issuers[0].name
    assert len(issuers) == 2

    new_issuer = issuer_dataclasses[1]
    new_issuer.default = True
    issuer_dataclasses.append(new_issuer)

    with pytest.raises(Exception) as ex:
        get_oidc_issuer_from_dataclass(issuer_dataclasses)
        assert ex.value == 'Default issuer is already set'


# noinspection DuplicatedCode
@pytest.mark.asyncio
async def test_cached_user(gateway, oidc_tokens):
    headers = {
        'Authorization': 'Bearer ' + oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            0
        ].name,
    }

    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get('/test', headers=headers)
    assert response.status_code == 200

    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get('/test', headers=headers)
    assert response.status_code == 200


# noinspection DuplicatedCode
@pytest.mark.asyncio
async def test_cached_user_token_from_wrong_issuer(gateway, oidc_tokens):
    headers = {
        'Authorization': 'Bearer ' + oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            0
        ].name,
    }

    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get('/test', headers=headers)
    assert response.status_code == 200

    headers = {
        'Authorization': 'Bearer ' + oidc_tokens[0],
        CONFIG.oidc_config.oidc_provider_header_name: CONFIG.oidc_config.issuers[
            1
        ].name,
    }
    async with AsyncClient(
            app=gateway, base_url='http://testserver', follow_redirects=True
    ) as ac:
        response: httpx.Response = await ac.get('/test', headers=headers)
    assert response.status_code == 401