import asyncio
import logging
import sys
from asyncio import sleep
from typing import Dict, Optional, Any, Union

import alembic.config
import sqlalchemy
from csirtmu.starlette_jwt_auth.routes import auth_backend_version_route
from csirtmu.starlette_jwt_auth.token_authentication_backend import TokenAuthenticationBackend
from csirtmu.strawberry_dmt.authorization import CapableUser, Authorization
from csirtmu.strawberry_dmt.custom_view import GraphQLCustomASGIApp
from csirtmu.strawberry_dmt.data_loader import CachingDataLoader
from csirtmu.strawberry_dmt.sqla_base import Base
from sqlalchemy.exc import IntegrityError, OperationalError
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import sessionmaker
from sqlalchemy_oso import SQLAlchemyOso
from starlette.applications import Starlette
from starlette.authentication import UnauthenticatedUser
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.middleware.cors import CORSMiddleware
from starlette.requests import Request
from starlette.responses import Response
from starlette.websockets import WebSocket
from starlette_prometheus import metrics
from tenacity import RetryError, after_log, retry, retry_if_exception_type, stop_after_attempt, wait_exponential, \
    wait_random

from csirtmu.ttp_database.config import BASE_DIR, CONFIG
from csirtmu.ttp_database.models import RelationUser
from csirtmu.ttp_database.schema import schema
from csirtmu.ttp_database.views import AppState, create_healthcheck_liveness, create_healthcheck_readiness

logger = logging.getLogger(__package__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(logging.Formatter("%(asctime)s [%(levelname)s]: %(message)s"))
logger.addHandler(handler)


def get_oso(oso_policy_list: str):
    oso = SQLAlchemyOso(Base)
    oso.register_class(Base)
    oso.register_class(UnauthenticatedUser, name='UnauthenticatedUser')
    oso.register_class(CapableUser, name='User')
    oso.register_class(RelationUser, name='RelationUser')
    oso.load_files(oso_policy_list.split(','))
    return oso


class CustomGraphQLCustomASGIApp(GraphQLCustomASGIApp):
    def __init__(self, async_session_factory: sqlalchemy.orm.sessionmaker = None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.async_session_factory = async_session_factory

    async def get_context(self, request: Union[Request, WebSocket], response: Optional[Response] = None) \
            -> Optional[Any]:
        context: Dict = await super().get_context(request, response)
        context.update(dict(
            loader=CachingDataLoader(),
            session_factory=self.session_factory,
            async_session_factory=self.async_session_factory,
            authorization=Authorization.from_(self.oso, request.user)
        ))
        return context


def asgi_app(oso: SQLAlchemyOso, state: AppState, debug: bool = True):
    session_maker = sessionmaker(CONFIG.database.engine, future=True)
    async_session_maker = sessionmaker(CONFIG.database.async_engine, future=True, class_=AsyncSession)
    app = Starlette(debug=debug, routes=[auth_backend_version_route])
    app.add_middleware(CORSMiddleware,
                       allow_headers=["*"],
                       allow_origins=["*"],
                       allow_methods=["*"])
    app.add_middleware(AuthenticationMiddleware, backend=TokenAuthenticationBackend(user_class=RelationUser))

    graphql_app = CustomGraphQLCustomASGIApp(
        schema=schema, session_factory=session_maker, debug=debug, oso=oso, async_session_factory=async_session_maker)
    app.add_route("/", graphql_app)
    app.add_route("/graphql", graphql_app)

    app.add_route(path='/healthcheck/liveness', route=create_healthcheck_liveness(state), methods=['GET'])
    app.add_route(path='/healthcheck/readiness', route=create_healthcheck_readiness(state), methods=['GET'])
    app.add_route(path='/metrics', route=metrics)

    return app


async def load_db(state: AppState) -> None:
    @retry(stop=stop_after_attempt(10),
           wait=wait_exponential(multiplier=1, min=2, max=10) + wait_random(0, 2),
           retry=retry_if_exception_type(OperationalError),
           retry_error_callback=lambda _: exit(1),
           after=after_log(logger, logging.INFO))
    async def load():
        try:
            alembic.config.main(['upgrade', 'head'])
        except IntegrityError:
            print("It seems that other worker is already running migrations")
            await sleep(5)

    try:
        await load()
    except RetryError:
        state.healthy = False


def app():
    oso = get_oso((BASE_DIR / 'config/authorization.polar').resolve().__str__())
    state = AppState()
    loop = asyncio.get_event_loop()
    if loop.is_running():
        loop.create_task(load_db(state))
    else:
        loop.run_until_complete(load_db(state))
    return asgi_app(oso=oso, state=state, debug=True)
