import random
import string
from typing import List, Optional

from csirtmu.starlette_jwt_auth.models import MappedUser
from csirtmu.strawberry_dmt.sqla_base import fk_for, AutoBase, m2m_for, AutoPolymorphicBase
from sqlalchemy import Column, String, Integer, DateTime, select
from sqlalchemy.orm import relationship, configure_mappers, declared_attr, column_property, Session
from sqlalchemy.sql.functions import func

from csirtmu.ttp_database.config import CONFIG

ttp_externalReference_association = m2m_for('TTP', 'ExternalReference')
techniqueMitigationRelationship_externalReference_association = m2m_for('TechniqueMitigationRelationship',
                                                                        'ExternalReference')
malware_externalReference_association = m2m_for('Malware', 'ExternalReference')
tool_externalReference_association = m2m_for('Tool', 'ExternalReference')
group_externalReference_association = m2m_for('Group', 'ExternalReference')
mitigation_externalReference_association = m2m_for('Mitigation', 'ExternalReference')
tactic_technique_association = m2m_for('Tactic', 'Technique')
technique_platform_association = m2m_for('Technique', 'Platform')
malware_platform_association = m2m_for('Malware', 'Platform')
tool_platform_association = m2m_for('Tool', 'Platform')
malware_alias_association = m2m_for('Malware', 'Alias')
tool_alias_association = m2m_for('Tool', 'Alias')
malware_group_association = m2m_for('Malware', 'Group')
tool_group_association = m2m_for('Tool', 'Group')
group_alias_association = m2m_for('Group', 'Alias')

ALPHABET = string.ascii_lowercase + string.digits


def get_unique_identifier(context):
    unique_identifier = "".join(random.choices(ALPHABET, k=8))
    with Session(bind=CONFIG.database.engine, expire_on_commit=False) as session:
        while session.query(TTP).filter_by(unique_identifier=unique_identifier).first() is not None:
            unique_identifier = random.choices(ALPHABET, k=8)
    return unique_identifier


class ExternalReference(AutoBase):
    source_name = Column(String, nullable=False)
    description = Column(String)
    url = Column(String)
    external_id = Column(String)


class TTP(AutoPolymorphicBase(on_column="ttp_type")):
    ttp_type = Column(String(50))
    description = Column(String)
    created = Column(DateTime(timezone=True), server_default=func.now())
    modified = Column(
        DateTime(timezone=True), onupdate=func.current_timestamp(), server_default=func.now())
    external_references = relationship('ExternalReference', secondary=ttp_externalReference_association,
                                       backref='ttps')
    unique_identifier = Column(String(8), unique=True, nullable=False, default=get_unique_identifier)


class Tactic(TTP):
    tactic_name = Column(String, nullable=False)

    techniques = relationship('Technique', secondary=tactic_technique_association, back_populates='tactics')


class Technique(TTP):
    technique_name = Column(String, nullable=False)
    tactics = relationship('Tactic', secondary=tactic_technique_association, back_populates='techniques')
    platforms = relationship('Platform', secondary=technique_platform_association, backref='techniques')

    parent_id = Column(Integer, fk_for('Technique'))

    @declared_attr
    def sub_technique_of(self):
        return relationship("Technique", backref='sub_techniques', remote_side=[self.id], foreign_keys=[self.parent_id])

    @declared_attr
    def is_subtechnique(self):
        return column_property(select(self.parent_id != None).correlate_except().scalar_subquery())


class Malware(AutoBase):
    name = Column(String, nullable=False)
    description = Column(String)
    created = Column(DateTime(timezone=True), server_default=func.now())
    modified = Column(
        DateTime(timezone=True), onupdate=func.current_timestamp(), server_default=func.now())
    platforms = relationship('Platform', secondary=malware_platform_association, backref='malware')
    aliases = relationship('Alias', secondary=malware_alias_association, backref='malware')
    groups = relationship('Group', secondary=malware_group_association, backref='malware')

    external_references = relationship('ExternalReference', secondary=malware_externalReference_association,
                                       backref='malware')


class Tool(AutoBase):
    name = Column(String, nullable=False)
    description = Column(String)
    created = Column(DateTime(timezone=True), server_default=func.now())
    modified = Column(
        DateTime(timezone=True), onupdate=func.current_timestamp(), server_default=func.now())
    platforms = relationship('Platform', secondary=tool_platform_association, backref='tools')
    aliases = relationship('Alias', secondary=tool_alias_association, backref='tools')
    groups = relationship('Group', secondary=tool_group_association, backref='tools')

    external_references = relationship('ExternalReference', secondary=tool_externalReference_association,
                                       backref='tools')


class Group(AutoBase):
    name = Column(String, nullable=False)
    description = Column(String)
    created = Column(DateTime(timezone=True), server_default=func.now())
    modified = Column(
        DateTime(timezone=True), onupdate=func.current_timestamp(), server_default=func.now())
    aliases = relationship('Alias', secondary=group_alias_association, backref='groups')

    external_references = relationship('ExternalReference', secondary=group_externalReference_association,
                                       backref='groups')


class Mitigation(AutoBase):
    name = Column(String, nullable=False)
    description = Column(String)
    created = Column(DateTime(timezone=True), server_default=func.now())
    modified = Column(
        DateTime(timezone=True), onupdate=func.current_timestamp(), server_default=func.now())

    external_references = relationship('ExternalReference', secondary=mitigation_externalReference_association,
                                       backref='mitigations')


class TechniqueMitigationRelationship(AutoBase):
    description = Column(String)
    technique_id = Column(Integer, fk_for('Technique'))
    technique = relationship('Technique', backref='mitigations', foreign_keys=[technique_id])
    mitigation_id = Column(Integer, fk_for('Mitigation'))
    mitigation = relationship('Mitigation', backref='techniques', foreign_keys=[mitigation_id])

    external_references = relationship('ExternalReference',
                                       secondary=techniqueMitigationRelationship_externalReference_association,
                                       backref='technique_mitigation_relationships')

    @declared_attr
    def name(self):
        return column_property(select(Mitigation.name)
                               .where(Mitigation.id == self.mitigation_id)
                               .correlate_except(Mitigation)
                               .scalar_subquery())


class GroupProcedure(TTP):
    group_technique_id = Column(Integer, fk_for('Technique'), nullable=False)
    group_technique = relationship('Technique', backref='group_procedures', foreign_keys=[group_technique_id])
    group_id = Column(Integer, fk_for('Group'), nullable=False)
    group = relationship('Group', backref='techniques_used', foreign_keys=[group_id])

    @declared_attr
    def group_procedure_name(self) -> str:
        property = column_property(select(Group.name)
                                   .where(Group.id == self.group_id)
                                   .correlate_except(Group)
                                   .scalar_subquery())
        setattr(property.columns[0], 'nullable', False)
        return property


class MalwareProcedure(TTP):
    malware_technique_id = Column(Integer, fk_for('Technique'), nullable=False)
    malware_technique = relationship('Technique', backref='malware_procedures', foreign_keys=[malware_technique_id])
    malware_id = Column(Integer, fk_for('Malware'), nullable=False)
    malware = relationship('Malware', backref='techniques_used', foreign_keys=[malware_id])

    @declared_attr
    def malware_procedure_name(self) -> str:
        property = column_property(select(Malware.name)
                                   .where(Malware.id == self.malware_id)
                                   .correlate_except(Malware)
                                   .scalar_subquery())

        setattr(property.columns[0], 'nullable', False)
        return property


class ToolProcedure(TTP):
    tool_technique_id = Column(Integer, fk_for('Technique'), nullable=False)
    tool_technique = relationship('Technique', backref='tool_procedures', foreign_keys=[tool_technique_id])
    tool_id = Column(Integer, fk_for('Tool'), nullable=False)
    tool = relationship('Tool', backref='techniques_used', foreign_keys=[tool_id])

    @declared_attr
    def tool_procedure_name(self) -> str:
        property = column_property(select(Tool.name)
                                   .where(Tool.id == self.tool_id)
                                   .correlate_except(Tool)
                                   .scalar_subquery())

        setattr(property.columns[0], 'nullable', False)
        return property


class Platform(AutoBase):
    value = Column(String, nullable=False)


class Alias(AutoBase):
    value = Column(String, nullable=False)


configure_mappers()


class RelationUser(MappedUser):
    uid: str
    capabilities: List[str]

    def __init__(self, uid: str, capabilities: Optional[List[str]] = None):
        super().__init__()
        self.uid = uid
        if capabilities:
            self.capabilities = capabilities
        else:
            self.capabilities = []

    @property
    def identity(self) -> str:
        return self.uid

    @property
    def is_authenticated(self) -> bool:
        return True

    @property
    def display_name(self) -> str:
        return self.uid

    @staticmethod
    def create_or_get(access_token: str, user_id: str, capabilities: List[str]) -> 'MappedUser':
        return RelationUser(uid=user_id, capabilities=capabilities)
