from typing import List, Optional

import strawberry
from csirtmu.strawberry_dmt.locale.locale import Languages
from csirtmu.strawberry_dmt.permissions import CAN_CREATE_ONE, CAN_QUERY_ONE
from csirtmu.strawberry_dmt.strawberry_types import MutateResult
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import selectinload
from strawberry.types import Info

from csirtmu.ttp_database import models
from csirtmu.ttp_database.resolvers_description import descriptions
from csirtmu.ttp_database.models import Technique as ModelTechnique
from csirtmu.ttp_database.permission_classes import is_authorized
from csirtmu.ttp_database.schema_auto_generated import Technique, mutations, queries, registry


@strawberry.type
class Query(queries):
    pass


@strawberry.type
class Mutation(mutations):
    @strawberry.field(permission_classes=[is_authorized(models.Technique, CAN_CREATE_ONE),
                                          is_authorized(models.Tactic, CAN_QUERY_ONE)],
                      description=descriptions["create_subtechnique"])
    async def create_subtechnique(self, info: Info, data: Technique.get_strawberry_info().create) -> Optional[
        MutateResult]:
        session: AsyncSession
        async with info.context['async_session_factory']() as session:

            if not data.sub_technique_of:
                raise ValueError("Foreign key for parent technique wasn't provided")

            parent_technique = (
                await session.execute(select(ModelTechnique).
                                      filter_by(id=int(data.sub_technique_of)).
                                      options(selectinload(ModelTechnique.tactics)))).scalar()
            if not parent_technique:
                raise ValueError("Parent technique with provided key does not exists")
            if parent_technique.parent_id:
                raise ValueError(
                    "Cannot assign subtechnique because technique with provided key is already subtechnique.")

            # Inherit from parent technique
            data.tactics = [tactic.id for tactic in parent_technique.tactics]

            result = await mutations.create_technique(None, info, data)
            return result

    @strawberry.field(permission_classes=[is_authorized(models.Technique, CAN_CREATE_ONE),
                                          is_authorized(models.Tactic, CAN_QUERY_ONE)],
                      description=descriptions["create_subtechnique_bulk"])
    async def create_subtechnique_bulk(self, info: Info, dataList: List[Technique.get_strawberry_info().create]) -> \
            Optional[List[MutateResult]]:

        session: AsyncSession
        async with info.context['async_session_factory']() as session:
            for subtechnique in dataList:
                if not subtechnique.sub_technique_of:
                    raise ValueError("Foreign key for parent technique wasn't provided")

                parent_technique = (await session.execute(
                    select(ModelTechnique).filter_by(id=int(subtechnique.sub_technique_of)).options(
                        selectinload(ModelTechnique.tactics)))).scalar()
                if not parent_technique:
                    raise ValueError("Parent technique with provided key does not exists")
                if parent_technique.parent_id:
                    raise ValueError(
                        "Cannot assign subtechnique because technique with provided key is already subtechnique.")

                # Inherit from parent technique
                subtechnique.tactics = [tactic.id for tactic in parent_technique.tactics]

            return await mutations.create_technique_bulk(self, info, dataList)


schema: strawberry.federation.Schema = registry.schema(Query, Mutation, docs_language=Languages.CS)
