import asyncio
from multiprocessing import Pool

from csirtmu.ttp_database.utils.mitre_adapter.dataclasses_models import Software
from csirtmu.ttp_database.utils.mitre_adapter.fetched_entities import FetchedEntities
from csirtmu.ttp_database.utils.mitre_adapter.handlers.default_handler import DefaultHandler
from csirtmu.ttp_database.utils.mitre_adapter.mutations import mutation_format
from csirtmu.ttp_database.utils.mitre_adapter.utils import verbose_print, InitializedObject, \
    create_entities, get_ids_by_names, parser


class SoftwareHandler(DefaultHandler):
    def __init__(self, fetched_entities: FetchedEntities, software_type: str, func):
        self.fetched_entities = fetched_entities
        super().__init__(software_type, func, Software)

    def _init_software(self, val):
        async def _init():
            aliases = list(filter(lambda x: x != val.get('name'), val.get("x_mitre_aliases", [])))
            aliases_ids = get_ids_by_names(aliases, self.fetched_entities.aliases, "value")
            filtered_platforms = list(
                filter(lambda platform: platform["value"] in val.get('x_mitre_platforms', []),
                       self.fetched_entities.platforms))
            platforms_ids = [platform["id"] for platform in filtered_platforms]
            external_references = get_ids_by_names([ref.get("url") for ref in val.get("external_references") if ref.get("url")],
                                                   self.fetched_entities.references, "url")
            return InitializedObject(base_entity={'name': val.get("name"),
                                                  'description': val.get("description"),
                                                  'aliases': aliases_ids,
                                                  'platforms': platforms_ids,
                                                  'external_references': external_references})

        return asyncio.run(_init())

    async def _seed_software(self, software, software_name: str):
        await self.fetched_entities.reload()

        with Pool() as pool:
            output = pool.map(self._init_software, [val for key, val in software])

        await create_entities(mutation_format(f'create{software_name}', parser([x.base_entity for x in output])))

    async def seed_malware(self, malware):
        verbose_print(f"Seeding malware...")
        await self._seed_software(malware, "Malware")

    async def seed_tools(self, tools):
        verbose_print(f"Seeding tools...")
        await self._seed_software(tools, "Tool")


class MalwareHandler(SoftwareHandler):
    def __init__(self, fetched_entities: FetchedEntities):
        super().__init__(fetched_entities, "malware", self.seed_malware)


class ToolHandler(SoftwareHandler):
    def __init__(self, fetched_entities: FetchedEntities):
        super().__init__(fetched_entities, "tool", self.seed_tools)
