import asyncio
from multiprocessing import Pool
from typing import List

from mitreattack import get_attack_id

from csirtmu.ttp_database.utils.mitre_adapter.client import post
from csirtmu.ttp_database.utils.mitre_adapter.dataclasses_models import AttackPattern
from csirtmu.ttp_database.utils.mitre_adapter.fetched_entities import FetchedEntities
from csirtmu.ttp_database.utils.mitre_adapter.handlers.default_handler import DefaultHandler
from csirtmu.ttp_database.utils.mitre_adapter.mutations import mutation_format
from csirtmu.ttp_database.utils.mitre_adapter.queries import get_techniques_with_attack_id_by_ids_query
from csirtmu.ttp_database.utils.mitre_adapter.relationships_helper import RelationshipsHelper
from csirtmu.ttp_database.utils.mitre_adapter.utils import verbose_print, InitializedObject, \
    create_entities, get_ids_by_names, prepare_mitigations, prepare_procedures, parser, \
    procedure_type_technique_column_name_map, divide_post_request


class SubtechniqueHandler(DefaultHandler):
    def __init__(self, fetched_entities: FetchedEntities, relationships_helper: RelationshipsHelper):
        self.fetched_entities = fetched_entities
        self.relationships_helper = relationships_helper
        super().__init__("attack-pattern", self.seed_subtechniques, AttackPattern)

    def _init_subtechniques(self, val):
        async def _init():
            technique = self.relationships_helper.get_related(val.get('id'), 'subtechnique-of', 'attack-pattern')[0]
            external_references = get_ids_by_names([ref.get("url") for ref in val.get("external_references") if ref.get("url")],
                                                   self.fetched_entities.references, "url")

            my_attack_id = get_attack_id(val)

            technique_attack_id = get_attack_id(technique)
            fetched_technique = {}
            for ft in self.fetched_entities.techniques:
                data = ft.get("externalReferences", {}).get("data")
                if data and data[0].get("externalId") == technique_attack_id:
                    fetched_technique = ft
                    break

            technique_id = fetched_technique.get("id")

            entity_types = ["tool", "malware", "intrusion-set"]
            tools_relationships, malware_relationships, groups_relationships = [
                self.relationships_helper.get_relations_with_related_entities(val.get('id'), 'uses', entity_type, reversed_relation=True) for entity_type in entity_types]

            mititgations_relationships = self.relationships_helper.get_relations_with_related_entities(val.get('id'), 'mitigates',
                                                                                  'course-of-action',
                                                                                  reversed_relation=True)
            mitigations = await prepare_mitigations(self.fetched_entities, mititgations_relationships)
            procedures = await prepare_procedures(self.fetched_entities, tools_relationships, malware_relationships,
                                                  groups_relationships)

            filtered_platforms = list(
                filter(lambda platform: platform["value"] in val.get('x_mitre_platforms', []),
                       self.fetched_entities.platforms))
            platforms_ids = [platform["id"] for platform in filtered_platforms]

            return InitializedObject(base_entity={'techniqueName': val.get("name"),
                                                  'description': val.get("description"),
                                                  'sub_technique_of': technique_id,
                                                  'platforms': platforms_ids,
                                                  'external_references': external_references},
                                     update_after_creation={"procedures": procedures, "mitigations": mitigations},
                                     attack_id=my_attack_id)

        return asyncio.run(_init())

    async def seed_subtechniques(self, subtechniques):
        verbose_print(f"Seeding subtechniques...")
        await self.fetched_entities.reload()

        with Pool() as pool:
            output: List[InitializedObject] = pool.map(self._init_subtechniques, [val for key, val in subtechniques])

        final = {initialized_object.attack_id: initialized_object for initialized_object in output}
        result = await create_entities(
            mutation_format('createSubtechnique', parser([x.base_entity for x in output])))

        created_subtechniques_ids = [dict_with_id["id"] for dict_with_id in result["createSubtechniqueBulk"]]

        query = await get_techniques_with_attack_id_by_ids_query(created_subtechniques_ids)
        result = await post(query)
        created_subtechniques = result["getTechniqueList"]["data"]

        for subtechnique in created_subtechniques:
            technique_id = subtechnique.get("id")
            data = subtechnique.get("externalReferences", {}).get("data")
            attack_id = data[0].get("externalId", None) if data else None
            local_technique = final.get(attack_id)
            if local_technique:
                procedures = local_technique.update_after_creation.get("procedures", {})
                for key, procedure_group in procedures.items():
                    for procedure in procedure_group:
                        # Set technique id for prepared procedures
                        procedure[procedure_type_technique_column_name_map[key]] = technique_id

                mitigations = local_technique.update_after_creation.get("mitigations", [])
                for mitigation in mitigations:
                    mitigation["technique"] = technique_id

        tool_procedures_to_create = []
        malware_procedures_to_create = []
        group_procedures_to_create = []
        technique_mitigation_relationship_to_create = []

        for out in output:
            procedures = out.update_after_creation.get("procedures", {})
            tool_procedures_to_create += procedures.get("tools_procedures", [])
            malware_procedures_to_create += procedures.get("malware_procedures", [])
            group_procedures_to_create += procedures.get("groups_procedures", [])
            technique_mitigation_relationship_to_create += out.update_after_creation.get("mitigations", [])

        await divide_post_request(tool_procedures_to_create, 'createToolProcedure')
        await divide_post_request(malware_procedures_to_create, 'createMalwareProcedure')
        await divide_post_request(group_procedures_to_create, 'createGroupProcedure')
        await divide_post_request(technique_mitigation_relationship_to_create, 'createTechniqueMitigationRelationship')
