import asyncio
from multiprocessing import Pool
from typing import List

from mitreattack import get_attack_id

from csirtmu.ttp_database.utils.mitre_adapter.client import post
from csirtmu.ttp_database.utils.mitre_adapter.dataclasses_models import AttackPattern
from csirtmu.ttp_database.utils.mitre_adapter.fetched_entities import FetchedEntities
from csirtmu.ttp_database.utils.mitre_adapter.handlers.default_handler import DefaultHandler
from csirtmu.ttp_database.utils.mitre_adapter.mutations import mutation_format
from csirtmu.ttp_database.utils.mitre_adapter.queries import get_techniques_with_attack_id_by_ids_query
from csirtmu.ttp_database.utils.mitre_adapter.relationships_helper import RelationshipsHelper
from csirtmu.ttp_database.utils.mitre_adapter.utils import verbose_print, InitializedObject, \
    create_entities, get_ids_by_names, prepare_mitigations, prepare_procedures, parser, get_tactics_id, \
    procedure_type_technique_column_name_map, divide_post_request


class TechniqueHandler(DefaultHandler):
    def __init__(self, fetched_entities: FetchedEntities, relationships_helper: RelationshipsHelper):
        self.fetched_entities = fetched_entities
        self.relationships_helper = relationships_helper

        super().__init__("attack-pattern", self.seed_techniques, AttackPattern)
    
    def _init_techniques(self, val):
        async def _init():
            external_references = get_ids_by_names([ref.get("url") for ref in val.get("external_references") if ref.get("url")],
                                                   self.fetched_entities.references, "url")
            mitre_attack_id = get_attack_id(val)
            tactic_ids = await get_tactics_id(val.get('kill_chain_phases'), self.fetched_entities.tactics)
            tools_relationships = self.relationships_helper.get_relations_with_related_entities(val.get('id'), 'uses', 'tool',
                                                                           reversed_relation=True)
            malware_relationships = self.relationships_helper.get_relations_with_related_entities(val.get('id'), 'uses', 'malware',
                                                                             reversed_relation=True)
            groups_relationships = self.relationships_helper.get_relations_with_related_entities(val.get('id'), 'uses', 'intrusion-set',
                                                                            reversed_relation=True)

            mititgations_relationships = self.relationships_helper.get_relations_with_related_entities(val.get('id'), 'mitigates',
                                                                                  'course-of-action',
                                                                                  reversed_relation=True)
            mitigations = await prepare_mitigations(self.fetched_entities, mititgations_relationships)
            filtered_platforms = list(
                filter(lambda platform: platform["value"] in val.get('x_mitre_platforms', []),
                       self.fetched_entities.platforms))
            platforms_ids = [platform["id"] for platform in filtered_platforms]

            procedures = await prepare_procedures(self.fetched_entities, tools_relationships, malware_relationships,
                                                  groups_relationships)

            return InitializedObject(base_entity={'techniqueName': val.get("name"),
                                                  'description': val.get("description"),
                                                  'tactics': tactic_ids,
                                                  'platforms': platforms_ids,
                                                  'external_references': external_references},
                                     objects_to_create={"external_references": external_references},
                                     update_after_creation={"procedures": procedures, "mitigations": mitigations},
                                     attack_id=mitre_attack_id)

        return asyncio.run(_init())

    async def seed_techniques(self, techniques):
        verbose_print(f"Seeding techniques...")
        await self.fetched_entities.reload()

        with Pool() as pool:
            output: List[InitializedObject] = pool.map(self._init_techniques, [val for key, val in techniques])

        # updated_output = await update_base_entity_relations(output)
        updated_output = output
        final = {initialized_object.attack_id: initialized_object for initialized_object in updated_output}
        result = await create_entities(
            mutation_format('createTechnique', parser([x.base_entity for x in updated_output])))

        created_techniques_ids = [dict_with_id["id"] for dict_with_id in result["createTechniqueBulk"]]

        query = await get_techniques_with_attack_id_by_ids_query(created_techniques_ids)
        result = await post(query)
        created_techniques = result["getTechniqueList"]["data"]

        for technique in created_techniques:
            technique_id = technique.get("id")
            data = technique.get("externalReferences", {}).get("data")
            attack_id = data[0].get("externalId", None) if data else None
            local_technique = final.get(attack_id)
            if local_technique:
                procedures = local_technique.update_after_creation.get("procedures", {})
                for key, procedure_group in procedures.items():
                    for procedure in procedure_group:
                        # Set technique id for prepared procedures
                        procedure[procedure_type_technique_column_name_map[key]] = technique_id

                mitigations = local_technique.update_after_creation.get("mitigations", {})
                for mitigation in mitigations:
                    mitigation["technique"] = technique_id

        tool_procedures_to_create = []
        malware_procedures_to_create = []
        group_procedures_to_create = []
        technique_mitigation_relationship_to_create = []

        for out in updated_output:
            procedures = out.update_after_creation.get("procedures", {})
            tool_procedures_to_create += procedures.get("tools_procedures", [])
            malware_procedures_to_create += procedures.get("malware_procedures", [])
            group_procedures_to_create += procedures.get("groups_procedures", [])
            technique_mitigation_relationship_to_create += out.update_after_creation.get("mitigations", [])

        await divide_post_request(tool_procedures_to_create, 'createToolProcedure')
        await divide_post_request(malware_procedures_to_create, 'createMalwareProcedure')
        await divide_post_request(group_procedures_to_create, 'createGroupProcedure')
        await divide_post_request(technique_mitigation_relationship_to_create, 'createTechniqueMitigationRelationship')
