import re
from typing import List, Tuple, Any

from deepdiff.model import TreeResult, DiffLevel, PrettyOrderedSet

from csirtmu.ttp_database.utils.mitre_adapter.utils import stix_to_mitre_name

ROOT_REGEX = r"root\[\d*\]"
ONLY_ROOT_REGEX = f"^{ROOT_REGEX}$"  # Matching root[3], root[123] etc.


class ItemDiff:
    diff_tree: TreeResult

    def __init__(self, diff_tree):
        self.diff_tree = diff_tree
        self.added_items = self.get_added_items()
        self.removed_items = self.get_removed_items()
        self.changed_items = self.get_changed_items()

    @staticmethod
    def _is_only_root(item: DiffLevel) -> bool:
        # if regex successfully matches pattern, return True
        return True if re.search(ONLY_ROOT_REGEX, item.path()) else False

    def get_added_items(self) -> List:
        added = self.diff_tree.get('iterable_item_added', PrettyOrderedSet())
        return added.items

    def get_removed_items(self) -> List:
        removed = self.diff_tree.get('iterable_item_removed', PrettyOrderedSet())
        return removed.items

    def get_changed_items(self) -> List:
        values_changed = self.diff_tree.get('values_changed', PrettyOrderedSet())
        # This acts as values_changed from empty value to specific, thus it can be united
        dictionary_item_added = self.diff_tree.get('dictionary_item_added', PrettyOrderedSet())
        return values_changed.union(dictionary_item_added).items

    def classify_added_items(self) -> Tuple[List, List]:
        # iterable_item_added can contain either completely new items or new item of the iterable attribute
        new_items = []
        new_attributes = []
        for item in self.added_items:
            if self._is_only_root(item):
                new_items.append((item.path(), item.t2))
            else:
                new_attributes.append(item)

        return new_items, new_attributes

    def get_base_object_diff(self, item: DiffLevel) -> DiffLevel:
        return item.all_up.down

    def get_atribute_to_change(self, item: DiffLevel) -> Tuple[str, Any]:
        atribute_name = stix_to_mitre_name(item.t2_child_rel.param)
        atribute_value = item.t2_child_rel.child
        return atribute_name, atribute_value
