import asyncio
import time
from functools import update_wrapper
from typing import List, Dict

import click
import requests

from csirtmu.ttp_database.utils.mitre_adapter.client import try_connection
from csirtmu.ttp_database.utils.mitre_adapter.config_handler import get_local_version, set_local_version
from csirtmu.ttp_database.utils.mitre_adapter.update_handler import UpdateHandler
from csirtmu.ttp_database.utils.mitre_adapter.utils import verbose_print

INDEX_URL = "https://raw.githubusercontent.com/mitre-attack/attack-stix-data/master/index.json"
ENTERPRISE_COLLECTION_NAME = "Enterprise ATT&CK"


def get_stix_by_version(version: str, collection):
    filtered = list(filter(lambda version_object: version_object['version'] == version, collection['versions']))
    if filtered:
        version_object = filtered[0]
        return requests.get(version_object['url']).json()
    else:
        return {"objects": {}}


async def update_database(update_handler: UpdateHandler):
    await update_handler.initial_updates()
    await update_handler.handle_tactics()
    await update_handler.handle_mitigations()
    await update_handler.handle_malware()
    await update_handler.handle_tools()
    await update_handler.handle_groups()
    await update_handler.handle_techniques()
    await update_handler.handle_subtechniques()


def main_async(f):
    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(f(*args, **kwargs))
    return update_wrapper(wrapper, f)


def get_enterprise_collection(collections: List[Dict]) -> Dict:
    for collection in collections:
        if collection.get("name") == ENTERPRISE_COLLECTION_NAME:
            return collection

    return {}


def get_latest_version_object(versions: List[Dict]) -> Dict:
    if versions:
        return sorted(versions, key=lambda version_object: float(version_object.get("version")), reverse=True)[0]
    return {}


def validate_version(version: str):
    if float(version) < 7:
        raise Exception("This adapter supports only version greater than 7 (included), due to STIX version 2")


@click.command()
@click.option('-v', '--verbose', is_flag=True, default=False, help='Enables verbose mode')
@click.option('--mode', required=True, type=click.Choice(['update', 'append-only'], case_sensitive=False), help='Select whether reflect all changes or add new objects only')
@click.argument('API_URL')
@click.argument('token')
@main_async
async def main(verbose, mode, api_url, token):
    await try_connection()
    verbose_print("Running MITRE adapter.")
    index = requests.get(INDEX_URL).json()
    enterprise_collection = get_enterprise_collection(index.get("collections", []))
    latest_version = get_latest_version_object(enterprise_collection.get("versions", []))
    latest_version_number = latest_version.get("version")
    validate_version(latest_version_number)
    local_version_number = get_local_version()
    if latest_version_number != local_version_number:
        verbose_print(f"Your version is: {local_version_number}. The latest version is: {latest_version_number}")
        verbose_print("Changes detected. Updating...")
        start_time = time.perf_counter()
        stix_new = get_stix_by_version(latest_version_number, enterprise_collection)
        stix_old = get_stix_by_version(local_version_number, enterprise_collection)
        uh = UpdateHandler(stix_new['objects'], stix_old['objects'])
        verbose_print(f"Initialized memory store in: {time.perf_counter() - start_time} seconds")
        await update_database(uh)
        verbose_print(f"Finished updating in: {time.perf_counter() - start_time} seconds")
        set_local_version(latest_version_number)
    verbose_print(f"Database is up to date. (Version {latest_version_number})")


if __name__ == "__main__":
    main()
