import json
from typing import List, Set

from csirtmu.ttp_database.utils.mitre_adapter.client import post


async def get_total_count(entity_type: str):
    query = f"""
            query {entity_type}ListQuery {{
              get{entity_type}List {{
                totalCount
              }}
            }}
           """
    result = await post(query)
    total_count = result[f'get{entity_type}List']['totalCount']
    return total_count


async def query_part(entity_type: str, fields: List = None):
    if not fields:
        fields = ['id']

    total_count = await get_total_count(entity_type)

    return f"""
              get{entity_type}List(limit: {total_count}) {{
                data {{
                    {" ".join(fields)}
                }}
            }}
        """


async def wrap_query_parts(query_parts: List[str]):
    joined_parts = "\n".join(query_parts)
    query = f"""
            query WrappedQuery{{
                {joined_parts}
            }}
    """
    return await post(query)


async def simple_query(entity_type: str, fields: List = None):
    if not fields:
        fields = ['id', 'name']

    total_count = await get_total_count(entity_type)

    query = f"""
            query {entity_type}ListQuery {{
              get{entity_type}List(limit: {total_count}) {{
                data {{
                    {" ".join(fields)}
                }}
              }}
            }}
           """
    result = await post(query)
    return result[f'get{entity_type}List']['data']


async def filtered_query(entity_type: str, filter_string: str, data_args: Set[str] = None):
    if data_args is None:
        data_args = {'id'}

    total_count = await get_total_count(entity_type)

    query = f"""
            query {entity_type}ListQuery {{
              get{entity_type}List(filter: {{AND: {{OR: {filter_string} }} }}, limit: {total_count}) {{
                data {{
                  {' '.join(data_args)}
                }}
              }}
            }}
           """
    return query


async def filtered_query_by_name(entity_type: str, names: list, arg: str = "name"):
    filter_string = f"{{ {arg}: {{in_:{json.dumps(names)} }} }}"
    data_args = {"id", arg}
    return await filtered_query(entity_type, filter_string, data_args)


async def get_entities_by_attack_id_query(entity_type: str, ids: list):
    filter_string = f'{{externalReferences: {{externalId: {{in_: {json.dumps(ids)} }} }} }}'
    idk = """
          externalReferences(filter: {AND: {OR: {sourceName: {exact: "mitre-attack"}}}}) {
            data {
              externalId
            }
          }
    """
    data_args = {"id", idk}
    return await filtered_query(entity_type, filter_string, data_args)


def get_query(query_name: str, names: List[str]):
    string_names = ",".join(['"{}"'.format(item) for item in names])
    return f"""
            query {query_name}sQuery {{
              get{query_name}List(filter: {{AND: {{OR: {{name: {{in_: [{string_names}]}}}}}}}}) {{
                data {{
                  id
                }}
              }}
            }}
           """


async def get_entity_by_attack_id_query(entity_type, attack_id):
    filter_string = f'{{externalReferences: {{externalId: {{exact: "{attack_id}"}} }} }}'
    return await filtered_query(entity_type, filter_string)


async def get_technique_by_attack_id_query(attack_id):
    return await get_entity_by_attack_id_query("Technique", attack_id)


async def get_tool_by_attack_id_query(attack_id):
    return await get_entity_by_attack_id_query("Tool", attack_id)


async def get_malware_by_attack_id_query(attack_id):
    return await get_entity_by_attack_id_query("Malware", attack_id)


async def get_group_by_attack_id_query(attack_id):
    return await get_entity_by_attack_id_query("TtpGroup", attack_id)


async def get_techniques_with_attack_id_by_ids_query(ids: List[str]):
    ids_string = ", ".join(ids)
    filter_string = f'{{id: {{in_: [{ids_string}]}} }}'
    external_references_filter = '(filter: [ {AND: {OR: {externalId: {isnull: false} }}}, {AND: {OR: {sourceName: {exact: "mitre-attack"} }}} ])'
    data_args = {"id", f'externalReferences{external_references_filter} {{data{{externalId}} }}'}

    return await filtered_query("Technique", filter_string, data_args)


def get_by_id(entity_type: str, id: str, data_args: Set[str] = None):
    if data_args is None:
        data_args = {'id'}
    query = f"""
            query {entity_type}Query {{
              get{entity_type}(id: {id}) {{
                {" ".join(data_args)}
              }}
            }}
           """
    return query


def get_filtered_external_references_by_entity_id(entity_type: str, id: str, source_names: list):
    foo = f"""
        externalReferences(filter: {{AND: {{OR: {{sourceName: {{in_: {json.dumps(source_names)} }} }} }} }}) {{
          data {{
            id
            sourceName
            url
          }}
        }}
    """

    return get_by_id(entity_type, id, {"id", foo})
