from collections import defaultdict
from multiprocessing import Pool
from typing import List, Dict, Optional, Union, Callable

import click
from csirtmu.strawberry_dmt.strawberry_types import RelationOperation
from deepdiff import DeepDiff
from deepdiff.helper import CannotCompare
from stix2 import MemoryStore, Filter

from csirtmu.ttp_database.utils.mitre_adapter.client import post
from csirtmu.ttp_database.utils.mitre_adapter.fetched_entities import FetchedEntities
from csirtmu.ttp_database.utils.mitre_adapter.handlers.default_handler import DefaultHandler
from csirtmu.ttp_database.utils.mitre_adapter.handlers.group_handler import GroupHandler
from csirtmu.ttp_database.utils.mitre_adapter.handlers.mitigation_handler import MitigationHandler
from csirtmu.ttp_database.utils.mitre_adapter.handlers.software_handler import MalwareHandler, \
    ToolHandler
from csirtmu.ttp_database.utils.mitre_adapter.handlers.subtechnique_handler import SubtechniqueHandler
from csirtmu.ttp_database.utils.mitre_adapter.handlers.tactic_handler import TacticHandler
from csirtmu.ttp_database.utils.mitre_adapter.handlers.technique_handler import TechniqueHandler
from csirtmu.ttp_database.utils.mitre_adapter.item_diff import ItemDiff
from csirtmu.ttp_database.utils.mitre_adapter.queries import simple_query, filtered_query_by_name, \
    get_entities_by_attack_id_query, get_filtered_external_references_by_entity_id
from csirtmu.ttp_database.utils.mitre_adapter.relationships_helper import RelationshipsHelper
from csirtmu.ttp_database.utils.mitre_adapter.utils import remove_revoked_and_deprecated, \
    default_filter, to_pascal_case, to_camel_case, is_primitive, update_mutation, stix_to_mitre_name, \
    make_singular_name, ObjectToUpdate, get_attack_id, create_missing_references, create_missing_platforms, \
    create_missing_aliases, delete_entities

ROOT_REGEX = r"root\[\d*\]"
ONLY_ROOT_REGEX = f"^{ROOT_REGEX}$"  # Matching root[3], root[123] etc.


class UpdateHandler:
    @staticmethod
    def _init_memory_store(objects):
        return MemoryStore(stix_data=remove_revoked_and_deprecated(objects))

    def __init__(self, new_objects, old_objects):
        with Pool() as pool:
            output = pool.map(self._init_memory_store, [new_objects, old_objects])
        self.ms_new = output[0]
        self.ms_old = output[1]
        self.fetched_entities = FetchedEntities()

        relationships = default_filter(self.ms_new, "relationship")
        self.relationships_helper = RelationshipsHelper(self.ms_new, relationships)

    @staticmethod
    def compare_func(x, y, level=None):
        try:
            if isinstance(x, Dict) and x.get('id') and isinstance(y, Dict) and y.get('id'):
                return x['id'] == y['id']
            elif isinstance(x, Dict) and x.get('url') and isinstance(y, Dict) and y.get('url'):
                if x.get('source_name') and y.get('source_name'):
                    return x['url'] == y['url'] and x['source_name'] == y['source_name']
                return x['url'] == y['url']
            elif isinstance(x, Dict) and x.get('source_name') and isinstance(y, Dict) and y.get('source_name'):
                return x['source_name'] == y['source_name']
            return x == y
        except Exception:
            raise CannotCompare() from None

    async def update_attributes(self, mitre_name, data_list: Dict[str, ObjectToUpdate]):
        # Fetch entities from DB and assign ids to local objects
        query = await get_entities_by_attack_id_query(to_pascal_case(mitre_name), list(data_list.keys()))
        result = await post(query)
        fetched_entities = result[f'get{to_pascal_case(mitre_name)}List']['data']

        for entity in fetched_entities:
            data = entity.get("externalReferences", {}).get("data")
            attack_id = data[0].get("externalId", None) if data else None
            if attack_id:
                object_to_update = data_list.get(attack_id)
                object_to_update.id = entity.get("id")

        all_keys = set([z for x, y in data_list.items() for z in (y.objects_to_change.keys())])
        for key in all_keys:
            ids = []
            for obj in data_list.values():
                if obj.objects_to_change.get(key):
                    if key == "platforms" or key == "aliases":
                        entities = self.fetched_entities.platforms if key == "platforms" else self.fetched_entities.aliases
                        entities_names = [x.get("value") for x in obj.objects_to_change.get(key)]
                        filtered = list(filter(lambda entity: entity.get("value") in entities_names, entities))
                    elif key == "externalReferences":
                        entities_names = [(x.get("source_name"), x.get("url")) for x in obj.objects_to_change.get(key)]
                        filtered = list(filter(lambda entity:(entity.get("source_name"), entity.get("url")) in entities_names, self.fetched_entities.references))
                    else:
                        entities = await simple_query(to_pascal_case(make_singular_name(key[0].upper() + key[1:])))
                        entities_names = [x.get("name") for x in obj.objects_to_change.get(key)]
                        filtered = list(filter(lambda entity: entity.get("name") in entities_names, entities))
                    ids.append(filtered)
                else:
                    ids.append([])

            for i, obj in enumerate(data_list.values()):
                if obj.objects_to_change.get(key):
                    obj.objects_to_change[key] = {"op": RelationOperation.ADD, "ids": [_id.get('id') for _id in ids[i]]}

        await update_mutation(to_pascal_case(mitre_name), data_list)

    async def _default_handler(self, handled_type: DefaultHandler, filters: Optional[List[Union[Filter, Callable]]] = None):
        if filters is None:
            filters = []
        mitre_name = stix_to_mitre_name(handled_type.diff_name)

        def get_diff():
            old_objects = sorted(default_filter(self.ms_old, handled_type.diff_name, *filters),
                                 key=lambda o: (o.get('created'), o.get('id')))
            new_objects = sorted(default_filter(self.ms_new, handled_type.diff_name, *filters),
                                 key=lambda o: (o.get('created'), o.get('id')))

            old_objects = [handled_type.dataclass_model.from_dict(x).to_dict() for x in old_objects]
            new_objects = [handled_type.dataclass_model.from_dict(x).to_dict() for x in new_objects]

            return DeepDiff(old_objects, new_objects,
                            report_repetition=True,
                            iterable_compare_func=self.compare_func)

        async def handle_additions(item_diff: ItemDiff):
            if item_diff.added_items:
                new_items, new_attributes = item_diff.classify_added_items()
                await handled_type.add_func(new_items)

                # Get item and update(add) atribute to ids of created new_attributes
                data_list = {}
                for item in new_attributes:
                    item_diff = item.all_up.down

                    attack_id = get_attack_id(item_diff.t1)
                    if attack_id:
                        object_to_update = data_list.setdefault(attack_id, ObjectToUpdate(attack_id))
                        attr_name = to_camel_case(stix_to_mitre_name(item_diff.t2_child_rel.param))
                        subitem = object_to_update.objects_to_change.setdefault(attr_name, [])
                        item_to_append = item.t2
                        if is_primitive(item_to_append):
                            item_to_append = {"value": item_to_append}
                        subitem.append(item_to_append)
                if data_list:
                    await self.update_attributes(mitre_name, data_list)

        async def handle_deletions(item_diff: ItemDiff):
            data_list: Dict[str, ObjectToUpdate] = {}
            delete_completly = defaultdict(dict)
            for item in item_diff.removed_items:
                base_object_diff = item_diff.get_base_object_diff(item)
                entity_attack_id = get_attack_id(base_object_diff.t1)
                object_to_update = data_list.setdefault(entity_attack_id, ObjectToUpdate(entity_attack_id))

                if item == base_object_diff:
                    delete_completly[entity_attack_id] = item.t1
                    continue
                attribute_name, attribute_value = item_diff.get_atribute_to_change(base_object_diff)
                if is_primitive(attribute_value):
                    object_to_update.objects_to_change.setdefault(to_camel_case(attribute_name), attribute_value)
                # We need to cut relationship
                else:
                    obj = object_to_update.objects_to_change.setdefault(to_camel_case(attribute_name), [])
                    obj.append(item.t1)

            if delete_completly:
                await delete_entities(delete_completly.items(), to_pascal_case(mitre_name))

            if data_list:
                query = await get_entities_by_attack_id_query(to_pascal_case(mitre_name), list(data_list.keys()))
                result = await post(query)
                fetched_entities = result[f'get{to_pascal_case(mitre_name)}List']['data']

                platforms = await simple_query("TtpPlatform", fields=["id", "value"])
                aliases = await simple_query("Alias", fields=["id", "value"])

                for entity in fetched_entities:
                    data = entity.get("externalReferences", {}).get("data")
                    attack_id = data[0].get("externalId", None) if data else None
                    if attack_id:
                        object_to_update = data_list[attack_id]
                        object_to_update.id = entity.get("id")

                for attack_id, obj_to_update in data_list.items():
                    for k, v in obj_to_update.objects_to_change.items():
                        if k == "platforms":
                            platforms_to_delete_ids = [x.get("id") for x in platforms if x.get("value") in v]
                            obj_to_update.objects_to_change[k] = {"op": RelationOperation.REM, "ids": platforms_to_delete_ids}


                        elif k == "aliases":
                            aliases_to_delete_ids = [x.get("id") for x in aliases if x.get("value") in v]
                            obj_to_update.objects_to_change[k] = {"op": RelationOperation.REM, "ids": aliases_to_delete_ids}

                        else:
                            erfs_source_names = [x.get("source_name") for x in v]
                            query = get_filtered_external_references_by_entity_id(to_pascal_case(mitre_name), obj_to_update.id, erfs_source_names)
                            result = await post(query)
                            external_references = result[f'get{to_pascal_case(mitre_name)}']['externalReferences']["data"]

                            external_references_to_delete_ids = [x.get("id") for x in external_references]
                            obj_to_update.objects_to_change[k] = {"op": RelationOperation.REM, "ids": external_references_to_delete_ids}

                await update_mutation(to_pascal_case(mitre_name), data_list)

        async def handle_updates(item_diff: ItemDiff):
            data_list: Dict[str, ObjectToUpdate] = {}
            non_related_updates = defaultdict(dict)
            for item in item_diff.changed_items:
                base_object_diff = item_diff.get_base_object_diff(item)
                attribute_name, attribute_value = item_diff.get_atribute_to_change(base_object_diff)
                if is_primitive(attribute_value):
                    entity_attack_id = get_attack_id(base_object_diff.t1)
                    object_to_update = data_list.setdefault(entity_attack_id, ObjectToUpdate(entity_attack_id))
                    object_to_update.objects_to_change.setdefault(to_camel_case(attribute_name), attribute_value)
                else:
                    entity_name = item.up.t1.get("source_name", None)
                    object_to_update = non_related_updates[attribute_name].setdefault(entity_name, ObjectToUpdate(entity_name))
                    attribute_name, attribute_value = item_diff.get_atribute_to_change(item.up)
                    object_to_update.objects_to_change.setdefault(to_camel_case(attribute_name), attribute_value)

            if data_list:
                query = await get_entities_by_attack_id_query(to_pascal_case(mitre_name), list(data_list.keys()))
                result = await post(query)
                fetched_entities = result[f'get{to_pascal_case(mitre_name)}List']['data']

                for entity in fetched_entities:
                    data = entity.get("externalReferences", {}).get("data")
                    attack_id = data[0].get("externalId", None) if data else None
                    if attack_id:
                        object_to_update = data_list[attack_id]
                        object_to_update.id = entity.get("id")

                await update_mutation(to_pascal_case(mitre_name), data_list)

            if non_related_updates:
                for k, v in non_related_updates.items():
                    k = make_singular_name(k)
                    query = await filtered_query_by_name(to_pascal_case(k), list(v.keys()), arg="sourceName")
                    result = await post(query)
                    fetched_entities = result[f'get{to_pascal_case(k)}List']['data']

                    for entity in fetched_entities:
                        object_to_update = v[entity.get("sourceName")]
                        object_to_update.id = entity.get("id")

                    await update_mutation(to_pascal_case(k), v)

        diff = get_diff()
        item_diff = ItemDiff(diff.tree)
        await handle_additions(item_diff)
        update_mode = click.get_current_context().params.get("mode")
        if update_mode == "update":
            await handle_deletions(item_diff)
            await handle_updates(item_diff)

    async def handle_tactics(self):
        await self._default_handler(TacticHandler(self.fetched_entities))

    async def handle_techniques(self):
        # There's a bug in MITRE ATT&CK 11.0 which doesn't allow to use object Filter hence we need to specify lambda
        filters = [lambda x: not x.get("x_mitre_is_subtechnique")]
        await self._default_handler(TechniqueHandler(self.fetched_entities, self.relationships_helper), filters)

    async def handle_subtechniques(self):
        filters = [Filter("x_mitre_is_subtechnique", "=", True)]
        await self._default_handler(SubtechniqueHandler(self.fetched_entities, self.relationships_helper), filters)

    async def handle_mitigations(self):
        await self._default_handler(MitigationHandler(self.fetched_entities))

    async def handle_malware(self):
        await self._default_handler(MalwareHandler(self.fetched_entities))

    async def handle_tools(self):
        await self._default_handler(ToolHandler(self.fetched_entities))

    async def handle_groups(self):
        await self._default_handler(GroupHandler(self.fetched_entities, self.relationships_helper))

    async def initial_updates(self):
        all_objects = self.ms_new.query()
        await create_missing_references(all_objects)
        await create_missing_platforms(all_objects)
        await create_missing_aliases(all_objects)
        await self.fetched_entities.reload()
