import enum
import json
from dataclasses import dataclass
from typing import List, Dict, Set, Tuple, Callable

import click
from dataclasses_json import dataclass_json
from stix2 import MemoryStore, Filter, KillChainPhase
from stix2.v20 import Relationship, _DomainObject

from csirtmu.ttp_database.utils.mitre_adapter.client import post
from csirtmu.ttp_database.utils.mitre_adapter.dataclasses_models import ExternalReference
from csirtmu.ttp_database.utils.mitre_adapter.enums import MitreStixNames
from csirtmu.ttp_database.utils.mitre_adapter.fetched_entities import FetchedEntities
from csirtmu.ttp_database.utils.mitre_adapter.mutations import mutation_format, delete_mutation_format
from csirtmu.ttp_database.utils.mitre_adapter.queries import filtered_query_by_name, simple_query, \
    get_entities_by_attack_id_query


class ObjectToUpdate:
    def __init__(self, attack_id):
        self.attack_id = attack_id
        self.objects_to_change = {}
        self.id = None


@dataclass_json
@dataclass
class InitializedObject:
    base_entity: Dict
    objects_to_create: Dict[str, List[Dict]] = None
    update_after_creation: Dict[str, List[Dict]] = None
    attack_id: str = None


def to_camel_case(snake_str):
    components = snake_str.split('_')
    return components[0] + ''.join(x.title() for x in components[1:])


def to_pascal_case(snake_str):
    return ''.join(snake_str.title().split('_'))


def make_singular_name(plural_name: str):
    if plural_name.endswith('ses'):
        return plural_name[:-2]
    return plural_name[:-1]


# There is an error in 7.0
# Malware Twitoor is marked as enterprise-attack but in fact it is mobile-attack
# But the external reference has correct domain (mobile-attack)
# Thus it needs to be manually excluded
EXCLUDED_OBJECT_IDS = [
    "malware--41e3fd01-7b83-471f-835d-d2b1dc9a770c"
]

procedure_type_technique_column_name_map = {
    "tools_procedures": "tool_technique",
    "malware_procedures": "malware_technique",
    "groups_procedures": "group_technique",
}


def remove_revoked_and_deprecated(objects):
    return list(filter(lambda obj: not obj.get("x_mitre_deprecated") and
                                   not obj.get("revoked") and
                                   "enterprise-attack" in obj.get("x_mitre_domains", [])
                                   and obj.get("id") not in EXCLUDED_OBJECT_IDS,
                       objects))


def default_filter(ms: MemoryStore, type_: str = None, *args, **kwargs):
    basic_filters: List[Filter] = [arg for arg in args if isinstance(arg, Filter)]
    lambda_filters: List[Callable] = [arg for arg in args if isinstance(arg, Callable)]
    if type_:
        basic_filters.append(Filter("type", "=", type_))
    query_result = ms.query(basic_filters)
    result = [obj for obj in query_result if all(map(lambda filter_: filter_(obj), lambda_filters))]
    return result


async def create_entities(mutation: str):
    return await post(mutation)


def parse_dict(dict: Dict):
    result = []
    for k, v in dict.items():
        if isinstance(v, Dict):
            result.append(f'{to_camel_case(k)}: {parse_dict(v)}')
        else:
            if isinstance(v, enum.Enum):
                result.append(f'{to_camel_case(k)}: {v.name}')
            else:
                result.append(f'{to_camel_case(k)}: {json.dumps(v)}')
    return f'{{{",".join(result)}}}'


def parser(field: List[Dict]):
    result = []
    for x in field:
        result.append(parse_dict(x))
    return f'[{",".join(result)}]'


def prepare_external_references(list_of_references):
    references = []
    if list_of_references:
        for ref in list_of_references:
            references.append(ExternalReference.from_dict(ref).to_dict())
    return references


def get_attack_id(entity: _DomainObject):
    external_references = entity.get("external_references", {})
    filtered = list(filter(lambda er: er.get("source_name", "") == "mitre-attack", external_references))
    if filtered:
        return filtered[0].get("external_id", None)
    else:
        return None


async def prepare_mitigations(fetched_entities: FetchedEntities, mititgations_relationships):
    datalist = []
    for mitigation_relationship, mitigation in mititgations_relationships:
        mitigation_attack_id = get_attack_id(mitigation)
        fetched_mitigation = {}
        for fm in fetched_entities.mitigations:
            data = fm.get("externalReferences", {}).get("data")
            if data and data[0].get("externalId") == mitigation_attack_id:
                fetched_mitigation = fm
                break

        if fetched_mitigation:
            mitigation_id = fetched_mitigation.get("id")
            external_references = get_ids_by_names(
                [ref.get("url") for ref in mitigation_relationship.get("external_references", [])],
                fetched_entities.references, "url")
            procedure = {"description": mitigation_relationship.get("description", ""), "mitigation": mitigation_id, "externalReferences": external_references}
            datalist.append(procedure)

    return datalist


async def prepare_procedures(fetched_entities,
                             tools_relationships: List[Tuple[Relationship, _DomainObject]],
                             malware_relationships: List[Tuple[Relationship, _DomainObject]],
                             groups_relationships: List[Tuple[Relationship, _DomainObject]]):
    datalist = {
        "tools_procedures": [],
        "malware_procedures": [],
        "groups_procedures": [],
    }

    def _common(entity_relationships: List[Tuple[Relationship, _DomainObject]], datalist_procedures: List, fetched_e, name: str):
        for entity_relationship, entity in entity_relationships:
            entity_attack_id = get_attack_id(entity)

            fetched_entity = None
            for x in fetched_e:
                external_references = x.get("externalReferences", {})
                data = external_references.get("data", [{}])
                first = data[0] if data else {}
                eid = first.get("externalId")
                if eid == entity_attack_id:
                    fetched_entity = x
                    break
            if fetched_entity:
                entity_id = fetched_entity["id"]
                external_references = get_ids_by_names([ref.get("url") for ref in entity_relationship.get("external_references", [])],
                                                       fetched_entities.references, "url")
                procedure = {"description": entity_relationship.get("description", ""), name: entity_id, "externalReferences": external_references}
                datalist_procedures.append(procedure)

    _common(tools_relationships, datalist["tools_procedures"], fetched_entities.tools, "tool")
    _common(malware_relationships, datalist["malware_procedures"], fetched_entities.malware, "malware")
    _common(groups_relationships, datalist["groups_procedures"], fetched_entities.groups, "group")

    return datalist


def related_names(related_entities: List):
    return [related.name for related in related_entities]


async def get_tactics_id(kill_chain_phases: List[KillChainPhase], fetched_tactics):
    tactics_names = [phase.get('phase_name') for phase in kill_chain_phases]
    filtered = list(
        filter(lambda tactic: tactic["tacticName"].replace(' ', '-').lower() in tactics_names, fetched_tactics))
    return [tactic["id"] for tactic in filtered]


async def get_platforms_id(platforms_names):
    query = await filtered_query_by_name("TtpPlatform", platforms_names, "value")
    result = await post(query)
    platforms = result[f'getTtpPlatformList']['data']
    return [platform["id"] for platform in platforms]


async def create_aliases(aliases: List):
    if not aliases:
        return []
    result = []
    for alias in aliases:
        result.append({'value': alias})
    result = await create_entities(mutation_format('createAlias', parser(result)))
    return [x['id'] for x in result['createAliasBulk']]


def get_ids_by_names(names: List[str], entities, name_arg="name"):
    filtered = list(
        filter(lambda entity: entity[name_arg] in names, entities))
    return [entity["id"] for entity in filtered]


async def delete_entities(to_remove, type):
    attack_ids = [attack_id for attack_id, entity in to_remove]
    query = await get_entities_by_attack_id_query(type, attack_ids)
    result = await post(query)
    ids = [x["id"] for x in result[f'get{type}List']['data']]
    await post(delete_mutation_format(type, json.dumps(ids)))


def is_primitive(obj):
    return isinstance(obj, (int, str, bool, float))


async def update_mutation(type: str, data_list: Dict[str, ObjectToUpdate]):
    updates = []
    for x in data_list.values():
        parsed_data = parse_dict(x.objects_to_change)
        updates.append(f"""
                         update{type}{x.id}: update{type}(id: {x.id}, data: {parsed_data}) {{
                          id
                        }}
                        """)
    str = f'{",".join(updates)}'
    mutation = f"""mutation {type}UpdateMutation {{{str}}}"""

    await post(mutation)


def stix_to_mitre_name(name: str):
    try:
        return MitreStixNames(name).name.lower()
    except ValueError:
        return name


def verbose_print(text_to_print):
    if click.get_current_context().params.get("verbose"):
        print(text_to_print)


async def get_or_create_multiple(to_get: Set, fetched_entities: Set):
    if set_difference := to_get.difference(fetched_entities):
        await create_entities(mutation_format('createAlias', parser([{"value": value} for value in set_difference])))


async def create_missing_aliases(entities):
    fetched_aliases_names = set(
        [fetched_alias.get('value') for fetched_alias in await simple_query('Alias', fields=['id', 'value'])])
    aliases_in_memory_store = set()

    for val in entities:
        aliases = set(val.get('aliases', []) + val.get('x_mitre_aliases', []))
        aliases.discard(val.get('name'))  # This removes official name from aliases
        aliases_in_memory_store.update(aliases)

    if set_difference := aliases_in_memory_store.difference(fetched_aliases_names):
        await create_entities(
            mutation_format('createAlias', parser([{'value': value} for value in set_difference])))


async def create_missing_platforms(entities):
    fetched_platforms = set([d.get('value') for d in await simple_query('TtpPlatform', fields=['id', 'value'])])
    platforms_in_memory_store = set()

    for val in entities:
        platforms = set(val.get('x_mitre_platforms', []) + val.get('platforms', []))
        platforms_in_memory_store.update(platforms)

    if set_difference := platforms_in_memory_store.difference(fetched_platforms):
        await create_entities(
            mutation_format('createTtpPlatform', parser([{'value': value} for value in set_difference])))


async def create_missing_references(entities):
    already_present = set([(er.get("sourceName"), er.get('url')) for er in await simple_query('ExternalReference', fields=['id', 'url', 'sourceName'])])
    to_create = []

    for val in entities:
        references = prepare_external_references([er for er in val.get('external_references', []) if (er.get("source_name"), er.get('url')) not in already_present])
        already_present.update([(er.get("source_name"), er.get('url')) for er in references])
        to_create += references

    await divide_post_request(to_create, 'createExternalReference')


async def divide_post_request(to_create: List, mutation_name: str):
    if to_create:
        # This piece of code is needed due to deployed instance in k8s
        # where request fails because of its size, hence it is divided
        request_items = 1000
        for i in range((len(to_create) // request_items) + 1):
            bottom_index = request_items * i
            top_index = request_items * (i + 1)
            to_create_part = to_create[bottom_index:top_index]
            await create_entities(mutation_format(mutation_name, parser(to_create_part)))