import asyncio
import os
from pathlib import Path
from typing import Generator

import alembic.config
import pytest
import pytest_asyncio
from csirtmu.strawberry_dmt.authorization import CapableUser
from csirtmu.strawberry_dmt.sqla_base import Base
from sqlalchemy import text
from sqlalchemy.orm import close_all_sessions, sessionmaker
from sqlalchemy_oso import SQLAlchemyOso
from starlette.authentication import UnauthenticatedUser
from starlette.testclient import TestClient

from csirtmu.ttp_database import __version__, app
from csirtmu.ttp_database.config import _load_config
from csirtmu.ttp_database.models import RelationUser
from csirtmu.ttp_database.views import AppState

BASE_DIR = Path(__file__).resolve().parent
ASSET_DIR = BASE_DIR / 'assets'
CONFIG = _load_config(BASE_DIR / "config/config-test.yaml")


def test_version():
    assert __version__ == '0.1.0'


BASE_DIR = Path(__file__).resolve().parent


@pytest_asyncio.fixture(autouse=True, scope='function')
async def migrate_tables():
    try:
        with CONFIG.database.engine.connect() as conn:
            conn.execute(text('DROP SCHEMA IF EXISTS "public" CASCADE;'))
            conn.commit()
    except Exception:
        pass
    try:
        cwd = os.getcwd()
        os.chdir(str(BASE_DIR.parent))
        print("Creating tables")
        with CONFIG.database.engine.connect() as conn:
            conn.execute(text('CREATE SCHEMA "public";'))
            conn.commit()
        alembic.config.main(['upgrade', 'head'])
        os.chdir(cwd)
    finally:
        yield
        close_all_sessions()
        with CONFIG.database.engine.connect() as conn:
            conn.execute(text('DROP SCHEMA IF EXISTS "public" CASCADE;'))
            conn.commit()
            conn.execute(text('CREATE SCHEMA "public";'))
            conn.commit()


@pytest.fixture(scope="session")
def strict_oso():
    oso = SQLAlchemyOso(Base)
    oso.register_class(Base)
    oso.register_class(CapableUser, name='User')
    oso.register_class(RelationUser, name='RelationUser')
    oso.load_files([(BASE_DIR / 'config/authorization-test-strict.polar').resolve().__str__()])
    return oso


@pytest.fixture(scope="session")
def oso():
    oso = SQLAlchemyOso(Base)
    oso.register_class(Base)
    oso.register_class(CapableUser, name='User')
    oso.register_class(RelationUser, name='RelationUser')
    oso.register_class(UnauthenticatedUser, name='UnauthenticatedUser')
    oso.load_files([(BASE_DIR / 'config/authorization-test.polar').resolve().__str__()])
    return oso


@pytest_asyncio.fixture(scope="session")
async def client(oso, event_loop):
    return TestClient(app.asgi_app(oso, AppState()))


@pytest_asyncio.fixture(scope="session")
async def client_strict_auth(strict_oso, event_loop):
    star_app = app.asgi_app(oso=strict_oso, state=AppState())
    with TestClient(app=star_app) as c:
        yield c


@pytest_asyncio.fixture(scope='function', autouse=True)
async def session(migrate_tables):
    session_factory = sessionmaker(CONFIG.database.engine, future=True)
    with session_factory() as s:
        yield s
    CONFIG.database.engine.dispose()
    await CONFIG.database.async_engine.dispose()


@pytest_asyncio.fixture
async def load_data(client):
    with open(ASSET_DIR / 'mock.graphql', 'r') as f:
        query = f.read()

    response = client.post('/graphql', json=dict(query=query, operationName='loadData'))

    data = response.json()


@pytest.fixture(scope="session", autouse=True)
def event_loop() -> Generator:
    loop = asyncio.get_event_loop_policy().new_event_loop()
    asyncio.set_event_loop(loop)
    yield loop
    loop.close()
