from csirtmu.ttp_database import models
from tests.conftest import BASE_DIR


class TestTTPService:
    def test_query_interface_type(self, client, session):
        tactic1 = models.Tactic(tactic_name='Initial Access')
        tactic2 = models.Tactic(tactic_name='Defense Evasion')
        technique1 = models.Technique(technique_name='Phishing')
        technique2 = models.Technique(technique_name='Valid Accounts')
        subtechnique1 = models.Technique(technique_name='Spearphishing Attachment', sub_technique_of=technique1)
        subtechnique2 = models.Technique(technique_name='Default Accounts', sub_technique_of=technique2)
        subtechnique3 = models.Technique(technique_name='Domain Accounts', sub_technique_of=technique2)

        session.add_all([tactic1, tactic2, technique1, technique2, subtechnique1, subtechnique2, subtechnique3])
        session.commit()

        with open(BASE_DIR / 'assets/queries.graphql', 'r') as f:
            query = f.read()

        response = client.post('/graphql', json=dict(query=query, operationName='getTTP'))
        assert response.status_code == 200

        result = response.json()
        assert result.get('errors') is None

        assert result['data']['getIttpList']['totalCount'] == 7
        assert result['data']['getIttpList']['data'][0]['tacticName'] == tactic1.tactic_name
        assert result['data']['getIttpList']['data'][3]['techniqueName'] == technique2.technique_name

        assert result['data']['getTacticList']['totalCount'] == 2
        assert result['data']['getTacticList']['data'][0]['tacticName'] == tactic1.tactic_name
        assert result['data']['getTacticList']['data'][1]['id'] == str(tactic2.id)

        assert result['data']['getTechniqueList']['totalCount'] == 2
        assert result['data']['getTechniqueList']['data'][0]['techniqueName'] == technique1.technique_name
        assert result['data']['getTechniqueList']['data'][1]['id'] == str(technique2.id)

        assert result['data']['getSubTechniqueList']['totalCount'] == 3
        assert result['data']['getSubTechniqueList']['data'][0]['techniqueName'] == subtechnique1.technique_name
        assert result['data']['getSubTechniqueList']['data'][1]['id'] == str(subtechnique2.id)
        assert result['data']['getSubTechniqueList']['data'][2]['ttpType'] == subtechnique3.ttp_type

    def test_query_ttp_relationships(self, client, session):
        tactic1 = models.Tactic(tactic_name='Initial Access')
        tactic2 = models.Tactic(tactic_name='Defense Evasion')
        technique1 = models.Technique(technique_name='Phishing', tactics=[tactic1])
        technique2 = models.Technique(technique_name='Valid Accounts', tactics=[tactic1, tactic2])
        subtechnique1 = models.Technique(technique_name='Spearphishing Attachment', sub_technique_of=technique1)
        subtechnique2 = models.Technique(technique_name='Default Accounts', sub_technique_of=technique2)
        subtechnique3 = models.Technique(technique_name='Domain Accounts', sub_technique_of=technique2)

        session.add_all([tactic1, tactic2, technique1, technique2, subtechnique1, subtechnique2, subtechnique3])
        session.commit()

        with open(BASE_DIR / 'assets/queries.graphql', 'r') as f:
            query = f.read()

        response = client.post('/graphql', json=dict(query=query, operationName='TTPRelationships'))
        assert response.status_code == 200

        result = response.json()
        assert result.get('errors') is None

        assert result['data']['getTacticList']['data'][0]['techniques']['totalCount'] == 2
        assert result['data']['getTacticList']['data'][1]['techniques']['totalCount'] == 1
        assert result['data']['getTacticList']['data'][0]['techniques']['data'][0]['techniqueName'] == 'Phishing'
        assert result['data']['getTacticList']['data'][0]['techniques']['data'][1]['subTechniques']['totalCount'] == 2
        assert result['data']['getTacticList']['data'][0]['techniques']['data'][1]['subTechniques']['data'][0]['techniqueName'] == 'Default Accounts'

        assert result['data']['getSubTechniqueList']['data'][2]['subTechniqueOf']['techniqueName'] == 'Valid Accounts'
        assert result['data']['getSubTechniqueList']['data'][2]['subTechniqueOf']['tactics']['totalCount'] == 2
        assert result['data']['getSubTechniqueList']['data'][2]['subTechniqueOf']['tactics']['data'][0]['tacticName'] == 'Initial Access'

    def test_query_ttp_with_inline_fragments(self, client, session):
        tactic1 = models.Tactic(tactic_name='Initial Access')
        technique1 = models.Technique(technique_name='Phishing', tactics=[tactic1])
        group1 = models.Group(name='Axiom')
        groupProcedure1 = models.GroupProcedure(group_technique=technique1, group=group1)

        session.add_all([tactic1, technique1, group1, groupProcedure1])
        session.commit()

        with open(BASE_DIR / 'assets/queries.graphql', 'r') as f:
            query = f.read()

        response = client.post('/graphql', json=dict(query=query, operationName='TTPWithInlineFragments'))
        assert response.status_code == 200

        result = response.json()
        assert result.get('errors') is None

        assert result['data']['getIttpList']['totalCount'] == 3
        assert 'groupProcedureName' in result['data']['getIttpList']['data'][2]
        assert result['data']['getIttpList']['data'][2]['group']['id'] == str(group1.id)
        assert 'tacticName' in result['data']['getIttpList']['data'][0]
        assert result['data']['getIttpList']['data'][0]['techniques']['totalCount'] == 1
        assert 'created' in result['data']['getIttpList']['data'][1]
        assert result['data']['getIttpList']['data'][1]['subTechniques']['totalCount'] == 0

