from typing import Union, Optional, Any, Dict

from csirtmu.starlette_jwt_auth.models import MappedUser
from csirtmu.starlette_jwt_auth.routes import auth_backend_version_route
from csirtmu.starlette_jwt_auth.token_authentication_backend import TokenAuthenticationBackend, on_auth_error
from oso import Oso
from starlette.applications import Starlette
from starlette.authentication import UnauthenticatedUser
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.middleware.cors import CORSMiddleware
from starlette.requests import Request
from starlette.responses import Response
from starlette.websockets import WebSocket
from strawberry.asgi import GraphQL
from strawberry.schema import BaseSchema

from csirtmu.user_profile_introspection_service.config import BASE_DIR
from csirtmu.user_profile_introspection_service.keycloak_client.keycloak_client import KeycloakClient
from csirtmu.user_profile_introspection_service.schema import schema
from csirtmu.user_profile_introspection_service.views import healthcheck_liveness, healthcheck_readiness
from starlette_prometheus import metrics


class GraphQLCustomASGIApp(GraphQL):
    def __init__(self, schema: BaseSchema, graphiql: bool = True, keep_alive: bool = False,
                 keep_alive_interval: float = 1, debug: bool = False, oso: Oso = None) -> None:
        super().__init__(schema, graphiql, keep_alive=keep_alive, keep_alive_interval=keep_alive_interval, debug=debug)
        self.oso = oso

    async def get_context(self, request: Union[Request, WebSocket], response: Optional[Response] = None) \
            -> Optional[Any]:
        context: Dict = await super().get_context(request, response)
        context['oso'] = self.oso
        return context


def app() -> Starlette:
    _app = Starlette(routes=[auth_backend_version_route])
    _app.add_middleware(AuthenticationMiddleware, backend=TokenAuthenticationBackend(), on_error=on_auth_error)
    _app.add_middleware(
        CORSMiddleware, allow_headers=["*"], allow_origins=["*"], allow_methods=["*"]
    )

    oso = Oso()
    oso.register_class(MappedUser, name='User')
    oso.register_class(UnauthenticatedUser)
    oso.register_class(KeycloakClient)
    oso.load_files([(BASE_DIR / 'config/authorization.polar').resolve().__str__()])

    _app.add_route(path='/healthcheck/liveness', route=healthcheck_liveness, methods=['GET'])
    _app.add_route(path='/healthcheck/readiness', route=healthcheck_readiness, methods=['GET'])
    _app.add_route(path='/metrics', route=metrics)

    _app.add_route(auth_backend_version_route.path, auth_backend_version_route.endpoint)

    graphql_app = GraphQLCustomASGIApp(schema, oso=oso)

    for route in ['/', '/graphql']:
        _app.add_route(route, graphql_app)
        _app.add_websocket_route(route, graphql_app)
    return _app
