import datetime
from typing import List, Optional

import httpx
from inflection import camelize, underscore

from csirtmu.user_profile_introspection_service.schema_models import UserInfoSearchInput, User


class KeycloakClient:

    def __init__(self, base_url: str, realm: str, username: str, password: str, client_id: str = 'admin-cli'):
        self.base_url = base_url
        self.client_id = client_id
        self.realm = realm
        self.username = username
        self.password = password
        self._auth_token = None
        self._auth_token_expiration = datetime.datetime.fromtimestamp(0)

    def _refresh_token(self):
        response = httpx.post(
            f'{self.base_url}/realms/{self.realm}/protocol/openid-connect/token',
            data={
                'client_id': self.client_id,
                'username': self.username,
                'password': self.password,
                'grant_type': 'password'
            })
        data = response.json()
        # TODO: This is somewhat ugly buy so is Keycloak's flow
        self._auth_token_expiration = datetime.datetime.now() + datetime.timedelta(seconds=data['expires_in'] - 1)
        self._auth_token = data['access_token']

    @property
    def auth_token(self):
        if datetime.datetime.now() >= self._auth_token_expiration:
            self._refresh_token()
        return self._auth_token

    async def search(self, search_input: UserInfoSearchInput) -> List[User]:
        parsed_params = {camelize(k, False): v for k, v in search_input.__dict__.items()
                         if v is not None}

        async with httpx.AsyncClient(headers={'Authorization': f'Bearer {self.auth_token}'}) as client:
            response = await client.get(
                f'{self.base_url}/admin/realms/{search_input.realm}/users',
                params=parsed_params)
        data = response.json()
        result = []
        for user_record in data:
            result.append(type('User', (object,), {underscore(k): v for k, v in user_record.items()})())
        return result

    async def get_user_by_id(self, user_id: str, realm: str) -> Optional[User]:
        async with httpx.AsyncClient(headers={'Authorization': f'Bearer {self.auth_token}'}) as client:
            response = await client.get(f'{self.base_url}/admin/realms/{realm}/users/{user_id}')
        data = response.json()

        if 'error' in data:
            return None
        return type('User', (object,), {underscore(k): v for k, v in data.items()})()
