import json
from typing import Optional, List

import requests
from requests import RequestException

from csirtmu.auth_keycloak_test_utils.utils import refresh_access_token, \
    get_auth_endpoint_for_realm, get_tokens_for_user


# This will save your life
# https://documenter.getpostman.com/view/7294517/SzmfZHnd#intro
class KeycloakTestClient:
    host: str
    port: str

    _master_user: str
    _master_pass: str

    master_token: str
    _refresh_token: str

    KEYCLOAK_MAIN_REALM = 'master'

    _public_url: str

    def __init__(self, host, port, master_user, master_pass, public_url: str = None, legacy: bool = False):
        self.host = host
        self.port = port
        self._master_user = master_user
        self._master_pass = master_pass
        self._public_url = public_url
        if self._public_url is None:
            self._public_url = f'http://{host}:{port}'

        self.legacy = legacy

        main_realm_auth_endpoint = get_auth_endpoint_for_realm(self.endpoint, self.KEYCLOAK_MAIN_REALM,
                                                               legacy=self.legacy)
        self.master_token, self._refresh_token = get_tokens_for_user(realm_auth_endpoint=main_realm_auth_endpoint,
                                                                     username=self._master_user,
                                                                     password=self._master_pass,
                                                                     client_id='admin-cli')

    @property
    def endpoint(self) -> str:
        legacy_auth_suffix = '/auth' if self.legacy else ''
        return f'http://{self.host}:{self.port}{legacy_auth_suffix}'

    @property
    def auth_header(self) -> dict:
        endpoint = get_auth_endpoint_for_realm(endpoint=self.endpoint, realm=self.KEYCLOAK_MAIN_REALM,
                                               legacy=self.legacy)
        try:
            self.master_token = refresh_access_token(realm_auth_endpoint=endpoint, refresh_token=self._refresh_token,
                                                     client_id='admin-cli')
        except RequestException:  # If refresh token is expired
            self.master_token, self._refresh_token = get_tokens_for_user(realm_auth_endpoint=endpoint,
                                                                         username=self._master_user,
                                                                         password=self._master_pass,
                                                                         client_id='admin-cli')
        return {'Authorization': f'Bearer {self.master_token}'}

    def get_current_user_info(self, realm: str) -> dict:
        endpoint = f'{self.endpoint}/realms/{realm}/protocol/openid-connect/userinfo'
        response = requests.get(endpoint, headers=self.auth_header)
        response.raise_for_status()
        return response.json()

    def create_realm(self, realm: str) -> bool:
        endpoint = f'{self.endpoint}/admin/realms'
        response = requests.post(endpoint, headers=self.auth_header, json={
            'realm': realm,
            'displayName': 'CSIRT MU Test realm',
            'enabled': True,
            'sslRequired': 'external',
            'registrationAllowed': True,
            'loginWithEmailAllowed': True,
            'resetPasswordAllowed': True,
            'editUsernameAllowed': False,
            'bruteForceProtected': True,
            'attributes': {
                "cibaBackchannelTokenDeliveryMode": "poll",
                "cibaAuthRequestedUserHint": "login_hint",
                "oauth2DevicePollingInterval": "5",
                "clientOfflineSessionMaxLifespan": "0",
                "clientSessionIdleTimeout": "0",
                "userProfileEnabled": "false",
                "clientOfflineSessionIdleTimeout": "0",
                "cibaInterval": "5",
                "cibaExpiresIn": "120",
                "oauth2DeviceCodeLifespan": "600",
                "parRequestUriLifespan": "60",
                "clientSessionMaxLifespan": "0",
                "frontendUrl": self._public_url
            }
        })
        response.raise_for_status()
        return True

    def create_client(self, realm: str, client: str, client_secret: str) -> bool:
        endpoint = f'{self.endpoint}/admin/realms/{realm}/clients'
        response = requests.post(endpoint, headers=self.auth_header, json={
            'clientId': client,
            'enabled': True,
            'protocol': 'openid-connect',
            'secret': client_secret
        })
        response.raise_for_status()
        return True

    def set_redirect_uris_for_client(self, realm: str, client: str, redirect_uris: List[str]) -> bool:
        endpoint = f'{self.endpoint}/admin/realms/{realm}/clients/{client}'
        response = requests.put(endpoint, headers=self.auth_header, json={'redirectUris': redirect_uris})
        return response.status_code == 204

    def set_post_logout_redirect_uris_for_client(self, realm: str, client: str, redirect_uris: List[str]) -> bool:
        endpoint = f'{self.endpoint}/admin/realms/{realm}/clients/{client}'
        response = requests.put(endpoint, headers=self.auth_header, json={
            'attributes': {
                'post.logout.redirect.uris': '##'.join(redirect_uris)
            }
        })
        return response.status_code == 204

    def set_web_origins_for_client(self, realm: str, client: str, web_origins: List[str]) -> bool:
        endpoint = f'{self.endpoint}/admin/realms/{realm}/clients/{client}'
        response = requests.put(endpoint, headers=self.auth_header, json={'webOrigins': web_origins})
        return response.status_code == 204

    def get_clients(self, realm: str) -> dict:
        endpoint = f'{self.endpoint}/admin/realms/{realm}/clients'
        response = requests.get(endpoint, headers=self.auth_header)
        response.raise_for_status()
        return response.json()

    def create_client_scope(self, realm: str, name: str) -> bool:
        endpoint = f'{self.endpoint}/admin/realms/{realm}/client-scopes'
        response = requests.post(endpoint, headers=self.auth_header, json={
            'name': name,
            'protocol': 'openid-connect',
        })
        response.raise_for_status()
        return True

    def get_client_scopes(self, realm: str) -> list:
        endpoint = f'{self.endpoint}/admin/realms/{realm}/client-scopes'
        response = requests.get(endpoint, headers=self.auth_header)
        response.raise_for_status()
        return response.json()

    def add_default_client_scope_to_client(self, realm: str, client_id: str, client_scope_id: str):
        endpoint = f'{self.endpoint}/admin/realms/{realm}/clients/{client_id}/default-client-scopes/' \
                   f'{client_scope_id}'
        response = requests.put(endpoint, headers=self.auth_header, data={
            'client': client_id,
            'clientScopeId': client_scope_id,
            'realm': realm
        })
        response.raise_for_status()
        return True

    def add_oidc_claim_to_client(self, realm: str, client_id: str, mapper_name: str, claim_name: str, claim_value: str):
        endpoint = f'{self.endpoint}/admin/realms/{realm}/clients/{client_id}/protocol-mappers/models'
        if not isinstance(claim_value, str):
            claim_value = json.dumps(claim_value)
        response = requests.post(endpoint, headers=self.auth_header, json={
            'protocol': 'openid-connect',
            'config': {
                'id.token.claim': True,
                'access.token.claim': True,
                'userinfo.token.claim': True,
                'access.tokenResponse.claim': False,
                'jsonType.label': 'JSON',
                'claim.name': claim_name,
                'claim.value': claim_value
            },
            'name': mapper_name,
            'protocolMapper': 'oidc-hardcoded-claim-mapper'
        })
        response.raise_for_status()
        return True

    def add_oidc_claim_to_client_scope(self, realm: str, client_scope_id: str, mapper_name: str, claim_name: str,
                                       claim_value: str):
        endpoint = f'{self.endpoint}/admin/realms/{realm}/client-scopes/{client_scope_id}/protocol-mappers/models'
        if not isinstance(claim_value, str):
            claim_value = json.dumps(claim_value)
        response = requests.post(endpoint, headers=self.auth_header, json={
            'protocol': 'openid-connect',
            'config': {
                'id.token.claim': True,
                'access.token.claim': True,
                'userinfo.token.claim': True,
                'access.tokenResponse.claim': False,
                'jsonType.label': 'JSON',
                'claim.name': claim_name,
                'claim.value': claim_value
            },
            'name': mapper_name,
            'protocolMapper': 'oidc-hardcoded-claim-mapper'
        })
        response.raise_for_status()
        return True

    def create_user(self, username: str, password: str, realm: str, first_name: str, last_name: str, email: str) \
            -> bool:
        endpoint = f'{self.endpoint}/admin/realms/{realm}/users'
        response = requests.post(endpoint, headers=self.auth_header, json={
            'username': username,
            'email': email,
            'firstName': first_name,
            'lastName': last_name,
            'credentials': [{
                'value': password,
                'type': 'password',
                'temporary': False
            }],
            'emailVerified': True,
            'enabled': True
        })
        response.raise_for_status()
        return True

    def get_roles_for_realm(self, realm: str):
        endpoint = f'{self.endpoint}/admin/realms/{realm}/roles'
        response = requests.get(endpoint, headers=self.auth_header)
        response.raise_for_status()
        return response.json()

    def get_users_in_realm(self, realm: str):
        endpoint = f'{self.endpoint}/admin/realms/{realm}/users'
        response = requests.get(endpoint, headers=self.auth_header)
        response.raise_for_status()
        return response.json()

    def get_user_id_in_realm_by_username(self, realm: str, username: str) -> Optional[str]:
        users = self.get_users_in_realm(realm)
        for user in users:
            if user['username'] == username:
                return user['id']
        return

    def remove_role_mapping_from_user(self, realm: str, username: str, role_name: str):
        user_id = self.get_user_id_in_realm_by_username(realm=realm, username=username)
        if not user_id:
            raise ValueError(f'User {username} not found in realm {realm}!')

        roles = self.get_roles_for_realm(realm)
        role_id = None
        for role in roles:
            if role['name'] == role_name:
                role_id = role['id']
        if not role_id:
            raise ValueError(f'Role {role_name} not found in realm {realm}!')

        endpoint = f'{self.endpoint}/admin/realms/{realm}/users/{user_id}/role-mappings/realm'
        response = requests.delete(endpoint, headers=self.auth_header, json=[{
            'id': role_id,
            'name': role_name
        }])

        response.raise_for_status()
        return True

    def add_role_mapping_to_user(self, realm: str, username: str, role_name: str):
        user_id = self.get_user_id_in_realm_by_username(realm=realm, username=username)
        if not user_id:
            raise ValueError(f'User {username} not found in realm {realm}!')

        roles = self.get_roles_for_realm(realm)
        role_id = None
        for role in roles:
            if role['name'] == role_name:
                role_id = role['id']
        if not role_id:
            raise ValueError(f'Role {role_name} not found in realm {realm}!')

        endpoint = f'{self.endpoint}/admin/realms/{realm}/users/{user_id}/role-mappings/realm'
        response = requests.post(endpoint, headers=self.auth_header, json=[{
            'id': role_id,
            'name': role_name
        }])

        response.raise_for_status()
        return True
