from aiohttp import DummyCookieJar
from aiohttp.client import ClientSession
from authlib.jose import JWTClaims, JsonWebKey, KeySet, jwt
from authlib.oidc.core import UserInfo
from authlib.oidc.discovery import OpenIDProviderMetadata, well_known


class OpenIDConnectIssuer:
    keyword = 'Bearer'

    def __init__(self, issuer_url: str, jwks_default_kid: str):
        self.issuer_url = issuer_url
        self.jwks_default_kid = jwks_default_kid
        self._metadata = None
        self._jwks = None

    async def get_userinfo(self, access_token: str) -> UserInfo:
        oidc_metadata = await self.get_metadata()
        async with ClientSession(headers={'Authorization': f'{self.keyword} {access_token}'},
                                 cookie_jar=DummyCookieJar()) as s:
            async with s.get(oidc_metadata['userinfo_endpoint']) as response:
                response.raise_for_status()
                return UserInfo(await response.json())

    async def decode_and_verify_token(self, jwt_token: str) -> JWTClaims:
        jwks = await self.get_jwks()
        jwks_default_kid = self.jwks_default_kid

        def load_key_custom(headers, _payload):
            kid = headers.get('kid', jwks_default_kid)
            return jwks.find_by_kid(kid)

        jwt_decoded: JWTClaims = jwt.decode(jwt_token, key=load_key_custom, claims_cls=JWTClaims)
        jwt_decoded.validate()
        return jwt_decoded

    async def get_metadata(self) -> OpenIDProviderMetadata:
        if self._metadata is None:
            url = well_known.get_well_known_url(self.issuer_url, external=True)
            async with ClientSession(cookie_jar=DummyCookieJar()) as s:
                async with s.get(url) as response:
                    response.raise_for_status()
                    self._metadata = OpenIDProviderMetadata(await response.json())
        return self._metadata

    async def get_jwks(self) -> KeySet:
        if self._jwks is None:
            oidc_metadata = await self.get_metadata()
            async with ClientSession(cookie_jar=DummyCookieJar()) as s:
                async with s.get(oidc_metadata['jwks_uri']) as response:
                    response.raise_for_status()
                    key_set = await response.json()
            self._jwks = JsonWebKey.import_key_set(key_set)
        return self._jwks
