import pytest
import pytest_asyncio
from aiohttp import ClientSession
from authlib.jose import jwk, jwt
from authlib.jose.errors import BadSignatureError
from authlib.oidc.discovery import OpenIDProviderMetadata
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa

from csirtmu.oidc.utils.oidc_issuer import OpenIDConnectIssuer
from tests.test_utils import MockRequest


class TestIssuer:
    @pytest.fixture(scope='function')
    def mock_response(self, monkeypatch):
        def mock_get(*args, **kwargs):
            return MockRequest({'name': 'Rodney McKay'})

        monkeypatch.setattr(ClientSession, "get", mock_get)

    # Tuple[private, public]
    @pytest.fixture
    def rsa_keys(self):
        private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
        return private_key.private_bytes(encoding=serialization.Encoding.PEM,
                                         format=serialization.PrivateFormat.PKCS8,
                                         encryption_algorithm=serialization.NoEncryption()), \
               private_key.public_key().public_bytes(
                   encoding=serialization.Encoding.PEM,
                   format=serialization.PublicFormat.SubjectPublicKeyInfo
               )

    # Tuple[private, public]
    @pytest.fixture
    def rsa_keys_invalid(self):
        private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
        return private_key.private_bytes(encoding=serialization.Encoding.PEM,
                                         format=serialization.PrivateFormat.PKCS8,
                                         encryption_algorithm=serialization.NoEncryption()), \
               private_key.public_key().public_bytes(
                   encoding=serialization.Encoding.PEM,
                   format=serialization.PublicFormat.SubjectPublicKeyInfo
               )

    @pytest_asyncio.fixture
    async def oidc_issuer(self, monkeypatch, rsa_keys) -> OpenIDConnectIssuer:
        def patched_constructor(self, issuer_url: str, jwks_default_kid: str):
            self.issuer_url = issuer_url
            self._metadata = OpenIDProviderMetadata({'userinfo_endpoint': 'http://127.0.0.1/userinfo'})
            self._jwks = jwk.JsonWebKey.import_key_set([rsa_keys[1]])

            self._jwks.keys = self._jwks.keys  # Force kid "generation"
            self.jwks_default_kid = self._jwks.keys[0].kid

        monkeypatch.setattr(OpenIDConnectIssuer, '__init__', patched_constructor)
        issuer = OpenIDConnectIssuer(issuer_url='', jwks_default_kid='test_key')
        return issuer

    @pytest.fixture
    def auth_token_valid(self, rsa_keys):
        return jwt.encode(
            header={'alg': 'RS256'},
            payload={'preferred_username': 'tester'},
            key=jwk.JsonWebKey.import_key(rsa_keys[0], {'kty': 'RSA'})
        )

    @pytest.fixture
    def auth_token_invalid_key(self, rsa_keys_invalid):
        return jwt.encode(
            header={'alg': 'RS256'},
            payload={'preferred_username': 'tester'},
            key=jwk.JsonWebKey.import_key(rsa_keys_invalid[0], {'kty': 'RSA'})
        )

    @pytest.mark.asyncio
    async def test_decode_valid_token(self, oidc_issuer: OpenIDConnectIssuer, auth_token_valid):
        claims = await oidc_issuer.decode_and_verify_token(auth_token_valid)
        assert claims.get('preferred_username') == 'tester'

    @pytest.mark.asyncio
    async def test_decode_token_invalid_key(self, oidc_issuer: OpenIDConnectIssuer, auth_token_invalid_key):
        with pytest.raises(BadSignatureError):
            await oidc_issuer.decode_and_verify_token(auth_token_invalid_key)
