from typing import Dict

import pytest

from csirtmu.oidc.utils.utils import extract_access_token, extract_auth_header


class MockResponse:
    def __init__(self, status: int, body: Dict):
        self.status = status
        self.body = body

    @property
    def ok(self):
        return self.status == 200

    async def json(self):
        return self.body


class MockRequest:
    def __init__(self, body: Dict, status: int = 200):
        self.status = status
        self.body = body

    async def __aenter__(self) -> MockResponse:
        return MockResponse(self.status, self.body)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return


@pytest.fixture
def token():
    return 'eyJhbGciOiJSUzI1NiIsInR5cCIgOiAiSldUIiwia2lkIiA6ICJybTJnRVNULXZZc1liU1hhX2MzeVlSM3ltNlk3R3Jubndh'


def test_extract_token_correct(token):
    extracted_token = extract_access_token(f'Bearer {token}', keyword='Bearer')
    assert extracted_token == token


def test_extract_token_invalid_keyword(token):
    extracted_token = extract_access_token(f'Boarer  {token}', keyword='Bearer')
    assert extracted_token is None


@pytest.mark.parametrize("token_fragment", ['', ' ', 'x', 'x '])
def test_extract_token_fragment_in_header(token_fragment):
    extracted_token = extract_access_token(token_fragment, keyword='Bearer')
    assert extracted_token is None


def test_extract_token_missing_token():
    with pytest.raises(ValueError):
        extract_access_token('Bearer ', keyword='Bearer')


@pytest.mark.asyncio
async def test_extract_auth_header_correct():
    # Header dicts are usually case insensitive
    class Request:
        headers = {
            'Authorization': f'Bearer {token}'
        }

    assert extract_auth_header(Request()) == f'Bearer {token}'


@pytest.mark.asyncio
async def test_extract_auth_header_missing_header():
    class Request:
        headers = {}

    assert extract_auth_header(Request()) is None
