# Starlette JWT Authorization Backend

This is an authorization backend used with the csirtmu-service-gateway authentication service to provide OIDC authorization based on access token capabilities.

---

## Content
- [Installation](#installation)
- [Starlette JWT Authorization Backend](#starlette-jwt-authorization-backend)
  - [OSO Instructions](#oso-instructions)
  - [Usage](#usage)

---

## Installation

### Add repository
```
$ poetry source add ORION  https://gitlab.orion.lan/api/v4/projects/2485/packages/pypi/simple
$ poetry config http-basic.ORION token <GITLAB_TOKEN>
```

### Add packages

```
$ poetry add --source ORION csirtmu-starlette-jwt-authorization-backend
```

---

## OSO Instructions

Oso uses .polar files to specify permission rules.
`csirtmu.starlette_jwt_auth` adds permissions to user tokens based on their capabilities, i.e.:

```python
from csirtmu.starlette_jwt_auth.models import MappedUser

user = MappedUser.create_or_get(access_token='none', user_id='none', capabilities=['first|value', 'second|value'])
```

~> 
```polar
has_permission(token, 'first|value', _);
has_permission(token, 'second|value', _);
```

## Dependencies
- csirtmu-auth-utils
- csirtmu-dataclass-config-loader

## Usage

1. Create Oso object and map sqlalchemy classes
2. Create a *\<policy file name\>.polar* and configure based on *tests/examples/sample_config.polar*
3. Add oso configuration files

```python
from oso.oso import Oso
from csirtmu.starlette_jwt_auth.models import MappedUser
import importlib.resources as pkg_resources
import csirtmu.starlette_jwt_auth

oso = Oso()
oso.register_class(MappedUser)
oso.register_class(...)

# IMPORTANT to add the oso_configuration from `csirtmu.starlette_jwt_auth` 
with pkg_resources.path(csirtmu.starlette_jwt_auth, 'oso_configuration.polar') as cfg_path:
    oso_config_files = [
        ... / 'config.polar',
        str(cfg_path),
    ]
oso.load_files(oso_config_files)

```

4. Create `csirtmu.starlette_jwt_auth` sqlalchemy tables

```python
from sqlalchemy.engine import create_engine
from csirtmu.starlette_jwt_auth.models import Base

engine = create_engine('sqlite:///')
Base.metadata.create_all(bind=engine)
```

6. Create *config.yaml* with content based on *tests/examples/config-example.yaml* and configure
7. Add TokenAuthenticationBackend to Starlette
```python
from starlette.applications import Starlette
from starlette.middleware.authentication import AuthenticationMiddleware
from csirtmu.starlette_jwt_auth.token_authentication_backend import TokenAuthenticationBackend
import csirtmu.starlette_jwt_auth.config as config
from sqlalchemy.orm.session import sessionmaker

star_app = Starlette(...)
config.conf_path = ... / 'config.yaml'
# Please note that the sessionmaker passed to `TokenAuthenticationBackend` does NOT use oso
star_app.add_middleware(AuthenticationMiddleware,
                        backend=TokenAuthenticationBackend(config=config.get_config(),
                                                           session_factory=sessionmaker(...)))
```
8. Require authenticated permissions from starlette (`@requires('authenticated')`) 
9. Add authorization checks based on used data (authorized_sessionmaker, ...) **IMPORTANT**. Without authorizing using any of the oso methods, no authorization is applied.
10. See [documentation for sqlalchemy](https://docs.osohq.com/python/reference/api/sqlalchemy.html) and [documentation](https://docs.osohq.com/guides/rbac.html)
