from datetime import datetime
from typing import Optional, Tuple, Type

import orjson.orjson
from asgiref.sync import sync_to_async
from authlib.jose import JWTClaims, JsonWebToken
from authlib.jose.errors import DecodeError, JoseError
from csirtmu.oidc.utils.authentication_cache import AuthenticationCache
from csirtmu.oidc.utils.utils import extract_access_token, extract_auth_header
from starlette.authentication import AuthCredentials, AuthenticationBackend, AuthenticationError, BaseUser
from starlette.requests import HTTPConnection, Request
from starlette.responses import JSONResponse

from csirtmu.starlette_jwt_auth.config import CONFIG, JwtAuthConfig
from csirtmu.starlette_jwt_auth.models import MappedUser

authlib_jwt = JsonWebToken(['HS512'])


def on_auth_error(request: Request, exc: Exception) -> JSONResponse:  # pragma: no cover
    return JSONResponse({"error": str(exc)}, status_code=401)


class CustomJWTClaims(JWTClaims):
    REGISTERED_CLAIMS = ['iss', 'aud', 'exp', 'nbf', 'iat', 'jti', 'scope', 'username']

    def validate(self, now=None, leeway=0):
        super(CustomJWTClaims, self).validate(now, leeway)
        for field in set(self.REGISTERED_CLAIMS) - set(super().REGISTERED_CLAIMS):
            self._validate_claim_value(field)


class TokenAuthenticationBackend(AuthenticationBackend):
    config: JwtAuthConfig

    def __init__(self, user_class: Type[MappedUser] = MappedUser) -> None:
        self.config = CONFIG
        self.user_class = user_class
        if not user_class or not issubclass(user_class, MappedUser):
            raise TypeError('User class must be a subclass of MappedUser')
        self.claims_options = {
            "iss": {
                "essential": True,
                "value": self.config.gateway.token_issuer
            },
            "aud": {
                "essential": True,
                "value": self.config.gateway.token_audience
            },
            "exp": {
                "essential": True,
            },
            "iat": {
                "essential": True,
            },
            "jti": {
                "essential": True,
            },
            "scope": {
                "essential": True,
            },
            "username": {
                "essential": True,
            },
        }
        self.cache = AuthenticationCache(max_size=2500, ttl_default=300)

    @staticmethod
    def _decode_payload(bytes_payload):
        try:
            payload = orjson.loads(bytes_payload)
        except ValueError:
            raise DecodeError('Invalid payload value')
        if not isinstance(payload, dict):
            raise DecodeError('Invalid payload type')
        return payload

    async def decode_and_validate_token(self, token: str) -> Optional[JWTClaims]:
        try:
            decoded_token = authlib_jwt.decode(token, key=self.config.gateway.token_key,
                                               claims_cls=CustomJWTClaims,
                                               claims_options=self.claims_options)
            decoded_token.validate(leeway=CONFIG.leeway)
            return decoded_token
        except JoseError:
            return None

    async def authenticate(self, request: HTTPConnection) -> Optional[Tuple[AuthCredentials, BaseUser]]:
        if not (auth_header := extract_auth_header(request, self.config.gateway.token_header_name)):
            return None

        try:
            access_token = extract_access_token(auth_header, self.config.gateway.token_prefix)
            if access_token is None:
                return None
        except ValueError as ex:
            raise AuthenticationError(ex)

        ut = self.cache.get(access_token)
        if ut is None:
            decoded_token = await self.decode_and_validate_token(access_token)
            if decoded_token is None:
                return None

            claim_capabilities = decoded_token.get('capabilities', [])
            ut = (access_token, decoded_token['username'], claim_capabilities)

            # A JSON numeric value representing the number of seconds from
            #   1970-01-01T00:00:00Z UTC until the specified UTC date/time, ignoring leap seconds.
            cache_ttl = int(decoded_token['exp'] - datetime.utcnow().timestamp()) + CONFIG.leeway
            self.cache.add(key=access_token, value=ut, ttl=cache_ttl)

        access, user_id, capabilities = ut

        user = await sync_to_async(self.user_class.create_or_get, thread_sensitive=False) \
            (access_token=access, user_id=user_id, capabilities=capabilities)
        return AuthCredentials(['authenticated']), user
