import uuid
from datetime import datetime, timedelta, timezone
from pathlib import Path

import jwt
import pytest
from oso import Oso
from sqlalchemy.orm import Session, sessionmaker
from sqlalchemy_oso import authorized_sessionmaker
from starlette.datastructures import Headers
from starlette.requests import Request

from csirtmu.starlette_jwt_auth.config import CONFIG, JwtAuthConfig
from csirtmu.starlette_jwt_auth.models import MappedUser
from tests.examples.example_application.app import app
from tests.examples.example_application.config import Config, DB
from tests.examples.example_application.models import Base, IPAddress


def _create_default_jwt(config):
    return {
        'iss': config.gateway.token_issuer,
        'aud': config.gateway.token_audience,
        'exp': datetime(3000, 1, 1, tzinfo=timezone.utc).timestamp(),
        'nbf': (datetime.utcnow() - timedelta(minutes=5)).timestamp(),
        'iat': (datetime.utcnow() - timedelta(minutes=5)).timestamp(),
        'jti': str(uuid.uuid4()),
        'scope': 'profile email',
    }


@pytest.fixture(scope="session")
def example_app():
    return app()


@pytest.fixture
def auth_config() -> JwtAuthConfig:
    return CONFIG


@pytest.fixture
def config(db) -> Config:
    return Config(db=db)


@pytest.fixture
def db() -> DB:
    return DB()


@pytest.fixture
def engine(db: DB):
    return db.engine


@pytest.fixture
def prepare_database(engine):
    Base.metadata.drop_all(engine)
    Base.metadata.create_all(engine)
    s = Session(engine)
    s.add(IPAddress())
    s.commit()


@pytest.fixture
def oso_config_files():
    return [
        (Path(__file__).parent / 'examples' / 'sample_config.polar').resolve().__str__(),
        (Path(__file__).parent / 'examples' / 'oso_configuration.polar').resolve().__str__(),
    ]


@pytest.fixture
def oso():
    oso = Oso()
    oso.register_class(IPAddress)
    oso.register_class(MappedUser)
    return oso


@pytest.fixture
def oso_session_factory(prepare_database, oso, oso_config_files, engine, auth_config):
    oso.load_files(oso_config_files)
    return authorized_sessionmaker(bind=engine,
                                   get_oso=lambda: oso,
                                   get_user=lambda: MappedUser.create_or_get(
                                       access_token='test',
                                       user_id=f'yss{auth_config.jwt_username_separator}zub',
                                       capabilities=["something"]),
                                   get_checked_permissions=lambda: {
                                       IPAddress: 'read'
                                   })


@pytest.fixture
def session_factory(prepare_database, engine):
    return sessionmaker(bind=engine)


@pytest.fixture
def empty_request():
    return Request({'type': 'http', 'headers': Headers()})


@pytest.fixture
def unauthorized_valid_token(auth_config):
    token = jwt.encode(
        {**_create_default_jwt(auth_config),
         'username': auth_config.gateway.token_issuer + '|frank'},
        key=auth_config.gateway.token_key,
        algorithm='HS512')

    return token


@pytest.fixture
def authorized_token(auth_config):
    token = jwt.encode({
        **_create_default_jwt(auth_config),
        'username': auth_config.gateway.token_issuer + '|frank',
        'capabilities': ['something']
    },
        key=auth_config.gateway.token_key,
        algorithm='HS512')

    return token


@pytest.fixture
def authorized_token_2(auth_config):
    token = jwt.encode({
        **_create_default_jwt(auth_config),
        'username': auth_config.gateway.token_issuer + '|frank',
        'capabilities': [
            'something',
            'something-else',
        ]
    },
        key=auth_config.gateway.token_key,
        algorithm='HS512')

    return token
