from pathlib import Path

from starlette.applications import Starlette
from starlette.authentication import requires
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.requests import Request
from starlette.routing import Route
from starlette.types import Receive, Scope, Send
from strawberry.asgi import GraphQL, HTTPHandler

from csirtmu.starlette_jwt_auth.models import MappedUser
from csirtmu.starlette_jwt_auth.routes import auth_backend_version_route
from csirtmu.starlette_jwt_auth.token_authentication_backend import TokenAuthenticationBackend
from tests.examples.example_application.config import CONFIG
from tests.examples.example_application.models import Base, DatabaseUser, IPAddress
from tests.examples.example_application.schema import schema


class AuthenticationHttpHandler(HTTPHandler):
    @requires('authenticated')
    def _request_authenticate(self, request):
        return

    async def handle(self, scope: Scope, receive: Receive, send: Send):
        self._request_authenticate(Request(scope=scope, receive=receive))
        return await super().handle(scope, receive, send)


def app() -> Starlette:
    config = CONFIG
    # Register models to Oso
    oso = config.oso
    oso.register_class(IPAddress)
    oso.register_class(MappedUser)
    # Load Oso polar config
    oso_config_files = [
        str((Path(__file__).parent.parent / 'oso_configuration.polar').resolve()),
        str((Path(__file__).parent.parent / 'sample_config.polar').resolve())
    ]
    oso.load_files(oso_config_files)
    # Prepare DB schema
    Base.metadata.create_all(bind=config.db.engine)

    graphql = GraphQL(schema)
    graphql.http_handler_class = AuthenticationHttpHandler

    star_app = Starlette(debug=True,
                         routes=[
                             Route('/', graphql),
                             auth_backend_version_route
                         ])
    star_app.add_middleware(AuthenticationMiddleware,
                            backend=TokenAuthenticationBackend(user_class=DatabaseUser))
    return star_app
