import jwt
import pytest
from starlette.authentication import AuthenticationError
from starlette.datastructures import Headers

from csirtmu.starlette_jwt_auth.token_authentication_backend import TokenAuthenticationBackend
from tests.conftest import _create_default_jwt


class TestTokenExtraction:
    @pytest.fixture
    def request_with_empty_token(self, auth_config, empty_request):
        token = jwt.encode(_create_default_jwt(auth_config), auth_config.gateway.token_key, algorithm='HS512')

        empty_request.scope['headers'] = Headers({
            **dict(empty_request.scope['headers'].items()),
            auth_config.gateway.token_header_name: f'{auth_config.gateway.token_prefix} {token}',
        }).raw

        return empty_request

    @pytest.fixture
    def request_with_unknown_user(self, auth_config, empty_request):
        token = jwt.encode(
            {**_create_default_jwt(auth_config), 'username': auth_config.gateway.token_issuer + '|missing'},
            key=auth_config.gateway.token_key, algorithm='HS512')

        empty_request.scope['headers'] = Headers({
            **dict(empty_request.scope['headers'].items()),
            auth_config.gateway.token_header_name: f'{auth_config.gateway.token_prefix} {token}',
        }).raw

        return empty_request

    @pytest.fixture
    def request_with_existing_token(self, auth_config, unauthorized_valid_token, empty_request):
        empty_request.scope['headers'] = Headers({
            **dict(empty_request.scope['headers'].items()),
            auth_config.gateway.token_header_name: f'{auth_config.gateway.token_prefix} {unauthorized_valid_token}',
        }).raw

        return empty_request

    @pytest.fixture
    def auth_backend(self):
        backend = TokenAuthenticationBackend()
        return backend

    @pytest.mark.asyncio
    async def test_auth_header_missing_in_request(self, auth_backend, empty_request):
        assert await auth_backend.authenticate(empty_request) is None

    @pytest.mark.asyncio
    async def test_empty_token(self, auth_backend, request_with_empty_token):
        result = await auth_backend.authenticate(request_with_empty_token)
        assert result is None

    @pytest.mark.asyncio
    async def test_missing_user(self, auth_backend, request_with_unknown_user):
        credentials, user = await auth_backend.authenticate(request_with_unknown_user)
        assert user is not None

    @pytest.mark.asyncio
    async def test_existing_user(self, auth_config, auth_backend, request_with_existing_token):
        credentials, user_token = await auth_backend.authenticate(request_with_existing_token)
        assert 'frank' in user_token.user_id

    @pytest.mark.asyncio
    async def test_invalid_issuer(self, auth_config, auth_backend, authorized_token, empty_request):
        token = jwt.encode({**jwt.decode(authorized_token, key=auth_config.gateway.token_key, algorithms=['HS512'],
                                         audience=auth_config.gateway.token_audience),
                            'iss': 'invalid-issuer'},
                           key=auth_config.gateway.token_key,
                           algorithm='HS512')
        empty_request.scope['headers'] = Headers({
            **dict(empty_request.scope['headers'].items()),
            auth_config.gateway.token_header_name: f'{auth_config.gateway.token_prefix} {token}',
        }).raw
        result = await auth_backend.authenticate(empty_request)
        assert result is None

    @pytest.mark.asyncio
    async def test_other_authorization_header(self, auth_config, auth_backend, authorized_token, empty_request):
        empty_request.scope['headers'] = Headers({
            **dict(empty_request.scope['headers'].items()),
            auth_config.gateway.token_header_name: f'Not-A-Jwt-Token {authorized_token}',
        }).raw
        result = await auth_backend.authenticate(empty_request)
        assert result is None

    @pytest.mark.asyncio
    async def test_empty_token_value(self, auth_config, auth_backend, authorized_token, empty_request):
        empty_request.scope['headers'] = Headers({
            **dict(empty_request.scope['headers'].items()),
            auth_config.gateway.token_header_name: f'{auth_config.gateway.token_prefix} ',
        }).raw
        with pytest.raises(AuthenticationError):
            await auth_backend.authenticate(empty_request)

    @pytest.mark.asyncio
    async def test_extra_whitespace_in_token_value(self, auth_config, auth_backend, authorized_token, empty_request):
        empty_request.scope['headers'] = Headers({
            **dict(empty_request.scope['headers'].items()),
            auth_config.gateway.token_header_name: f'{auth_config.gateway.token_prefix} anything goes',
        }).raw
        with pytest.raises(AuthenticationError):
            await auth_backend.authenticate(empty_request)
