import { HttpErrorResponse, HttpStatusCode } from '@angular/common/http';
import { ModuleWithProviders, NgModule, Optional, SkipSelf } from '@angular/core';
import { ApolloLink, InMemoryCache } from '@apollo/client/core';
import { onError } from '@apollo/client/link/error';
import { ApiConfig, APOLLO_CLIENT_NAME } from '@orion/common';
import { FragmentTypes } from '@orion/fragment-matcher';
import { ErrorHandlerService } from '@orion/shared/services/global';
import { retrieveApolloContextErrorTitle, retrieveApolloContextRetryable } from '@csirtmu-orion/microframework/graphql';
import { Apollo, ApolloModule } from 'apollo-angular';
import { HttpLink } from 'apollo-angular/http';
import { extractFiles } from 'extract-files';
import { RetryLink } from '@apollo/client/link/retry';

@NgModule({ imports: [ApolloModule] })
export class ApolloApiModule {
  constructor(
    @Optional() @SkipSelf() parentModule: ApolloApiModule,
    apollo: Apollo,
    httpLink: HttpLink,
    config: ApiConfig,
    errorHandler: ErrorHandlerService
  ) {
    if (parentModule) {
      throw new Error('ApolloApiModule is already loaded. Import it only once in single module hierarchy.');
    }

    const name = APOLLO_CLIENT_NAME;
    const http = httpLink.create({ uri: config.baseUrl, extractFiles: extractFiles });
    const errorLink = onError(({ graphQLErrors, networkError, operation }) => {
      if (graphQLErrors) {
        errorHandler.emit(graphQLErrors, `${retrieveApolloContextErrorTitle(operation.getContext())} failed`);
      }
      if (networkError instanceof HttpErrorResponse && networkError.status !== HttpStatusCode.Unauthorized) {
        errorHandler.emit(networkError, '');
      }
    });
    const retryLink = new RetryLink({
      delay: {
        initial: 300,
        max: Infinity,
        jitter: true,
      },
      attempts: {
        max: 5,
        retryIf: (error, operation) => !!error && retrieveApolloContextRetryable(operation.getContext()),
      },
    });

    // order important, otherwise retry attempts will display error popups
    const link: ApolloLink = errorLink.concat(retryLink).concat(http);
    apollo.createNamed(name, {
      link,
      cache: new InMemoryCache({
        possibleTypes: FragmentTypes.possibleTypes,
      }),
    });
  }

  static forRoot(config: ApiConfig): ModuleWithProviders<ApolloApiModule> {
    return {
      ngModule: ApolloApiModule,
      providers: [{ provide: ApiConfig, useValue: config }],
    };
  }
}
