import { Directive, OnInit, Optional } from '@angular/core';
import { AssetTypeOverviewService } from '@orion/asset-services';
import { FilterChip } from '@csirtmu-orion/microframework/filter';
import { FilterChipsService, FilterService } from '@csirtmu-orion/microframework/filter-services';
import { IPaginationEvent } from '@sentinel/common/pagination';
import { PaginationService } from '@csirtmu-orion/microframework/pagination-services';
import { EntityPaginableSmartOverviewDirective } from '@csirtmu-orion/microframework/entity-smart-directives';
import { Observable } from 'rxjs';
import { shareReplay, takeUntil } from 'rxjs/operators';

@Directive()
export abstract class AssetSmartOverviewDirective<Entity, Filter>
  extends EntityPaginableSmartOverviewDirective<Entity>
  implements OnInit
{
  filterChangedByUser = false;
  filterLocalState$?: Observable<Filter>;
  filterChips$?: Observable<FilterChip[]>;
  protected constructor(
    // No point in introducing a generic for query, query variables, orderable, and api filter in this component
    // as the types context doesn't matter to this component.
    // eslint-disable-next-line
    protected entityService: AssetTypeOverviewService<any, any, Entity, Filter, any>,
    protected filterService: FilterService<Filter>,
    protected paginationService: PaginationService,
    @Optional() protected chipsService?: FilterChipsService<Filter>
  ) {
    super(entityService, paginationService);
  }

  ngOnInit(): void {
    super.ngOnInit();
    this.filterLocalState$ = this.filterService.filterLocalState$;
    this.filterChips$ = this.chipsService?.chips$;
    this.filterService.filter$
      .pipe(shareReplay({ refCount: true }), takeUntil(this.destroy$))
      .subscribe((filter) => this.onFilterChange(filter));
  }

  onFilterChange(filter: Filter): void {
    this.resetPaginationOnUserFilterChange();
    this.entityService.update(this.paginationService.get(), filter);
  }

  onRemoveChip(chip: FilterChip): void {
    this.chipsService?.removeChip(chip);
  }

  onClearFilters(): void {
    this.filterService.clear();
  }

  onSubmitFilter(): void {
    this.filterService.submit();
  }

  protected onPaginationChange(pagination: IPaginationEvent): void {
    this.entityService.update(pagination, this.filterService.get());
  }

  protected resetPaginationOnUserFilterChange(): void {
    // initial filter change is always filter init (from route or default), all subsequent changes are done by user
    if (this.filterChangedByUser) {
      this.paginationService.resetPosition();
    } else {
      this.filterChangedByUser = true;
    }
  }
}
