import { AfterViewInit, ChangeDetectionStrategy, Component, ElementRef, ViewChild } from '@angular/core';
import { ActivatedRoute, Data, NavigationEnd, Router } from '@angular/router';
import { ASSETS_PATH } from '@orion/asset-paths';
import { CASES_PATH } from '@orion/case-paths';
import { toAuthPath } from '@orion/common';
import { IPFIX_PATH } from '@orion/data-platform-paths';
import { LABELS_PATH } from '@orion/label-paths';
import { OBSERVABLES_PATH } from '@orion/observable-paths';
import { ApolloRefreshPageDataService, LoadingService, OrionLogoutService } from '@orion/shared/services/global';
import { TTP_PATH } from '@orion/ttps-paths';
import { USER_ACCOUNTING_PATH } from '@orion/user-accounting-paths';
import { CurrentUserFragment, CurrentUserService, UserNavigator } from '@orion/user-services';
import { WORKFLOW_TEMPLATE_PATH } from '@orion/workflow-template-paths';
import { Optional } from '@sentinel/common/utils';
import { Agenda, AgendaContainer, SentinelUser } from '@sentinel/layout';
import { SentinelBreadcrumb, SentinelBreadcrumbBuilder } from '@sentinel/layout/breadcrumbs';
import { Observable } from 'rxjs';
import { filter, map, startWith, take } from 'rxjs/operators';
import { environment } from '../../environments/environment';
import { FullWidthPageContentService } from '../shared/services/global/full-width-page-content.service';
import { SidebarState, SidebarStateService } from '@sentinel/layout/sidebar';

@Component({
  selector: 'orion-auth-section',
  templateUrl: './auth-section.component.html',
  styleUrls: ['./auth-section.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush,
})
export class AuthSectionComponent implements AfterViewInit {
  @ViewChild('pageContainer') pageContainer?: ElementRef<HTMLDivElement>;
  @ViewChild('pageContent') pageContent?: ElementRef<HTMLDivElement>;

  agendaContainers: AgendaContainer[] = [];
  pageHeadline$?: Observable<string>;
  headlineIcon$?: Observable<string>;
  isLoading$?: Observable<boolean>;
  activeUser$?: Observable<Optional<SentinelUser>>;
  userProfilePath$?: Observable<Optional<string>>;
  breadcrumbs$?: Observable<SentinelBreadcrumb[]>;
  fullWidth$?: Observable<boolean>;
  sidebarState$?: Observable<SidebarState>;
  version = environment.version;
  toolbarHeadlineMargin = '0px';

  private pageContainerResizeObserver?: ResizeObserver;

  constructor(
    private router: Router,
    private activatedRoute: ActivatedRoute,
    private authService: OrionLogoutService,
    private userService: CurrentUserService,
    private userPathResolver: UserNavigator,
    private loadingService: LoadingService,
    private fullWidthModeService: FullWidthPageContentService,
    private sidebarStateService: SidebarStateService,
    private refreshPageDataService: ApolloRefreshPageDataService
  ) {}

  ngOnInit(): void {
    this.initUser();
    this.initAgendas();
    this.initBreadcrumbs();
    this.pageHeadline$ = this.getCurrRouteData().pipe(map((data) => data.headline));
    this.headlineIcon$ = this.getCurrRouteData().pipe(map((data) => data.headlineIcon));
    this.isLoading$ = this.loadingService.isLoading$;
    this.fullWidth$ = this.fullWidthModeService.fullWidth$;
    this.sidebarState$ = this.sidebarStateService.state$;
  }

  ngAfterViewInit(): void {
    if (!this.pageContainerResizeObserver && this.pageContainer) {
      this.pageContainerResizeObserver = new ResizeObserver((entries) => {
        for (let entry of entries) {
          const cr = entry.contentRect;
          const contentWidth = this.pageContent?.nativeElement?.offsetWidth ?? 0;
          // divide by two because the page content is centered, +20 because of toolbar padding
          const calculatedMargin = Math.max((cr.width - contentWidth) / 2 + 20, 0);
          this.toolbarHeadlineMargin = `${calculatedMargin}px`;
        }
      });
      this.pageContainerResizeObserver.observe(this.pageContainer.nativeElement);
    }
  }

  onLogout(): void {
    this.authService.logout();
  }

  refreshPageData(): void {
    this.refreshPageDataService.refreshData().pipe(take(1)).subscribe();
  }

  private initUser(): void {
    this.activeUser$ = this.userService.data$?.pipe(map((user) => (user ? this.mapToSentinelUser(user) : undefined)));
    this.userProfilePath$ = this.userService.data$?.pipe(
      map((user) => (user ? this.userPathResolver.toDetail(user.id) : undefined))
    );
  }

  private initAgendas(): void {
    this.agendaContainers = [
      new AgendaContainer('Overview', [
        new Agenda('Assets', toAuthPath(ASSETS_PATH)),
        new Agenda('Observables', toAuthPath(OBSERVABLES_PATH)),
      ]),
      new AgendaContainer('Monitoring', [
        new Agenda('IPFIX', toAuthPath(IPFIX_PATH)),
        new Agenda('User Accounting', toAuthPath(USER_ACCOUNTING_PATH)),
      ]),
      new AgendaContainer('Case Management', [
        new Agenda('Cases', toAuthPath(CASES_PATH)),
        new Agenda('Labels', toAuthPath(LABELS_PATH)),
        new Agenda('Templates', toAuthPath(WORKFLOW_TEMPLATE_PATH)),
      ]),
      new AgendaContainer('Knowledge Base', [new Agenda('TTPs', toAuthPath(TTP_PATH))]),
    ];
  }

  private initBreadcrumbs(): void {
    const navigationEndEvent$ = this.router.events.pipe(
      startWith(new NavigationEnd(-1, '', '')), // to trigger init on the first app load
      filter((event) => event instanceof NavigationEnd)
    );
    this.breadcrumbs$ = navigationEndEvent$.pipe(
      map(() => SentinelBreadcrumbBuilder.build(this.activatedRoute.snapshot))
    );
  }

  private getCurrRouteData(): Observable<Data> {
    return this.router.events.pipe(
      startWith(new NavigationEnd(-1, '', '')), // to trigger init on the first app load
      filter((event) => event instanceof NavigationEnd),
      map(() => {
        let route = this.activatedRoute;
        while (route.firstChild) {
          route = route.firstChild;
        }
        return route;
      }),
      filter((route) => route.outlet === 'primary'),
      map((route) => route.snapshot),
      map((snapshot) => snapshot.data)
    );
  }

  private mapToSentinelUser(user: CurrentUserFragment): SentinelUser {
    return { name: user.name, login: user.id, picture: user.iconSrc };
  }
}
