import { ChangeDetectionStrategy, Component, OnInit } from '@angular/core';
import {
  ExternalReferenceActionsService,
  ExternalReferenceViewStateService,
} from '@orion/external-reference-services/overview';
import { CaseManagementExternalReferenceOrdEnum } from '@orion/graphql-types';
import { Destroyable } from '@sentinel/common';
import { ApolloResultType } from '@csirtmu-orion/microframework/graphql';
import {
  INIT_PAGINATION_EVENT_BASE_TOKEN,
  PaginationBaseInit,
  PaginationService,
} from '@csirtmu-orion/microframework/pagination-services';
import { SentinelControlItem } from '@sentinel/components/controls';
import { BehaviorSubject, defer, Observable, of } from 'rxjs';
import { map, takeUntil } from 'rxjs/operators';
import { ExternalReferenceListService } from './external-reference-list.service';
import { ExternalReferenceListElementFragment } from './external-reference-selection';

@Component({
  selector: 'orion-add-external-reference',
  templateUrl: './add-external-reference.component.html',
  styleUrls: ['./add-external-reference.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush,
  providers: [
    ExternalReferenceListService,
    PaginationService,
    {
      provide: INIT_PAGINATION_EVENT_BASE_TOKEN,
      useValue: new PaginationBaseInit(50, CaseManagementExternalReferenceOrdEnum.URL, 'asc'),
    },
  ],
})
export class AddExternalReferenceComponent extends Destroyable() implements OnInit {
  externalReferences$?: Observable<ExternalReferenceListElementFragment[]>;
  isLoading$?: Observable<boolean>;
  hasError$?: Observable<boolean>;
  isSaving$?: Observable<boolean>;
  selected$: BehaviorSubject<ExternalReferenceListElementFragment | undefined> = new BehaviorSubject<
    ExternalReferenceListElementFragment | undefined
  >(undefined);
  controls: SentinelControlItem[] = [];

  constructor(
    private externalReferenceService: ExternalReferenceListService,
    private actionService: ExternalReferenceActionsService,
    private viewStateService: ExternalReferenceViewStateService,
    private paginationService: PaginationService
  ) {
    super();
  }

  ngOnInit(): void {
    this.externalReferences$ = this.externalReferenceService.data$.pipe(map((paginated) => paginated.elements));
    this.isLoading$ = this.externalReferenceService.isLoading$;
    this.hasError$ = this.externalReferenceService.hasError$;
    this.isSaving$ = this.actionService.addingInProgress$;
    this.controls = [
      new SentinelControlItem(
        'CANCEL',
        'Cancel',
        'warn',
        of(false),
        defer(() => this.onCancel()),
        this.isSaving$
      ),
      new SentinelControlItem(
        'SAVE',
        'Save',
        'primary',
        this.selected$.asObservable().pipe(map((reference) => reference === undefined)),
        defer(() => this.onSave()),
        this.isSaving$
      ),
    ];
  }

  onSave(): void {
    const selected = this.selected$.getValue();
    if (selected) {
      this.actionService
        .addAssociation(selected.id)
        .pipe(takeUntil(this.destroy$))
        .subscribe((result) => {
          if (result.type === ApolloResultType.SUCCESS) {
            this.viewStateService.hideComponent();
          }
        });
    }
  }

  onCancel(): void {
    this.viewStateService.hideComponent();
  }

  onSelectionChange(selected: ExternalReferenceListElementFragment): void {
    this.selected$.next(selected);
  }

  onSearch(value: string): void {
    this.externalReferenceService.update(this.paginationService.get(), { url: value });
  }

  onTryAgain(): void {
    this.externalReferenceService.recoverAfterError();
  }
}
