import { RelationOperation } from '@orion/graphql-types';
import { namedOperations } from '@orion/named-operations';
import { NotificationService } from '@orion/shared/services/global';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  ApolloResultType,
  createApolloContextErrorTitle,
} from '@csirtmu-orion/microframework/graphql';
import { ENTITY_NOT_LOADING_STATE, EntityLoadingState } from '@sentinel/common/utils';
import { SentinelNotificationTypeEnum } from '@sentinel/layout/notification';
import { BehaviorSubject, Observable } from 'rxjs';
import { catchError, concatMap, map, tap } from 'rxjs/operators';
import { UpdateExternalReferenceMutationService, UpdateExternalReferenceMutationVariables } from './graphql';

export abstract class ExternalReferenceActionsService {
  private removalInProgressSubject$: BehaviorSubject<EntityLoadingState> = new BehaviorSubject<EntityLoadingState>(
    ENTITY_NOT_LOADING_STATE
  );
  private addingInProgressSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  removalInProgress$: Observable<EntityLoadingState> = this.removalInProgressSubject$.asObservable();
  addingInProgress$: Observable<boolean> = this.addingInProgressSubject$.asObservable();

  protected constructor(
    protected notificationService: NotificationService,
    protected updateMutation: UpdateExternalReferenceMutationService
  ) {}

  protected abstract createVars(
    ticketId: string,
    op: RelationOperation
  ): Observable<UpdateExternalReferenceMutationVariables>;
  removeAssociation(ticketId: string): Observable<ApolloResult> {
    this.removalInProgressSubject$.next({ isLoading: true, id: ticketId });
    return this.handleOperation(ticketId, 'remove').pipe(
      tap({
        next: () => this.removalInProgressSubject$.next(ENTITY_NOT_LOADING_STATE),
        error: () => this.removalInProgressSubject$.next(ENTITY_NOT_LOADING_STATE),
      })
    );
  }

  addAssociation(ticketId: string): Observable<ApolloResult> {
    this.addingInProgressSubject$.next(true);
    return this.handleOperation(ticketId, 'add').pipe(
      tap({
        next: () => this.addingInProgressSubject$.next(false),
        error: () => this.addingInProgressSubject$.next(false),
      })
    );
  }

  protected handleOperation(ticketId: string, operation: 'remove' | 'add') {
    const errorTitle = `${operation === 'remove' ? 'Removing' : 'Adding'} external reference association`;
    const op = operation === 'remove' ? RelationOperation.REM : RelationOperation.ADD;
    return this.createVars(ticketId, op).pipe(
      concatMap((vars) =>
        this.updateMutation.mutate(vars, {
          refetchQueries: [namedOperations.Query.getExternalReferences],
          awaitRefetchQueries: true,
          context: createApolloContextErrorTitle(errorTitle),
        })
      ),
      map((res) => ApolloOperationResultResolver.resolve(res)),
      tap((res) => this.handleOperationResult(res, operation)),
      catchError((err) => ApolloOperationResultResolver.handleError(err))
    );
  }

  protected handleOperationResult(result: ApolloResult, operation: 'remove' | 'add'): void {
    if (result.type === ApolloResultType.SUCCESS) {
      const title = `External ticket was ${operation ? 'removed' : 'added'}`;
      this.notificationService.emit(SentinelNotificationTypeEnum.Success, title);
    }
  }
}
