import { Injectable } from '@angular/core';
import { ActivatedRoute } from '@angular/router';
import {
  CaseFileViewStateService,
  CreateEditCaseFileItemForm,
  UpdateDirectoryMutationService,
  UpdateFileMutationService,
} from '@orion/case-components/case-file';
import { namedOperations } from '@orion/named-operations';
import { EditEntityService } from '@orion/shared-services';
import { NotificationService } from '@orion/shared/services/global';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  createApolloContextErrorTitle,
} from '@csirtmu-orion/microframework/graphql';
import { Observable, of } from 'rxjs';
import { catchError, map, tap } from 'rxjs/operators';

@Injectable()
export class EditCaseFileItemService extends EditEntityService<CreateEditCaseFileItemForm> {
  private type: 'File' | 'Directory' = 'Directory';

  constructor(
    protected notificationService: NotificationService,
    protected activatedRoute: ActivatedRoute,
    private updateDirMutation: UpdateDirectoryMutationService,
    private updateFileMutation: UpdateFileMutationService,
    private viewStateService: CaseFileViewStateService
  ) {
    super('Directory', notificationService, activatedRoute);
  }

  protected createForm(): Observable<CreateEditCaseFileItemForm> {
    return of(new CreateEditCaseFileItemForm());
  }

  init(id: string, name: string, type: 'File' | 'Directory'): Observable<unknown> {
    this.entityName = type;
    this.type = type;
    this.entityId = id;
    return this.form$.pipe(tap((form) => form.setName(name)));
  }

  protected updateEntity(form: CreateEditCaseFileItemForm): Observable<ApolloResult> {
    if (this.entityId) {
      if (this.type === 'Directory') {
        return this.updateDirectory(this.entityId, form);
      } else {
        return this.updateFile(this.entityId, form);
      }
    } else {
      throw new Error(`ID of edited ${this.entityName} is no properly set up`);
    }
  }

  private updateDirectory(id: string, form: CreateEditCaseFileItemForm): Observable<ApolloResult> {
    const vars = { id: id, patch: { name: form.getData().name } };
    return this.updateDirMutation
      .mutate(vars, {
        refetchQueries: [namedOperations.Query.getActiveCaseFileItem],
        awaitRefetchQueries: true,
        context: createApolloContextErrorTitle('Updating directory'),
      })
      .pipe(
        map((res) => ApolloOperationResultResolver.resolve(res)),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  private updateFile(id: string, form: CreateEditCaseFileItemForm): Observable<ApolloResult> {
    const vars = { id: id, patch: { displayName: form.getData().name } };
    return this.updateFileMutation
      .mutate(vars, {
        refetchQueries: [namedOperations.Query.getActiveCaseFileItem],
        awaitRefetchQueries: true,
        context: createApolloContextErrorTitle('Updating file'),
      })
      .pipe(
        map((res) => ApolloOperationResultResolver.resolve(res)),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  protected handleCancel(): Observable<boolean> {
    this.viewStateService.clearEdit();
    return of(true);
  }

  protected onSaveError(): Observable<boolean> {
    return of(false);
  }

  protected onSaveSuccess() {
    this.viewStateService.clearEdit();
    return of(true);
  }
}
