import { Injectable } from '@angular/core';
import { CreateFileDTO } from '@orion/case-types';
import { NotificationService } from '@orion/shared/services/global';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  ApolloResultType,
  wrapApolloContextForFileUpload,
} from '@csirtmu-orion/microframework/graphql';
import { BehaviorSubject, forkJoin, Observable, of } from 'rxjs';
import { catchError, map, tap } from 'rxjs/operators';
import { CreateFileMutationService } from '../../../graphql';
import {
  ActiveCaseFileItemQueryService,
  ActiveCaseFileItemStateService,
  CaseFileViewStateService,
} from '../../../services';

@Injectable()
export class UploadCaseFilesService {
  protected savingSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  saving$: Observable<boolean> = this.savingSubject$.asObservable();

  constructor(
    protected viewState: CaseFileViewStateService,
    private activeCaseFile: ActiveCaseFileItemStateService,
    private createFilesMutation: CreateFileMutationService,
    private notificationService: NotificationService,
    private activeCaseFileQuery: ActiveCaseFileItemQueryService
  ) {}

  cancel(): void {
    this.viewState.hideUploadFileComponent();
  }

  upload(files: File[]): Observable<ApolloResult> {
    const parent = this.activeCaseFile.getActiveCaseFile();
    if (parent?.__typename === 'Directory') {
      const createFileDTOs = files.map((file) => new CreateFileDTO(file, parent.id));
      this.savingSubject$.next(true);
      return this.uploadMultipleFiles(createFileDTOs).pipe(
        tap(
          () => this.onUploadFinished(),
          () => this.onUploadFinished()
        )
      );
    } else {
      this.onUploadFinished();
      this.notificationService.emit(
        'error',
        'File parent directory was not initialized. Please report this bug to the developers'
      );
      return of(new ApolloResult(ApolloResultType.ERROR));
    }
  }

  private uploadMultipleFiles(dtos: CreateFileDTO[]): Observable<ApolloResult> {
    const uploads$ = dtos.map((dto) => this.uploadSingleFile(dto));
    return forkJoin(uploads$).pipe(map((results) => this.handleUploadResults(results)));
  }

  private uploadSingleFile(dto: CreateFileDTO): Observable<ApolloResult> {
    return this.createFilesMutation
      .mutate(
        { file: dto.file, filename: dto.name, extension: dto.extension, parentId: dto.parent },
        {
          context: wrapApolloContextForFileUpload({}),
        }
      )
      .pipe(
        map((res) => ApolloOperationResultResolver.resolve(res)),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  private handleUploadResults(results: ApolloResult[]): ApolloResult {
    if (results.some((result) => result.type === ApolloResultType.ERROR)) {
      this.notificationService.emit('error', 'Upload of one or more files failed.');
      return new ApolloResult(ApolloResultType.ERROR);
    } else {
      this.notificationService.emit('success', 'Files were successfully uploaded');
      return new ApolloResult(ApolloResultType.SUCCESS);
    }
  }

  private onUploadFinished(): void {
    this.savingSubject$.next(false);
    this.viewState.hideUploadFileComponent();
    this.activeCaseFileQuery.recoverAfterError();
  }
}
