import { Injectable } from '@angular/core';
import { MatDialog } from '@angular/material/dialog';
import { DeleteCaseFileItemsMutationService } from '@orion/case-components/case-file';
import { namedOperations } from '@orion/named-operations';
import { NotificationService } from '@orion/shared/services/global';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  ApolloResultType,
  createApolloContextErrorTitle,
} from '@csirtmu-orion/microframework/graphql';
import {
  SentinelConfirmationDialogComponent,
  SentinelConfirmationDialogConfig,
  SentinelDialogResultEnum,
} from '@sentinel/components/dialogs';
import { EMPTY, Observable } from 'rxjs';
import { catchError, concatMap, map, tap } from 'rxjs/operators';

@Injectable()
export class DeleteFileItemService {
  constructor(
    private dialog: MatDialog,
    private deleteCaseFileItemMutation: DeleteCaseFileItemsMutationService,
    private notification: NotificationService
  ) {}

  delete(ids: string[]): Observable<ApolloResult> {
    return this.confirmDeletion(ids.length).pipe(
      concatMap((dialogResult) => this.handleDeletionDialogResult(dialogResult, ids))
    );
  }

  private confirmDeletion(itemCount: number): Observable<SentinelDialogResultEnum> {
    const singleItem = itemCount === 1;
    const title = singleItem ? 'Delete selected item' : `Delete ${itemCount} selected items`;
    const message = `Do you really want to delete ${singleItem ? 'this item' : 'these items'}`;
    const dialogConfig = new SentinelConfirmationDialogConfig(title, message, 'Cancel', 'Confirm');
    const dialogRef = this.dialog.open(SentinelConfirmationDialogComponent, { data: dialogConfig });
    return dialogRef
      .afterClosed()
      .pipe(
        map((res) =>
          res === SentinelDialogResultEnum.CONFIRMED
            ? SentinelDialogResultEnum.CONFIRMED
            : SentinelDialogResultEnum.DISMISSED
        )
      );
  }

  private handleDeletionDialogResult(dialogResult: SentinelDialogResultEnum, ids: string[]): Observable<ApolloResult> {
    return dialogResult === SentinelDialogResultEnum.CONFIRMED ? this.handleDeletion(ids) : EMPTY;
  }

  private handleDeletion(ids: string[]): Observable<ApolloResult> {
    const bulkAction = ids.length > 1;
    return this.deleteCaseFileItemMutation
      .mutate(
        { ids: ids },
        {
          refetchQueries: [namedOperations.Query.getActiveCaseFileItem],
          awaitRefetchQueries: true,
          context: createApolloContextErrorTitle(bulkAction ? 'Deleting selected items' : 'Deleting selected item'),
        }
      )
      .pipe(
        map((result) => ApolloOperationResultResolver.resolve(result)),
        tap((result) => this.handleDeleteMutationResult(result, bulkAction)),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  private handleDeleteMutationResult(result: ApolloResult, bulkAction = false): void {
    if (result.type === ApolloResultType.SUCCESS) {
      const message = bulkAction ? 'Selected items were deleted' : 'Selected item was deleted';
      this.notification.emit('success', message);
    }
  }
}
