import { Injectable } from '@angular/core';
import { CreateCaseRelationship } from '@orion/case-types';
import { namedOperations } from '@orion/named-operations';
import { NotificationService } from '@orion/shared/services/global';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  ApolloResultType,
  createApolloContextErrorTitle,
} from '@csirtmu-orion/microframework/graphql';
import { ENTITY_NOT_LOADING_STATE, EntityLoadingState } from '@sentinel/common/utils';
import { BehaviorSubject, Observable } from 'rxjs';
import { catchError, map, tap } from 'rxjs/operators';
import {
  CreateCaseRelationshipsMutationService,
  CreateCaseRelationshipsMutationVariables,
  DeleteCaseRelationshipMutationService,
} from './graphql';

@Injectable()
export class CaseRelationshipActionsService {
  private deletingInProgressSubject$: BehaviorSubject<EntityLoadingState> = new BehaviorSubject<EntityLoadingState>(
    ENTITY_NOT_LOADING_STATE
  );
  deletingInProgress$: Observable<EntityLoadingState> = this.deletingInProgressSubject$.asObservable();

  constructor(
    private notificationService: NotificationService,
    private createRelationshipMutation: CreateCaseRelationshipsMutationService,
    private deleteRelationshipService: DeleteCaseRelationshipMutationService
  ) {}

  create(relationships: CreateCaseRelationship[]): Observable<ApolloResult> {
    const vars: CreateCaseRelationshipsMutationVariables = { relationships: relationships };
    return this.createRelationshipMutation
      .mutate(vars, {
        refetchQueries: [namedOperations.Query.getCaseDetail],
        awaitRefetchQueries: true,
        context: createApolloContextErrorTitle('Creating case relationship'),
      })
      .pipe(
        map((result) => ApolloOperationResultResolver.resolve(result)),
        tap((result) => this.handleActionResult(result, 'create')),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  delete(id: string): Observable<ApolloResult> {
    this.deletingInProgressSubject$.next({ isLoading: true, id: id });
    return this.deleteRelationshipService
      .mutate(
        { id: id },
        {
          refetchQueries: [namedOperations.Query.getCaseDetail],
          awaitRefetchQueries: true,
          context: createApolloContextErrorTitle('Deleting case relationship'),
        }
      )
      .pipe(
        tap({
          next: () => this.deletingInProgressSubject$.next(ENTITY_NOT_LOADING_STATE),
          error: () => this.deletingInProgressSubject$.next(ENTITY_NOT_LOADING_STATE),
        }),
        map((result) => ApolloOperationResultResolver.resolve(result)),
        tap((result) => this.handleActionResult(result, 'delete')),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  private handleActionResult(result: ApolloResult, type: 'create' | 'delete'): void {
    if (result.type === ApolloResultType.SUCCESS) {
      this.onSuccess(type);
    }
  }

  private onSuccess(type: 'create' | 'delete'): void {
    const message =
      type === 'create' ? 'Relationship was successfully created' : 'Relationship was successfully deleted';
    this.notificationService.emit('success', message);
  }
}
