import { ChangeDetectionStrategy, Component, EventEmitter, Input, OnInit, Output } from '@angular/core';
import { CaseListElementFragment, CaseListService } from '@orion/case-components/shared';
import { CASES_PATH } from '@orion/case-paths';
import { resolveFromIdOrUrl } from '@orion/common';
import { CaseOrdEnum } from '@orion/graphql-types';
import {
  INIT_PAGINATION_EVENT_BASE_TOKEN,
  PaginationBaseInit,
  PaginationService,
} from '@csirtmu-orion/microframework/pagination-services';
import { trackByIndex } from '@sentinel/common/utils';
import { Observable } from 'rxjs';
import { map } from 'rxjs/operators';
import { CaseRelationshipViewType } from '../../case-relationship-view-type';
import { CaseRelationshipSavedEvent } from './case-relationship-saved-event';

@Component({
  selector: 'orion-create-case-relationship',
  templateUrl: './create-case-relationship.component.html',
  styleUrls: ['./create-case-relationship.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush,
  providers: [
    CaseListService,
    PaginationService,
    {
      provide: INIT_PAGINATION_EVENT_BASE_TOKEN,
      useValue: new PaginationBaseInit(50, CaseOrdEnum.NAME, 'asc'),
    },
  ],
})
export class CreateCaseRelationshipComponent implements OnInit {
  @Input() caseId?: string;
  @Input() possibleRelationshipTypes: CaseRelationshipViewType[] = [];
  @Output() save: EventEmitter<CaseRelationshipSavedEvent> = new EventEmitter<CaseRelationshipSavedEvent>();
  @Output() cancel: EventEmitter<void> = new EventEmitter<void>();
  casesToSelect$?: Observable<CaseListElementFragment[]>;
  casesLoading$?: Observable<boolean>;
  casesHasError$?: Observable<boolean>;
  selectedType?: CaseRelationshipViewType;
  selectedTargetCases: CaseListElementFragment[] = [];
  saveDisabled = true;
  trackByFn = trackByIndex;

  constructor(private casesService: CaseListService, private paginationService: PaginationService) {}

  ngOnInit(): void {
    this.casesToSelect$ = this.casesService.data$?.pipe(map((resource) => resource.elements));
    this.casesLoading$ = this.casesService.isLoading$;
    this.casesHasError$ = this.casesService.hasError$;
  }

  onSave(): void {
    if (this.selectedType && this.selectedTargetCases && this.selectedTargetCases.length > 0) {
      const event = new CaseRelationshipSavedEvent(
        this.selectedTargetCases.map((targetCase) => targetCase.id),
        this.selectedType
      );
      this.save.emit(event);
    }
  }

  onCaseSelectionChange(selectedCases: CaseListElementFragment[]): void {
    this.selectedTargetCases = selectedCases;
    this.updateSaveDisabled();
  }

  onCaseSearch(value: string): void {
    const maybeId = resolveFromIdOrUrl(value, CASES_PATH);
    this.casesService.update(this.paginationService.get(), { id: maybeId, exceptId: this.caseId });
  }

  onCaseSearchTryAgain(): void {
    this.casesService.recoverAfterError();
  }

  onTypeSelectionChange(): void {
    this.updateSaveDisabled();
  }

  onCancel(): void {
    this.cancel.emit();
  }

  private updateSaveDisabled(): void {
    this.saveDisabled = this.selectedType === undefined || this.selectedTargetCases.length < 1;
  }
}
