import { Injectable } from '@angular/core';
import { FetchResult } from '@apollo/client/core';
import { namedOperations } from '@orion/named-operations';
import { NotificationService } from '@orion/shared/services/global';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  ApolloResultType,
  createApolloContextErrorTitle,
} from '@csirtmu-orion/microframework/graphql';
import { BehaviorSubject, Observable } from 'rxjs';
import { catchError, map, tap } from 'rxjs/operators';
import { RejectCaseMutationService, ResolveCaseMutationService } from './graphql';

@Injectable()
export class CaseDetailActionService {
  constructor(
    private rejectCaseMutation: RejectCaseMutationService,
    private resolveCaseMutation: ResolveCaseMutationService,
    private notificationService: NotificationService
  ) {}

  private resolveInProgressSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  resolveInProgress$: Observable<boolean> = this.resolveInProgressSubject$.asObservable();

  private rejectInProgressSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  rejectInProgress$: Observable<boolean> = this.rejectInProgressSubject$.asObservable();

  resolve(id: string): Observable<ApolloResult> {
    this.resolveInProgressSubject$.next(true);
    return this.resolveCaseMutation
      .mutate(
        { caseId: id },
        {
          refetchQueries: [namedOperations.Query.getCaseDetail],
          context: createApolloContextErrorTitle('Resolving case'),
          awaitRefetchQueries: true,
        }
      )
      .pipe(
        map((res) => this.handleActionResponse(res, 'resolve')),
        tap({
          next: () => this.resolveInProgressSubject$.next(false),
          error: () => this.resolveInProgressSubject$.next(false),
        }),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  reject(id: string): Observable<ApolloResult> {
    this.rejectInProgressSubject$.next(true);
    return this.rejectCaseMutation
      .mutate(
        { caseId: id },
        {
          refetchQueries: [namedOperations.Query.getCaseDetail],
          context: createApolloContextErrorTitle('Rejecting case'),
          awaitRefetchQueries: true,
        }
      )
      .pipe(
        map((res) => this.handleActionResponse(res, 'reject')),
        catchError((err) => ApolloOperationResultResolver.handleError(err)),
        tap({
          next: () => this.rejectInProgressSubject$.next(false),
          error: () => this.rejectInProgressSubject$.next(false),
        })
      );
  }

  private handleActionResponse(fetchResult: FetchResult, action: 'resolve' | 'reject'): ApolloResult {
    const result = ApolloOperationResultResolver.resolve(fetchResult);
    if (result.type === ApolloResultType.SUCCESS) {
      const pastTenseActionName = action === 'resolve' ? 'resolved' : 'rejected';
      this.notificationService.emit('success', `Case was successfully ${pastTenseActionName}`);
    }
    return result;
  }
}
