import {
  ChangeDetectionStrategy,
  Component,
  EventEmitter,
  Inject,
  InjectionToken,
  Input,
  OnInit,
  Output,
} from '@angular/core';
import { AssigneeListElementFragment, AssigneesListService } from '@orion/case-services';
import { EntityOperationEventToRelationUpdateConverter } from '@orion/case-types';
import { CaseDbUserOrdEnum, CasePriorityEnum, CaseUpdate, LabelOrdEnum } from '@orion/graphql-types';
import { LabelListElementFragment, LabelListService } from '@orion/label-services';
import { TimeService } from '@orion/shared/services/global';
import { CurrentUserFragment, CurrentUserService } from '@orion/user-services';
import { Destroyable } from '@sentinel/common';
import { EntityOperationEvent } from '@sentinel/common/entity';
import { EntityUpdateEvent } from '@csirtmu-orion/microframework/graphql';
import { PaginationBaseInit, PaginationService } from '@csirtmu-orion/microframework/pagination-services';
import { SidebarState, SidebarStateService } from '@sentinel/layout/sidebar';
import { Observable } from 'rxjs';
import { map } from 'rxjs/operators';
import { CaseDetailSidebarFragment } from './case-detail-sidebar.fragment.generated';

const assigneesPaginationServiceToken = new InjectionToken<PaginationService>('assigneesPaginationService');
const labelsPaginationServiceToken = new InjectionToken<PaginationService>('labelsPaginationService');

@Component({
  selector: 'orion-case-detail-sidebar',
  templateUrl: './case-detail-sidebar.component.html',
  styleUrls: ['./case-detail-sidebar.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush,
  providers: [
    AssigneesListService,
    LabelListService,
    {
      provide: assigneesPaginationServiceToken,
      useFactory: () => new PaginationService(new PaginationBaseInit(50, CaseDbUserOrdEnum.NAME, 'asc')),
    },
    {
      provide: labelsPaginationServiceToken,
      useFactory: () => new PaginationService(new PaginationBaseInit(50, LabelOrdEnum.NAME, 'asc')),
    },
  ],
})
export class CaseDetailSidebarComponent extends Destroyable() implements OnInit {
  @Input() case?: CaseDetailSidebarFragment;
  @Output() updateCase: EventEmitter<EntityUpdateEvent<CaseUpdate>> = new EventEmitter<EntityUpdateEvent<CaseUpdate>>();
  possibleAssignees$?: Observable<AssigneeListElementFragment[]>;
  possibleLabels$?: Observable<LabelListElementFragment[]>;
  currUser$?: Observable<CurrentUserFragment>;
  today$?: Observable<Date>;
  state$?: Observable<SidebarState>;

  constructor(
    @Inject(assigneesPaginationServiceToken) private assigneesPaginationService: PaginationService,
    @Inject(labelsPaginationServiceToken) private labelsPaginationService: PaginationService,
    private assigneesService: AssigneesListService,
    private labelsService: LabelListService,
    private currUserService: CurrentUserService,
    private timeService: TimeService,
    private sidebarStateService: SidebarStateService
  ) {
    super();
  }

  ngOnInit(): void {
    this.possibleAssignees$ = this.assigneesService.data$?.pipe(map((resource) => resource.elements));
    this.possibleLabels$ = this.labelsService.data$?.pipe(map((resource) => resource.elements));
    this.today$ = this.timeService.today$;
    this.state$ = this.sidebarStateService.state$;
    this.currUser$ = this.currUserService.data$;
  }

  onAssigneeSearch(value: string): void {
    this.assigneesService.update(this.assigneesPaginationService.get(), { name: value });
  }

  onLabelSearch(value: string): void {
    this.labelsService.update(this.labelsPaginationService.get(), { name: value });
  }

  onAssigneeSave(saveResult: EntityOperationEvent<AssigneeListElementFragment>): void {
    if (this.case) {
      const assigneesUpdate = EntityOperationEventToRelationUpdateConverter.convert(saveResult);
      this.updateCase.next({ id: this.case.id, patch: { assignees: assigneesUpdate } });
    }
  }

  onLabelsSave(saveResult: EntityOperationEvent<LabelListElementFragment>): void {
    if (this.case) {
      const labelsUpdate = EntityOperationEventToRelationUpdateConverter.convert(saveResult);
      this.updateCase.next({ id: this.case.id, patch: { labels: labelsUpdate } });
    }
  }

  onPrioritySave(selectedPriority: CasePriorityEnum): void {
    if (this.case) {
      this.updateCase.next({ id: this.case.id, patch: { priority: selectedPriority } });
    }
  }

  onDueDateSave(date?: Date): void {
    if (this.case) {
      this.updateCase.next({ id: this.case.id, patch: { dueDate: date } });
    }
  }

  onSidebarClick(): void {
    this.sidebarStateService.toggleState();
  }
}
