import { Injectable } from '@angular/core';
import { ActivatedRoute, ParamMap, Params, Router } from '@angular/router';
import { AssigneeListElementFragment, GetAssigneeListQueryService } from '@orion/case-services';
import { CaseClientFilter } from '@orion/case-types';
import {
  CaseDbUserFilter1AndFilterEntry,
  CasePriorityEnum,
  CaseStateEnum,
  CaseTypeEnum,
  LabelFilter1AndFilterEntry,
} from '@orion/graphql-types';
import { GetLabelListQueryService, LabelListElementFragment } from '@orion/label-services';
import { QueryParamFilterService } from '@csirtmu-orion/microframework/filter-services';
import { enumFromValue, Optional } from '@sentinel/common/utils';
import { forkJoin, Observable, of } from 'rxjs';
import { catchError, map } from 'rxjs/operators';

@Injectable()
export class CaseFilterService extends QueryParamFilterService<CaseClientFilter> {
  constructor(
    protected router: Router,
    protected activatedRoute: ActivatedRoute,
    protected queryAssignees: GetAssigneeListQueryService,
    protected queryLabels: GetLabelListQueryService
  ) {
    super(router, activatedRoute);
  }

  addLabel(toAdd: LabelListElementFragment): void {
    const currFilter = this.filterSubject$.getValue();
    if (currFilter.labels && currFilter.labels.find((label) => label.id === toAdd.id)) {
      return;
    }
    const newLabels: LabelListElementFragment[] = [];
    if (currFilter.labels) {
      newLabels.push(...currFilter.labels, toAdd);
    } else {
      newLabels.push(toAdd);
    }
    this.updateAndSubmit({ labels: newLabels });
  }

  removeLabel(toRemove: LabelListElementFragment): void {
    const currFilter = this.filterSubject$.getValue();
    const newLabels = currFilter.labels?.filter((label) => label.id !== toRemove.id);
    this.updateAndSubmit({ labels: newLabels });
  }

  removeAssignee(toRemove: AssigneeListElementFragment): void {
    const currFilter = this.filterSubject$.getValue();
    const newAssignees = currFilter.assignees?.filter((assignee) => assignee.id !== toRemove.id);
    this.updateAndSubmit({ assignees: newAssignees });
  }

  protected createEmptyFilter(): CaseClientFilter {
    return {
      name: undefined,
      state: undefined,
      type: undefined,
      priority: undefined,
      from: undefined,
      to: undefined,
      dueDateUntil: undefined,
      labels: undefined,
      assignees: undefined,
    };
  }

  protected toQueryParams(filter: CaseClientFilter): Params {
    const params = super.toQueryParams(filter);
    delete params.assignees;
    const assigneeIds = filter.assignees?.map((assignee) => assignee.id);
    delete params.labels;
    const labelIds = filter.labels?.map((label) => label.id);
    if (filter.from) {
      params.from = filter.from.valueOf();
    }
    if (filter.to) {
      params.to = filter.to.valueOf();
    }
    if (filter.dueDateUntil) {
      params.dueDateUntil = filter.dueDateUntil.valueOf();
    }
    return { ...params, assigneeIds, labelIds };
  }

  protected createFilterFromRoute(queryParamMap: ParamMap, paramMap: ParamMap): Observable<CaseClientFilter> {
    const assigneeIds = queryParamMap.getAll('assigneeIds');
    const labelIds = queryParamMap.getAll('labelIds');
    const stateParam = queryParamMap.get('state');
    const priorityParam = queryParamMap.get('priority');
    const typeParam = queryParamMap.get('type');
    const nameParam = queryParamMap.get('name');
    const from = this.deslugifyTimestamp(queryParamMap.get('from'));
    const to = this.deslugifyTimestamp(queryParamMap.get('to'));
    const dueDate = this.deslugifyTimestamp(queryParamMap.get('dueDateUntil'));

    if (
      this.isParamMapEmpty(nameParam, stateParam, priorityParam, typeParam, assigneeIds, labelIds, from, to, dueDate)
    ) {
      return of(this.EMPTY_FILTER);
    }
    const name = this.deslugify(nameParam);
    const caseState = this.resolveState(this.deslugify(stateParam));
    const priority = this.resolvePriority(this.deslugify(priorityParam));
    const caseType = this.resolveType(this.deslugify(typeParam));
    const syncFilter = this.createBaseFilter(name, caseState, priority, caseType, from, to, dueDate);
    if (this.hasAsyncParams(assigneeIds, labelIds)) {
      return this.extendFilterOfAsyncParams(syncFilter, assigneeIds, labelIds);
    } else {
      return of(syncFilter);
    }
  }

  private isParamMapEmpty(
    nameParam: string | null,
    stateParam: string | null,
    priorityParam: string | null,
    typeParam: string | null,
    assigneeIds: string[],
    labelIds: string[],
    from?: Date,
    to?: Date,
    dueDate?: Date
  ): boolean {
    return (
      nameParam === null &&
      stateParam === null &&
      priorityParam === null &&
      typeParam === null &&
      from === undefined &&
      to === undefined &&
      dueDate === undefined &&
      assigneeIds.length === 0 &&
      labelIds.length === 0
    );
  }

  private resolveState(value?: string): Optional<CaseStateEnum> {
    return value ? enumFromValue(CaseStateEnum, value) : undefined;
  }

  private resolvePriority(value?: string): Optional<CasePriorityEnum> {
    return value ? enumFromValue(CasePriorityEnum, value) : undefined;
  }

  private resolveType(value?: string): Optional<CaseTypeEnum> {
    return value ? enumFromValue(CaseTypeEnum, value) : undefined;
  }

  private createBaseFilter(
    name?: string,
    state?: CaseStateEnum,
    priority?: CasePriorityEnum,
    type?: CaseTypeEnum,
    from?: Date,
    to?: Date,
    dueDateUntil?: Date
  ): CaseClientFilter {
    return {
      name,
      state,
      type,
      priority,
      from,
      to,
      dueDateUntil,
    };
  }

  private hasAsyncParams(assigneeIds: string[], labelIds: string[]): boolean {
    return assigneeIds.length > 0 || labelIds.length > 0;
  }

  private extendFilterOfAsyncParams(
    syncFilter: CaseClientFilter,
    assigneeIds: string[],
    labelIds: string[]
  ): Observable<CaseClientFilter> {
    const extendOfAssignees$ = this.extendFilterOfAsyncAssignees(syncFilter, assigneeIds);
    const extendOfLabels$ = this.extendFilterOfAsyncLabels(syncFilter, labelIds);
    if (assigneeIds.length > 0 && labelIds.length > 0) {
      return forkJoin([extendOfAssignees$, extendOfLabels$]).pipe(
        map((filters) => {
          return { ...filters[0], ...filters[1] };
        })
      );
    }
    if (assigneeIds.length > 0) {
      return extendOfAssignees$;
    } else {
      return extendOfLabels$;
    }
  }

  private extendFilterOfAsyncAssignees(
    syncFilter: CaseClientFilter,
    assigneeIds: string[]
  ): Observable<CaseClientFilter> {
    return this.fetchAssignees(assigneeIds).pipe(
      map((users) => this.extendFilterOfAssignees(syncFilter, users)),
      catchError(() => of(syncFilter))
    );
  }

  private extendFilterOfAsyncLabels(syncFilter: CaseClientFilter, labelIds: string[]): Observable<CaseClientFilter> {
    return this.fetchLabels(labelIds).pipe(
      map((labels) => this.extendFilterOfLabels(syncFilter, labels)),
      catchError(() => of(syncFilter))
    );
  }

  private fetchAssignees(ids: string[]): Observable<Optional<AssigneeListElementFragment[]>> {
    const filter: CaseDbUserFilter1AndFilterEntry = { AND: [{ OR: { id: { in_: ids } } }] };
    return this.queryAssignees.fetch({ filter: [filter] }).pipe(
      map((queryResult) => {
        if (queryResult?.data?.getCaseDbUserList?.data) {
          return queryResult?.data?.getCaseDbUserList?.data;
        } else {
          return undefined;
        }
      })
    );
  }

  private fetchLabels(ids: string[]): Observable<Optional<LabelListElementFragment[]>> {
    const filter: LabelFilter1AndFilterEntry = { AND: [{ OR: { id: { in_: ids } } }] };
    return this.queryLabels.fetch({ filter: [filter] }).pipe(
      map((queryResult) => {
        if (queryResult.data.getLabelList?.data) {
          return queryResult.data.getLabelList?.data;
        } else {
          return undefined;
        }
      })
    );
  }

  private extendFilterOfAssignees(
    baseFilter: CaseClientFilter,
    assignees?: AssigneeListElementFragment[]
  ): CaseClientFilter {
    return { ...baseFilter, assignees: assignees };
  }

  private extendFilterOfLabels(baseFilter: CaseClientFilter, labels?: LabelListElementFragment[]): CaseClientFilter {
    return { ...baseFilter, labels: labels };
  }
}
