import {
  ChangeDetectionStrategy,
  Component,
  EventEmitter,
  Inject,
  InjectionToken,
  Input,
  OnInit,
  Output,
} from '@angular/core';
import { CaseFilterService } from '@orion/case-components/filters';
import { AssigneeListElementFragment, AssigneesListService } from '@orion/case-services';
import { CaseClientFilter } from '@orion/case-types';
import { CaseDbUserOrdEnum, CasePriorityEnum, CaseStateEnum, LabelOrdEnum } from '@orion/graphql-types';
import { LabelChipFragment } from '@orion/label-components/shared';
import { LabelListElementFragment, LabelListService } from '@orion/label-services';
import { PaginationBaseInit, PaginationService } from '@csirtmu-orion/microframework/pagination-services';
import { SentinelControlItem } from '@sentinel/components/controls';
import { EntityPaginableFilterableOverviewDirective } from '@csirtmu-orion/microframework/entity-smart-directives';
import { Observable } from 'rxjs';
import { map, take } from 'rxjs/operators';
import { CaseOverviewElementFragment } from './case-overview-element.fragment.generated';

const assigneesPaginationServiceToken = new InjectionToken<PaginationService>('assigneesPaginationService');
const labelsPaginationServiceToken = new InjectionToken<PaginationService>('labelsPaginationService');

@Component({
  selector: 'orion-case-overview',
  templateUrl: './case-overview.component.html',
  styleUrls: ['./case-overview.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush,
  providers: [
    AssigneesListService,
    LabelListService,
    {
      provide: assigneesPaginationServiceToken,
      useFactory: () => new PaginationService(new PaginationBaseInit(50, CaseDbUserOrdEnum.NAME, 'asc')),
    },
    {
      provide: labelsPaginationServiceToken,
      useFactory: () => new PaginationService(new PaginationBaseInit(50, LabelOrdEnum.NAME, 'asc')),
    },
  ],
})
export class CaseOverviewComponent
  extends EntityPaginableFilterableOverviewDirective<CaseOverviewElementFragment, CaseClientFilter>
  implements OnInit
{
  @Input() controls: SentinelControlItem[] = [];
  @Output() submitFilter: EventEmitter<void> = new EventEmitter<void>();
  assignees$?: Observable<AssigneeListElementFragment[]>;
  labels$?: Observable<LabelListElementFragment[]>;

  constructor(
    @Inject(assigneesPaginationServiceToken) private assigneesPaginationService: PaginationService,
    @Inject(labelsPaginationServiceToken) private labelsPaginationService: PaginationService,
    private filterService: CaseFilterService,
    private assigneesService: AssigneesListService,
    private labelsService: LabelListService
  ) {
    super();
  }

  ngOnInit(): void {
    this.assignees$ = this.assigneesService.data$.pipe(map((paginated) => paginated.elements));
    this.labels$ = this.labelsService.data$.pipe(map((paginated) => paginated.elements));
  }

  getAssignees(name: string): void {
    this.assigneesService.update(this.assigneesPaginationService.get(), { name: name });
  }

  getLabels(name: string): void {
    this.labelsService.update(this.labelsPaginationService.get(), { name: name });
  }

  onControlsEvent(clickedControlItem: SentinelControlItem): void {
    clickedControlItem.result$?.pipe(take(1)).subscribe();
  }

  onLabelClicked(label: LabelChipFragment): void {
    this.filterService.addLabel(label);
  }

  onPriorityClicked(priority: CasePriorityEnum): void {
    this.filterService.updateAndSubmit({ priority: priority });
  }

  onStateClicked(state: CaseStateEnum): void {
    this.filterService.updateAndSubmit({ state: state });
  }

  onSubmit(): void {
    this.submitFilter.next();
  }
}
