import {
  ChangeDetectionStrategy,
  Component,
  EventEmitter,
  Inject,
  InjectionToken,
  Input,
  OnChanges,
  OnInit,
  Optional,
  Output,
  SimpleChanges,
} from '@angular/core';
import { CommentForm } from '@orion/comment-types';
import { Destroyable } from '@sentinel/common';
import { SentinelControlItem } from '@sentinel/components/controls';
import { FileAction, FileTypeResolver, MarkdownEditorAction } from '@sentinel/components/markdown-editor';
import { defer, of, Subject } from 'rxjs';
import { debounceTime, map, skipWhile, takeUntil } from 'rxjs/operators';
import { EntityCommentFileUploadService } from '../../entity-comment-file-upload.service';
import { CommentAutosaveService } from './comment-autosave.service';
import { TagAction } from './tag-action';

export const COMMENT_EDITOR_WATCHED_TAGS_TOKEN = new InjectionToken<string[]>('COMMENT_EDITOR_WATCHED_TAGS_TOKEN');

@Component({
  selector: 'orion-comment-editor',
  templateUrl: './comment-editor.component.html',
  styleUrls: ['./comment-editor.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush,
})
export class CommentEditorComponent extends Destroyable() implements OnInit, OnChanges {
  @Input() autosaveIdToken?: string;
  @Input() content?: string;
  @Input() cancellable = false;
  @Input() isLoading = false;
  @Input() saveCommentLabel = 'Comment';
  @Output() cancel: EventEmitter<void> = new EventEmitter<void>();
  @Output() saved: EventEmitter<string> = new EventEmitter<string>();
  controls: SentinelControlItem[] = [];
  commentForm?: CommentForm;
  actionDispatcher: Subject<MarkdownEditorAction> = new Subject<MarkdownEditorAction>();
  activeTag?: string;
  activeTagText?: string;

  constructor(
    private commentAutosave: CommentAutosaveService,
    private fileUploadService: EntityCommentFileUploadService<unknown, unknown>,
    @Optional() @Inject(COMMENT_EDITOR_WATCHED_TAGS_TOKEN) public watchedTags?: string[]
  ) {
    super();
  }

  ngOnInit(): void {
    const initValue = this.content ? this.content : this.commentAutosave.get(this.autosaveIdToken) ?? '';
    this.commentForm = new CommentForm(initValue);
    this.commentForm.formGroup.valueChanges
      .pipe(
        takeUntil(this.destroy$),
        debounceTime(300),
        skipWhile(
          (control) => this.autosaveIdToken === undefined || control.content === undefined || control.content === ''
        )
      )
      .subscribe((control) => {
        if (control.content) {
          this.commentAutosave.update(control.content, this.autosaveIdToken);
        }
      });
    this.controls = this.createControls();
    this.commentForm.formGroup.updateValueAndValidity();
  }

  ngOnChanges(changes: SimpleChanges): void {
    if ('content' in changes && this.content) {
      this.onContentChange(this.content);
    }
    if ('autosaveIdToken' in changes && this.autosaveIdToken) {
      this.commentForm?.setContent(this.commentAutosave.get(this.autosaveIdToken) ?? '');
    }
  }

  onControlItemClicked(item: SentinelControlItem): void {
    item.result$?.pipe(takeUntil(this.destroy$)).subscribe();
  }

  onContentChange(content: string): void {
    this.commentForm?.setContent(content);
  }

  private onSave(): void {
    this.saved.emit(this.commentForm?.getContent());
    this.commentForm?.reset();
    if (this.autosaveIdToken) {
      this.commentAutosave.clear(this.autosaveIdToken);
    }
  }

  private createControls(): SentinelControlItem[] {
    const cancel = new SentinelControlItem(
      'CANCEL',
      'Cancel',
      'warn',
      of(false),
      defer(() => this.cancel.emit())
    );
    const save = new SentinelControlItem(
      'SAVE',
      this.saveCommentLabel,
      'accent',
      this.commentForm?.valid$.pipe(map((valid) => !valid)),
      defer(() => this.onSave())
    );
    if (this.cancellable) {
      return [cancel, save];
    } else {
      return [save];
    }
  }

  onUploadFiles(files: File[]): void {
    this.fileUploadService
      .uploadFiles(files)
      .pipe(takeUntil(this.destroy$))
      .subscribe((results) =>
        results.forEach((result) =>
          this.actionDispatcher.next(
            new FileAction(result.url, result.file.name, FileTypeResolver.isImage(result.file))
          )
        )
      );
  }

  onActiveTagChange(activeTag: string): void {
    this.activeTag = activeTag;
  }

  onActiveTagTextChange(tagText: string): void {
    this.activeTagText = tagText;
  }

  onTagSelected(tagId: string): void {
    if (this.activeTag) {
      this.actionDispatcher.next(new TagAction(this.activeTag, tagId, this.activeTagText));
      this.activeTag = undefined;
      this.activeTagText = undefined;
    }
  }
}
