import { Directive, EventEmitter, Input, OnChanges, Output, SimpleChanges } from '@angular/core';
import {
  AddCommentEvent,
  CommentActionType,
  CommentEvent,
  DeleteCommentEvent,
  UpdateCommentEvent,
} from '@orion/comment-types';

@Directive()
export abstract class SmartCommentDirective<T extends { id: string }> implements OnChanges {
  @Input() currUserId?: string;
  @Input() comment?: T;
  @Input() isLoading: boolean = false;
  @Output() commentEvent: EventEmitter<CommentEvent> = new EventEmitter<CommentEvent>();
  actions: CommentActionType[] = [];
  editMode = false;
  isReplying = false;

  abstract resolveActions(comment: T): CommentActionType[];

  ngOnChanges(changes: SimpleChanges): void {
    if ('comment' in changes && this.comment) {
      this.actions = this.resolveActions(this.comment);
    }
    if ('currUserId' in changes && this.comment) {
      this.actions = this.resolveActions(this.comment);
    }
  }

  onAction(actionType: CommentActionType): void {
    if (actionType === CommentActionType.EDIT) {
      this.editMode = true;
    }
    if (actionType === CommentActionType.REPLY) {
      this.isReplying = true;
    }
    if (actionType === CommentActionType.DELETE) {
      this.onDeleteAction();
    }
  }

  onSaveEdit(content: string): void {
    if (this.comment) {
      const event = new UpdateCommentEvent(this.comment.id, content);
      this.commentEvent.emit(event);
      this.editMode = false;
    }
  }

  onAddReply(event: AddCommentEvent | string): void {
    let addEvent: AddCommentEvent;
    if (event instanceof AddCommentEvent) {
      addEvent = event;
    } else {
      addEvent = new AddCommentEvent(event);
    }
    addEvent.payload.replyToId = this.comment?.id;
    this.commentEvent.emit(addEvent);
    this.isReplying = false;
  }

  onCancelReply(): void {
    this.isReplying = false;
  }

  onCancelEdit(): void {
    this.editMode = false;
  }

  protected resolveSharedActionTypes(authorId: string, currUserId?: string): CommentActionType[] {
    if (currUserId && authorId === currUserId) {
      return [CommentActionType.EDIT, CommentActionType.DELETE];
    } else {
      return [];
    }
  }

  private onDeleteAction(): void {
    if (this.comment) {
      const event = new DeleteCommentEvent(this.comment.id);
      this.commentEvent.emit(event);
    }
  }
}
