import { FailedFileUpload, FileUploadResult, SuccessfulFileUpload } from '@orion/comment-types';
import { FilenameWithExtension, fileToNameAndExtension } from '@orion/common';
import { NotificationService } from '@orion/shared/services/global';
import * as Apollo from 'apollo-angular';
import { MutationResult } from 'apollo-angular';
import { forkJoin, Observable, of } from 'rxjs';
import { catchError, concatMap, map } from 'rxjs/operators';
import { wrapApolloContextForFileUpload } from '@csirtmu-orion/microframework/graphql';

export interface CommentFileRefs {
  entityId: string;
  authorId: string;
}

export abstract class EntityCommentFileUploadService<Mutation, Vars> {
  protected constructor(
    protected notificationService: NotificationService,
    protected uploadMutation: Apollo.Mutation<Mutation, Vars>
  ) {}

  protected abstract getRefs(): Observable<CommentFileRefs>;
  protected abstract createVars(file: File, fileMeta: FilenameWithExtension, entityId: string, authorId: string): Vars;
  protected abstract resolveUrlFromResponse(response: Mutation): string | undefined;

  uploadFiles(files: File[]): Observable<SuccessfulFileUpload[]> {
    return this.getRefs().pipe(
      concatMap((refs) => forkJoin(files.map((file) => this.uploadFile(file, refs)))),
      map((uploadResults) => {
        this.reportFailedUploads(uploadResults.filter((result) => result instanceof FailedFileUpload));
        return uploadResults.filter((result) => result instanceof SuccessfulFileUpload) as SuccessfulFileUpload[];
      })
    );
  }

  protected uploadFile(file: File, refs: CommentFileRefs): Observable<FileUploadResult> {
    const fileMeta = fileToNameAndExtension(file);
    const vars: Vars = this.createVars(file, fileMeta, refs.entityId, refs.authorId);
    return this.uploadMutation.mutate(vars, { context: wrapApolloContextForFileUpload({}) }).pipe(
      map((result) => this.resolveResponse(result, file)),
      catchError(() => of(new FailedFileUpload(file)))
    );
  }

  private resolveResponse(result: MutationResult<Mutation>, file: File): FileUploadResult {
    if (result.data) {
      const url = this.resolveUrlFromResponse(result.data);
      if (url) {
        return new SuccessfulFileUpload(file, encodeURI(url));
      }
    }
    return new FailedFileUpload(file);
  }

  private reportFailedUploads(failedFileUploads: FileUploadResult[]): void {
    if (failedFileUploads.length > 0) {
      this.notificationService.emit('warning', `${failedFileUploads.length} file(s) failed to upload`);
    }
  }
}
