import { ChangeDetectionStrategy, Component, Input, OnChanges, OnInit, SimpleChanges } from '@angular/core';
import { CommentFilterService, EntityCommentsActionService } from '@orion/comment-services';
import {
  AddCommentEvent,
  AddCommentPayload,
  CommentClientFilter,
  CommentEvent,
  CommentRefs,
  DeleteCommentEvent,
  DeleteCommentPayload,
  UpdateCommentEvent,
  UpdateCommentPayload,
} from '@orion/comment-types';
import { CommentOrdEnum } from '@orion/graphql-types';
import { Destroyable } from '@sentinel/common';
import { IPaginationEvent } from '@sentinel/common/pagination';
import {
  INIT_PAGINATION_EVENT_BASE_TOKEN,
  PaginationBaseInit,
  PaginationService,
} from '@csirtmu-orion/microframework/pagination-services';
import { EntityLoadingState } from '@sentinel/common/utils';
import { BehaviorSubject, defer, Observable } from 'rxjs';
import { distinctUntilChanged, map, shareReplay, takeUntil, tap } from 'rxjs/operators';
import { CommentAutosaveService, TopLevelCommentFragment } from './comment-list';
import { CommentSort } from './comments-sort';
import { CommentsService } from './comments.service';

export const initialSort = CommentOrdEnum.CREATED;
export const initialSortDir = 'asc';
export const pageSize = 50;

@Component({
  selector: 'orion-entity-comments',
  templateUrl: './entity-comments.component.html',
  styleUrls: ['./entity-comments.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush,
  providers: [
    CommentsService,
    CommentAutosaveService,
    PaginationService,
    {
      provide: INIT_PAGINATION_EVENT_BASE_TOKEN,
      useValue: new PaginationBaseInit(pageSize, initialSort, initialSortDir),
    },
  ],
})
export class EntityCommentsComponent extends Destroyable() implements OnInit, OnChanges {
  @Input() currUserId?: string;

  // display loading only after manual change of sort
  private isLoadingSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  isLoading$?: Observable<boolean> = this.isLoadingSubject$.asObservable().pipe(distinctUntilChanged());
  addingInProgress$?: Observable<boolean>;
  updatingInProgress$?: Observable<EntityLoadingState>;
  deletingInProgress$?: Observable<EntityLoadingState>;
  hasError$?: Observable<boolean>;
  showMoreDisplayed$?: Observable<boolean>;
  comments$?: Observable<TopLevelCommentFragment[]>;
  onTryAgain$: Observable<void> = defer(() => this.tryAgain());
  canWriteComment = false;
  commentSort: CommentSort = initialSortDir === 'asc' ? 'oldest' : 'newest';

  constructor(
    private commentActionService: EntityCommentsActionService,
    private paginationService: PaginationService,
    private filterService: CommentFilterService,
    private commentListService: CommentsService
  ) {
    super();
  }

  ngOnInit(): void {
    this.comments$ = this.commentListService.data$.pipe(
      map((data) => data.elements),
      shareReplay({ refCount: true }),
      tap({
        next: () => this.isLoadingSubject$.next(false),
        error: () => this.isLoadingSubject$.next(false),
        complete: () => this.isLoadingSubject$.next(false),
      })
    );
    this.showMoreDisplayed$ = this.commentListService.data$.pipe(
      map((data) => data.pagination.totalElements > data.pagination.numberOfElements),
      distinctUntilChanged(),
      shareReplay({ refCount: true })
    );
    this.hasError$ = this.commentListService.hasError$?.pipe(shareReplay({ refCount: true }));
    this.paginationService.pagination$
      .pipe(takeUntil(this.destroy$), shareReplay({ refCount: true }))
      .subscribe((pagination) => this.onPaginationChange(pagination));
    this.filterService.filter$
      .pipe(takeUntil(this.destroy$), shareReplay({ refCount: true }))
      .subscribe((filter) => this.onFilterChange(filter));
    this.addingInProgress$ = this.commentActionService.addingInProgress$.pipe(shareReplay({ refCount: true }));
    this.updatingInProgress$ = this.commentActionService.updatingInProgress$.pipe(shareReplay({ refCount: true }));
    this.deletingInProgress$ = this.commentActionService.deletingInProgress$.pipe(shareReplay({ refCount: true }));
  }

  ngOnChanges(changes: SimpleChanges): void {
    if ('currUserId' in changes) {
      this.canWriteComment = this.currUserId !== undefined;
    }
  }

  onShowMore(): void {
    const currPagination = this.paginationService.get();
    currPagination.size += pageSize;
    this.paginationService.update(currPagination);
  }

  onCommentEvent(commentEvent: CommentEvent): void {
    if (commentEvent instanceof AddCommentEvent) {
      this.addComment(commentEvent.payload);
    } else if (commentEvent instanceof UpdateCommentEvent) {
      this.updateComment(commentEvent.payload);
    } else if (commentEvent instanceof DeleteCommentEvent) {
      this.deleteComment(commentEvent.payload);
    }
  }

  onCommentSortChange(sort: CommentSort): void {
    this.paginationService.update({ ...this.paginationService.get(), sortDir: sort === 'newest' ? 'desc' : 'asc' });
  }

  private tryAgain(): void {
    this.commentListService.recoverAfterError();
    this.isLoadingSubject$.next(true);
  }

  private onPaginationChange(pagination: IPaginationEvent): void {
    this.commentSort = pagination.sortDir === 'asc' ? 'oldest' : 'newest';
    this.commentListService.update(pagination, this.filterService.get());
    this.isLoadingSubject$.next(true);
  }

  private onFilterChange(filter: CommentClientFilter): void {
    this.commentListService.update(this.paginationService.get(), filter);
  }

  private addComment(payload: AddCommentPayload): void {
    if (this.currUserId) {
      const refs: CommentRefs = { authorId: this.currUserId, replyToId: payload.replyToId };
      this.commentActionService.add(payload.content, refs).pipe(takeUntil(this.destroy$)).subscribe();
    }
  }

  private updateComment(payload: UpdateCommentPayload): void {
    this.commentActionService.update(payload.id, payload.content).pipe(takeUntil(this.destroy$)).subscribe();
  }

  private deleteComment(payload: DeleteCommentPayload): void {
    this.commentActionService.delete(payload.id).pipe(takeUntil(this.destroy$)).subscribe();
  }
}
