import { MatDialog } from '@angular/material/dialog';
import { namedOperations } from '@orion/named-operations';
import { NotificationService } from '@orion/shared/services/global';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  ApolloResultType,
  createApolloContextErrorTitle,
} from '@csirtmu-orion/microframework/graphql';
import { ENTITY_NOT_LOADING_STATE, EntityLoadingState } from '@sentinel/common/utils';
import {
  SentinelConfirmationDialogComponent,
  SentinelConfirmationDialogConfig,
  SentinelDialogResultEnum,
} from '@sentinel/components/dialogs';
import { SentinelNotificationTypeEnum } from '@sentinel/layout/notification';
import { BehaviorSubject, EMPTY, Observable } from 'rxjs';
import { catchError, concatMap, map, tap } from 'rxjs/operators';
import { CommentRefs, CreateCommentDTO, EntityCommentRefs } from '../../types';
import {
  CreateCommentMutationService,
  CreateCommentMutationVariables,
  DeleteCommentMutationService,
  DeleteCommentMutationVariables,
  UpdateCommentMutationService,
  UpdateCommentMutationVariables,
} from './comment.mutation.generated';

export abstract class EntityCommentsActionService {
  private addingInProgressSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  addingInProgress$: Observable<boolean> = this.addingInProgressSubject$.asObservable();
  private updatingInProgressSubject$: BehaviorSubject<EntityLoadingState> = new BehaviorSubject<EntityLoadingState>(
    ENTITY_NOT_LOADING_STATE
  );
  updatingInProgress$: Observable<EntityLoadingState> = this.updatingInProgressSubject$.asObservable();
  private deletingInProgressSubject$: BehaviorSubject<EntityLoadingState> = new BehaviorSubject<EntityLoadingState>(
    ENTITY_NOT_LOADING_STATE
  );
  deletingInProgress$: Observable<EntityLoadingState> = this.deletingInProgressSubject$.asObservable();

  protected constructor(
    protected dialog: MatDialog,
    protected notificationService: NotificationService,
    protected addCommentService: CreateCommentMutationService,
    protected updateCommentService: UpdateCommentMutationService,
    protected deleteCommentService: DeleteCommentMutationService
  ) {}

  abstract createEntityRefs(refs: CommentRefs): Observable<EntityCommentRefs>;

  add(content: string, refs: CommentRefs): Observable<ApolloResult> {
    this.addingInProgressSubject$.next(true);
    return this.createEntityRefs(refs).pipe(
      concatMap((refs) =>
        this.addCommentService.mutate(this.createCommentVars(content, refs), {
          refetchQueries: [namedOperations.Query.getComments],
          context: createApolloContextErrorTitle('Adding comment'),
          awaitRefetchQueries: true,
        })
      ),
      map((res) => ApolloOperationResultResolver.resolve(res)),
      tap({
        next: () => this.addingInProgressSubject$.next(false),
        error: () => this.addingInProgressSubject$.next(false),
      }),
      catchError((err) => ApolloOperationResultResolver.handleError(err))
    );
  }

  update(id: string, updatedContent: string): Observable<ApolloResult> {
    this.updatingInProgressSubject$.next({ isLoading: true, id: id });
    const vars: UpdateCommentMutationVariables = { id: id, input: { content: updatedContent } };
    return this.updateCommentService
      .mutate(vars, {
        refetchQueries: [namedOperations.Query.getComments],
        context: createApolloContextErrorTitle('Updating comment'),
        awaitRefetchQueries: true,
      })
      .pipe(
        map((res) => ApolloOperationResultResolver.resolve(res)),
        tap({
          next: () => this.updatingInProgressSubject$.next(ENTITY_NOT_LOADING_STATE),
          error: () => this.updatingInProgressSubject$.next(ENTITY_NOT_LOADING_STATE),
        }),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  delete(id: string): Observable<ApolloResult> {
    return this.confirmDeletion().pipe(
      concatMap((dialogResult) => this.handleDeletionDialogResult(dialogResult, id)),
      tap({
        next: () => this.deletingInProgressSubject$.next(ENTITY_NOT_LOADING_STATE),
        error: () => this.deletingInProgressSubject$.next(ENTITY_NOT_LOADING_STATE),
      }),
      catchError((err) => ApolloOperationResultResolver.handleError(err)),
      tap((result) => this.handleDeletionResult(result))
    );
  }

  private createCommentVars(content: string, entityRefs: EntityCommentRefs): CreateCommentMutationVariables {
    const comment = new CreateCommentDTO(content, entityRefs);
    return { comment: comment };
  }

  private confirmDeletion(): Observable<SentinelDialogResultEnum> {
    const title = 'Delete comment';
    const message = 'Do you really want to delete this comment?';
    const confirmationConfig = new SentinelConfirmationDialogConfig(title, message, 'Cancel', 'Confirm');
    const dialogRef = this.dialog.open(SentinelConfirmationDialogComponent, { data: confirmationConfig });
    return dialogRef
      .afterClosed()
      .pipe(
        map((result) =>
          result === SentinelDialogResultEnum.CONFIRMED
            ? SentinelDialogResultEnum.CONFIRMED
            : SentinelDialogResultEnum.DISMISSED
        )
      );
  }

  private handleDeletionDialogResult(dialogResult: SentinelDialogResultEnum, id: string): Observable<ApolloResult> {
    if (dialogResult === SentinelDialogResultEnum.CONFIRMED) {
      this.deletingInProgressSubject$.next({ id: id, isLoading: true });
      const vars: DeleteCommentMutationVariables = { id: id };
      return this.deleteCommentService
        .mutate(vars, {
          refetchQueries: [namedOperations.Query.getComments],
          context: createApolloContextErrorTitle('Deleting comment'),
          awaitRefetchQueries: true,
        })
        .pipe(
          tap({
            next: () => this.updatingInProgressSubject$.next(ENTITY_NOT_LOADING_STATE),
            error: () => this.updatingInProgressSubject$.next(ENTITY_NOT_LOADING_STATE),
          }),
          map((res) => ApolloOperationResultResolver.resolve(res)),
          catchError((err) => ApolloOperationResultResolver.handleError(err))
        );
    } else {
      return EMPTY;
    }
  }

  private handleDeletionResult(result: ApolloResult): void {
    if (result.type === ApolloResultType.SUCCESS) {
      const message = 'Comment was deleted';
      this.notificationService.emit(SentinelNotificationTypeEnum.Success, message);
    }
  }
}
