import { Injectable } from '@angular/core';
import { ApolloQueryResult } from '@apollo/client/core';
import { CaseTableFragment } from '@orion/case-components/shared';
import { TaskListElementFragment } from '@orion/task-components/shared';
import { CurrentUserService } from '@orion/user-services';
import { ApolloQueryService } from '@csirtmu-orion/microframework/apollo-services';
import { DashboardQuery, DashboardQueryService, DashboardQueryVariables } from './graphql';
import { toIsoDateFormat } from '@orion/common';

export interface NumberCard {
  name: string;
  value: number;
}

export interface DashboardQueryResult {
  caseTodo: CaseTableFragment[];
  taskTodo: TaskListElementFragment[];
  numberCards: NumberCard[];
}

@Injectable()
export class DashboardService extends ApolloQueryService<
  DashboardQuery,
  DashboardQueryVariables,
  DashboardQueryResult
> {
  now: Date;
  lastWeek: Date;

  constructor(getQuery: DashboardQueryService, private activeUserService: CurrentUserService) {
    super(getQuery, 'Fetching Dashboard');
    this.now = new Date();
    const weekInMillis = 7 * 24 * 60 * 60 * 1000;
    this.lastWeek = new Date(this.now.getTime() - weekInMillis);
  }

  update(): void {
    this.updateVariables();
    if (!this.initialized$.getValue()) {
      this.initialize();
    }
    this.updateQuery();
  }

  protected createVariables(): DashboardQueryVariables {
    const id = this.activeUserService.getCurrentUser()?.id;
    if (id) {
      return {
        userId: id,
        today: toIsoDateFormat(this.now, true),
        lastWeek: toIsoDateFormat(this.lastWeek, true),
      };
    } else {
      throw Error('Active user was not initialized');
    }
  }

  protected toData(queryRes: ApolloQueryResult<DashboardQuery>): DashboardQueryResult {
    const casesOpenThisWeek = queryRes?.data?.caseOpenInTimeRange?.totalCount;
    const casesClosedThisWeek = queryRes?.data?.caseClosedInTimeRange?.totalCount;
    const casesPastDue = queryRes?.data?.casesPastDueDate?.totalCount;
    const tasksOpenThisWeek = queryRes?.data?.taskOpenInTimeRange?.totalCount;
    const tasksClosedThisWeek = queryRes?.data?.taskClosedInTimeRange?.totalCount;
    const tasksPastDue = queryRes?.data?.tasksPastDueDate?.totalCount;
    return {
      caseTodo: queryRes?.data?.caseTodo?.data ?? [],
      taskTodo: queryRes?.data?.taskTodo?.data ?? [],
      numberCards: [
        {
          name: 'Cases open this week',
          value: casesOpenThisWeek,
        },
        {
          name: 'Cases closed this week',
          value: casesClosedThisWeek,
        },
        {
          name: 'Cases past due',
          value: casesPastDue,
        },
        {
          name: 'Tasks open this week',
          value: tasksOpenThisWeek,
        },
        {
          name: 'Tasks closed this week',
          value: tasksClosedThisWeek,
        },
        {
          name: 'Tasks past due',
          value: tasksPastDue,
        },
      ],
    };
  }
}
