import { Injectable } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { LabelForm } from '@orion/label-types';
import { CreateEditEntityService } from '@orion/shared-services';
import { NotificationService } from '@orion/shared/services/global';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  createApolloContextErrorTitle,
} from '@csirtmu-orion/microframework/graphql';
import { EMPTY, from, Observable, of } from 'rxjs';
import { catchError, concatMap, map, take, tap } from 'rxjs/operators';
import { CreateLabelMutationService, LabelEditFragment, UpdateLabelMutationService } from './graphql';
import { LABEL_PARAM_SELECTOR } from '@orion/label-paths';
import { Optional } from '@sentinel/common/utils';
import { GetLabelEditQueryService } from '@orion/label-services';

@Injectable()
export class CreateEditLabelService extends CreateEditEntityService<LabelForm> {
  constructor(
    protected notificationService: NotificationService,
    protected activatedRoute: ActivatedRoute,
    private router: Router,
    private getLabelQuery: GetLabelEditQueryService,
    private createMutation: CreateLabelMutationService,
    private updateMutation: UpdateLabelMutationService
  ) {
    super('Label', notificationService, activatedRoute);
  }

  protected createForm(): Observable<LabelForm> {
    return this.getEntityId().pipe(
      take(1),
      concatMap((id) => {
        if (id === undefined) {
          return of(undefined);
        } else {
          return this.getLabelQuery.fetch({ id: id }).pipe(map((queryResult) => queryResult?.data?.getLabel));
        }
      }),
      tap((label) => this.resolveEditMode(label)),
      map((label) => new LabelForm(label))
    );
  }

  protected getEntityId(): Observable<Optional<string>> {
    return this.activatedRoute.paramMap.pipe(
      take(1),
      map((paramMap) => paramMap.get(LABEL_PARAM_SELECTOR) ?? undefined)
    );
  }

  protected updateEntity(form: LabelForm): Observable<ApolloResult> {
    return this.entityId$.pipe(
      map((entityId) => {
        return entityId ? { id: entityId, patch: form.getData() } : undefined;
      }),
      concatMap((vars) =>
        vars !== undefined
          ? this.updateMutation.mutate(vars, { context: createApolloContextErrorTitle('Updating label') })
          : EMPTY
      ),
      map((res) => ApolloOperationResultResolver.resolve(res)),
      catchError((err) => ApolloOperationResultResolver.handleError(err))
    );
  }

  protected createEntity(form: LabelForm): Observable<ApolloResult> {
    return this.createMutation
      .mutate({ label: form.getData() }, { context: createApolloContextErrorTitle('Creating label') })
      .pipe(
        map((res) => ApolloOperationResultResolver.resolve(res)),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  protected handleCancel(): Observable<boolean> {
    return this.navigateToOverview();
  }

  protected onSaveError(): Observable<boolean> {
    return of(false);
  }

  protected onSaveSuccess(): Observable<boolean> {
    return this.navigateToOverview();
  }

  private resolveEditMode(label?: LabelEditFragment): void {
    this.editMode = label !== undefined && label !== null;
  }

  private navigateToOverview(): Observable<boolean> {
    const command = this.editMode ? '../../' : '../';
    return from(this.router.navigate([command], { relativeTo: this.activatedRoute }));
  }
}
