import { Injectable } from '@angular/core';
import { MatDialog } from '@angular/material/dialog';
import { ActivatedRoute, Router } from '@angular/router';
import { DeleteLabelsMutationService, LabelOverviewFragment } from '@orion/label-components/overview';
import { namedOperations } from '@orion/named-operations';
import { NotificationService } from '@orion/shared/services/global';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  ApolloResultType,
  createApolloContextErrorTitle,
} from '@csirtmu-orion/microframework/graphql';
import { ENTITY_NOT_LOADING_STATE, EntityLoadingState } from '@sentinel/common/utils';
import {
  SentinelConfirmationDialogComponent,
  SentinelConfirmationDialogConfig,
  SentinelDialogResultEnum,
} from '@sentinel/components/dialogs';
import { SentinelNotificationTypeEnum } from '@sentinel/layout/notification';
import { BehaviorSubject, EMPTY, from, Observable } from 'rxjs';
import { catchError, concatMap, map, tap } from 'rxjs/operators';
import { LabelTableFragment } from '../components/overview/label-table/label-table.fragment.generated';

type InputFragments = LabelOverviewFragment | LabelTableFragment;

@Injectable()
export class LabelOverviewActionsService {
  private deletingInProgressSubject$: BehaviorSubject<EntityLoadingState> = new BehaviorSubject<EntityLoadingState>(
    ENTITY_NOT_LOADING_STATE
  );
  deletingInProgress$: Observable<EntityLoadingState> = this.deletingInProgressSubject$.asObservable();

  constructor(
    private router: Router,
    private activatedRoute: ActivatedRoute,
    private notificationService: NotificationService,
    private dialog: MatDialog,
    private deleteMutation: DeleteLabelsMutationService
  ) {}

  create(): Observable<boolean> {
    return from(this.router.navigate(['create'], { relativeTo: this.activatedRoute }));
  }

  edit(label: InputFragments): Observable<boolean> {
    return from(this.router.navigate([label?.id, 'edit'], { relativeTo: this.activatedRoute }));
  }

  delete(labels: InputFragments[]): Observable<ApolloResult> {
    const singleLabel = labels.length === 1;
    return this.confirmDeletion(labels).pipe(
      concatMap((dialogResult) => this.handleDeletionDialogResult(dialogResult, labels)),
      tap((result) => this.handleDeletionResult(result, singleLabel))
    );
  }

  private confirmDeletion(labels: InputFragments[]): Observable<SentinelDialogResultEnum> {
    const singleLabel = labels.length === 1;
    const title = singleLabel ? `Delete label ${labels[0].name}` : `Delete ${labels.length} labels`;
    const message = `Do you really want to delete ${singleLabel ? 'this label' : 'these labels'}?`;
    const confirmationConfig = new SentinelConfirmationDialogConfig(title, message, 'Cancel', 'Confirm');
    const dialogRef = this.dialog.open(SentinelConfirmationDialogComponent, {
      data: confirmationConfig,
    });
    return dialogRef
      .afterClosed()
      .pipe(
        map((res) =>
          res === SentinelDialogResultEnum.CONFIRMED
            ? SentinelDialogResultEnum.CONFIRMED
            : SentinelDialogResultEnum.DISMISSED
        )
      );
  }

  private handleDeletionDialogResult(
    dialogResult: SentinelDialogResultEnum,
    labels: InputFragments[]
  ): Observable<ApolloResult> {
    return dialogResult === SentinelDialogResultEnum.CONFIRMED ? this.handleDeletion(labels) : EMPTY;
  }

  private handleDeletion(labels: InputFragments[]): Observable<ApolloResult> {
    const ids = labels.map((label) => label.id);
    this.deletingInProgressSubject$.next({ isLoading: true, ids: ids });
    return this.deleteMutation
      .mutate(
        { ids: ids },
        {
          refetchQueries: [namedOperations.Query.getLabelOverview],
          awaitRefetchQueries: true,
          context: createApolloContextErrorTitle('Deleting label'),
        }
      )
      .pipe(
        tap({
          next: () => this.deletingInProgressSubject$.next(ENTITY_NOT_LOADING_STATE),
          error: () => this.deletingInProgressSubject$.next(ENTITY_NOT_LOADING_STATE),
        }),
        map((res) => ApolloOperationResultResolver.resolve(res)),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  private handleDeletionResult(result: ApolloResult, singleLabel: boolean): void {
    if (result.type === ApolloResultType.SUCCESS) {
      const message = singleLabel ? 'Label was deleted' : 'Labels were deleted';
      this.notificationService.emit(SentinelNotificationTypeEnum.Success, message);
    }
  }
}
