import { SecurityContext } from '@angular/core';
import { ASSETS_PATH } from '@orion/asset-paths';
import { CASES_PATH } from '@orion/case-paths';
import { enumToPath, toAuthPath } from '@orion/common';
import { EXTERNAL_REFERENCES_PATH } from '@orion/external-reference-paths';
import { KnownAssetsEnum } from '@orion/graphql-types';
import { OBSERVABLES_PATH } from '@orion/observable-paths';
import { TTP_PATH } from '@orion/ttps-paths';
import { USERS_PATH } from '@orion/user-paths';
import { enumFromValue } from '@sentinel/common/utils';
import { MarkdownConfig } from '@sentinel/components/markdown-editor';
import { MarkedOptions } from '@sentinel/components/markdown-view';
import { Renderer } from 'marked';

const defaultRenderer = new Renderer();

const renderer = {
  // replace absolute links within the frontend app and replace them with relative links
  // to avoid triggering browser full reload navigation
  link(href: string | null, title: string | null, text: string): string {
    const renderedLinkTag = defaultRenderer.link(href, title, text);
    if (renderedLinkTag.includes(window.location.hostname)) {
      const absoluteUrlRegex = new RegExp(
        `(http://|https://)?(${window.location.hostname})(:${window.location.port})?(\\S+)">`
      );
      const regexResult = absoluteUrlRegex.exec(renderedLinkTag);
      if (regexResult && regexResult.length > 0) {
        const relativeUrl = regexResult[regexResult.length - 1];
        if (relativeUrl) {
          return renderedLinkTag.replace(absoluteUrlRegex, relativeUrl + '">');
        }
      }
    }
    return renderedLinkTag;
  },
  //
  // // wrap image requests with authorization tokens and load them lazily
  // image(href: string | null, title: string | null, text: string) {
  //   if (href === null) {
  //     return text;
  //   }
  //
  //   let out = '<img src="' + href + '" alt="' + text;
  //   if (title) {
  //     out += ' title="' + title + '"';
  //   }
  //   // @ts-ignore
  //   out += this.options.xhtml ? '/>' : '>';
  //   console.log(out);
  //   return out;
  // }
};

function userIdToHref(id: string): string {
  return toAuthPath(`${USERS_PATH}/${id}`);
}

function caseIdToHref(id: string): string {
  return toAuthPath(`${CASES_PATH}/${id}`);
}

function taskIdToHref(id: string): string {
  return toAuthPath(`task/${id}`);
}

function assetTypeIdToHref(id: string, assetType: string): string {
  const type = enumToPath(enumFromValue(KnownAssetsEnum, assetType));
  return toAuthPath(`${ASSETS_PATH}/${type}/${id}`);
}

function observableIdToHref(id: string): string {
  return toAuthPath(`${OBSERVABLES_PATH}/${id}`);
}

function ttpIdToHref(id: string): string {
  return toAuthPath(`${TTP_PATH}/${id}`);
}

function externalReferenceIdToHref(id: string): string {
  return toAuthPath(`${EXTERNAL_REFERENCES_PATH}/${id}`);
}

function createTagStartRegex(tagChar: string): RegExp {
  return new RegExp(`^[\\${tagChar}]|\\s[\\${tagChar}]`);
}

function createTagTokenRegex(tagChar: string): RegExp {
  return new RegExp(`^[\\${tagChar}]\\d+|^\\s[\\${tagChar}]\\d+`);
}

// @ts-ignore
function renderAnchor(token): string {
  // @ts-ignore
  return ` <a href="${token?.href}" class="tag-anchor">${token.text}</a>`;
}

// @ts-ignore
const userTag = {
  name: 'userTag',
  level: 'inline',
  start(src: string) {
    return src.match(createTagStartRegex('@'))?.index;
  },
  // @ts-ignore
  tokenizer(src: string, tokens: any[]) {
    const match = createTagTokenRegex('@').exec(src);
    if (match) {
      return {
        type: 'userTag',
        raw: match[0],
        href: userIdToHref(match[0].split('@')[1]),
        text: match[0].trim(),
        tokens: [],
      };
    }
    return;
  },
  // @ts-ignore
  renderer(token) {
    return renderAnchor(token);
  },
};

// @ts-ignore
const caseTag = {
  name: 'caseTag',
  level: 'inline',
  start(src: string) {
    return src.match(createTagStartRegex('#'))?.index;
  },
  // @ts-ignore
  tokenizer(src: string, tokens: any[]) {
    const match = createTagTokenRegex('#').exec(src);
    if (match) {
      return {
        type: 'caseTag',
        raw: match[0],
        href: caseIdToHref(match[0].split('#')[1]),
        text: match[0].trim(),
      };
    }
    return;
  },
  // @ts-ignore
  renderer(token) {
    return renderAnchor(token);
  },
};

// @ts-ignore
const taskTag = {
  name: 'taskTag',
  level: 'inline',
  start(src: string) {
    return src.match(createTagStartRegex('$'))?.index;
  },
  // @ts-ignore
  tokenizer(src: string, tokens: any[]) {
    const match = createTagTokenRegex('$').exec(src);
    if (match) {
      return {
        type: 'taskTag',
        raw: match[0],
        href: taskIdToHref(match[0].split('$')[1]),
        text: match[0].trim(),
      };
    }
    return;
  },
  // @ts-ignore
  renderer(token) {
    return renderAnchor(token);
  },
};

// @ts-ignore
const assetTag = {
  name: 'assetTag',
  level: 'inline',
  start(src: string) {
    return src.match(createTagStartRegex('%'))?.index;
  },
  // @ts-ignore
  tokenizer(src: string, tokens: any[]) {
    const match = new RegExp(`^[%]\\w+[/]\\d+|^\\s[%]\\w+[/]\\d+`).exec(src);
    if (!match) {
      return;
    }
    const assetNavToken = match[0].split('%')[1];
    if (assetNavToken) {
      const assetNavTokenParts = assetNavToken.split('/');
      const assetType = assetNavTokenParts[0];
      const assetId = assetNavTokenParts[1];
      if (assetId && assetType) {
        return {
          type: 'assetTag',
          raw: match[0],
          href: assetTypeIdToHref(assetId, assetType),
          text: match[0].trim(),
        };
      }
    }
    return;
  },
  // @ts-ignore
  renderer(token) {
    return renderAnchor(token);
  },
};

// @ts-ignore
const observableTag = {
  name: 'observableTag',
  level: 'inline',
  start(src: string) {
    return src.match(createTagStartRegex('^'))?.index;
  },
  // @ts-ignore
  tokenizer(src: string, tokens: any[]) {
    const match = createTagTokenRegex('^').exec(src);
    if (match) {
      return {
        type: 'observableTag',
        raw: match[0],
        href: observableIdToHref(match[0].split('^')[1]),
        text: match[0].trim(),
      };
    }
    return;
  },
  // @ts-ignore
  renderer(token) {
    return renderAnchor(token);
  },
};

// @ts-ignore
const ttpTag = {
  name: 'ttpTag',
  level: 'inline',
  start(src: string) {
    return src.match(createTagStartRegex('&'))?.index;
  },
  // @ts-ignore
  tokenizer(src: string, tokens: any[]) {
    const match = createTagTokenRegex('&').exec(src);
    if (match) {
      return {
        type: 'ttpTag',
        raw: match[0],
        href: ttpIdToHref(match[0].split('&')[1]),
        text: match[0].trim(),
      };
    }
    return;
  },
  // @ts-ignore
  renderer(token) {
    return renderAnchor(token);
  },
};

// @ts-ignore
const externalReferenceTag = {
  name: 'externalReferenceTag',
  level: 'inline',
  start(src: string) {
    return src.match(createTagStartRegex('!'))?.index;
  },
  // @ts-ignore
  tokenizer(src: string, tokens: any[]) {
    const match = createTagTokenRegex('!').exec(src);
    if (match) {
      return {
        type: 'externalReferenceTag',
        raw: match[0],
        href: externalReferenceIdToHref(match[0].split('!')[1]),
        text: match[0].trim(),
      };
    }
    return;
  },
  // @ts-ignore
  renderer(token) {
    return renderAnchor(token);
  },
};

export const orionMarkdownConfig: MarkdownConfig = {
  markdownParser: {
    sanitize: SecurityContext.HTML,
    markedOptions: {
      provide: MarkedOptions,
      useValue: {
        gfm: true,
        breaks: false,
        pedantic: false,
        smartLists: true,
        smartypants: false,
        renderer: renderer,
        extensions: [userTag, caseTag, taskTag, assetTag, observableTag, ttpTag, externalReferenceTag],
      },
    },
  },
};
