import { AbstractControl, ValidationErrors, ValidatorFn } from '@angular/forms';
import Ajv, { Options, SchemaObject, ValidateFunction } from 'ajv';
import Ajv2019 from 'ajv/dist/2019';
import Ajv2020 from 'ajv/dist/2020';

type AnyAjv = Ajv | Ajv2019 | Ajv2020;
const opts: Options = { addUsedSchema: false, strict: false };

export class JSONSchemaValidator {
  static ajv7 = new Ajv(opts);
  static ajv2019 = new Ajv2019(opts);
  static ajv2020 = new Ajv2020(opts);

  static validator(schema: string, forceCompileSchema = false): ValidatorFn {
    return (control: AbstractControl): ValidationErrors | null => {
      if (control.value === null || control.value === undefined) {
        return null;
      }
      const errors = JSONSchemaValidator.validate(control.value, schema, forceCompileSchema);
      if (errors) {
        {
          return { jsonSchema: errors };
        }
      } else {
        return null;
      }
    };
  }

  static validate(value: string, schema: string, forceCompileSchema = false): string[] | null {
    try {
      const parsedValue = JSON.parse(value);
      const parsedSchema: SchemaObject = JSON.parse(schema);
      const ajv = this.selectAjvInstance(parsedSchema);
      const validate = this.getValidateFn(ajv, parsedSchema, forceCompileSchema);
      if (!validate) {
        return null;
      } else {
        const valid = validate(parsedValue);
        if (!valid) {
          return validate.errors?.map((error) => error.message ?? '') ?? null;
        } else {
          return null;
        }
      }
    } catch (e) {
      return null;
    }
  }

  private static selectAjvInstance(schema: SchemaObject): AnyAjv {
    if (schema.$schema) {
      if (schema.$schema.includes('draft/2020-12')) {
        return this.ajv2020;
      }
      if (schema.$schema.includes('draft/2019-09')) {
        return this.ajv2019;
      }
    }
    return this.ajv7;
  }

  private static getValidateFn(
    ajv: AnyAjv,
    schema: SchemaObject,
    forceCompileSchema = false
  ): ValidateFunction | undefined {
    if (forceCompileSchema) {
      return ajv.compile(schema);
    } else {
      const schemaId = schema.$id ?? btoa(schema.toString());
      if (!ajv.getSchema(schemaId)) {
        ajv.addSchema(schema, schemaId);
      }
      return ajv.getSchema(schemaId);
    }
  }
}
