import { HttpErrorResponse } from '@angular/common/http';
import { Injectable } from '@angular/core';
import { ApolloError } from '@apollo/client/core';
import {
  SentinelNotification,
  SentinelNotificationResult,
  SentinelNotificationService,
  SentinelNotificationTypeEnum,
} from '@sentinel/layout/notification';
import { GraphQLError } from 'graphql';
import { Observable } from 'rxjs';
import { map } from 'rxjs/operators';

export type OrionError = ApolloError | HttpErrorResponse | ReadonlyArray<GraphQLError> | undefined;

@Injectable({ providedIn: 'root' })
export class ErrorHandlerService {
  constructor(private notificationService: SentinelNotificationService) {}

  private static createAdditionalInfoFromHttpError(err: HttpErrorResponse): string[] {
    if (err === null || err === undefined || err?.status === 0 || err.error === null || err.error === undefined) {
      return ['Unknown error. Please check your internet connection or report the issue to developers', err?.message];
    } else {
      return [err.message];
    }
  }

  private static createAdditionalInfoFromApolloError(err: ApolloError): string[] {
    return [err.name, err.message];
  }

  private static createAdditionalInfoFromGraphQLErrors(errors: ReadonlyArray<GraphQLError>): string[] {
    return errors.map((err) => err.message);
  }

  emit(err: OrionError, operation: string, action?: string): Observable<boolean> {
    const notification: SentinelNotification = {
      type: SentinelNotificationTypeEnum.Error,
      title: operation,
    };
    if (action !== undefined) {
      notification.action = action;
    }
    if (err === undefined) {
      notification.additionalInfo = ['Unknown error. Please report the issue to developers'];
    } else if (err instanceof ApolloError) {
      notification.additionalInfo = ErrorHandlerService.createAdditionalInfoFromApolloError(err);
    } else if (err instanceof HttpErrorResponse) {
      notification.additionalInfo = ErrorHandlerService.createAdditionalInfoFromHttpError(err);
    } else {
      notification.additionalInfo = ErrorHandlerService.createAdditionalInfoFromGraphQLErrors(err);
    }

    return this.notificationService
      .emit(notification)
      .pipe(map((result) => result === SentinelNotificationResult.CONFIRMED));
  }
}
