import { HTTP_INTERCEPTORS, HttpClient } from '@angular/common/http';
import { HttpClientTestingModule, HttpTestingController } from '@angular/common/http/testing';
import { TestBed, waitForAsync } from '@angular/core/testing';
import { OrionConfig } from '@orion/common';
import { OrionDynamicEnvironment } from '@orion/env';
import { SentinelAuthContext } from '@sentinel/auth';
import { take } from 'rxjs/operators';
import { OidcProviderHeaderInterceptor } from './oidc-provider-header.interceptor';

describe('OidcProviderHeaderInterceptor', () => {
  const testUrl = 'http://mytesturl.com/test';
  const testConfig: OrionConfig = {
    apiConfig: {
      baseUrl: '',
    },
    authConfig: {
      oidcProviderHeaderName: 'TEST-HEADER-NAME',
      oidcProviderHeaderNameInterceptorAllowedUrls: [testUrl],
      guardLoginPageRedirect: '',
      guardMainPageRedirect: '',
      providers: [],
    },
  };
  const testSelectedProvider = {
    label: 'test',
    oidcConfig: { options: { oidcProviderHeaderValue: 'customHeaderValue' } },
  };

  let httpMock: HttpTestingController;
  let http: HttpClient;
  let authContextSpy: jasmine.SpyObj<SentinelAuthContext>;

  beforeEach(waitForAsync(() => {
    authContextSpy = jasmine.createSpyObj('SentinelAuthContext', ['getProviderFromStorage']);
    TestBed.configureTestingModule({
      imports: [HttpClientTestingModule],
      providers: [
        { provide: SentinelAuthContext, useValue: authContextSpy },
        { provide: HTTP_INTERCEPTORS, useClass: OidcProviderHeaderInterceptor, multi: true },
      ],
    });
    http = TestBed.inject(HttpClient);
    httpMock = TestBed.inject(HttpTestingController);
  }));

  afterEach(() => {
    httpMock.verify();
  });

  it('should add OIDC provider header to HTTP request if configured and target URL is in allowed URLs', (done) => {
    spyOn(OrionDynamicEnvironment, 'getConfig').and.returnValue(testConfig);
    authContextSpy.getProviderFromStorage.and.returnValue(testSelectedProvider);

    http
      .get(testUrl)
      .pipe(take(1))
      .subscribe({ next: (_) => {}, error: () => fail() });
    const incoming = httpMock.expectOne(testUrl);

    expect(incoming.request.headers.has(testConfig.authConfig.oidcProviderHeaderName)).toBeTrue();
    expect(incoming.request.headers.get(testConfig.authConfig.oidcProviderHeaderName)).toEqual(
      testSelectedProvider.oidcConfig.options.oidcProviderHeaderValue
    );
    done();
  });

  it('should not add OIDC provider header to HTTP request if configured and target URL is in not in allowed URLs', (done) => {
    spyOn(OrionDynamicEnvironment, 'getConfig').and.returnValue(testConfig);
    authContextSpy.getProviderFromStorage.and.returnValue(testSelectedProvider);
    const differentUrl = 'http://not-allowed.com/';

    http
      .get(differentUrl)
      .pipe(take(1))
      .subscribe({
        next: (_) => {},
        error: () => fail(),
      });
    const incoming = httpMock.expectOne(differentUrl);

    expect(incoming.request.headers.has(testConfig.authConfig.oidcProviderHeaderName)).toBeFalse();
    done();
  });
});
