import { HTTP_INTERCEPTORS, HttpClient, HttpStatusCode } from '@angular/common/http';
import { HttpClientTestingModule, HttpTestingController } from '@angular/common/http/testing';
import { TestBed, waitForAsync } from '@angular/core/testing';
import { NotificationService } from '@orion/shared/services/global';
import { SentinelAuthService } from '@sentinel/auth';
import { of } from 'rxjs';
import { take } from 'rxjs/operators';
import { UnauthorizedInterceptor } from './unauthorized.interceptor';

describe('UnauthorizedInterceptor', () => {
  const testUrl = 'http://testurl.com';
  let httpMock: HttpTestingController;
  let http: HttpClient;
  let authSpy: jasmine.SpyObj<SentinelAuthService>;
  let notificationSpy: jasmine.SpyObj<NotificationService>;

  beforeEach(waitForAsync(() => {
    authSpy = jasmine.createSpyObj('SentinelAuthService', ['logout']);
    notificationSpy = jasmine.createSpyObj('NotificationService', ['emit']);
    TestBed.configureTestingModule({
      imports: [HttpClientTestingModule],
      providers: [
        { provide: SentinelAuthService, useValue: authSpy },
        { provide: NotificationService, useValue: notificationSpy },
        { provide: HTTP_INTERCEPTORS, useClass: UnauthorizedInterceptor, multi: true },
      ],
    });
    http = TestBed.inject(HttpClient);
    httpMock = TestBed.inject(HttpTestingController);
  }));

  afterEach(() => {
    httpMock.verify();
  });

  it('should do nothing if successful status code is returned', (done) => {
    http
      .get(testUrl)
      .pipe(take(1))
      .subscribe({
        next: (response) => {
          expect(notificationSpy.emit).toHaveBeenCalledTimes(0);
          expect(authSpy.logout).toHaveBeenCalledTimes(0);
          done();
        },
        error: () => fail(),
      });
    const testRequest = httpMock.expectOne(testUrl);
    testRequest.flush('', { status: HttpStatusCode.Ok, statusText: 'OK' });
  });

  it('should do nothing if error status code different than Unauthorized is returned', (done) => {
    http
      .get(testUrl)
      .pipe(take(1))
      .subscribe({
        next: (response) => {
          fail();
        },
        error: () => {
          expect(notificationSpy.emit).toHaveBeenCalledTimes(0);
          expect(authSpy.logout).toHaveBeenCalledTimes(0);
          done();
        },
      });
    const testRequest = httpMock.expectOne(testUrl);
    testRequest.flush('', { status: HttpStatusCode.Forbidden, statusText: 'Forbidden' });
  });

  it('should display notification and logout if status code is Unauthorized', (done) => {
    notificationSpy.emit.and.returnValue(of(true));
    authSpy.logout.and.callThrough();
    http
      .get(testUrl)
      .pipe(take(1))
      .subscribe({
        next: (response) => {
          fail();
        },
        error: () => {
          expect(notificationSpy.emit).toHaveBeenCalledTimes(1);
          expect(authSpy.logout).toHaveBeenCalledTimes(1);
          done();
        },
      });
    const testRequest = httpMock.expectOne(testUrl);
    testRequest.flush('', { status: HttpStatusCode.Unauthorized, statusText: 'Unauthorized' });
  });
});
