import { ActivatedRoute } from '@angular/router';
import { NotificationService } from '@orion/shared/services/global';
import { BaseReactiveForm } from '@sentinel/common';
import { ApolloResult, ApolloResultType } from '@csirtmu-orion/microframework/graphql';
import { SentinelNotificationTypeEnum } from '@sentinel/layout/notification';
import { BehaviorSubject, Observable } from 'rxjs';
import { concatMap, distinctUntilChanged, map, shareReplay, switchMap, tap } from 'rxjs/operators';

export abstract class EntityModifyBaseService<Form extends BaseReactiveForm<any>> {
  protected entityName: string;
  protected isSavingSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  isSaving$: Observable<boolean> = this.isSavingSubject$.asObservable().pipe(distinctUntilChanged());
  form$: Observable<Form>;
  canDeactivate$: Observable<boolean>;

  protected constructor(
    entityName: string,
    protected notificationService: NotificationService,
    protected activatedRoute: ActivatedRoute
  ) {
    this.entityName = entityName;
    this.form$ = this.createForm().pipe(shareReplay({ refCount: true, bufferSize: 1 }));
    this.canDeactivate$ = this.form$.pipe(
      switchMap((form) => form.dirty$),
      map((dirty) => !dirty),
      distinctUntilChanged()
    );
  }

  protected abstract createForm(): Observable<Form>;

  cancel(): Observable<boolean> {
    return this.form$.pipe(
      tap((form) => (form.dirty = false)),
      concatMap(() => this.handleCancel())
    );
  }

  abstract save(): Observable<boolean>;

  protected abstract handleCancel(): Observable<boolean>;

  protected abstract onSaveSuccess(): Observable<boolean>;

  protected abstract onSaveError(): Observable<boolean>;

  protected abstract createSuccessSaveMessage(): string;

  protected abstract createErrorSaveMessage(): string;

  protected handleSaveResult(form: Form, result: ApolloResult): Observable<boolean> {
    if (result.type === ApolloResultType.SUCCESS) {
      form.dirty = false;
      this.notificationService.emit(SentinelNotificationTypeEnum.Success, this.createSuccessSaveMessage());
      return this.onSaveSuccess();
    } else {
      return this.onSaveError();
    }
  }
}
