import { TestBed, waitForAsync } from '@angular/core/testing';
import {
  StageDetailTabsFragment,
  StageOrWorkflowNestedTasksService,
  WorkflowDetailTabsFragment,
} from '@orion/stage-workflow-shared-components/detail';
import { take } from 'rxjs/operators';

describe('StageOrWorkflowNestedTaskService', () => {
  let service: StageOrWorkflowNestedTasksService;

  beforeEach(waitForAsync(() => {
    TestBed.configureTestingModule({
      providers: [StageOrWorkflowNestedTasksService],
    });
    service = TestBed.inject(StageOrWorkflowNestedTasksService);
  }));

  it('should be created', () => {
    expect(service).toBeTruthy();
  });

  it('should emit empty task ids if stage has no nested tasks', (done) => {
    const stageWithNoTasks = createStageFragment('1', []);
    service.update(stageWithNoTasks);
    service.ids$.pipe(take(1)).subscribe({
      next: (ids) => {
        expect(ids).toEqual([]);
        done();
      },
      error: () => fail(),
    });
  });

  it('should emit task ids if stage has nested tasks', (done) => {
    const stageWithTasks = createStageFragment('1', ['1', '2', '3']);
    service.update(stageWithTasks);
    service.ids$.pipe(take(1)).subscribe({
      next: (ids) => {
        expect(ids).toEqual(['1', '2', '3']);
        done();
      },
      error: () => fail(),
    });
  });

  it('should emit empty task ids if workflow has no stages', (done) => {
    const workflowWithNoStages = createWorkflowFragment('1', []);
    service.update(workflowWithNoStages);
    service.ids$.pipe(take(1)).subscribe({
      next: (ids) => {
        expect(ids).toEqual([]);
        done();
      },
      error: () => fail(),
    });
  });

  it('should emit empty task ids if workflow has stages with no tasks', (done) => {
    const stage1 = createStageFragment('1', []);
    const stage2 = createStageFragment('2', []);
    const workflowWithNoTasks = createWorkflowFragment('1', [stage1, stage2]);
    service.update(workflowWithNoTasks);
    service.ids$.pipe(take(1)).subscribe({
      next: (ids) => {
        expect(ids).toEqual([]);
        done();
      },
      error: () => fail(),
    });
  });

  it('should emit task ids if workflow has stages with tasks', (done) => {
    const stage1 = createStageFragment('1', ['1', '2', '3']);
    const stage2 = createStageFragment('2', ['4', '5']);
    const workflow = createWorkflowFragment('1', [stage1, stage2]);
    service.update(workflow);
    service.ids$.pipe(take(1)).subscribe({
      next: (ids) => {
        expect(ids).toEqual(['1', '2', '3', '4', '5']);
        done();
      },
      error: () => fail(),
    });
  });

  it('should update task ids if stage was updated with different stage', (done) => {
    const stage1 = createStageFragment('1', ['1', '2', '3']);
    const stage2 = createStageFragment('2', ['7', '8', '9']);
    service.update(stage1);
    service.update(stage2);
    let resultsIndex = 0;
    const expectedResults = [
      ['1', '2', '3'],
      ['7', '8', '9'],
    ];
    service.ids$.pipe(take(2)).subscribe({
      next: (ids) => {
        expect(ids).toEqual(expectedResults[resultsIndex]);
        resultsIndex++;
        if (resultsIndex === 2) {
          done();
        }
      },
      error: () => fail(),
    });
  });

  it('should update task ids if workflow was updated with different workflow', (done) => {
    const stage1 = createStageFragment('1', ['1', '2', '3']);
    const stage2 = createStageFragment('2', ['4', '5']);
    const stage3 = createStageFragment('3', ['6']);
    const stage4 = createStageFragment('4', ['7']);
    const workflow1 = createWorkflowFragment('1', [stage1, stage2]);
    const workflow2 = createWorkflowFragment('2', [stage3, stage4]);

    let resultsIndex = 0;
    const expectedResults = [
      ['1', '2', '3', '4', '5'],
      ['6', '7'],
    ];
    service.update(workflow1);
    service.update(workflow2);
    service.ids$.pipe(take(2)).subscribe({
      next: (ids) => {
        expect(ids).toEqual(expectedResults[resultsIndex]);
        resultsIndex++;
        if (resultsIndex === 2) {
          done();
        }
      },
      error: () => fail(),
    });
  });

  it('should update task ids if workflow was updated with stage', (done) => {
    const stage1 = createStageFragment('1', ['1', '2', '3']);
    const stage2 = createStageFragment('2', ['4', '5']);
    const workflow = createWorkflowFragment('1', [stage1, stage2]);
    const stageUpdate = createStageFragment('3', ['6', '7']);

    let resultsIndex = 0;
    const expectedResults = [
      ['1', '2', '3', '4', '5'],
      ['6', '7'],
    ];
    service.update(workflow);
    service.update(stageUpdate);
    service.ids$.pipe(take(2)).subscribe({
      next: (ids) => {
        expect(ids).toEqual(expectedResults[resultsIndex]);
        resultsIndex++;
        if (resultsIndex === 2) {
          done();
        }
      },
      error: () => fail(),
    });
  });
});

function createWorkflowFragment(workflowId: string, stages: StageDetailTabsFragment[]): WorkflowDetailTabsFragment {
  return {
    id: workflowId,
    description: '',
    __typename: 'Workflow',
    stagesWithTasksIds: {
      data: stages,
    },
  };
}

function createStageFragment(stageId: string, taskIds: string[]): StageDetailTabsFragment {
  return {
    id: stageId,
    description: '',
    __typename: 'Stage',
    tasksIds: {
      data: createTaskFragments(taskIds),
    },
  };
}

function createTaskFragments(taskIds: string[]): { id: string }[] {
  return taskIds.map((id) => {
    return {
      id: id,
    };
  });
}
