import {
  ChangeDetectionStrategy,
  Component,
  EventEmitter,
  Input,
  OnChanges,
  OnInit,
  Output,
  SimpleChanges,
} from '@angular/core';
import { CaseNameFragment } from '@orion/case-components/shared';
import { AssigneeListElementFragment, AssigneesListService } from '@orion/case-services';
import { EntityOperationEventToRelationUpdateConverter } from '@orion/case-types';
import { CaseDbUserOrdEnum } from '@orion/graphql-types';
import { TimeService } from '@orion/shared/services/global';
import { StageOrWorkflowUpdate } from '@orion/stage-workflow-types';
import { CurrentUserFragment, CurrentUserService } from '@orion/user-services';
import { Destroyable } from '@sentinel/common';
import { EntityOperationEvent } from '@sentinel/common/entity';
import { EntityUpdateEvent } from '@csirtmu-orion/microframework/graphql';
import {
  INIT_PAGINATION_EVENT_BASE_TOKEN,
  PaginationBaseInit,
  PaginationService,
} from '@csirtmu-orion/microframework/pagination-services';
import { SidebarState, SidebarStateService } from '@sentinel/layout/sidebar';
import { Observable } from 'rxjs';
import { map } from 'rxjs/operators';
import {
  StageDetailSidebarFragment,
  WorkflowDetailSidebarFragment,
} from './stage-or-workflow-detail-sidebar.fragment.generated';

@Component({
  selector: 'orion-stage-or-workflow-detail-sidebar',
  templateUrl: './stage-or-workflow-detail-sidebar.component.html',
  styleUrls: ['./stage-or-workflow-detail-sidebar.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush,
  providers: [
    PaginationService,
    AssigneesListService,
    {
      provide: INIT_PAGINATION_EVENT_BASE_TOKEN,
      useValue: new PaginationBaseInit(30, CaseDbUserOrdEnum.NAME, 'asc'),
    },
  ],
})
export class StageOrWorkflowDetailSidebarComponent extends Destroyable() implements OnInit, OnChanges {
  @Input() stageOrWorkflow?: StageDetailSidebarFragment | WorkflowDetailSidebarFragment;
  @Output() patch: EventEmitter<EntityUpdateEvent<StageOrWorkflowUpdate>> = new EventEmitter<
    EntityUpdateEvent<StageOrWorkflowUpdate>
  >();
  possibleAssignees$?: Observable<AssigneeListElementFragment[]>;
  currUser$?: Observable<CurrentUserFragment>;
  today$?: Observable<Date>;
  state$?: Observable<SidebarState>;
  caseFragment?: CaseNameFragment;

  constructor(
    private assigneesService: AssigneesListService,
    private paginationService: PaginationService,
    private timeService: TimeService,
    private currUserService: CurrentUserService,
    private sidebarStateService: SidebarStateService
  ) {
    super();
    this.possibleAssignees$ = this.assigneesService.data$?.pipe(map((resource) => resource.elements));
  }

  ngOnInit(): void {
    this.possibleAssignees$ = this.assigneesService.data$?.pipe(map((resource) => resource.elements));
    this.today$ = this.timeService.today$;
    this.state$ = this.sidebarStateService.state$;
    this.currUser$ = this.currUserService.data$;
  }

  ngOnChanges(changes: SimpleChanges): void {
    if ('stageOrWorkflow' in changes && this.stageOrWorkflow) {
      if (this.isWorkflow(this.stageOrWorkflow)) {
        this.caseFragment = this.stageOrWorkflow?.case;
      } else {
        this.caseFragment = this.stageOrWorkflow?.workflow?.case;
      }
    }
  }

  onAssigneeSearch(value: string): void {
    this.assigneesService.update(this.paginationService.get(), { name: value });
  }

  onAssigneeSave(saveResult: EntityOperationEvent<AssigneeListElementFragment>): void {
    if (this.stageOrWorkflow) {
      const assigneesUpdate = EntityOperationEventToRelationUpdateConverter.convert(saveResult);
      this.patch.emit({ id: this.stageOrWorkflow.id, patch: { assignees: assigneesUpdate } });
    }
  }

  onDueDateSave(date?: Date): void {
    if (this.stageOrWorkflow) {
      this.patch.emit({ id: this.stageOrWorkflow.id, patch: { dueDate: date } });
    }
  }

  onSidebarClick(): void {
    this.sidebarStateService.toggleState();
  }

  private isWorkflow(
    stageOrWorkflow: StageDetailSidebarFragment | WorkflowDetailSidebarFragment
  ): stageOrWorkflow is WorkflowDetailSidebarFragment {
    return stageOrWorkflow.__typename === 'Workflow';
  }
}
