import {
  ChangeDetectionStrategy,
  Component,
  Input,
  OnChanges,
  OnDestroy,
  SimpleChanges,
  ViewChild,
} from '@angular/core';
import { FormGroup } from '@angular/forms';
import { EntityFormModel } from '@orion/common';
import { KnownInputOutputTypes } from '@orion/graphql-types';
import { TaskListElementFragment } from '@orion/task-components/shared';
import { CreateTaskForm } from '@orion/task-types';
import { trackByIndex } from '@sentinel/common/utils';
import { MatStepper } from '@angular/material/stepper';
import { TaskFormStepperService } from '../task-form-stepper.service';
import { takeUntil } from 'rxjs/operators';
import { Destroyable } from '@sentinel/common';
import { StepperSelectionEvent } from '@angular/cdk/stepper';

@Component({
  selector: 'orion-create-task-form',
  templateUrl: './create-task-form.component.html',
  styleUrls: ['./create-task-form.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush,
})
export class CreateTaskFormComponent extends Destroyable() implements OnChanges, OnDestroy {
  @Input() stageId?: string;
  @Input() form?: CreateTaskForm;
  @ViewChild('stepper', { static: false }) private taskStepper?: MatStepper;

  inputOutputTypes = Object.values(KnownInputOutputTypes);
  sharedFormGroup?: FormGroup<EntityFormModel>;
  predecessorsSearchExcludedIds: string[] = [];
  trackByFn = trackByIndex;

  constructor(private taskFormStepper: TaskFormStepperService) {
    super();
    this.taskFormStepper.previous$.pipe(takeUntil(this.destroy$)).subscribe(() => this.goBack());
    this.taskFormStepper.next$.pipe(takeUntil(this.destroy$)).subscribe(() => this.goForward());
  }

  ngOnChanges(changes: SimpleChanges): void {
    if ('form' in changes && this.form) {
      this.sharedFormGroup = this.form.formGroup.controls.basicInfoGroup as unknown as FormGroup<EntityFormModel>;
    }
  }

  ngOnDestroy(): void {
    this.form?.onDestroy();
  }

  goBack() {
    if (this.taskStepper) {
      this.taskStepper.previous();
    }
  }

  goForward() {
    if (this.taskStepper) {
      this.taskStepper.next();
    }
  }

  addPredecessor(selectedPredecessor: TaskListElementFragment): void {
    if (this.form) {
      this.form.addPredecessors(selectedPredecessor);
      this.updateExcludedIds();
    }
  }

  removePredecessor(index: number): void {
    if (this.form) {
      this.form.formGroup.controls.predecessorsGroup?.controls?.predecessors.removeAt(index);
      this.updateExcludedIds();
    }
  }

  private updateExcludedIds(): void {
    if (this.form) {
      this.predecessorsSearchExcludedIds =
        this.form.formGroup.controls?.predecessorsGroup?.controls?.predecessors.value.map(
          (predecessorFormGroup) => predecessorFormGroup?.predecessor?.id ?? '-1'
        ) ?? [];
    }
  }

  onInputValueChange(value: string | undefined): void {
    if (value) {
      this.form?.formGroup.controls.inputGroup?.controls?.input.setValue(value);
    } else {
      this.form?.formGroup?.controls.inputGroup?.controls?.input.setValue(null);
    }
  }

  onInputValidityChange(valid: boolean): void {
    this.form?.onInputValidityChange(valid);
  }

  onStepManuallySelected(event: StepperSelectionEvent): void {
    this.taskFormStepper.setStep(event.selectedIndex);
  }
}
