import { Injectable } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { TaskStateEnum } from '@orion/graphql-types';
import { CreateEntityService } from '@orion/shared-services';
import { NotificationService } from '@orion/shared/services/global';
import {
  GetTaskListQuery,
  GetTaskListQueryService,
  GetTaskListQueryVariables,
  TaskListElementFragment,
} from '@orion/task-components/shared';
import {
  CreateParentlessTask,
  CreateParentlessTaskData,
  CreateTask,
  CreateTaskData,
  CreateTaskForm,
  CreateTaskWithParentData,
  isParentlessTask,
  TaskClientFilter,
  TaskFilterConverter,
  TaskRefs,
} from '@orion/task-types';
import { CurrentUserService } from '@orion/user-services';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  createApolloContextErrorTitle,
} from '@csirtmu-orion/microframework/graphql';
import { from, Observable, of } from 'rxjs';
import { catchError, concatMap, map, take } from 'rxjs/operators';
import { STAGE_PARAM_SELECTOR } from '@orion/case-paths';
import { CreateTaskMutationService, CreateTaskParentlessMutationService } from './graphql';

@Injectable()
export class CreateTaskService extends CreateEntityService<CreateTaskForm> {
  constructor(
    protected notificationService: NotificationService,
    protected activatedRoute: ActivatedRoute,
    private router: Router,
    private taskListQuery: GetTaskListQueryService,
    private createTaskMutation: CreateTaskMutationService,
    private createParentlessTaskMutation: CreateTaskParentlessMutationService,
    private activeUserService: CurrentUserService
  ) {
    super('Task', notificationService, activatedRoute);
  }

  protected createForm(): Observable<CreateTaskForm> {
    return this.activatedRoute.paramMap.pipe(
      map((paramMap) => paramMap.get(STAGE_PARAM_SELECTOR) ?? undefined),
      concatMap((stageId) => this.isTaskParentless(stageId)),
      map((isParentless) => new CreateTaskForm(isParentless))
    );
  }

  protected createEntity(form: CreateTaskForm): Observable<ApolloResult> {
    return this.getRefs().pipe(concatMap((refs) => this.createTask(form.getData(), refs)));
  }

  protected handleCancel(): Observable<boolean> {
    return this.navigateToOverview();
  }

  protected onSaveError(): Observable<boolean> {
    return of(false);
  }

  protected onSaveSuccess(): Observable<boolean> {
    return this.navigateToOverview();
  }

  private getRefs(): Observable<TaskRefs> {
    return this.activatedRoute.paramMap.pipe(
      take(1),
      map((paramMap) => {
        const authorId = this.activeUserService.getCurrentUser()?.id;
        const stageId = paramMap.get(STAGE_PARAM_SELECTOR) ?? undefined;
        if (stageId !== undefined && authorId !== undefined) {
          return { stageId, authorId };
        } else {
          throw 'error';
        }
      })
    );
  }

  private createTask(taskData: CreateTaskData, refs: TaskRefs): Observable<ApolloResult> {
    if (isParentlessTask(taskData)) {
      return this.createParentlessTask(taskData, refs);
    } else {
      return this.createTaskWithParent(taskData, refs);
    }
  }

  private createTaskWithParent(data: CreateTaskWithParentData, refs: TaskRefs): Observable<ApolloResult> {
    const createTask = new CreateTask(data, refs);
    const vars = { task: createTask };
    return this.createTaskMutation.mutate(vars, { context: createApolloContextErrorTitle('Creating task') }).pipe(
      map((res) => ApolloOperationResultResolver.resolve(res)),
      catchError((err) => ApolloOperationResultResolver.handleError(err))
    );
  }

  private createParentlessTask(data: CreateParentlessTaskData, refs: TaskRefs): Observable<ApolloResult> {
    const createTask = new CreateParentlessTask(data, refs);
    const vars = { task: createTask };
    return this.createParentlessTaskMutation
      .mutate(vars, { context: { useMultipart: true, ...createApolloContextErrorTitle('Creating task') } })
      .pipe(
        map((res) => ApolloOperationResultResolver.resolve(res)),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  private navigateToOverview(): Observable<boolean> {
    return from(this.router.navigate(['../'], { relativeTo: this.activatedRoute }));
  }

  private isTaskParentless(stageId?: string): Observable<boolean> {
    if (stageId === undefined) {
      throw 'error';
    }
    const taskFilter: TaskClientFilter = { inStageIds: [stageId] };
    const converter = new TaskFilterConverter();
    const vars: GetTaskListQueryVariables = { filter: converter.convert(taskFilter) };
    return this.taskListQuery
      .fetch(vars, { fetchPolicy: 'network-only' })
      .pipe(map((queryRes) => this.shouldNewTaskBeParentless(queryRes.data)));
  }

  private shouldNewTaskBeParentless(queryRes: GetTaskListQuery): boolean {
    const stageHasOtherTask = queryRes.getTaskList.data.length > 0;
    if (stageHasOtherTask) {
      const activeTasks = queryRes.getTaskList.data.filter((task) => !this.isTaskRejectedOrSkipped(task));
      return activeTasks.length === 0;
    } else {
      return true;
    }
  }

  private isTaskRejectedOrSkipped(task: TaskListElementFragment): boolean {
    return task.state === TaskStateEnum.SKIPPED || task.state === TaskStateEnum.REJECTED;
  }
}
