import { Injectable } from '@angular/core';
import { FetchResult } from '@apollo/client/core';
import { namedOperations } from '@orion/named-operations';
import { NotificationService } from '@orion/shared/services/global';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  ApolloResultType,
  createApolloContextErrorTitle,
} from '@csirtmu-orion/microframework/graphql';
import { BehaviorSubject, combineLatest, Observable } from 'rxjs';
import { catchError, concatMap, map, tap } from 'rxjs/operators';
import {
  RejectTaskMutationService,
  ResolveTaskMutationService,
  ResolveTaskWithDecisionMutationService,
  TriggerTaskMutationService,
  UpdateOutputFromBlockingTasksMutationService,
} from './graphql';

type BaseTaskActionNames = 'reject' | 'resolve' | 'trigger' | 'outputs';

@Injectable()
export class TaskActionsService {
  constructor(
    private rejectTaskMutation: RejectTaskMutationService,
    private resolveTaskMutation: ResolveTaskMutationService,
    private updateOutputsTaskMutation: UpdateOutputFromBlockingTasksMutationService,
    private triggerTaskMutation: TriggerTaskMutationService,
    private decisionMutation: ResolveTaskWithDecisionMutationService,
    private notificationService: NotificationService
  ) {}

  private resolveInProgressSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  resolveInProgress$: Observable<boolean> = this.resolveInProgressSubject$.asObservable();
  private rejectInProgressSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  rejectInProgress$: Observable<boolean> = this.rejectInProgressSubject$.asObservable();
  private updateOutputsInProgressSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  updateOutputsInProgress$: Observable<boolean> = this.updateOutputsInProgressSubject$.asObservable();
  private isTriggeringSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  isTriggering$: Observable<boolean> = this.isTriggeringSubject$.asObservable();
  private trueDecisionInProgressSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  trueDecisionInProgress$: Observable<boolean> = this.trueDecisionInProgressSubject$.asObservable();
  private falseDecisionInProgressSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  falseDecisionInProgress$: Observable<boolean> = this.falseDecisionInProgressSubject$.asObservable();
  isLoading$ = combineLatest([
    this.resolveInProgress$,
    this.rejectInProgress$,
    this.isTriggering$,
    this.falseDecisionInProgress$,
    this.trueDecisionInProgress$,
    this.updateOutputsInProgress$,
  ]).pipe(map((progressStates) => progressStates.some((isInProgress) => isInProgress)));

  resolve(id: string): Observable<ApolloResult> {
    this.resolveInProgressSubject$.next(true);
    return this.resolveTaskMutation
      .mutate(
        { id: id },
        {
          refetchQueries: [namedOperations.Query.getTaskDetail],
          awaitRefetchQueries: true,
          context: createApolloContextErrorTitle('Resolving task'),
        }
      )
      .pipe(
        map((res) => this.handleActionResponse(res, 'resolve')),
        tap(
          () => this.resolveInProgressSubject$.next(false),
          () => this.resolveInProgressSubject$.next(false)
        ),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  decide(value: boolean, taskId: string): Observable<ApolloResult> {
    value ? this.trueDecisionInProgressSubject$.next(true) : this.falseDecisionInProgressSubject$.next(true);
    return this.decisionMutation
      .mutate(
        { id: taskId, decision: value },
        {
          refetchQueries: [namedOperations.Query.getTaskDetail],
          awaitRefetchQueries: true,
          context: createApolloContextErrorTitle('Resolving decision task'),
        }
      )
      .pipe(
        map((res) => this.handleActionResponse(res, 'resolve')),
        tap(
          () => {
            this.trueDecisionInProgressSubject$.next(false);
            this.falseDecisionInProgressSubject$.next(false);
          },
          () => {
            this.trueDecisionInProgressSubject$.next(false);
            this.falseDecisionInProgressSubject$.next(false);
          }
        ),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  reject(id: string): Observable<ApolloResult> {
    this.rejectInProgressSubject$.next(true);
    return this.rejectTaskMutation
      .mutate(
        { id: id },
        {
          refetchQueries: [namedOperations.Query.getTaskDetail],
          awaitRefetchQueries: true,
          context: createApolloContextErrorTitle('Rejecting task'),
        }
      )
      .pipe(
        map((res) => this.handleActionResponse(res, 'reject')),
        tap(
          () => this.rejectInProgressSubject$.next(false),
          () => this.rejectInProgressSubject$.next(false)
        ),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  trigger(taskId: string): Observable<ApolloResult> {
    this.isTriggeringSubject$.next(true);
    return this.triggerTaskMutation
      .mutate(
        { taskId: taskId },
        {
          refetchQueries: [namedOperations.Query.getTaskDetail],
          context: createApolloContextErrorTitle('Triggering task'),
          awaitRefetchQueries: true,
        }
      )
      .pipe(
        map((res) => this.handleActionResponse(res, 'trigger')),
        tap({
          next: () => this.isTriggeringSubject$.next(false),
          error: () => this.isTriggeringSubject$.next(false),
        }),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  updateOutputs(id: string): Observable<ApolloResult> {
    this.updateOutputsInProgressSubject$.next(true);
    return this.updateOutputsTaskMutation
      .mutate(
        { id: id },
        {
          refetchQueries: [namedOperations.Query.getTaskDetail],
          awaitRefetchQueries: true,
          context: createApolloContextErrorTitle('Updating outputs of blocking tasks'),
        }
      )
      .pipe(
        map((res) => this.handleActionResponse(res, 'outputs')),
        tap(
          () => this.updateOutputsInProgressSubject$.next(false),
          () => this.updateOutputsInProgressSubject$.next(false)
        ),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  updateOutputsAndResolve(id: string): Observable<ApolloResult> {
    return this.updateOutputs(id).pipe(concatMap(() => this.resolve(id)));
  }

  private handleActionResponse(fetchResult: FetchResult, action: BaseTaskActionNames): ApolloResult {
    this.resolveInProgressSubject$.next(false);
    const result = ApolloOperationResultResolver.resolve(fetchResult);
    if (result.type === ApolloResultType.SUCCESS) {
      if (action === 'outputs') {
        this.notificationService.emit('success', `Task outputs were successfully updated`);
      } else {
        const pastTenseActionName = this.actionNameToPastTense(action);
        this.notificationService.emit('success', `Task was successfully ${pastTenseActionName}`);
      }
    }
    return result;
  }

  private actionNameToPastTense(name: BaseTaskActionNames): string {
    if (name === 'reject') {
      return 'rejected';
    } else if (name === 'resolve') {
      return 'resolved';
    } else {
      return 'triggered';
    }
  }
}
