import { Injectable } from '@angular/core';
import { TaskStateEnum } from '@orion/graphql-types';
import { ControlsService } from '@orion/shared-services';
import { Optional } from '@sentinel/common/utils';
import {
  SentinelControlItem,
  SentinelControlMenuItem,
  SentinelExpandableControlItem,
} from '@sentinel/components/controls';
import { BehaviorSubject, defer, of } from 'rxjs';
import { map } from 'rxjs/operators';
import { TaskActionsService } from './task-actions.service';
import { TaskDetailMainContentFragment } from './task-detail-main-content';

@Injectable()
export class TaskControls extends ControlsService {
  task$: BehaviorSubject<Optional<TaskDetailMainContentFragment>> = new BehaviorSubject<
    Optional<TaskDetailMainContentFragment>
  >(undefined);

  constructor(private taskActionsService: TaskActionsService) {
    super();
    this.controls$ = this.task$.pipe(map((task) => (task ? this.createControls(task) : [])));
  }

  onTaskChanged(task: TaskDetailMainContentFragment): void {
    this.task$.next(task);
  }

  private createControls(task: TaskDetailMainContentFragment): SentinelControlItem[] {
    const rejectActionItem = new SentinelControlItem(
      'REJECT_TASK',
      'Reject',
      'warn',
      this.taskActionsService.isLoading$,
      defer(() => this.taskActionsService.reject(task.id)),
      this.taskActionsService.rejectInProgress$
    );
    const resolveActionItem = new SentinelControlItem(
      'RESOLVE_TASK',
      'Resolve',
      'primary',
      this.taskActionsService.isLoading$,
      defer(() => this.taskActionsService.resolve(task.id)),
      this.taskActionsService.resolveInProgress$
    );
    const triggerTask = new SentinelControlItem(
      'TRIGGER_TASK',
      'Trigger',
      'accent',
      this.taskActionsService.isLoading$,
      defer(() => this.taskActionsService.trigger(task.id)),
      this.taskActionsService.isTriggering$
    );
    const updateOutputFromBlockingTasks = new SentinelExpandableControlItem(
      'JOIN_OUTPUTS',
      'Join Outputs',
      'primary',
      of(false),
      [
        new SentinelControlMenuItem(
          'MERGE_OUTPUTS',
          'Join outputs of blocking tasks',
          'primary',
          this.taskActionsService.isLoading$,
          defer(() => this.taskActionsService.updateOutputs(task.id)),
          'merge',
          this.taskActionsService.updateOutputsInProgress$
        ),
        new SentinelControlMenuItem(
          'MERGE_OUTPUTS_AND_RESOLVE',
          'Join outputs and resolve',
          'primary',
          this.taskActionsService.isLoading$,
          defer(() => this.taskActionsService.updateOutputsAndResolve(task.id)),
          'done',
          this.taskActionsService.updateOutputsInProgress$
        ),
      ]
    );
    const decideFalse = new SentinelControlItem(
      'DECIDE_FALSE',
      'Resolve to false',
      'warn',
      this.taskActionsService.isLoading$,
      defer(() => this.taskActionsService.decide(false, task.id)),
      this.taskActionsService.falseDecisionInProgress$
    );
    const decideTrue = new SentinelControlItem(
      'DECIDE_TRUE',
      'Resolve to true',
      'primary',
      this.taskActionsService.isLoading$,
      defer(() => this.taskActionsService.decide(true, task.id)),
      this.taskActionsService.trueDecisionInProgress$
    );
    if (task.state === TaskStateEnum.OPEN) {
      const controls: SentinelControlItem[] = [rejectActionItem];
      if (this.canBeDecided(task)) {
        controls.push(...[decideFalse, decideTrue]);
      }
      if (this.canBeResolved(task)) {
        controls.push(resolveActionItem);
      }
      if (this.canBeTriggered(task)) {
        controls.push(triggerTask);
      }
      if (this.canJoinOutputs(task)) {
        controls.push(updateOutputFromBlockingTasks);
      }
      return controls;
    }
    return [];
  }

  private hasInput(task: TaskDetailMainContentFragment): boolean {
    return (
      task.input !== undefined &&
      task.input !== null &&
      task.input.stringifiedValue !== undefined &&
      task.input.stringifiedValue !== null
    );
  }

  private hasOutput(task: TaskDetailMainContentFragment): boolean {
    return (
      task.output !== undefined &&
      task.output !== null &&
      task.output.stringifiedValue !== undefined &&
      task.output.stringifiedValue !== null
    );
  }

  private canBeDecided(task: TaskDetailMainContentFragment): boolean {
    return task.isDecision === true && (task.decision === undefined || task.decision === null);
  }

  private canBeResolved(task: TaskDetailMainContentFragment): boolean {
    return (
      task.output !== undefined &&
      !task.isAutomated &&
      task.isDecision === false &&
      (task.aggregatesResults !== true || task.outputIsJoined === true)
    );
  }

  private canBeTriggered(task: TaskDetailMainContentFragment): boolean {
    return task.isAutomated === true && this.hasInput(task) && !this.hasOutput(task);
  }

  private canJoinOutputs(task: TaskDetailMainContentFragment): boolean {
    return task.aggregatesResults === true && task.outputIsJoined !== true;
  }
}
