import { Injectable } from '@angular/core';
import { NotificationService } from '@orion/shared/services/global';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  createApolloContextErrorTitle,
  wrapApolloContextForFileUpload,
} from '@csirtmu-orion/microframework/graphql';
import { BehaviorSubject, Observable } from 'rxjs';
import { catchError, map, tap } from 'rxjs/operators';
import {
  CreateTaskInputMutationService,
  CreateTaskInputMutationVariables,
  SetTaskFileInputMutationService,
  SetTaskFileInputMutationVariables,
  SetTaskInputMutationService,
  SetTaskInputMutationVariables,
} from '../graphql';
import { CreateTaskInputOutputData } from './task-input-output-panel';
import { namedOperations } from '@orion/named-operations';

@Injectable()
export class SetTaskInputService {
  private isLoadingSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  isLoading$: Observable<boolean> = this.isLoadingSubject$.asObservable();

  constructor(
    private inputMutation: SetTaskInputMutationService,
    private fileInputMutation: SetTaskFileInputMutationService,
    private createTaskInputMutation: CreateTaskInputMutationService,
    private notificationService: NotificationService
  ) {}

  setInput(taskId: string, input: string): Observable<ApolloResult> {
    this.isLoadingSubject$.next(true);
    const vars: SetTaskInputMutationVariables = {
      taskId: taskId,
      input: input,
    };
    return this.inputMutation
      .mutate(vars, {
        refetchQueries: [namedOperations.Query.getTaskDetail],
        awaitRefetchQueries: true,
        context: createApolloContextErrorTitle('Setting task input'),
      })
      .pipe(
        map((res) => ApolloOperationResultResolver.resolve(res)),
        tap(
          () => this.onSuccess('Task input was successfully set'),
          () => this.isLoadingSubject$.next(false)
        ),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  setFileInput(taskId: string, fileInput: File): Observable<ApolloResult> {
    this.isLoadingSubject$.next(true);
    const vars: SetTaskFileInputMutationVariables = {
      taskId: taskId,
      fileInput: fileInput,
    };

    return this.fileInputMutation
      .mutate(vars, {
        refetchQueries: [namedOperations.Query.getTaskDetail],
        awaitRefetchQueries: true,
        context: wrapApolloContextForFileUpload(createApolloContextErrorTitle('Setting task input')),
      })
      .pipe(
        map((res) => ApolloOperationResultResolver.resolve(res)),
        catchError((err) => ApolloOperationResultResolver.handleError(err)),
        tap(
          () => this.onSuccess('Task input was successfully set'),
          () => this.isLoadingSubject$.next(false)
        )
      );
  }

  createTaskInput(taskId: string, input: CreateTaskInputOutputData): Observable<ApolloResult> {
    this.isLoadingSubject$.next(true);
    const vars: CreateTaskInputMutationVariables = {
      taskId: taskId,
      type: input.type,
      description: input.description,
      inputFormat: input.validation,
    };
    return this.createTaskInputMutation
      .mutate(vars, {
        refetchQueries: [namedOperations.Query.getTaskDetail],
        awaitRefetchQueries: true,
        context: wrapApolloContextForFileUpload(createApolloContextErrorTitle('Setting task input')),
      })
      .pipe(
        map((res) => ApolloOperationResultResolver.resolve(res)),
        catchError((err) => ApolloOperationResultResolver.handleError(err)),
        tap(
          () => this.onSuccess('Task input type was successfully set'),
          () => this.isLoadingSubject$.next(false)
        )
      );
  }

  private onSuccess(message: string): void {
    this.isLoadingSubject$.next(false);
    this.notificationService.emit('success', message);
  }
}
