import { Injectable } from '@angular/core';
import { NotificationService } from '@orion/shared/services/global';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  createApolloContextErrorTitle,
  wrapApolloContextForFileUpload,
} from '@csirtmu-orion/microframework/graphql';
import { BehaviorSubject, Observable } from 'rxjs';
import { catchError, map, tap } from 'rxjs/operators';
import {
  CreateTaskOutputMutationService,
  CreateTaskOutputMutationVariables,
  SetTaskFileOutputMutationService,
  SetTaskFileOutputMutationVariables,
  SetTaskOutputMutationService,
  SetTaskOutputMutationVariables,
} from '../graphql';
import { CreateTaskInputOutputData } from './task-input-output-panel';
import { namedOperations } from '@orion/named-operations';

@Injectable()
export class SetTaskOutputService {
  private isLoadingSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  isLoading$: Observable<boolean> = this.isLoadingSubject$.asObservable();

  constructor(
    private outputMutation: SetTaskOutputMutationService,
    private fileOutputMutation: SetTaskFileOutputMutationService,
    private createTaskOutputMutation: CreateTaskOutputMutationService,
    private notificationService: NotificationService
  ) {}

  setOutput(taskId: string, output: string): Observable<ApolloResult> {
    this.isLoadingSubject$.next(true);
    const vars: SetTaskOutputMutationVariables = {
      taskId: taskId,
      output: output,
    };
    return this.outputMutation
      .mutate(vars, {
        refetchQueries: [namedOperations.Query.getTaskDetail],
        awaitRefetchQueries: true,
        context: createApolloContextErrorTitle('Setting task output'),
      })
      .pipe(
        map((res) => ApolloOperationResultResolver.resolve(res)),
        tap(
          () => this.onSuccess('Task output was successfully set'),
          () => this.isLoadingSubject$.next(false)
        ),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  setFileOutput(taskId: string, fileOutput: File): Observable<ApolloResult> {
    this.isLoadingSubject$.next(true);
    const vars: SetTaskFileOutputMutationVariables = {
      taskId: taskId,
      fileOutput: fileOutput,
    };

    return this.fileOutputMutation
      .mutate(vars, {
        refetchQueries: [namedOperations.Query.getTaskDetail],
        awaitRefetchQueries: true,
        context: wrapApolloContextForFileUpload(createApolloContextErrorTitle('Setting task output')),
      })
      .pipe(
        map((res) => ApolloOperationResultResolver.resolve(res)),
        tap(
          () => this.onSuccess('Task output was successfully set'),
          () => this.isLoadingSubject$.next(false)
        ),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  createTaskOutput(taskId: string, output: CreateTaskInputOutputData): Observable<ApolloResult> {
    this.isLoadingSubject$.next(true);
    const vars: CreateTaskOutputMutationVariables = {
      taskId: taskId,
      type: output.type,
      description: output.description,
      outputFormat: output.validation,
    };
    return this.createTaskOutputMutation
      .mutate(vars, {
        refetchQueries: [namedOperations.Query.getTaskDetail],
        awaitRefetchQueries: true,
        context: wrapApolloContextForFileUpload(createApolloContextErrorTitle('Setting task output')),
      })
      .pipe(
        map((res) => ApolloOperationResultResolver.resolve(res)),
        catchError((err) => ApolloOperationResultResolver.handleError(err)),
        tap(
          () => this.onSuccess('Task output type was successfully set'),
          () => this.isLoadingSubject$.next(false)
        )
      );
  }

  private onSuccess(message: string): void {
    this.isLoadingSubject$.next(false);
    this.notificationService.emit('success', message);
  }
}
