import { ValidatorFn, Validators } from '@angular/forms';
import { IPValidators, jsonValidator, urlValidator } from '@orion/common';
import { KnownInputOutputTypes } from '@orion/graphql-types';
import { TaskIOTypename, TaskIOTypenameOrEnum } from '@orion/task-types';

export type TaskMode = 'input' | 'output';

export const BOOL_FORM_TYPES: TaskIOTypename[] = ['BoolInputOutput'];
export const ANY_DATE_FORM_TYPES: TaskIOTypename[] = ['DateInputOutput', 'DateTimeInputOutput'];
export const FILE_FORM_TYPES: TaskIOTypename[] = ['FileInputOutput', 'EmailMessageInputOutput'];
export const JSON_FORM_TYPES: TaskIOTypename[] = ['JSONInputOutput'];
export const CUSTOM_VALIDATION_TYPES: TaskIOTypename[] = ['JSONInputOutput', 'StringInputOutput'];

export const BOOL_FORM_ENUMS: KnownInputOutputTypes[] = [KnownInputOutputTypes.BOOL];
export const ANY_DATE_FORM_ENUMS: KnownInputOutputTypes[] = [
  KnownInputOutputTypes.DATE,
  KnownInputOutputTypes.DATETIME,
];
export const FILE_FORM_ENUMS: KnownInputOutputTypes[] = [
  KnownInputOutputTypes.FILE,
  KnownInputOutputTypes.EMAIL_MESSAGE,
];
export const JSON_FORM_ENUMS: KnownInputOutputTypes[] = [KnownInputOutputTypes.JSON];
export const CUSTOM_VALIDATION_ENUMS: KnownInputOutputTypes[] = [
  KnownInputOutputTypes.JSON,
  KnownInputOutputTypes.STRING,
];

export function isInTypenamesOrEnumTypes(
  typeOrEnum: TaskIOTypenameOrEnum | undefined,
  typenames: TaskIOTypename[],
  enums: KnownInputOutputTypes[]
): boolean {
  if (typeOrEnum) {
    if (typenames.find((typename) => typename === typeOrEnum)) {
      return true;
    } else {
      return enums.find((enumType) => enumType === typeOrEnum) !== undefined;
    }
  } else {
    return false;
  }
}

export function taskIOTypenameToValidators(type: TaskIOTypenameOrEnum): ValidatorFn[] {
  if (isJSONType(type)) {
    return [jsonValidator];
  }
  if (
    type === 'FileRefInputOutput' ||
    type === 'URLInputOutput' ||
    type === KnownInputOutputTypes.FILE_REF ||
    type === KnownInputOutputTypes.URL
  ) {
    return [urlValidator];
  }
  if (type === 'EmailAddressInputOutput' || type === KnownInputOutputTypes.EMAIL_ADDRESS) {
    return [Validators.email];
  }
  if (type === 'IPInputOutput' || type === KnownInputOutputTypes.IP) {
    return [IPValidators.IPAddress];
  }
  return [];
}

export function isBoolType(typeOrEnum?: TaskIOTypenameOrEnum): boolean {
  return isInTypenamesOrEnumTypes(typeOrEnum, BOOL_FORM_TYPES, BOOL_FORM_ENUMS);
}

export function isFileType(typeOrEnum?: TaskIOTypenameOrEnum): boolean {
  return isInTypenamesOrEnumTypes(typeOrEnum, FILE_FORM_TYPES, FILE_FORM_ENUMS);
}

export function isJSONType(typeOrEnum?: TaskIOTypenameOrEnum): boolean {
  return isInTypenamesOrEnumTypes(typeOrEnum, JSON_FORM_TYPES, JSON_FORM_ENUMS);
}

export function isStringType(typeOrEnum?: TaskIOTypenameOrEnum): boolean {
  return isInTypenamesOrEnumTypes(typeOrEnum, ['StringInputOutput'], [KnownInputOutputTypes.STRING]);
}

export function isStringLikeType(typeOrEnum?: TaskIOTypenameOrEnum): boolean {
  if (typeOrEnum === undefined || typeOrEnum === null) {
    return false;
  }
  const otherFormTypes = [...BOOL_FORM_TYPES, ...ANY_DATE_FORM_TYPES, ...FILE_FORM_TYPES, ...JSON_FORM_TYPES];
  const otherFormEnums = [...BOOL_FORM_ENUMS, ...ANY_DATE_FORM_ENUMS, ...FILE_FORM_ENUMS, ...JSON_FORM_ENUMS];
  return !isInTypenamesOrEnumTypes(typeOrEnum, otherFormTypes, otherFormEnums);
}

export function isAnyDateType(typeOrEnum?: TaskIOTypenameOrEnum): boolean {
  return isInTypenamesOrEnumTypes(typeOrEnum, ANY_DATE_FORM_TYPES, ANY_DATE_FORM_ENUMS);
}

export function isDateType(typeOrEnum?: TaskIOTypenameOrEnum): boolean {
  return isInTypenamesOrEnumTypes(typeOrEnum, ['DateInputOutput'], [KnownInputOutputTypes.DATE]);
}

export function typeHasCustomValidation(typeOrEnum?: TaskIOTypenameOrEnum): boolean {
  return isInTypenamesOrEnumTypes(typeOrEnum, CUSTOM_VALIDATION_TYPES, CUSTOM_VALIDATION_ENUMS);
}

export function typenameToKnownInputOutput(typename: TaskIOTypename): KnownInputOutputTypes {
  switch (typename) {
    case 'BaseTaskInputOutput': {
      return KnownInputOutputTypes.OTHER;
    }
    case 'BoolInputOutput': {
      return KnownInputOutputTypes.BOOL;
    }
    case 'DateInputOutput': {
      return KnownInputOutputTypes.DATE;
    }
    case 'DateTimeInputOutput': {
      return KnownInputOutputTypes.DATETIME;
    }
    case 'DomainInputOutput': {
      return KnownInputOutputTypes.DOMAIN;
    }
    case 'EmailAddressInputOutput': {
      return KnownInputOutputTypes.EMAIL_ADDRESS;
    }
    case 'EmailMessageInputOutput': {
      return KnownInputOutputTypes.EMAIL_MESSAGE;
    }
    case 'FileInputOutput': {
      return KnownInputOutputTypes.FILE;
    }
    case 'FileRefInputOutput': {
      return KnownInputOutputTypes.FILE_REF;
    }
    case 'HashInputOutput': {
      return KnownInputOutputTypes.HASH;
    }
    case 'IPInputOutput': {
      return KnownInputOutputTypes.IP;
    }
    case 'JSONInputOutput': {
      return KnownInputOutputTypes.JSON;
    }
    case 'StringInputOutput': {
      return KnownInputOutputTypes.STRING;
    }
    case 'URLInputOutput': {
      return KnownInputOutputTypes.URL;
    }
  }
}
