import { FormControl, FormGroup, Validators } from '@angular/forms';
import { KnownInputOutputTypes } from '@orion/graphql-types';
import { BaseReactiveForm, SentinelValidators } from '@sentinel/common';
import { Subscription } from 'rxjs';
import { hasCustomValidation } from '../../io-validation/task-io-validation.pipe';
import { TaskInputOutputPanelFragment } from '../../../../../graphql';
import { taskIOTypenameToValidators, typeHasCustomValidation, typenameToKnownInputOutput } from '../../io-utils';

interface CreateInputOutputTaskFormModel {
  type: FormControl<KnownInputOutputTypes>;
  description: FormControl<string | null>;
  addCustomValidation: FormControl<boolean | null>;
  validation: FormControl<string | null>;
}

export interface CreateTaskInputOutputData {
  type: KnownInputOutputTypes;
  description?: string;
  validation?: string;
}

export class CreateInputOutputForm extends BaseReactiveForm<CreateInputOutputTaskFormModel> {
  validationAvailable = false;
  private subscriptions: Subscription[] = [];
  constructor(io?: TaskInputOutputPanelFragment) {
    const formGroup = new FormGroup<CreateInputOutputTaskFormModel>({
      type: new FormControl<KnownInputOutputTypes>(KnownInputOutputTypes.OTHER, {
        nonNullable: true,
        initialValueIsDefault: true,
      }),
      description: new FormControl<string | null>(io?.description ?? null),
      addCustomValidation: new FormControl<boolean | null>(null),
      validation: new FormControl<string | null>(null),
    });
    formGroup.controls.type.setValidators([Validators.required]);
    super(formGroup);
    this.subscriptions.push(this.formGroup.controls.type.valueChanges.subscribe((type) => this.onTypeChanged(type)));
    this.subscriptions.push(
      this.formGroup.controls.addCustomValidation.valueChanges.subscribe((customInputValidationActivated) =>
        this.onAddValidationChanged(customInputValidationActivated)
      )
    );
    if (io) {
      const hasValidation =
        hasCustomValidation(io) && io.validation !== undefined && io.validation !== null && io.validation !== '';
      const customValidation = hasValidation ? io?.validation : null;
      const defaultType = io
        ? typenameToKnownInputOutput(io.__typename ?? 'BaseTaskInputOutput')
        : KnownInputOutputTypes.OTHER;
      formGroup.controls.type.disable();
      formGroup.controls.type.setValue(defaultType);
      if (hasValidation) {
        formGroup.controls.addCustomValidation.setValue(hasValidation);
        formGroup.controls.validation.setValue(customValidation ?? null);
      }
    }
  }

  toData(): CreateTaskInputOutputData {
    return {
      type: this.formGroup.controls.type.value,
      description: this.formGroup.controls.description.value ?? undefined,
      validation: this.formGroup.controls.validation.value ?? undefined,
    };
  }

  onDestroy(): void {
    this.subscriptions.forEach((subscription) => subscription.unsubscribe());
  }

  private onTypeChanged(type: KnownInputOutputTypes): void {
    this.validationAvailable = typeHasCustomValidation(type);
    this.formGroup.controls.addCustomValidation.setValue(false);
  }

  private onAddValidationChanged(value: boolean | null): void {
    if (!value) {
      this.formGroup.controls.validation.setValue(null);
      this.formGroup.controls.validation.setValidators([]);
    } else {
      const validators = [SentinelValidators.noWhitespace];
      const type = this.formGroup.controls.type.value;
      if (type) {
        validators.push(...taskIOTypenameToValidators(type));
      }
      this.formGroup.controls.validation.setValidators(validators);
    }
    this.formGroup.controls.validation.updateValueAndValidity();
  }
}
