import {
  ChangeDetectionStrategy,
  Component,
  EventEmitter,
  Input,
  OnChanges,
  OnInit,
  Output,
  QueryList,
  SimpleChanges,
  ViewChildren,
} from '@angular/core';
import { FormArray, FormControl, Validators } from '@angular/forms';
import { MatCheckbox, MatCheckboxChange } from '@angular/material/checkbox';
import { jsonValidator } from '@orion/common';
import { Destroyable } from '@sentinel/common';
import { trackByIndex } from '@sentinel/common/utils';
import { distinctUntilChanged, filter, map, skipWhile, takeUntil } from 'rxjs/operators';

@Component({
  selector: 'orion-task-json-form-input-output-edit-form-view',
  templateUrl: './task-json-form-input-output-edit-form-view.component.html',
  styleUrls: ['./task-json-form-input-output-edit-form-view.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush,
})
export class TaskJsonFormInputOutputEditFormViewComponent extends Destroyable() implements OnInit, OnChanges {
  @Input() items?: string[];
  @Input() required = false;
  @Input() disabled = false;
  @Output() valueChange: EventEmitter<string[]> = new EventEmitter<string[]>();
  @Output() validityChange: EventEmitter<boolean> = new EventEmitter<boolean>();

  @ViewChildren(MatCheckbox) checkboxes?: QueryList<MatCheckbox>;

  readonly VALIDATORS = [Validators.required, jsonValidator];

  trackByFn = trackByIndex;
  formArray: FormArray<FormControl<string | null>> = new FormArray<FormControl<string | null>>([], []);

  ngOnInit(): void {
    this.formArray.valueChanges
      .pipe(
        skipWhile((value, index) => value === null && index === 0),
        takeUntil(this.destroy$),
        distinctUntilChanged(),
        map((items) => items.filter((item) => item !== null) as string[])
      )
      .subscribe((value) => this.onFormValueChange(value));
    this.formArray.statusChanges
      .pipe(
        takeUntil(this.destroy$),
        filter((status) => status === 'VALID' || status === 'INVALID'),
        map((status) => status === 'VALID'),
        distinctUntilChanged()
      )
      .subscribe((value) => this.validityChange.emit(value));
    this.formArray.updateValueAndValidity();
  }

  ngOnChanges(changes: SimpleChanges): void {
    if ('items' in changes && this.items) {
      this.setItems(this.items);
    }
  }

  hasError(errorCode: string): boolean {
    return this.formArray.controls.some((c) => c.hasError(errorCode));
  }

  onItemSelectionChange(change: MatCheckboxChange, index: number): void {
    if (change.checked) {
      this.formArray?.controls[index]?.enable();
    } else {
      this.formArray?.controls[index]?.disable();
    }
  }

  private setItems(jsonItems: string[]): void {
    this.formArray.clear({ emitEvent: false });
    jsonItems.forEach((jsonItem) => this.formArray.push(new FormControl<string>(jsonItem, this.VALIDATORS)));
    this.checkboxes?.forEach((checkbox) => (checkbox.checked = true));
  }

  /**
   * Normally form array filters the selected values for us by omitting disabled controls, however,
   * due to implementation detail of FormArray, where all its controls are disabled, it switches to
   * disabled state as well and emits every nested control value, which is something we do not want in this case.
   *
   */
  private onFormValueChange(value: string[]): void {
    this.formArray.enabled ? this.valueChange.emit(value) : this.valueChange.emit([]);
  }
}
