import {
  ChangeDetectionStrategy,
  ChangeDetectorRef,
  Component,
  EventEmitter,
  Input,
  OnChanges,
  Output,
  SimpleChanges,
} from '@angular/core';
import { JSONSchemaValidator } from '@orion/common';
import { NotificationService } from '@orion/shared/services/global';
import { Destroyable } from '@sentinel/common';
import {
  SentinelControlItem,
  SentinelControlMenuItem,
  SentinelExpandableControlItem,
} from '@sentinel/components/controls';
import { BehaviorSubject, defer, of } from 'rxjs';
import { map, take, takeUntil } from 'rxjs/operators';

export type JSONFormViewMode = 'select' | 'raw';

@Component({
  selector: 'orion-task-json-form-input-output-edit',
  templateUrl: './task-json-form-input-output-edit.component.html',
  styleUrls: ['./task-json-form-input-output-edit.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush,
})
export class TaskJsonFormInputOutputEditComponent extends Destroyable() implements OnChanges {
  @Input() initValue?: string;
  @Input() lastSavedValue?: string;
  @Input() required = false;
  @Input() disabled = false;
  @Input() validationValue?: string;
  @Output() valueChange: EventEmitter<string> = new EventEmitter<string>();
  @Output() validityChange: EventEmitter<boolean> = new EventEmitter<boolean>();

  formItems?: string[];
  formItemsSchemaError: string[] | null = null;
  rawJSON?: string;
  currValue?: string;
  viewMode$: BehaviorSubject<JSONFormViewMode> = new BehaviorSubject<JSONFormViewMode>('raw');
  controls: SentinelControlItem[] = [];
  private initFromLastSavedValue = false;

  constructor(private notification: NotificationService, private cdr: ChangeDetectorRef) {
    super();
    this.controls = [
      new SentinelExpandableControlItem('MODE', 'Mode', 'accent', of(false), [
        new SentinelControlMenuItem(
          'SELECT',
          'Select',
          'accent',
          of(false),
          defer(() => this.changeMode('select')),
          'check_box'
        ),
        new SentinelControlMenuItem(
          'RAW',
          'Raw',
          'accent',
          of(false),
          defer(() => this.changeMode('raw')),
          'data_object'
        ),
      ]),
      new SentinelControlItem(
        'RESET',
        'Reset',
        'accent',
        of(false),
        defer(() => this.reset())
      ),
      new SentinelControlItem(
        'ADD',
        'Add',
        'accent',
        this.viewMode$.asObservable().pipe(map((mode) => mode !== 'select')),
        defer(() => this.add())
      ),
    ];
  }

  ngOnChanges(changes: SimpleChanges): void {
    if ('lastSavedValue' in changes && this.lastSavedValue) {
      this.initFromLastSavedValue = true;
      this.resolveValues(this.viewMode$.getValue(), this.lastSavedValue);
    }
    if ('initValue' in changes && !this.initFromLastSavedValue) {
      this.resolveValues(this.viewMode$.getValue(), this.initValue);
    }
  }

  changeMode(mode: JSONFormViewMode): void {
    if (mode === this.viewMode$.getValue()) {
      return;
    }
    this.viewMode$.next(mode);
    if (this.currValue) {
      this.resolveValues(mode, this.currValue);
    }
  }

  reset(): void {
    if (this.viewMode$.getValue() === 'select') {
      this.tryToConvertToJSONFormItems(this.initValue ?? '[{}]');
    } else {
      // trick we need to do to trigger changes in child component because the initValue is always the same
      // thus not registered by Angular change detection
      this.rawJSON = '';
      this.cdr.detectChanges();
      this.rawJSON = this.initValue ?? '{}';
      this.cdr.detectChanges();
    }
  }

  add(): void {
    if (this.formItems) {
      this.formItems = [...this.formItems, '{\n\n}'];
    }
  }

  onItemsChange(items: string[]): void {
    this.currValue = this.convertSelectedItemsToJSON(items);
    this.valueChange.emit(this.currValue);
    if (this.validationValue) {
      this.formItemsSchemaError = JSONSchemaValidator.validate(this.currValue, this.validationValue);
      const isValid = this.formItemsSchemaError === null;
      this.validityChange.emit(isValid);
    }
  }

  onValueChange(value: string): void {
    this.currValue = value;
    this.valueChange.emit(value);
  }

  onValidityChange(valid: boolean): void {
    this.validityChange.emit(valid);
  }

  onMenuItemClicked(item: SentinelControlItem): void {
    item.result$?.pipe(take(1), takeUntil(this.destroy$)).subscribe();
  }

  private resolveValues(mode: JSONFormViewMode, value?: string): void {
    if (mode === 'select' && value) {
      this.tryToConvertToJSONFormItems(value);
    } else {
      this.rawJSON = value;
    }
  }

  private tryToConvertToJSONFormItems(value: string): void {
    try {
      const jsonItems: string[] = JSON.parse(value);
      this.formItems = jsonItems.map((item) => JSON.stringify(item, null, '\t'));
    } catch (e) {
      this.notification.emit('warning', 'Failed to convert the JSON file to form view');
      this.viewMode$.next('raw');
      this.resolveValues(this.viewMode$.getValue(), value);
    }
  }

  private convertSelectedItemsToJSON(items: string[]): string {
    let selectedJsonItems = '[';
    items.forEach((item, index) => {
      if (item) {
        const isLast = items.length - 1 === index;
        const separator = isLast ? '' : ',\n';
        selectedJsonItems += `${item}${separator}`;
      }
    });
    selectedJsonItems += ']';
    return selectedJsonItems;
  }
}
