import {
  ChangeDetectionStrategy,
  ChangeDetectorRef,
  Component,
  EventEmitter,
  Input,
  OnChanges,
  Output,
  SimpleChanges,
} from '@angular/core';
import { TaskInputOutputEditCanDeactivateService } from '@orion/task-services';
import { TaskIOTypename } from '@orion/task-types';
import { isFileType, TaskMode } from '../../io-utils';
import { TaskInputOutputLocalAutosaveService } from './task-input-output-local-autosave.service';

export interface IOTaskSave {
  resolveTask: boolean;
  value: string;
}

export interface IOTaskFileSave {
  resolveTask: boolean;
  value: File;
}

@Component({
  selector: 'orion-task-input-output-set',
  templateUrl: './task-input-output-set.component.html',
  styleUrls: ['./task-input-output-set.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush,
  providers: [TaskInputOutputLocalAutosaveService],
})
export class TaskInputOutputSetComponent implements OnChanges {
  @Input() id?: string;
  @Input() canResolve?: boolean;
  @Input() initialValue?: string;
  @Input() taskMode?: TaskMode;
  @Input() type?: TaskIOTypename;
  @Input() disabled = false;
  @Input() validationValue?: string;
  @Output() saveValue: EventEmitter<IOTaskSave> = new EventEmitter<IOTaskSave>();
  @Output() saveFile: EventEmitter<IOTaskFileSave> = new EventEmitter<IOTaskFileSave>();
  canSave = false;
  currValue?: string;
  currFile?: File;
  initializedFromLocalStorage = false;
  lastSavedValue?: string;

  constructor(
    private inputOutputEditCanDeactivateService: TaskInputOutputEditCanDeactivateService,
    private taskStorageService: TaskInputOutputLocalAutosaveService,
    private cdr: ChangeDetectorRef
  ) {}

  ngOnChanges(changes: SimpleChanges): void {
    if ('id' in changes && this.id !== undefined) {
      if (changes['id'].previousValue !== changes['id'].currentValue) {
        this.initializedFromLocalStorage = false;
      }
    }
    if (
      ('id' in changes || 'taskMode' in changes) &&
      this.id !== undefined &&
      this.taskMode &&
      !this.initializedFromLocalStorage
    ) {
      const storageValue = this.taskStorageService.get(this.id, this.taskMode);
      if (storageValue) {
        this.lastSavedValue = storageValue;
      } else {
        this.lastSavedValue = this.initialValue;
      }
      this.currValue = this.lastSavedValue;
      this.initializedFromLocalStorage = true;
    }
    if ('initialValue' in changes && this.initialValue && !this.initializedFromLocalStorage) {
      this.lastSavedValue = this.initialValue;
    }
  }

  onValueChanged(value: string): void {
    this.currValue = value;
    if (this.id && this.taskMode && value !== undefined && value !== null) {
      this.taskStorageService.update(this.id, this.taskMode, value);
    }
  }

  onFileChanged(value: File | undefined): void {
    this.currFile = value;
    this.inputOutputEditCanDeactivateService.update(false);
  }

  onValidityChanged(valid: boolean): void {
    this.canSave = valid;
    this.cdr.detectChanges(); // save button does not initialize properly otherwise
  }

  onSaved(resolve: boolean): void {
    if (this.type && isFileType(this.type) && this.currFile) {
      this.saveFile.emit({ value: this.currFile, resolveTask: resolve });
    } else if (this.currValue) {
      this.saveValue.emit({ value: this.currValue, resolveTask: resolve });
    }
    if (this.id && this.taskMode) {
      this.taskStorageService.remove(this.id, this.taskMode);
    }
  }
}
