import { Directive, EventEmitter, Input, OnChanges, OnInit, Output, SimpleChanges } from '@angular/core';
import { FormControl, ValidatorFn, Validators } from '@angular/forms';
import { Destroyable } from '@sentinel/common';
import { distinctUntilChanged, filter, map, skipWhile, takeUntil } from 'rxjs/operators';

@Directive()
export abstract class TaskInputOutputEditDirective<ValueType> extends Destroyable() implements OnInit, OnChanges {
  @Input() lastSavedValue?: ValueType;
  @Input() required = false;
  @Input() disabled = false;
  @Output() valueChange: EventEmitter<ValueType> = new EventEmitter<ValueType>();
  @Output() validityChange: EventEmitter<boolean> = new EventEmitter<boolean>();

  formControl = new FormControl<ValueType | null>(null, []);

  ngOnInit(): void {
    this.formControl.valueChanges
      .pipe(
        skipWhile((value, index) => value === null && index === 0), // skip initial value emission
        takeUntil(this.destroy$),
        distinctUntilChanged()
      )
      .subscribe((value) => {
        if (value !== null) {
          this.valueChange.emit(value);
        }
      });
    this.formControl.statusChanges
      .pipe(
        takeUntil(this.destroy$),
        filter((status) => status === 'VALID' || status === 'INVALID'),
        map((status) => status === 'VALID'),
        distinctUntilChanged()
      )
      .subscribe((value) => this.validityChange.emit(value));
    if (this.lastSavedValue !== undefined) {
      this.formControl.setValue(this.lastSavedValue);
    }
  }

  ngOnChanges(changes: SimpleChanges): void {
    if ('lastSavedValue' in changes && this.lastSavedValue !== undefined && this.lastSavedValue !== null) {
      this.formControl.setValue(this.lastSavedValue);
    }
    if ('required' in changes) {
      if (this.required) {
        this.addValidators(Validators.required);
      } else {
        this.formControl.removeValidators(Validators.required);
        this.formControl.updateValueAndValidity();
      }
    }
    if ('disabled' in changes) {
      this.disabled ? this.formControl.disable() : this.formControl.enable();
    }
  }

  addValidators(validators: ValidatorFn | ValidatorFn[]): void {
    this.formControl.addValidators(validators);
    this.formControl.updateValueAndValidity();
  }

  removeValidators(validators: ValidatorFn | ValidatorFn[]): void {
    this.formControl.removeValidators(validators);
    this.formControl.updateValueAndValidity();
  }
}
