import { ChangeDetectionStrategy, Component, Input, OnChanges, OnInit, SimpleChanges } from '@angular/core';
import { TaskInputOutputEditCanDeactivateService } from '@orion/task-services';
import { TaskInputOutputUpdate } from '@orion/task-types';
import { CurrentUserService } from '@orion/user-services';
import { Destroyable } from '@sentinel/common';
import { ApolloResult, ApolloResultType } from '@csirtmu-orion/microframework/graphql';
import { SentinelControlItem } from '@sentinel/components/controls';
import { combineLatest, Observable, of } from 'rxjs';
import { concatMap, distinctUntilChanged, map, takeUntil, tap } from 'rxjs/operators';
import { TaskOverviewTabFragment } from '../task-overview-tab.fragment.generated';
import { SetTaskInputService } from './set-task-input.service';
import { SetTaskOutputService } from './set-task-output.service';
import { CreateTaskInputOutputData, IOTaskFileSave, IOTaskSave } from './task-input-output-panel';
import { TaskIoUpdateService } from './task-io-update.service';
import { TaskIOViewStateService } from './task-io-view-state.service';
import { TaskActionsService } from '../../../../task-actions.service';

@Component({
  selector: 'orion-task-io-panels',
  templateUrl: './task-io-panels.component.html',
  styleUrls: ['./task-io-panels.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush,
  providers: [SetTaskInputService, SetTaskOutputService, TaskIoUpdateService, TaskIOViewStateService],
})
export class TaskIOPanelsComponent extends Destroyable() implements OnInit, OnChanges {
  @Input() task?: TaskOverviewTabFragment;
  canResolve = false;
  canCreateInput = false;
  inputIsLoading$?: Observable<boolean>;
  inputDisabled$?: Observable<boolean>;
  outputIsLoading$?: Observable<boolean>;
  outputDisabled$?: Observable<boolean>;
  inputControls$?: Observable<SentinelControlItem[]>;
  outputControls$?: Observable<SentinelControlItem[]>;
  editInputDisplayed$?: Observable<boolean>;
  editOutputDisplayed$?: Observable<boolean>;

  constructor(
    private inputService: SetTaskInputService,
    private outputService: SetTaskOutputService,
    private taskActionService: TaskActionsService,
    private activeUserService: CurrentUserService,
    private ioUpdateService: TaskIoUpdateService,
    private taskIOViewStateService: TaskIOViewStateService,
    private ioEditCanDeactivateService: TaskInputOutputEditCanDeactivateService
  ) {
    super();
  }

  ngOnInit(): void {
    this.activeUserId$ = this.activeUserService.data$.pipe(map((user) => user?.id));
    this.inputIsLoading$ = this.inputService.isLoading$;
    this.outputIsLoading$ = this.outputService.isLoading$;
    this.inputDisabled$ = combineLatest([this.inputIsLoading$, this.taskIOViewStateService.taskNotOpen$]).pipe(
      map(([isLoading, taskNotOpen]) => isLoading || taskNotOpen),
      distinctUntilChanged()
    );
    this.outputDisabled$ = combineLatest([
      this.outputIsLoading$,
      this.taskIOViewStateService.inputIsOpenForEdit$,
      this.taskIOViewStateService.taskNotOpen$,
    ]).pipe(
      map(([isLoading, inputOpenForEdit, taskNotOpen]) => isLoading || inputOpenForEdit || taskNotOpen),
      distinctUntilChanged()
    );
    this.inputControls$ = this.taskIOViewStateService.inputControls$;
    this.outputControls$ = this.taskIOViewStateService.outputControls$;
    this.editInputDisplayed$ = this.taskIOViewStateService.editInputDisplayed$;
    this.editOutputDisplayed$ = this.taskIOViewStateService.editOutputDisplayed$;
  }

  ngOnChanges(changes: SimpleChanges): void {
    if ('task' in changes && this.task) {
      this.taskIOViewStateService.update(this.task);
      this.canResolve = this.task.isDecision === false && this.task.isAutomated === false;
      this.canCreateInput =
        this.task.output === null ||
        this.task.output === undefined ||
        this.task.output.stringifiedValue === undefined ||
        this.task.output.stringifiedValue === null;
    }
  }

  setTaskInput(input: IOTaskSave): void {
    if (this.task) {
      const setInput$ = this.inputService.setInput(this.task.id, input.value);
      this.setInputOrOutput(this.task.id, setInput$, input.resolveTask);
    }
  }

  setTaskFileInput(input: IOTaskFileSave): void {
    if (this.task) {
      const setFileInput$ = this.inputService
        .setFileInput(this.task.id, input.value)
        .pipe(tap(() => this.ioEditCanDeactivateService.update(true)));
      this.setInputOrOutput(this.task.id, setFileInput$, input.resolveTask);
    }
  }

  setTaskOutput(output: IOTaskSave): void {
    if (this.task) {
      const setOutput$ = this.outputService.setOutput(this.task.id, output.value);
      this.setInputOrOutput(this.task.id, setOutput$, output.resolveTask);
    }
  }

  setTaskFileOutput(output: IOTaskFileSave): void {
    if (this.task) {
      const setFileOutput$ = this.outputService
        .setFileOutput(this.task.id, output.value)
        .pipe(tap(() => this.ioEditCanDeactivateService.update(true)));
      this.setInputOrOutput(this.task.id, setFileOutput$, output.resolveTask);
    }
  }

  onCreateInput(input: CreateTaskInputOutputData): void {
    if (this.task) {
      this.inputService.createTaskInput(this.task.id, input).pipe(takeUntil(this.destroy$)).subscribe();
    }
  }

  onCreateOutput(output: CreateTaskInputOutputData): void {
    if (this.task) {
      this.outputService.createTaskOutput(this.task.id, output).pipe(takeUntil(this.destroy$)).subscribe();
    }
  }

  onSaveEdit(data: TaskInputOutputUpdate, type: 'input' | 'output'): void {
    this.ioUpdateService
      .update(data, type)
      .pipe(takeUntil(this.destroy$))
      .subscribe(() => this.taskIOViewStateService.hideEditComponent(type));
  }

  onCancelEdit(type: 'input' | 'output'): void {
    this.taskIOViewStateService.hideEditComponent(type);
  }

  private setInputOrOutput(taskId: string, saveIO$: Observable<ApolloResult>, resolveTask: boolean): void {
    if (resolveTask) {
      saveIO$
        .pipe(
          concatMap((res) => {
            if (res.type === ApolloResultType.SUCCESS) {
              return this.taskActionService.resolve(taskId);
            } else {
              return of(res);
            }
          }),
          takeUntil(this.destroy$)
        )
        .subscribe();
    } else {
      saveIO$.pipe(takeUntil(this.destroy$)).subscribe();
    }
  }
}
