import { Injectable } from '@angular/core';
import { NotificationService } from '@orion/shared/services/global';
import { TaskInputOutputUpdate } from '@orion/task-types';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  ApolloResultType,
  createApolloContextErrorTitle,
} from '@csirtmu-orion/microframework/graphql';
import { MutationResult } from 'apollo-angular';
import { BehaviorSubject, Observable, of } from 'rxjs';
import { catchError, map, tap } from 'rxjs/operators';
import {
  UpdateBoolIoMutationService,
  UpdateDateIoMutationService,
  UpdateDateTimeIoMutationService,
  UpdateDomainIoMutationService,
  UpdateEmailAddressIoMutationService,
  UpdateEmailMessageIoMutationService,
  UpdateFileIoMutationService,
  UpdateFileRefIoMutationService,
  UpdateHashIoMutationService,
  UpdateIpioMutationService,
  UpdateJsonIoMutationService,
  UpdateStringIoMutationService,
  UpdateUrlIoMutationService,
} from '../graphql';
import { namedOperations } from '@orion/named-operations';

@Injectable()
export class TaskIoUpdateService {
  private inputIsLoadingSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  inputIsLoading$: Observable<boolean> = this.inputIsLoadingSubject$.asObservable();

  private outputIsLoadingSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  outputIsLoading$: Observable<boolean> = this.outputIsLoadingSubject$.asObservable();

  constructor(
    private boolMutation: UpdateBoolIoMutationService,
    private dateMutation: UpdateDateIoMutationService,
    private dateTimeMutation: UpdateDateTimeIoMutationService,
    private domainMutation: UpdateDomainIoMutationService,
    private emailAddressMutation: UpdateEmailAddressIoMutationService,
    private emailMessageMutation: UpdateEmailMessageIoMutationService,
    private fileMutation: UpdateFileIoMutationService,
    private fileRefMutation: UpdateFileRefIoMutationService,
    private hashMutation: UpdateHashIoMutationService,
    private ipMutation: UpdateIpioMutationService,
    private jsonMutation: UpdateJsonIoMutationService,
    private stringMutation: UpdateStringIoMutationService,
    private urlMutation: UpdateUrlIoMutationService,
    private notificationService: NotificationService
  ) {}

  update(task: TaskInputOutputUpdate, type: 'input' | 'output'): Observable<ApolloResult> {
    let update$: Observable<MutationResult>;
    if (type === 'input') {
      this.inputIsLoadingSubject$.next(true);
    } else {
      this.outputIsLoadingSubject$.next(true);
    }
    const mutationOptions = {
      refetchQueries: [namedOperations.Query.getTaskDetail],
      awaitRefetchQueries: true,
      context: createApolloContextErrorTitle(`Updating task ${type}`),
    };
    switch (task.type) {
      case 'BoolInputOutput': {
        const boolValue = task.stringifiedValue?.toLowerCase() === 'true';
        update$ = this.boolMutation.mutate(
          {
            id: task?.id,
            data: { description: task?.description, booleanValue: boolValue },
          },
          mutationOptions
        );
        break;
      }
      case 'DateInputOutput': {
        update$ = this.dateMutation.mutate(
          {
            id: task?.id,
            data: { description: task?.description, dateValue: task?.stringifiedValue },
          },
          mutationOptions
        );
        break;
      }
      case 'DateTimeInputOutput': {
        update$ = this.dateTimeMutation.mutate(
          {
            id: task?.id,
            data: { description: task?.description, datetimeValue: task?.stringifiedValue },
          },
          mutationOptions
        );
        break;
      }
      case 'DomainInputOutput': {
        update$ = this.domainMutation.mutate(
          {
            id: task?.id,
            data: { description: task?.description, domainValue: task?.stringifiedValue },
          },
          mutationOptions
        );
        break;
      }
      case 'EmailAddressInputOutput': {
        update$ = this.emailAddressMutation.mutate(
          {
            id: task?.id,
            data: { description: task?.description, emailAddressValue: task?.stringifiedValue },
          },
          mutationOptions
        );
        break;
      }
      case 'EmailMessageInputOutput': {
        update$ = this.emailMessageMutation.mutate(
          {
            id: task?.id,
            data: { description: task?.description, emailMessageValue: task?.stringifiedValue },
          },
          mutationOptions
        );
        break;
      }
      case 'FileInputOutput': {
        update$ = this.fileMutation.mutate(
          {
            id: task?.id,
            data: { description: task?.description, fileValue: task?.stringifiedValue },
          },
          mutationOptions
        );
        break;
      }
      case 'FileRefInputOutput': {
        update$ = this.fileRefMutation.mutate(
          {
            id: task?.id,
            data: { description: task?.description, fileRefValue: task?.stringifiedValue },
          },
          mutationOptions
        );
        break;
      }
      case 'HashInputOutput': {
        update$ = this.hashMutation.mutate(
          {
            id: task?.id,
            data: { description: task?.description, hashValue: task?.stringifiedValue },
          },
          mutationOptions
        );
        break;
      }
      case 'IPInputOutput': {
        update$ = this.ipMutation.mutate(
          {
            id: task?.id,
            data: { description: task?.description, ipValue: task?.stringifiedValue },
          },
          mutationOptions
        );
        break;
      }
      case 'JSONInputOutput': {
        update$ = this.jsonMutation.mutate(
          {
            id: task?.id,
            data: { description: task?.description, jsonValue: task?.stringifiedValue, jsonSchema: task?.validation },
          },
          mutationOptions
        );
        break;
      }
      case 'StringInputOutput': {
        update$ = this.stringMutation.mutate(
          {
            id: task?.id,
            data: {
              description: task?.description,
              stringValue: task?.stringifiedValue,
              stringRegex: task?.validation,
            },
          },
          mutationOptions
        );
        break;
      }
      case 'URLInputOutput': {
        update$ = this.urlMutation.mutate(
          {
            id: task?.id,
            data: { description: task?.description, urlValue: task?.stringifiedValue },
          },
          mutationOptions
        );
        break;
      }
      default: {
        return of(new ApolloResult(ApolloResultType.ERROR));
      }
    }
    return update$.pipe(
      map((res) => ApolloOperationResultResolver.resolve(res)),
      catchError((err) => ApolloOperationResultResolver.handleError(err)),
      tap({
        next: () => {
          this.stopLoading(type);
          this.notificationService.emit('success', `Task ${type}  was updated`);
        },
        error: () => this.stopLoading(type),
      })
    );
  }

  private stopLoading(type: 'input' | 'output'): void {
    if (type === 'input') {
      this.inputIsLoadingSubject$.next(false);
    } else {
      this.outputIsLoadingSubject$.next(false);
    }
  }
}
