import { Injectable } from '@angular/core';
import { MatDialog } from '@angular/material/dialog';
import { TaskStateEnum } from '@orion/graphql-types';
import { SentinelControlItem } from '@sentinel/components/controls';
import {
  SentinelConfirmationDialogComponent,
  SentinelConfirmationDialogConfig,
  SentinelDialogResultEnum,
} from '@sentinel/components/dialogs';
import { BehaviorSubject, defer, Observable, of } from 'rxjs';
import { concatMap, map } from 'rxjs/operators';
import { TaskInputOutputPanelFragment } from '../graphql';
import { TaskIoUpdateService } from './task-io-update.service';
import { TaskOverviewTabFragment } from '../task-overview-tab.fragment.generated';

@Injectable()
export class TaskIOViewStateService {
  private editInputDisplayedSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  private editOutputDisplayedSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  private inputIsOpenForEditSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  private taskNotOpenSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  private inputControlsSubject$: BehaviorSubject<SentinelControlItem[]> = new BehaviorSubject<SentinelControlItem[]>(
    []
  );
  private outputControlsSubject$: BehaviorSubject<SentinelControlItem[]> = new BehaviorSubject<SentinelControlItem[]>(
    []
  );
  editInputDisplayed$: Observable<boolean> = this.editInputDisplayedSubject$.asObservable();
  editOutputDisplayed$: Observable<boolean> = this.editOutputDisplayedSubject$.asObservable();
  inputIsOpenForEdit$: Observable<boolean> = this.inputIsOpenForEditSubject$.asObservable();
  taskNotOpen$: Observable<boolean> = this.taskNotOpenSubject$.asObservable();
  inputControls$: Observable<SentinelControlItem[]> = this.inputControlsSubject$.asObservable();
  outputControls$: Observable<SentinelControlItem[]> = this.outputControlsSubject$.asObservable();

  constructor(private ioUpdateService: TaskIoUpdateService, private dialog: MatDialog) {}

  update(task: TaskOverviewTabFragment): void {
    this.inputIsOpenForEditSubject$.next(this.inputIsOpenForEdit(task.input));
    this.taskNotOpenSubject$.next(task.state !== TaskStateEnum.OPEN);
    this.inputControlsSubject$.next(this.createControls('input', task));
    this.outputControlsSubject$.next(this.createControls('output', task));
  }

  hideEditComponent(type: 'input' | 'output'): void {
    if (type === 'input') {
      this.editInputDisplayedSubject$.next(false);
    } else {
      this.editOutputDisplayedSubject$.next(false);
    }
  }

  private inputIsOpenForEdit(input?: TaskInputOutputPanelFragment): boolean {
    if (input === undefined || input === null) {
      return false;
    }
    return input.stringifiedValue === undefined || input.stringifiedValue === null;
  }

  private createControls(type: 'input' | 'output', task: TaskOverviewTabFragment): SentinelControlItem[] {
    if (task.state !== TaskStateEnum.OPEN) {
      return [];
    }
    if (type === 'input') {
      if (task.input && task.input.stringifiedValue && (!task.output || !task.output.stringifiedValue)) {
        return [
          new SentinelControlItem(
            'edit',
            'Edit',
            'primary',
            this.ioUpdateService.inputIsLoading$,
            defer(() => this.displayEditComponent(type, task.isAutomated)),
            this.ioUpdateService.inputIsLoading$
          ),
        ];
      } else {
        return [];
      }
    } else {
      if (task.output && task.output.stringifiedValue) {
        return [
          new SentinelControlItem(
            'edit',
            'Edit',
            'primary',
            this.ioUpdateService.outputIsLoading$,
            defer(() => this.displayEditComponent(type, task.isAutomated)),
            this.ioUpdateService.outputIsLoading$
          ),
        ];
      } else {
        return [];
      }
    }
  }

  private displayEditComponent(type: 'input' | 'output', isAutomated = false): Observable<boolean> {
    if (isAutomated) {
      return this.displayWarning(type).pipe(
        concatMap((result) => {
          if (result === SentinelDialogResultEnum.CONFIRMED) {
            return this.setDisplayEditComponent(type);
          } else {
            return of(false);
          }
        })
      );
    } else {
      return this.setDisplayEditComponent(type);
    }
  }

  private setDisplayEditComponent(type: 'input' | 'output'): Observable<boolean> {
    if (type === 'input') {
      this.editInputDisplayedSubject$.next(true);
    } else {
      this.editOutputDisplayedSubject$.next(true);
    }
    return of(true);
  }

  private displayWarning(type: 'input' | 'output'): Observable<SentinelDialogResultEnum> {
    const dialogConfig = new SentinelConfirmationDialogConfig(
      `Edit ${type} of automated task`,
      `Do you want to edit ${type} of automated task? It might break the automation pipeline.`,
      'Cancel',
      'Confirm'
    );
    const dialogRef = this.dialog.open(SentinelConfirmationDialogComponent, { data: dialogConfig });
    return dialogRef
      .afterClosed()
      .pipe(
        map((res) =>
          res === SentinelDialogResultEnum.CONFIRMED
            ? SentinelDialogResultEnum.CONFIRMED
            : SentinelDialogResultEnum.DISMISSED
        )
      );
  }
}
