import { ChangeDetectionStrategy, Component, EventEmitter, Input, OnInit, Output } from '@angular/core';
import { resolveFromIdOrUrl } from '@orion/common';
import { TaskOrdEnum } from '@orion/graphql-types';
import { TaskListElementFragment, TaskListService } from '@orion/task-components/shared';
import {
  INIT_PAGINATION_EVENT_BASE_TOKEN,
  PaginationBaseInit,
  PaginationService,
} from '@csirtmu-orion/microframework/pagination-services';
import { trackByIndex } from '@sentinel/common/utils';
import { Observable } from 'rxjs';
import { map } from 'rxjs/operators';
import { CreateTaskReference } from './create-task-reference';
import { TaskReferenceDirectionEnum } from './task-reference-direction.enum';
import { TASK_PATH } from '@orion/case-paths';

@Component({
  selector: 'orion-create-task-reference',
  templateUrl: './create-task-reference.component.html',
  styleUrls: ['./create-task-reference.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush,
  providers: [
    TaskListService,
    PaginationService,
    {
      provide: INIT_PAGINATION_EVENT_BASE_TOKEN,
      useValue: new PaginationBaseInit(50, TaskOrdEnum.NAME, 'asc'),
    },
  ],
})
export class CreateTaskReferenceComponent implements OnInit {
  @Input() taskId?: string;
  @Output() save: EventEmitter<CreateTaskReference[]> = new EventEmitter<CreateTaskReference[]>();
  @Output() cancel: EventEmitter<void> = new EventEmitter<void>();
  referenceDirections = Object.values(TaskReferenceDirectionEnum);
  tasksToSelect$?: Observable<TaskListElementFragment[]>;
  tasksIsLoading$?: Observable<boolean>;
  tasksHasError$?: Observable<boolean>;
  selectedDirection?: TaskReferenceDirectionEnum;
  selectedTargetTasks: TaskListElementFragment[] = [];
  saveDisabled = true;
  trackByFn = trackByIndex;

  constructor(private taskListService: TaskListService, private paginationService: PaginationService) {}

  ngOnInit(): void {
    this.tasksToSelect$ = this.taskListService.data$.pipe(map((paginated) => paginated.elements));
    this.tasksIsLoading$ = this.taskListService.isLoading$;
    this.tasksHasError$ = this.taskListService.hasError$;
  }

  onSave(): void {
    if (this.taskId !== undefined && this.selectedDirection !== undefined) {
      const narrowedTaskId = this.taskId;
      const narrowedDir = this.selectedDirection;
      const taskEvents = this.selectedTargetTasks.map(
        (selectedTask) => new CreateTaskReference(narrowedTaskId, selectedTask.id, narrowedDir)
      );
      this.save.emit(taskEvents);
    }
  }

  onTaskSelectionChange(selectedTasks: TaskListElementFragment[]): void {
    this.selectedTargetTasks = selectedTasks;
    this.updateSaveDisabled();
  }

  onTaskSearch(value: string): void {
    const maybeId = resolveFromIdOrUrl(value, TASK_PATH);
    this.taskListService.update(this.paginationService.get(), {
      id: maybeId,
      exceptIds: this.taskId ? [this.taskId] : [],
    });
  }

  onDirectionSelectionChange(): void {
    this.updateSaveDisabled();
  }

  onCancel(): void {
    this.cancel.emit();
  }

  private updateSaveDisabled(): void {
    this.saveDisabled = this.selectedDirection === undefined || this.selectedTargetTasks.length < 1;
  }

  onTaskSearchTryAgain(): void {
    this.taskListService.recoverAfterError();
  }
}
