import { Injectable } from '@angular/core';
import { TaskReferenceTypeEnum } from '@orion/graphql-types';
import { NotificationService } from '@orion/shared/services/global';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  ApolloResultType,
  createApolloContextErrorTitle,
} from '@csirtmu-orion/microframework/graphql';
import { ENTITY_NOT_LOADING_STATE, EntityLoadingState } from '@sentinel/common/utils';
import { BehaviorSubject, Observable } from 'rxjs';
import { catchError, map, tap } from 'rxjs/operators';
import { CreateTaskReference } from './create-task-reference';
import {
  CreateTaskReferencesMutationService,
  CreateTaskReferencesMutationVariables,
  DeleteTaskReferencesMutationService,
} from './graphql';
import { TaskReferenceViewStateService } from './task-reference-view-state.service';
import { namedOperations } from '@orion/named-operations';

@Injectable()
export class TaskReferenceActionsService {
  private deletingInProgressSubject$: BehaviorSubject<EntityLoadingState> = new BehaviorSubject<EntityLoadingState>(
    ENTITY_NOT_LOADING_STATE
  );
  deletingInProgress$: Observable<EntityLoadingState> = this.deletingInProgressSubject$.asObservable();

  constructor(
    private notificationService: NotificationService,
    private viewStateService: TaskReferenceViewStateService,
    private createReferenceMutation: CreateTaskReferencesMutationService,
    private deleteReferenceMutation: DeleteTaskReferencesMutationService
  ) {}

  create(references: CreateTaskReference[]): Observable<ApolloResult> {
    const vars: CreateTaskReferencesMutationVariables = {
      data: references.map((reference) => {
        return {
          source: reference.sourceId,
          target: reference.targetId,
          relationshipType: TaskReferenceTypeEnum.REFERS_TO,
        };
      }),
    };
    return this.createReferenceMutation
      .mutate(vars, {
        refetchQueries: [namedOperations.Query.getTaskDetail],
        awaitRefetchQueries: true,
        context: createApolloContextErrorTitle('Creating task reference'),
      })
      .pipe(
        map((result) => ApolloOperationResultResolver.resolve(result)),
        tap((result) => {
          this.handleActionResult(result, 'create');
          this.viewStateService.hideAddComponent();
        }),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  delete(id: string): Observable<ApolloResult> {
    this.deletingInProgressSubject$.next({ isLoading: true, id: id });
    return this.deleteReferenceMutation
      .mutate(
        { ids: [id] },
        {
          refetchQueries: [namedOperations.Query.getTaskDetail],
          awaitRefetchQueries: true,
          context: createApolloContextErrorTitle('Deleting task reference'),
        }
      )
      .pipe(
        tap({
          next: () => this.deletingInProgressSubject$.next(ENTITY_NOT_LOADING_STATE),
          error: () => this.deletingInProgressSubject$.next(ENTITY_NOT_LOADING_STATE),
        }),
        map((result) => ApolloOperationResultResolver.resolve(result)),
        tap((result) => this.handleActionResult(result, 'delete')),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  private handleActionResult(result: ApolloResult, type: 'create' | 'delete'): void {
    if (result.type === ApolloResultType.SUCCESS) {
      this.onSuccess(type);
    }
  }

  private onSuccess(type: 'create' | 'delete'): void {
    const message = type === 'create' ? 'Reference was successfully created' : 'Reference was successfully deleted';
    this.notificationService.emit('success', message);
  }
}
