import { Injectable } from '@angular/core';
import { ActivatedRoute, ParamMap, Params, Router } from '@angular/router';
import { AssigneeListElementFragment, GetAssigneeListQueryService } from '@orion/case-services';
import { CaseDbUserFilter1AndFilterEntry, TaskStateEnum } from '@orion/graphql-types';
import { TaskClientFilter } from '@orion/task-types';
import { QueryParamFilterService } from '@csirtmu-orion/microframework/filter-services';
import { enumFromValue, Optional } from '@sentinel/common/utils';
import { Observable, of } from 'rxjs';
import { catchError, map } from 'rxjs/operators';

@Injectable()
export class TaskFilterService extends QueryParamFilterService<TaskClientFilter> {
  constructor(
    protected router: Router,
    protected activatedRoute: ActivatedRoute,
    private queryAssignees: GetAssigneeListQueryService
  ) {
    super(router, activatedRoute);
  }

  removeAssignee(toRemove: AssigneeListElementFragment): void {
    const currFilter = this.filterSubject$.getValue();
    const newAssignees = currFilter.assignees?.filter((assignee) => assignee.id !== toRemove.id);
    this.update({ assignees: newAssignees });
  }

  removeState(toRemove: TaskStateEnum): void {
    const currFilter = this.filterSubject$.getValue();
    const newStates = currFilter.states?.filter((state) => state !== toRemove);
    this.update({ states: newStates });
  }

  protected createFilterFromRoute(queryParamMap: ParamMap, paramMap: ParamMap): Observable<TaskClientFilter> {
    const assigneeIds = queryParamMap.getAll('assigneeIds');
    const states = queryParamMap.getAll('states');
    const nameParam = queryParamMap.get('name');
    const from = this.deslugifyTimestamp(queryParamMap.get('from'));
    const to = this.deslugifyTimestamp(queryParamMap.get('to'));
    const dueDate = this.deslugifyTimestamp(queryParamMap.get('dueDateUntil'));
    if (this.isParamMapEmpty(nameParam, assigneeIds, states, from, to, dueDate)) {
      return of(this.EMPTY_FILTER);
    }
    const name = this.deslugify(nameParam);
    const taskStates = this.resolveStates(states);
    const syncFilter = this.createBaseFilter(name, taskStates, from, to, dueDate);
    if (this.hasAsyncParams(assigneeIds)) {
      return this.extendFilterOfAsyncParams(syncFilter, assigneeIds);
    } else {
      return of(syncFilter);
    }
  }

  protected createEmptyFilter(): TaskClientFilter {
    return {
      name: undefined,
      from: undefined,
      to: undefined,
      dueDateUntil: undefined,
      states: [],
      assignees: [],
    };
  }

  protected toQueryParams(filter: TaskClientFilter): Params {
    const params = super.toQueryParams(filter);
    delete params.exceptId;
    delete params.stageId;
    delete params.assignees;
    const assigneeIds = filter.assignees?.map((assignee) => assignee.id);
    if (filter.from) {
      params.from = filter.from.valueOf();
    }
    if (filter.to) {
      params.to = filter.to.valueOf();
    }
    if (filter.dueDateUntil) {
      params.dueDateUntil = filter.dueDateUntil.valueOf();
    }
    return { ...params, assigneeIds };
  }

  private isParamMapEmpty(
    nameParam: string | null,
    assigneeIds: string[],
    stateParams: string[],
    from?: Date,
    to?: Date,
    dueDate?: Date
  ): boolean {
    return (
      nameParam === null &&
      from === undefined &&
      to === undefined &&
      dueDate === undefined &&
      assigneeIds.length === 0 &&
      stateParams.length === 0
    );
  }

  private resolveStates(states: string[]): TaskStateEnum[] {
    const maybeStates = states.map((state) => this.resolveState(this.deslugify(state)));
    return maybeStates.filter((state) => state !== undefined) as TaskStateEnum[];
  }

  private resolveState(value?: string): Optional<TaskStateEnum> {
    return value ? enumFromValue(TaskStateEnum, value) : undefined;
  }

  private createBaseFilter(
    name?: string,
    states?: TaskStateEnum[],
    from?: Date,
    to?: Date,
    dueDateUntil?: Date
  ): TaskClientFilter {
    return {
      name,
      states,
      from,
      to,
      dueDateUntil,
    };
  }

  private hasAsyncParams(assigneeIds: string[]): boolean {
    return assigneeIds.length > 0;
  }

  private extendFilterOfAsyncParams(syncFilter: TaskClientFilter, assigneeIds: string[]): Observable<TaskClientFilter> {
    return this.extendFilterOfAsyncAssignees(syncFilter, assigneeIds);
  }

  private extendFilterOfAsyncAssignees(
    syncFilter: TaskClientFilter,
    assigneeIds: string[]
  ): Observable<TaskClientFilter> {
    return this.fetchAssignees(assigneeIds).pipe(
      map((users) => this.extendFilterOfAssignees(syncFilter, users)),
      catchError(() => of(syncFilter))
    );
  }

  private fetchAssignees(ids: string[]): Observable<Optional<AssigneeListElementFragment[]>> {
    const filter: CaseDbUserFilter1AndFilterEntry = { AND: [{ OR: { id: { in_: ids } } }] };
    return this.queryAssignees.fetch({ filter: [filter] }).pipe(
      map((queryResult) => {
        if (queryResult?.data?.getCaseDbUserList?.data) {
          return queryResult?.data?.getCaseDbUserList?.data;
        } else {
          return undefined;
        }
      })
    );
  }

  private extendFilterOfAssignees(
    baseFilter: TaskClientFilter,
    assignees?: AssigneeListElementFragment[]
  ): TaskClientFilter {
    return { ...baseFilter, assignees: assignees };
  }
}
