import { FormArray, FormControl, FormControlStatus, FormGroup, Validators } from '@angular/forms';
import { defaultEntityControls, EntityFormModel } from '@orion/common';
import { KnownInputOutputTypes, TaskRelationshipTypeEnum } from '@orion/graphql-types';
import { BaseReactiveForm, SentinelValidators } from '@sentinel/common';
import { Observable, of, Subscription } from 'rxjs';
import { CreateParentlessTaskData, CreateTaskData, CreateTaskWithParentData } from './create-task-data';
import { taskIOTypenameToValidators, typeHasCustomValidation } from '@orion/task-components/task-detail';
import { TaskListElementFragment } from '@orion/task-components/shared';
import { distinctUntilChanged, filter, map, startWith } from 'rxjs/operators';

export interface TaskPredecessorFormGroupModel {
  predecessor: FormControl<TaskListElementFragment | null>;
  predecessorRelationship: FormControl<TaskRelationshipTypeEnum | null>;
}

export interface CreateTaskBasicInfoModel extends EntityFormModel {
  aggregateResults: FormControl<boolean | null>;
  isDecision: FormControl<boolean | null>;
}

export interface CreateTaskInputModel {
  input: FormControl<string | null>;
  inputType: FormControl<KnownInputOutputTypes | null>;
  inputDescription: FormControl<string | null>;
  addCustomInputValidation: FormControl<boolean | null>;
  inputValidation: FormControl<string | null>;
}

export interface CreateTaskOutputModel {
  outputType: FormControl<KnownInputOutputTypes | null>;
  outputDescription: FormControl<string | null>;
  addCustomOutputValidation: FormControl<boolean | null>;
  outputValidation: FormControl<string | null>;
}

export interface CreateTaskPredecessorsModel {
  predecessors: FormArray<FormGroup<TaskPredecessorFormGroupModel | null>>;
}

export interface CreateTaskFormModel {
  basicInfoGroup: FormGroup<CreateTaskBasicInfoModel | null>;
  inputGroup: FormGroup<CreateTaskInputModel | null>;
  outputGroup: FormGroup<CreateTaskOutputModel | null>;
  predecessorsGroup: FormGroup<CreateTaskPredecessorsModel | null>;
}

export class CreateTaskForm extends BaseReactiveForm<CreateTaskFormModel> {
  readonly isParentless: boolean;
  readonly steps: number;
  private subscriptions: Subscription[] = [];
  inputCustomValidationAvailable = false;
  outputCustomValidationAvailable = false;

  constructor(isParentless = true) {
    const formGroup = new FormGroup<CreateTaskFormModel>({
      basicInfoGroup: new FormGroup<CreateTaskBasicInfoModel | null>({
        ...defaultEntityControls(undefined, undefined, true),
        aggregateResults: new FormControl<boolean | null>(false),
        isDecision: new FormControl<boolean | null>(false),
      }),
      inputGroup: new FormGroup<CreateTaskInputModel | null>({
        input: new FormControl<string | null>(null),
        inputType: new FormControl<KnownInputOutputTypes | null>(null),
        inputDescription: new FormControl<string | null>(null),
        addCustomInputValidation: new FormControl<boolean | null>(false),
        inputValidation: new FormControl<string | null>(null),
      }),
      outputGroup: new FormGroup<CreateTaskOutputModel | null>({
        outputType: new FormControl<KnownInputOutputTypes | null>(null),
        outputDescription: new FormControl<string | null>(null),
        addCustomOutputValidation: new FormControl<boolean | null>(false),
        outputValidation: new FormControl<string | null>(null),
      }),
      predecessorsGroup: new FormGroup<CreateTaskPredecessorsModel | null>({
        predecessors: new FormArray<FormGroup<TaskPredecessorFormGroupModel | null>>(
          [
            new FormGroup<TaskPredecessorFormGroupModel | null>({
              predecessor: new FormControl<TaskListElementFragment | null>(
                null,
                isParentless ? [] : [Validators.required]
              ),
              predecessorRelationship: new FormControl<TaskRelationshipTypeEnum | null>(
                TaskRelationshipTypeEnum.ON_COMPLETION,
                isParentless ? [] : [Validators.required]
              ),
            }),
          ],
          isParentless ? [] : [Validators.required, Validators.minLength(1)]
        ),
      }),
    });
    super(formGroup);
    this.isParentless = isParentless;
    this.steps = this.isParentless ? 3 : 4;
    if (this.formGroup.controls.basicInfoGroup.controls) {
      this.subscriptions.push(
        this.formGroup.controls.basicInfoGroup.controls.aggregateResults.valueChanges.subscribe((aggregateResults) => {
          this.onAggregateResultChanged(aggregateResults);
        })
      );
    }
    if (this.formGroup.controls.inputGroup.controls) {
      this.subscriptions.push(
        this.formGroup.controls.inputGroup.controls.inputType.valueChanges.subscribe((type) =>
          this.onInputTypeChanged(type)
        )
      );
      this.subscriptions.push(
        this.formGroup.controls.inputGroup.controls.addCustomInputValidation.valueChanges.subscribe(
          (customInputValidationActivated) => this.onAddInputCustomValidationChange(customInputValidationActivated)
        )
      );
    }
    if (this.formGroup.controls.outputGroup.controls) {
      this.subscriptions.push(
        this.formGroup.controls.outputGroup.controls.outputType.valueChanges.subscribe((type) =>
          this.onOutputTypeChanged(type)
        )
      );
      this.subscriptions.push(
        this.formGroup.controls.outputGroup.controls.addCustomOutputValidation.valueChanges.subscribe(
          (customOutputValidationActivated) => this.onAddOutputCustomValidationChange(customOutputValidationActivated)
        )
      );
    }
  }

  onDestroy(): void {
    this.subscriptions.forEach((subscription) => subscription.unsubscribe());
  }

  addPredecessors(selectedTask: TaskListElementFragment): void {
    if (this.formGroup.controls.predecessorsGroup.controls) {
      if (
        this.formGroup.controls.predecessorsGroup.controls.predecessors.length === 1 &&
        this.formGroup.controls.predecessorsGroup.controls.predecessors.controls[0]?.controls?.predecessor?.value ===
          null
      ) {
        this.formGroup.controls.predecessorsGroup.controls.predecessors.controls[0].controls.predecessor.setValue(
          selectedTask
        );
      } else {
        this.formGroup.controls.predecessorsGroup.controls.predecessors.push(
          new FormGroup<TaskPredecessorFormGroupModel | null>({
            predecessor: new FormControl<TaskListElementFragment | null>(selectedTask, [Validators.required]),
            predecessorRelationship: new FormControl<TaskRelationshipTypeEnum | null>(
              TaskRelationshipTypeEnum.ON_COMPLETION,
              [Validators.required]
            ),
          })
        );
      }
    }
  }

  getData(): CreateTaskData {
    return this.isParentless ? this.createParentlessTask() : this.createTask();
  }

  getStepValid(step: number): Observable<boolean> {
    let statusChange$: Observable<FormControlStatus> = of('INVALID');
    if (step === 0) {
      statusChange$ = this.formGroup.controls.basicInfoGroup.statusChanges.pipe(
        startWith(this.formGroup.controls.basicInfoGroup.status)
      );
    }
    if (step === 1) {
      statusChange$ = this.formGroup.controls.inputGroup.statusChanges.pipe(
        startWith(this.formGroup.controls.inputGroup.status)
      );
    }
    if (step === 2) {
      statusChange$ = this.formGroup.controls.outputGroup.statusChanges.pipe(
        startWith(this.formGroup.controls.outputGroup.status)
      );
    }
    return statusChange$.pipe(
      filter((status) => status === 'VALID' || status === 'INVALID'),
      distinctUntilChanged(),
      map((status) => status === 'VALID')
    );
  }

  /**
   * Need to change it manually as we do not control inputs validators
   *
   */
  onInputValidityChange(valid: boolean): void {
    this.formGroup.controls.inputGroup.setErrors(valid ? null : { incorrect: true });
    this.formGroup.updateValueAndValidity();
  }

  private createParentlessTask(): CreateParentlessTaskData {
    return {
      isDecision: this.formGroup.controls.basicInfoGroup?.controls?.isDecision.value ?? false,
      aggregatesResults: this.formGroup.controls.basicInfoGroup?.controls?.aggregateResults.value ?? false,
      name: this.formGroup.controls?.basicInfoGroup?.controls?.name.value ?? '',
      description: this.formGroup.controls?.basicInfoGroup?.controls?.description.value ?? '',
      input: this.formGroup.controls.inputGroup?.controls?.input.value ?? undefined,
      inputType: this.formGroup.controls.inputGroup?.controls?.inputType.value ?? undefined,
      inputDescription: this.formGroup.controls.inputGroup?.controls?.inputDescription.value ?? undefined,
      inputFormat: this.formGroup.controls.inputGroup?.controls?.inputValidation?.value ?? undefined,
      outputType: this.formGroup.controls.outputGroup?.controls?.outputType.value ?? undefined,
      outputDescription: this.formGroup.controls.outputGroup?.controls?.outputDescription.value ?? undefined,
      outputFormat: this.formGroup.controls.outputGroup?.controls?.outputValidation?.value ?? undefined,
    };
  }

  private createTask(): CreateTaskWithParentData {
    return {
      isDecision: this.formGroup.controls.basicInfoGroup?.controls?.isDecision.value ?? false,
      aggregatesResults: this.formGroup.controls.basicInfoGroup?.controls?.aggregateResults.value ?? false,
      name: this.formGroup.controls?.basicInfoGroup?.controls?.name.value ?? '',
      description: this.formGroup.controls?.basicInfoGroup?.controls?.description.value ?? '',
      input: this.formGroup.controls.inputGroup?.controls?.input.value ?? undefined,
      inputType: this.formGroup.controls.inputGroup?.controls?.inputType.value ?? undefined,
      inputDescription: this.formGroup.controls.inputGroup?.controls?.inputDescription.value ?? undefined,
      inputFormat: this.formGroup.controls.inputGroup?.controls?.inputValidation?.value ?? undefined,
      outputType: this.formGroup.controls.outputGroup?.controls?.outputType.value ?? undefined,
      outputDescription: this.formGroup.controls.outputGroup?.controls?.outputDescription.value ?? undefined,
      outputFormat: this.formGroup.controls.outputGroup?.controls?.outputValidation?.value ?? undefined,
      predecessors:
        this.formGroup.controls.predecessorsGroup?.controls?.predecessors.controls.map((fromGroup) => {
          return {
            id: fromGroup?.controls?.predecessor?.value?.id ?? '-1',
            condition: fromGroup?.controls?.predecessorRelationship?.value ?? TaskRelationshipTypeEnum.ON_COMPLETION,
          };
        }) ?? [],
    };
  }

  private onAggregateResultChanged(aggregateResults: boolean | null): void {
    if (aggregateResults) {
      this.formGroup.controls.outputGroup.controls?.outputType.setValue(KnownInputOutputTypes.JSON);
      this.formGroup.controls.outputGroup.controls?.outputType.disable();
    } else {
      this.formGroup.controls.outputGroup.controls?.outputType.setValue(null);
      this.formGroup.controls.outputGroup.controls?.outputType.enable();
    }
  }

  private onInputTypeChanged(type: KnownInputOutputTypes | null): void {
    const inputValidators = type ? taskIOTypenameToValidators(type) : [];
    this.formGroup.controls?.inputGroup?.controls?.input.setValidators(inputValidators);
    this.formGroup.controls?.inputGroup?.controls?.input.updateValueAndValidity();
    this.inputCustomValidationAvailable = type !== null && typeHasCustomValidation(type);
    this.formGroup.controls.inputGroup?.controls?.addCustomInputValidation.setValue(false);
  }

  private onOutputTypeChanged(type: KnownInputOutputTypes | null): void {
    this.outputCustomValidationAvailable = type !== null && typeHasCustomValidation(type);
    this.formGroup.controls?.outputGroup?.controls?.addCustomOutputValidation.setValue(false);
  }

  private onAddInputCustomValidationChange(value: boolean | null): void {
    if (!value) {
      this.formGroup.controls?.inputGroup?.controls?.inputValidation.setValue(null);
      this.formGroup.controls?.inputGroup?.controls?.inputValidation.setValidators([]);
    } else {
      const validators = [SentinelValidators.noWhitespace];
      const inputType = this.formGroup.controls?.inputGroup?.controls?.inputType.value;
      if (inputType) {
        validators.push(...taskIOTypenameToValidators(inputType));
      }
      this.formGroup.controls.inputGroup?.controls?.inputValidation.setValidators(validators);
    }
    this.formGroup.controls.inputGroup?.controls?.inputValidation.updateValueAndValidity();
  }

  private onAddOutputCustomValidationChange(value: boolean | null): void {
    if (!value) {
      this.formGroup.controls?.outputGroup?.controls?.outputValidation.setValue(null);
      this.formGroup.controls?.outputGroup?.controls?.outputValidation.setValidators([]);
    } else {
      const validators = [SentinelValidators.noWhitespace];
      const outputType = this.formGroup.controls?.outputGroup?.controls?.outputType.value;
      if (outputType) {
        validators.push(...taskIOTypenameToValidators(outputType));
      }
      this.formGroup.controls?.outputGroup?.controls?.outputValidation.setValidators(validators);
    }
    this.formGroup.controls?.outputGroup?.controls?.outputValidation.updateValueAndValidity();
  }
}
