import { Injectable } from '@angular/core';
import { TtpIdFragment } from '@orion/ttps-types';
import { Optional } from '@sentinel/common/utils';
import { OrionNestedTreeControl, TreeNode } from '@csirtmu-orion/microframework/nested-tree';
import { EMPTY, Observable } from 'rxjs';
import { filter, mergeMap, tap } from 'rxjs/operators';
import { SelectedTtpFragment, SelectedTtpTechniqueFragment, TtpNodeFragment } from './graphql';
import { isSelectedTtpSubTechniqueFragment, isSelectedTtpTechniqueFragment } from './model';
import { TTPTreeNodeManagerService } from './ttp-tree-node-manager.service';

@Injectable()
export class TTPTreeExpanderService {
  constructor(private ttpNodeManager: TTPTreeNodeManagerService) {}

  expand(
    selectedTTP: SelectedTtpFragment,
    treeControl: OrionNestedTreeControl<TtpNodeFragment, string>
  ): Optional<Observable<TreeNode<TtpNodeFragment>[]>> {
    if (isSelectedTtpTechniqueFragment(selectedTTP)) {
      return this.expandTreeToTechnique(selectedTTP, treeControl);
    }
    if (isSelectedTtpSubTechniqueFragment(selectedTTP)) {
      return this.expandTreeToSubTechnique(selectedTTP, treeControl);
    } else {
      return undefined;
    }
  }

  private expandTreeToTechnique(
    selectedTTP: SelectedTtpTechniqueFragment,
    treeControl: OrionNestedTreeControl<TtpNodeFragment, string>
  ): Observable<TreeNode<TtpNodeFragment>[]> {
    const ttpIds = selectedTTP.tactics?.data ?? [];
    return this.ttpNodeManager.rootNodes$.pipe(
      tap((roots) => {
        const matching = this.matchNodesByTTPIds(ttpIds, roots);
        if (matching) {
          treeControl?.toggle(matching);
        }
      })
    );
  }

  private expandTreeToSubTechnique(
    selectedTTP: SelectedTtpTechniqueFragment,
    treeControl: OrionNestedTreeControl<TtpNodeFragment, string>
  ): Observable<TreeNode<TtpNodeFragment>[]> {
    const tacticIds = selectedTTP.subTechniqueOf?.tactics?.data ?? [];
    const techniqueIds = [selectedTTP.subTechniqueOf] as TtpIdFragment[];
    return this.ttpNodeManager.rootNodes$.pipe(
      mergeMap((roots) => {
        const matching = this.matchNodesByTTPIds(tacticIds, roots);
        if (matching) {
          treeControl?.toggle(matching);
          return matching.children$.pipe(filter((children) => children.length > 0));
        } else {
          return EMPTY;
        }
      }),
      tap((children) => {
        const matching = this.matchNodesByTTPIds(techniqueIds, children);
        if (matching) {
          treeControl?.toggle(matching);
        }
      })
    );
  }

  private matchNodesByTTPIds(
    ttpIds: TtpIdFragment[],
    nodes: TreeNode<TtpNodeFragment>[]
  ): Optional<TreeNode<TtpNodeFragment>> {
    return nodes.find(
      (node) => ttpIds.find((ttpId) => ttpId.id === node.data.id && ttpId.ttpType === node.data.ttpType) !== undefined
    );
  }
}
