import { Injectable } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { CreateEditEntityService } from '@orion/shared-services';
import { NotificationService } from '@orion/shared/services/global';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  createApolloContextErrorTitle,
} from '@csirtmu-orion/microframework/graphql';
import { EMPTY, from, Observable, of } from 'rxjs';
import { catchError, concatMap, map, take, tap } from 'rxjs/operators';
import {
  CreateWorkflowTemplateMutationService,
  EditWorkflowTemplateFragment,
  UpdateWorkflowTemplateMutationService,
} from './graphql';
import { WorkflowTemplateForm } from './workflow-template-form';
import { Optional } from '@sentinel/common/utils';
import { WORKFLOW_TEMPLATE_PARAM_SELECTOR } from '@orion/workflow-template-paths';
import { GetWorkflowTemplateEditQueryService } from '@orion/workflow-template-services';

@Injectable()
export class CreateEditWorkflowTemplateService extends CreateEditEntityService<WorkflowTemplateForm> {
  constructor(
    protected notificationService: NotificationService,
    protected activatedRoute: ActivatedRoute,
    private router: Router,
    private getWorkflowTemplateQuery: GetWorkflowTemplateEditQueryService,
    private createMutation: CreateWorkflowTemplateMutationService,
    private updateMutation: UpdateWorkflowTemplateMutationService
  ) {
    super('Workflow Template', notificationService, activatedRoute);
  }

  protected createEntity(form: WorkflowTemplateForm): Observable<ApolloResult> {
    return this.createMutation
      .mutate(
        { workflowTemplate: form.getData() },
        { context: createApolloContextErrorTitle('Creating workflow template') }
      )
      .pipe(
        map((res) => ApolloOperationResultResolver.resolve(res)),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  protected createForm(): Observable<WorkflowTemplateForm> {
    return this.getEntityId().pipe(
      take(1),
      concatMap((id) => {
        if (id === undefined) {
          return of(undefined);
        } else {
          return this.getWorkflowTemplateQuery
            .fetch({ id: id })
            .pipe(map((queryResult) => queryResult?.data?.getWorkflowTemplate));
        }
      }),
      tap((template) => this.resolveEditMode(template)),
      map((template) => new WorkflowTemplateForm(template))
    );
  }

  protected getEntityId(): Observable<Optional<string>> {
    return this.activatedRoute.paramMap.pipe(
      take(1),
      map((paramMap) => paramMap.get(WORKFLOW_TEMPLATE_PARAM_SELECTOR) ?? undefined)
    );
  }

  protected handleCancel(): Observable<boolean> {
    return this.navigateToOverview();
  }

  protected onSaveError(): Observable<boolean> {
    return of(false);
  }

  protected onSaveSuccess(): Observable<boolean> {
    return this.navigateToOverview();
  }

  protected updateEntity(form: WorkflowTemplateForm): Observable<ApolloResult> {
    return this.entityId$.pipe(
      map((entityId) => {
        return entityId ? { id: entityId, patch: form.getData() } : undefined;
      }),
      concatMap((vars) =>
        vars !== undefined
          ? this.updateMutation.mutate(vars, { context: createApolloContextErrorTitle('Updating workflow template') })
          : EMPTY
      ),
      map((res) => ApolloOperationResultResolver.resolve(res)),
      catchError((err) => ApolloOperationResultResolver.handleError(err))
    );
  }

  private navigateToOverview(): Observable<boolean> {
    const command = this.editMode ? '../../' : '../';
    return from(this.router.navigate([command], { relativeTo: this.activatedRoute }));
  }

  private resolveEditMode(template?: EditWorkflowTemplateFragment): void {
    this.editMode = template !== undefined && template !== null;
  }
}
