import { Injectable } from '@angular/core';
import { MatDialog } from '@angular/material/dialog';
import { ActivatedRoute, Router } from '@angular/router';
import { namedOperations } from '@orion/named-operations';
import { FileDownloadService } from '@orion/shared-services';
import { NotificationService } from '@orion/shared/services/global';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  ApolloResultType,
  createApolloContextErrorTitle,
  wrapApolloContextForFileUpload,
} from '@csirtmu-orion/microframework/graphql';
import { ENTITY_NOT_LOADING_STATE, EntityLoadingState } from '@sentinel/common/utils';
import {
  SentinelConfirmationDialogComponent,
  SentinelConfirmationDialogConfig,
  SentinelDialogResultEnum,
} from '@sentinel/components/dialogs';
import { SentinelNotificationTypeEnum } from '@sentinel/layout/notification';
import { BehaviorSubject, EMPTY, from, Observable, of } from 'rxjs';
import { catchError, concatMap, map, tap } from 'rxjs/operators';
import { DeleteWorkflowTemplatesMutationService, UploadWorkflowTemplateMutationService } from './graphql';
import { WorkflowTemplateTableElementFragment } from './workflow-template-table';

@Injectable()
export class WorkflowTemplateOverviewActionsService {
  private deletingInProgressSubject$: BehaviorSubject<EntityLoadingState> = new BehaviorSubject<EntityLoadingState>(
    ENTITY_NOT_LOADING_STATE
  );
  deletingInProgress$: Observable<EntityLoadingState> = this.deletingInProgressSubject$.asObservable();
  private uploadInProgressSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  uploadInProgress$: Observable<boolean> = this.uploadInProgressSubject$.asObservable();

  showUploadFileComponent$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);

  constructor(
    private router: Router,
    private activatedRoute: ActivatedRoute,
    private deleteMutationService: DeleteWorkflowTemplatesMutationService,
    private downloadService: FileDownloadService,
    private uploadMutationService: UploadWorkflowTemplateMutationService,
    private notificationService: NotificationService,
    private dialog: MatDialog
  ) {}

  create(): Observable<boolean> {
    return from(this.router.navigate(['create'], { relativeTo: this.activatedRoute }));
  }

  edit(id: string): Observable<boolean> {
    return from(this.router.navigate([id, 'edit'], { relativeTo: this.activatedRoute }));
  }

  delete(ids: string[]): Observable<ApolloResult> {
    return this.confirmDeletion().pipe(
      concatMap((dialogResult) => this.handleDeletionDialogResult(dialogResult, ids)),
      tap((result) => this.handleDeletionResult(result))
    );
  }

  showUploadComponent(): void {
    return this.showUploadFileComponent$.next(true);
  }

  hideUploadComponent(): void {
    return this.showUploadFileComponent$.next(false);
  }

  upload(file: File): Observable<ApolloResult> {
    this.uploadInProgressSubject$.next(true);
    return this.uploadMutationService
      .mutate(
        { file: file },
        {
          refetchQueries: [namedOperations.Query.getWorkflowTemplateOverview],
          context: wrapApolloContextForFileUpload(createApolloContextErrorTitle('Uploading workflow template')),
        }
      )
      .pipe(
        tap({
          next: () => {
            this.uploadInProgressSubject$.next(false);
            this.hideUploadComponent();
          },
          error: () => this.uploadInProgressSubject$.next(false),
        }),
        map((res) => ApolloOperationResultResolver.resolve(res)),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  download(workflowTemplate?: WorkflowTemplateTableElementFragment): Observable<ApolloResult> {
    if (workflowTemplate?.src) {
      return this.downloadService.download(workflowTemplate.src, `${workflowTemplate.name}.json`).pipe(
        tap((result) => {
          if (result.type === ApolloResultType.SUCCESS) {
            this.notificationService.emit('success', 'File was successfully downloaded');
          } else {
            this.notificationService.emit('error', 'Download failed');
          }
        })
      );
    } else {
      return of(new ApolloResult(ApolloResultType.ERROR));
    }
  }

  private confirmDeletion(): Observable<SentinelDialogResultEnum> {
    const title = 'Delete Workflow Template';
    const message = `Do you really want to delete selected workflow template?`;
    const confirmationConfig = new SentinelConfirmationDialogConfig(title, message, 'Cancel', 'Confirm');
    const dialogRef = this.dialog.open(SentinelConfirmationDialogComponent, {
      data: confirmationConfig,
    });
    return dialogRef
      .afterClosed()
      .pipe(
        map((res) =>
          res === SentinelDialogResultEnum.CONFIRMED
            ? SentinelDialogResultEnum.CONFIRMED
            : SentinelDialogResultEnum.DISMISSED
        )
      );
  }

  private handleDeletionDialogResult(dialogResult: SentinelDialogResultEnum, ids: string[]): Observable<ApolloResult> {
    return dialogResult === SentinelDialogResultEnum.CONFIRMED ? this.handleDeletion(ids) : EMPTY;
  }

  private handleDeletion(ids: string[]): Observable<ApolloResult> {
    this.deletingInProgressSubject$.next({ isLoading: true, ids: ids });
    return this.deleteMutationService
      .mutate(
        { ids: ids },
        {
          refetchQueries: [namedOperations.Query.getWorkflowTemplateOverview],
          awaitRefetchQueries: true,
          context: createApolloContextErrorTitle('Deleting workflow template'),
        }
      )
      .pipe(
        tap({
          next: () => this.deletingInProgressSubject$.next(ENTITY_NOT_LOADING_STATE),
          error: () => this.deletingInProgressSubject$.next(ENTITY_NOT_LOADING_STATE),
        }),
        map((res) => ApolloOperationResultResolver.resolve(res)),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  private handleDeletionResult(result: ApolloResult): void {
    if (result.type === ApolloResultType.SUCCESS) {
      this.notificationService.emit(SentinelNotificationTypeEnum.Success, 'Deleted workflow template');
    }
  }
}
