import { Injectable } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { CreateEntityService } from '@orion/shared-services';
import { NotificationService } from '@orion/shared/services/global';
import { CurrentUserService } from '@orion/user-services';
import { CreateWorkflowDTO, CreateWorkflowForm, WorkflowRefs } from '@orion/workflow-types';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  createApolloContextErrorTitle,
} from '@csirtmu-orion/microframework/graphql';
import { from, Observable, of } from 'rxjs';
import { catchError, concatMap, map, take } from 'rxjs/operators';
import { CreateWorkflowMutationService } from './graphql';
import { CASE_PARAM_SELECTOR } from '@orion/case-paths';

@Injectable()
export class CreateWorkflowService extends CreateEntityService<CreateWorkflowForm> {
  constructor(
    protected notificationService: NotificationService,
    protected activatedRoute: ActivatedRoute,
    private router: Router,
    private createWorkflowMutation: CreateWorkflowMutationService,
    private activeUser: CurrentUserService
  ) {
    super('Workflow', notificationService, activatedRoute);
  }

  protected createForm(): Observable<CreateWorkflowForm> {
    return of(new CreateWorkflowForm());
  }

  protected createEntity(form: CreateWorkflowForm): Observable<ApolloResult> {
    return this.createWorkflowRefs().pipe(
      map((refs) => new CreateWorkflowDTO(form.getData(), refs)),
      concatMap((dto) =>
        this.createWorkflowMutation.mutate(
          { workflow: dto },
          { context: createApolloContextErrorTitle('Creating workflow') }
        )
      ),
      map((res) => ApolloOperationResultResolver.resolve(res)),
      catchError((err) => ApolloOperationResultResolver.handleError(err))
    );
  }

  protected handleCancel(): Observable<boolean> {
    return this.navigateToCase();
  }

  protected onSaveError(): Observable<boolean> {
    return of(false);
  }

  protected onSaveSuccess(): Observable<boolean> {
    return this.navigateToCase();
  }

  private createWorkflowRefs(): Observable<WorkflowRefs> {
    return this.activatedRoute.paramMap.pipe(
      take(1),
      map((paramMap) => {
        const caseId = paramMap.get(CASE_PARAM_SELECTOR) ?? undefined;
        const authorId = this.activeUser.getCurrentUser()?.id;
        if (caseId !== undefined && authorId !== undefined) {
          return { caseId, authorId };
        } else {
          throw 'error';
        }
      })
    );
  }

  private navigateToCase(): Observable<boolean> {
    return from(this.router.navigate(['../../'], { relativeTo: this.activatedRoute }));
  }
}
