import { Injectable } from '@angular/core';
import { FetchResult } from '@apollo/client/core';
import { namedOperations } from '@orion/named-operations';
import { NotificationService } from '@orion/shared/services/global';
import {
  ApolloOperationResultResolver,
  ApolloResult,
  ApolloResultType,
  createApolloContextErrorTitle,
} from '@csirtmu-orion/microframework/graphql';
import { BehaviorSubject, Observable } from 'rxjs';
import { catchError, map, tap } from 'rxjs/operators';
import { ExportToWorkflowTemplateMutationService, RejectWorkflowMutationService } from './graphql';

@Injectable()
export class WorkflowActionService {
  private rejectInProgressSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  rejectInProgress$: Observable<boolean> = this.rejectInProgressSubject$.asObservable();

  private exportInProgressSubject$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);
  exportInProgress$: Observable<boolean> = this.exportInProgressSubject$.asObservable();

  constructor(
    private rejectWorkflowMutation: RejectWorkflowMutationService,
    private exportWorkflowMutation: ExportToWorkflowTemplateMutationService,
    private notificationService: NotificationService
  ) {}

  reject(id: string): Observable<ApolloResult> {
    this.rejectInProgressSubject$.next(true);
    return this.rejectWorkflowMutation
      .mutate(
        { workflowId: id },
        {
          refetchQueries: [namedOperations.Query.getWorkflowDetail],
          context: createApolloContextErrorTitle('Rejecting workflow'),
          awaitRefetchQueries: true,
        }
      )
      .pipe(
        map((res) => this.handleActionResponse(res, 'Workflow was successfully rejected')),
        tap({
          next: () => this.rejectInProgressSubject$.next(false),
          error: () => this.rejectInProgressSubject$.next(false),
        }),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  exportToTemplate(id: string): Observable<ApolloResult> {
    this.exportInProgressSubject$.next(true);
    return this.exportWorkflowMutation
      .mutate(
        { workflowId: id },
        {
          context: createApolloContextErrorTitle('Exporting to template'),
          awaitRefetchQueries: true,
        }
      )
      .pipe(
        map((res) => this.handleActionResponse(res, 'Workflow was exported to template')),
        tap({
          next: () => this.exportInProgressSubject$.next(false),
          error: () => this.exportInProgressSubject$.next(false),
        }),
        catchError((err) => ApolloOperationResultResolver.handleError(err))
      );
  }

  private handleActionResponse(fetchResult: FetchResult, message: string): ApolloResult {
    const result = ApolloOperationResultResolver.resolve(fetchResult);
    if (result.type === ApolloResultType.SUCCESS) {
      this.notificationService.emit('success', message);
    }
    return result;
  }
}
