import { ApolloQueryResult, NetworkStatus } from '@apollo/client/core';
import { WatchQueryFetchPolicy } from '@apollo/client/core/watchQueryOptions';
import { createApolloContextErrorTitle, createApolloContextRetryable } from '@csirtmu-orion/microframework/graphql';
import * as Apollo from 'apollo-angular';
import { QueryRef } from 'apollo-angular';
import { BehaviorSubject, EMPTY, merge, Observable, of, Subject } from 'rxjs';
import { catchError, distinctUntilChanged, filter, map, shareReplay, switchMap, tap } from 'rxjs/operators';

export interface ApolloQueryServiceOptions {
  fetchPolicy?: WatchQueryFetchPolicy;
  pollInterval?: number;
  retryable?: boolean;
}

const DEFAULT_OPTIONS: ApolloQueryServiceOptions = {
  fetchPolicy: 'network-only',
  pollInterval: undefined,
  retryable: true,
};

export abstract class ApolloQueryService<Query, QueryVariables, Data> {
  queryRef?: QueryRef<Query, QueryVariables>;
  operationName: string;
  data$: Observable<Data>;
  isLoading$: Observable<boolean>;
  isRefetching$: Observable<boolean>;
  hasError$?: Observable<boolean>;

  protected loadingInternal$?: Observable<boolean>;
  protected hasErrorInternal$?: Observable<boolean>;
  protected isLoadingSubject$: Subject<boolean> = new Subject<boolean>();
  protected variables?: QueryVariables;
  protected initialized$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);

  private isInErrorState = false;
  private options: ApolloQueryServiceOptions;
  protected abstract toData(queryRes: ApolloQueryResult<Query>): Data;

  protected constructor(
    public getQuery: Apollo.Query<Query, QueryVariables>,
    operationName: string,
    options?: ApolloQueryServiceOptions
  ) {
    this.operationName = operationName;
    this.options = options ?? DEFAULT_OPTIONS;
    this.data$ = this.initialized$.pipe(
      shareReplay({ refCount: true }),
      switchMap(
        () =>
          this.queryRef?.valueChanges.pipe(
            tap((queryRes) => this.checkAndHandleQueryError(queryRes)),
            // ignoring all responses not containing data
            filter((queryRes) => queryRes.data !== undefined && queryRes.networkStatus === NetworkStatus.ready),
            map((queryRes) => this.toData(queryRes)),
            catchError(() => EMPTY)
          ) ?? EMPTY
      )
    );
    this.isRefetching$ = this.initialized$.pipe(
      switchMap(
        () =>
          this.queryRef?.valueChanges.pipe(
            map(
              (queryRes) =>
                (queryRes.networkStatus === NetworkStatus.refetch && queryRes.loading) ||
                queryRes.error !== undefined ||
                false
            ),
            catchError(() => of(false)),
            distinctUntilChanged()
          ) ?? EMPTY
      )
    );
    this.isLoading$ = this.initialized$.pipe(switchMap(() => this.loadingInternal$ ?? EMPTY));
    this.hasError$ = this.initialized$.pipe(switchMap(() => this.hasErrorInternal$ ?? EMPTY));
  }

  public recoverAfterError(): void {
    this.initialize();
  }

  protected abstract createVariables(...args: unknown[]): QueryVariables;

  protected updateVariables(...args: unknown[]): void {
    this.variables = this.createVariables(args);
  }

  protected initialize(): void {
    this.initQueryWatch();
    this.initLoading();
    this.initHasError();
    this.initialized$.next(true);
  }

  protected initQueryWatch(): void {
    this.queryRef = this.getQuery.watch(this.variables, {
      useInitialLoading: true,
      pollInterval: this.options?.pollInterval ?? DEFAULT_OPTIONS.pollInterval,
      fetchPolicy: this.options.fetchPolicy ?? DEFAULT_OPTIONS.fetchPolicy,
      nextFetchPolicy: 'network-only',
      errorPolicy: 'all',
      returnPartialData: true,
      notifyOnNetworkStatusChange: true, // to get all updates on loading state
      context: {
        ...createApolloContextErrorTitle(this.operationName),
        ...createApolloContextRetryable(
          this.options?.retryable !== undefined ? this.options?.retryable : DEFAULT_OPTIONS.retryable ?? false
        ),
      },
    });
  }

  protected updateQuery(): void {
    if (this.isInErrorState) {
      this.initialize();
    }
    this.isLoadingSubject$.next(true);
    if (this.queryRef && this.variables) {
      this.queryRef.setVariables(this.variables);
    }
  }

  protected initLoading(): void {
    if (this.queryRef) {
      const ref = this.queryRef;
      this.loadingInternal$ = merge(
        this.isLoadingSubject$.asObservable(),
        ref.valueChanges.pipe(
          filter((queryRes) => queryRes.networkStatus !== NetworkStatus.refetch),
          // we ignore refetch loadings before it causes bad ux.
          map((queryRes) => queryRes.loading || queryRes.error !== undefined),
          catchError(() => of(false))
        )
      ).pipe(distinctUntilChanged());
    }
  }

  protected initHasError(): void {
    if (this.queryRef) {
      const ref = this.queryRef;
      this.hasErrorInternal$ = ref.valueChanges.pipe(
        map((queryRes) => this.hasError(queryRes)),
        catchError(() => of(true)),
        distinctUntilChanged(),
        tap((hasError) => (this.isInErrorState = hasError))
      );
    }
  }

  protected checkAndHandleQueryError(queryRes: ApolloQueryResult<Query>): void {
    if (this.hasError(queryRes)) {
      this.isLoadingSubject$.next(false);
    }
  }

  protected hasError(queryRes: ApolloQueryResult<Query>): boolean {
    return (
      queryRes.error !== undefined || queryRes.errors !== undefined || queryRes.networkStatus === NetworkStatus.error
    );
  }
}
