import { TestBed, waitForAsync } from '@angular/core/testing';
import { take } from 'rxjs/operators';
import { FilterService } from './filter.service';

interface TestFilter {
  filterValue?: string;
  filterArray?: string[];
  filterNestedValue?: { a: string; b: number };
}

class FilterServiceTestImpl extends FilterService<TestFilter> {
  protected createEmptyFilter(): TestFilter {
    return { filterArray: undefined, filterValue: undefined, filterNestedValue: undefined };
  }
}

describe('FilterService', () => {
  let service: FilterServiceTestImpl;

  beforeEach(waitForAsync(() => {
    TestBed.configureTestingModule({
      providers: [FilterServiceTestImpl],
    });
    service = TestBed.inject(FilterServiceTestImpl);
  }));

  it('should create', () => {
    expect(service).toBeTruthy();
  });

  it('should only emit if set filter is different from prev', (done) => {
    const filter1 = { ...service.EMPTY_FILTER, filterValue: 'hello' };
    const filter2 = { ...filter1, filterArray: ['hello', 'world'] };
    const filter3 = { ...filter2, filterNestedValue: { a: '123', b: 999 } };
    const expectedFilters: TestFilter[] = [{ ...service.EMPTY_FILTER }, filter1, filter2, filter3];
    let resultIndex = 0;
    const expectedEmissions = expectedFilters.length;
    service.filter$.pipe(take(expectedEmissions)).subscribe({
      next: (filter) => {
        expect(Object.assign({}, filter)).toEqual(Object.assign({}, expectedFilters[resultIndex]));
        resultIndex++;
        if (expectedEmissions === resultIndex) {
          done();
        }
      },
      error: () => fail(),
    });
    service.set({ ...filter1 });
    service.submit();
    service.set({ ...filter2 });
    service.submit();
    service.set({ ...filter2 }); // emitting the same filter two times
    service.submit();
    service.set({ ...filter3 });
    service.submit();
  });

  it('should only emit if partially submitted filter is different from prev', (done) => {
    const updatedValue1 = 'hello';
    const updatedValue2 = ['hello', 'world'];
    const updatedValue3 = { a: '123', b: 999 };
    const updatedValue4 = { a: 'xxx', b: 888 };
    const filter1 = { ...service.EMPTY_FILTER, filterValue: updatedValue1 };
    const filter2 = { ...filter1, filterArray: updatedValue2 };
    const filter3 = { ...filter2, filterNestedValue: updatedValue3 };
    const filter4 = { ...filter3, filterNestedValue: updatedValue4 };
    const expectedFilters: TestFilter[] = [{ ...service.EMPTY_FILTER }, filter1, filter2, filter3, filter4];
    let resultIndex = 0;
    const expectedEmissions = expectedFilters.length;
    service.filter$.pipe(take(expectedEmissions)).subscribe({
      next: (filter) => {
        expect(Object.assign({}, filter)).toEqual(Object.assign({}, expectedFilters[resultIndex]));
        resultIndex++;
        if (expectedEmissions === resultIndex) {
          done();
        }
      },
      error: () => fail(),
    });
    service.updateAndSubmit({ filterValue: updatedValue1 });
    service.updateAndSubmit({ filterArray: updatedValue2 });
    service.updateAndSubmit({ filterArray: updatedValue2 }); // emitting the same filter two times
    service.updateAndSubmit({ filterNestedValue: updatedValue3 });
    service.updateAndSubmit({ filterNestedValue: updatedValue4 });
  });

  it('should emit empty filter after clear', (done) => {
    const filter1 = { ...service.EMPTY_FILTER, filterValue: 'hello' };
    const expectedFilters: TestFilter[] = [{ ...service.EMPTY_FILTER }, filter1, { ...service.EMPTY_FILTER }, filter1];
    let resultIndex = 0;
    const expectedEmissions = expectedFilters.length;
    service.filter$.pipe(take(expectedEmissions)).subscribe({
      next: (filter) => {
        expect(Object.assign({}, filter)).toEqual(Object.assign({}, expectedFilters[resultIndex]));
        resultIndex++;
        if (expectedEmissions === resultIndex) {
          done();
        }
      },
      error: () => fail(),
    });
    service.set({ ...filter1 });
    service.submit();
    service.clear();
    service.set({ ...filter1 });
    service.submit();
  });
});
