import { Injectable } from '@angular/core';
import { TestBed, waitForAsync } from '@angular/core/testing';
import { ActivatedRoute, convertToParamMap, Data, ParamMap, Router } from '@angular/router';
import { Observable, of } from 'rxjs';
import { skip, take } from 'rxjs/operators';
import { QueryParamFilterService } from './query-param-filter.service';

interface TestFilter {
  filterStr?: string;
  filterNumber?: number;
}

@Injectable()
class FilterServiceTestImpl extends QueryParamFilterService<TestFilter> {
  constructor(protected override router: Router, protected override activatedRoute: ActivatedRoute) {
    super(router, activatedRoute);
  }
  protected createEmptyFilter(): TestFilter {
    return { filterStr: undefined, filterNumber: undefined };
  }

  protected createFilterFromRoute(queryParamMap: ParamMap): Observable<TestFilter> {
    const str = queryParamMap.get('filterStr') ?? undefined;
    const num = queryParamMap.get('filterNumber') ?? undefined;
    return of({ filterStr: str, filterNumber: num ? Number.parseInt(num) : undefined });
  }
}

describe('QueryParamFilterService', () => {
  const INIT_FILTER_STR = 'initFilterValue';
  const INIT_FILTER_NUM = 123;
  let service: FilterServiceTestImpl;
  let routerSpy: jasmine.SpyObj<Router>;
  let mockQueryParamMap: ParamMap;
  let mockParamMap: ParamMap;
  let mockData: Data;

  beforeEach(waitForAsync(() => {
    mockQueryParamMap = convertToParamMap({
      filterStr: INIT_FILTER_STR,
      filterNumber: INIT_FILTER_NUM,
    });
    routerSpy = jasmine.createSpyObj('Router', ['navigate']);
    TestBed.configureTestingModule({
      providers: [
        FilterServiceTestImpl,
        { provide: Router, useValue: routerSpy },
        {
          provide: ActivatedRoute,
          useValue: {
            queryParamMap: of(mockQueryParamMap),
            paramMap: of(mockParamMap),
            data: of(mockData),
          },
        },
      ],
    });
    service = TestBed.inject(FilterServiceTestImpl);
  }));

  it('should create', () => {
    expect(service).toBeTruthy();
  });

  it('should init from route params', (done) => {
    const expectedFilter: TestFilter = { filterStr: INIT_FILTER_STR, filterNumber: INIT_FILTER_NUM };
    routerSpy.navigate.and.returnValue(Promise.resolve(true));
    service.filter$.pipe(take(1)).subscribe({
      next: (pagination) => {
        expect(Object.assign({}, pagination)).toEqual(Object.assign({}, expectedFilter));
        done();
      },
      error: () => fail(),
    });
  });

  it('should update route params if set is different from the previous', (done) => {
    routerSpy.navigate.and.returnValue(Promise.resolve(true));
    const filter1: TestFilter = {
      filterStr: 'someFilter1',
      filterNumber: 999,
    };
    const filter2: TestFilter = {
      filterStr: 'someFilter2',
      filterNumber: 888,
    };
    const filter3: TestFilter = {
      filterStr: 'someFilter3',
      filterNumber: 777,
    };

    const expectedFilters = [filter1, filter2, filter3, filter2];
    const expectedResultsCount = expectedFilters.length;
    let resultIndex = 0;

    // we skip the initial filter
    service.filter$.pipe(skip(1), take(expectedResultsCount)).subscribe({
      next: (filter) => {
        // we need to test it in timeout because the service navigates in timeout with zero
        // delay to avoid bugs in router updates
        setTimeout(() => {
          expect(Object.assign({}, filter)).toEqual(Object.assign({}, expectedFilters[resultIndex]));
          expect(routerSpy.navigate).toHaveBeenCalledTimes(resultIndex + 2); // first time for init, second time for update
          resultIndex++;
          if (resultIndex === expectedResultsCount) {
            done();
          }
        }, 0);
      },
      error: () => fail(),
    });
    service.set(filter1);
    service.submit();
    service.set(filter2);
    service.submit();
    service.set(filter2); // should not emit
    service.submit();
    service.set(filter3);
    service.submit();
    service.set(filter2);
    service.submit();
  });

  it('should clear route params', (done) => {
    routerSpy.navigate.and.returnValue(Promise.resolve(true));
    const filter1: TestFilter = {
      filterStr: 'someFilter1',
      filterNumber: 999,
    };
    const filter2: TestFilter = {
      filterStr: 'someFilter2',
      filterNumber: 888,
    };
    const expectedFilters = [filter1, service.EMPTY_FILTER, filter2];
    const expectedResultsCount = expectedFilters.length;
    let resultIndex = 0;

    // we skip the initial filter
    service.filter$.pipe(skip(1), take(expectedResultsCount)).subscribe({
      next: (filter) => {
        // we need to test it in timeout because the service navigates in timeout with zero
        // delay to avoid bugs in router updates
        setTimeout(() => {
          expect(Object.assign({}, filter)).toEqual(Object.assign({}, expectedFilters[resultIndex]));
          expect(routerSpy.navigate).toHaveBeenCalledTimes(resultIndex + 2); // first time for init, second time for update
          resultIndex++;
          if (resultIndex === expectedResultsCount) {
            done();
          }
        }, 0);
      },
      error: () => fail(),
    });
    service.set(filter1);
    service.submit();
    service.clear();
    service.set(filter2);
    service.submit();
  });

  it('should partially update route params if submitted filter is different from the previous', (done) => {
    routerSpy.navigate.and.returnValue(Promise.resolve(true));
    const updatedValue1 = 'filterValue1';
    const updatedValue2 = 555;
    const updatedValue3 = 'filterValue2';
    const initFilter: TestFilter = { filterStr: INIT_FILTER_STR, filterNumber: INIT_FILTER_NUM };
    const filter1: TestFilter = { filterStr: updatedValue1, filterNumber: initFilter.filterNumber };
    const filter2: TestFilter = {
      filterStr: filter1.filterStr,
      filterNumber: updatedValue2,
    };
    const filter3: TestFilter = {
      filterStr: updatedValue3,
      filterNumber: filter2.filterNumber,
    };

    const expectedFilters = [filter1, filter2, filter3];
    const expectedResultsCount = expectedFilters.length;
    let resultIndex = 0;

    // we skip the init emission
    service.filter$.pipe(skip(1), take(expectedResultsCount)).subscribe({
      next: (filter) => {
        // we need to test it in timeout because the service navigates in timeout with zero
        // delay to avoid bugs in router updates
        setTimeout(() => {
          expect(Object.assign({}, filter)).toEqual(Object.assign({}, expectedFilters[resultIndex]));
          expect(routerSpy.navigate).toHaveBeenCalledTimes(resultIndex + 2); // one for skipped init, second for the actual emission
          resultIndex++;
          if (resultIndex === expectedResultsCount) {
            done();
          }
        }, 0);
      },
      error: () => fail(),
    });
    service.updateAndSubmit({ filterStr: updatedValue1 });
    service.updateAndSubmit({ filterNumber: updatedValue2 });
    service.updateAndSubmit({ filterNumber: updatedValue2 }); // should not emit
    service.updateAndSubmit({ filterStr: updatedValue3 });
  });
});
