import { ActivatedRoute, Data, ParamMap, Params, QueryParamsHandling, Router } from '@angular/router';
import { Optional } from '@sentinel/common/utils';
import equal from 'fast-deep-equal/es6';
import { combineLatest, merge, Observable } from 'rxjs';
import { concatMap, distinctUntilChanged, shareReplay, skip, tap } from 'rxjs/operators';
import { FilterService } from './filter.service';

export abstract class QueryParamFilterService<Filter> extends FilterService<Filter> {
  queryParamsHandling: QueryParamsHandling = 'merge';
  replaceUrl = true;
  override filter$: Observable<Filter>;

  protected constructor(protected router: Router, protected activatedRoute: ActivatedRoute) {
    super();
    this.filter$ = this.createFilterObservable();
  }
  protected abstract createFilterFromRoute(queryParamMap: ParamMap, paramMap: ParamMap, data: Data): Observable<Filter>;

  protected toQueryParams(filter: Filter): Params {
    const params: Params = {};
    const entries = Object.entries(filter);
    entries.forEach((entry) => (params[entry[0]] = this.slugify(entry[1])));
    return params;
  }

  protected updateQueryParams(filter: Filter): void {
    setTimeout(() => {
      this.router
        .navigate([], {
          queryParams: this.toQueryParams(filter),
          queryParamsHandling: this.queryParamsHandling,
          replaceUrl: this.replaceUrl,
        })
        .catch(() => {
          console.error('Failed to change filter query params');
        });
    }, 0);
  }

  protected slugify(value: string | undefined | null): Optional<string> {
    if (value === null || value === undefined) {
      return undefined;
    } else {
      return encodeURIComponent(value);
    }
  }

  protected deslugify(value: string | undefined | null): Optional<string> {
    if (value === null || value === undefined) {
      return undefined;
    } else {
      return decodeURIComponent(value);
    }
  }

  protected deslugifyTimestamp(value: string | undefined | null): Optional<Date> {
    if (value === null || value === undefined) {
      return undefined;
    } else {
      const deslugified = this.deslugify(value);
      if (deslugified) {
        const timestamp = Number.parseInt(deslugified);
        if (!Number.isNaN(timestamp)) {
          return new Date(timestamp);
        }
      }
      return undefined;
    }
  }

  private createFilterObservable(): Observable<Filter> {
    const fromRoute$ = this.fromRoute().pipe(tap((filter) => this.updateAndSubmit(filter)));
    const fromSubject$ = this.createFilterSubjectObservable();
    return merge(fromSubject$, fromRoute$).pipe(
      distinctUntilChanged((prev, curr) => equal(prev, curr)),
      tap((filter) => this.updateQueryParams(filter)),
      shareReplay({ refCount: true })
    );
  }

  private createFilterSubjectObservable(): Observable<Filter> {
    // skip subject init from route
    return this.filterSubject$.asObservable().pipe(shareReplay({ refCount: true }), skip(1));
  }

  private fromRoute(): Observable<Filter> {
    return combineLatest([
      this.activatedRoute.queryParamMap,
      this.activatedRoute.paramMap,
      this.activatedRoute.data,
    ]).pipe(
      concatMap((routeAsyncAttrs) =>
        this.createFilterFromRoute(routeAsyncAttrs[0], routeAsyncAttrs[1], routeAsyncAttrs[2])
      )
    );
  }
}
