import { CollectionViewer } from '@angular/cdk/collections';
import { MatTreeNestedDataSource } from '@angular/material/tree';
import { Observable, Subscription } from 'rxjs';
import { filter, map, mergeMap, switchMap, take } from 'rxjs/operators';
import { OrionNestedTreeControl } from './orion-nested-tree-control';
import { TreeNode } from './tree-node';
import { OrionTreeNodeManagerService } from './orion-tree-node-manager.service';
import { TreeRootEventSource, TreeRootEventType } from './tree-root-event';

export class OrionNestedTreeDataSource<NodeData, NodeDataId = NodeData> extends MatTreeNestedDataSource<
  TreeNode<NodeData>
> {
  private expandSub?: Subscription;
  private loadMoreSub?: Subscription;
  private loadMoreRootsSub?: Subscription;
  private treeControl: OrionNestedTreeControl<NodeData, NodeDataId>;
  private treeNodeManager: OrionTreeNodeManagerService<NodeData>;

  isLoadingRoots$: Observable<boolean>;
  isLoadingMoreRoots$: Observable<boolean>;
  hasMoreRoots$: Observable<boolean>;
  loadRootsHasError$: Observable<boolean>;
  loadMoreRootsHasError$: Observable<boolean>;

  constructor(
    treeControl: OrionNestedTreeControl<NodeData, NodeDataId>,
    treeNodeManager: OrionTreeNodeManagerService<NodeData>
  ) {
    super();
    this.treeControl = treeControl;
    this.treeNodeManager = treeNodeManager;
    this.isLoadingRoots$ = this.treeNodeManager.rootEvent$.pipe(
      filter((event) => event.type === TreeRootEventType.LOADING && event.source === TreeRootEventSource.INIT_LOAD),
      map((event) => event.value)
    );
    this.isLoadingMoreRoots$ = this.treeNodeManager.rootEvent$.pipe(
      filter((event) => event.type === TreeRootEventType.LOADING && event.source === TreeRootEventSource.LOAD_MORE),
      map((event) => event.value)
    );
    this.hasMoreRoots$ = this.treeNodeManager.rootEvent$.pipe(
      filter((event) => event.type === TreeRootEventType.HAS_MORE),
      map((event) => event.value)
    );
    this.loadRootsHasError$ = this.treeNodeManager.rootEvent$.pipe(
      filter((event) => event.type === TreeRootEventType.ERROR && event.source === TreeRootEventSource.INIT_LOAD),
      map((event) => event.value)
    );
    this.loadMoreRootsHasError$ = this.treeNodeManager.rootEvent$.pipe(
      filter((event) => event.type === TreeRootEventType.ERROR && event.source === TreeRootEventSource.LOAD_MORE),
      map((event) => event.value)
    );
  }

  override connect(collectionViewer: CollectionViewer): Observable<TreeNode<NodeData>[]> {
    this.expandSub = this.treeControl.loadChildren$
      .pipe(mergeMap((node) => this.treeNodeManager.getChildren(node)))
      .subscribe();
    this.loadMoreSub = this.treeControl.loadMoreChildren$
      .pipe(mergeMap((node) => this.treeNodeManager.loadMoreChildren(node)))
      .subscribe();
    this.loadMoreRootsSub = this.treeControl.loadMoreRoots$
      .pipe(mergeMap(() => this.treeNodeManager.loadMoreRoots()))
      .subscribe();
    this.treeNodeManager.getRoots().pipe(take(1)).subscribe();
    return collectionViewer.viewChange.pipe(switchMap(() => this.treeNodeManager.rootNodes$));
  }

  override disconnect(): void {
    if (this.expandSub) {
      this.expandSub.unsubscribe();
    }
    if (this.loadMoreSub) {
      this.loadMoreSub.unsubscribe();
    }
    if (this.loadMoreRootsSub) {
      this.loadMoreRootsSub.unsubscribe();
    }
    super.disconnect();
  }
}
