import { BehaviorSubject, Observable } from 'rxjs';
import { catchError, map, skipWhile, tap } from 'rxjs/operators';
import { NodeEventSource } from './node-event-source';
import { RootNodes } from './root-nodes';
import { OrionTreeDataService } from './orion-tree-data.service';
import { TreeNode } from './tree-node';
import { TreeRootEvent, TreeRootEventSource, TreeRootEventType } from './tree-root-event';

export abstract class OrionTreeNodeManagerService<NodeData> {
  superiorEntityId?: string;
  private rootEventSubject$: BehaviorSubject<TreeRootEvent> = new BehaviorSubject<TreeRootEvent>(
    new TreeRootEvent(false, TreeRootEventType.LOADING, TreeRootEventSource.INIT_LOAD)
  );
  rootEvent$: Observable<TreeRootEvent> = this.rootEventSubject$.asObservable();
  private rootNodesSubject$: BehaviorSubject<TreeNode<NodeData>[]> = new BehaviorSubject<TreeNode<NodeData>[]>([]);
  rootNodes$: Observable<TreeNode<NodeData>[]> = this.rootNodesSubject$
    .asObservable()
    .pipe(skipWhile((roots) => roots.length === 0));

  private readonly BATCH_SIZE;
  private rootsOffset = 0;
  private totalRootsCount = 0;
  protected constructor(protected nodeService: OrionTreeDataService<NodeData>, batchSize = 50) {
    this.BATCH_SIZE = batchSize;
  }

  getRoots(): Observable<TreeNode<NodeData>[]> {
    this.handleBeforeLoadRoots(TreeRootEventSource.INIT_LOAD);
    const batchSize = Math.max(this.BATCH_SIZE, this.rootNodesSubject$.getValue().length);
    return this.nodeService.getRoots(batchSize, this.rootsOffset, this.superiorEntityId).pipe(
      map((roots) => this.mapToRootNodes(roots)),
      tap(
        (nodes) => this.handleLoadRootsSuccess(nodes, TreeRootEventSource.INIT_LOAD, batchSize),
        () => this.handleRootsError(TreeRootEventSource.INIT_LOAD)
      ),
      catchError(() => [])
    );
  }

  loadMoreRoots(): Observable<TreeNode<NodeData>[]> {
    this.handleBeforeLoadRoots(TreeRootEventSource.LOAD_MORE);
    return this.nodeService.getRoots(this.BATCH_SIZE, this.rootsOffset, this.superiorEntityId).pipe(
      map((roots) => this.mapToRootNodes(roots)),
      tap(
        (nodes) => this.handleLoadRootsSuccess(nodes, TreeRootEventSource.LOAD_MORE),
        () => this.handleRootsError(TreeRootEventSource.LOAD_MORE)
      ),
      catchError(() => [])
    );
  }

  getChildren(node: TreeNode<NodeData>): Observable<TreeNode<NodeData>[]> {
    const size = Math.max(this.BATCH_SIZE, node.getChildrenCount());
    this.beforeLoadChildrenRequest(node, NodeEventSource.EXPAND_NODE);
    return this.nodeService.getChildren(node, size, 0).pipe(
      tap(() => this.resolveHasMoreChildren(node, size, 0)),
      map((children) => this.mapToNodes(children)),
      tap(
        (children) => this.handleLoadChildrenSuccess(node, children, true),
        () => this.handleNodeError(node, NodeEventSource.EXPAND_NODE)
      ),
      catchError(() => [])
    );
  }

  loadMoreChildren(node: TreeNode<NodeData>): Observable<TreeNode<NodeData>[]> {
    const offset = node.getChildrenCount();
    this.beforeLoadChildrenRequest(node, NodeEventSource.LOAD_MODE);
    return this.nodeService.getChildren(node, this.BATCH_SIZE, offset).pipe(
      tap(() => this.resolveHasMoreChildren(node, this.BATCH_SIZE, offset)),
      map((children) => this.mapToNodes(children)),
      tap(
        (children) => this.handleLoadChildrenSuccess(node, children),
        () => this.handleNodeError(node, NodeEventSource.LOAD_MODE)
      ),
      catchError(() => [])
    );
  }

  protected abstract getTotalChildrenCount(parentNode: TreeNode<NodeData>): number;

  protected abstract createNodeInstance(data: NodeData): TreeNode<NodeData>;

  private handleBeforeLoadRoots(source: TreeRootEventSource) {
    this.rootEventSubject$.next(new TreeRootEvent(true, TreeRootEventType.LOADING, source));
    this.rootEventSubject$.next(new TreeRootEvent(false, TreeRootEventType.ERROR, source));
  }

  private handleLoadRootsSuccess(
    roots: TreeNode<NodeData>[],
    source: TreeRootEventSource,
    batchSize: number = this.BATCH_SIZE
  ): void {
    this.rootEventSubject$.next(new TreeRootEvent(false, TreeRootEventType.LOADING, source));
    this.resolveHasMoreRoots(batchSize, this.rootsOffset, source);
    this.incrementOffset(roots.length);
    if (source === TreeRootEventSource.INIT_LOAD) {
      this.setRoots(roots);
    } else {
      this.appendRoots(roots);
    }
  }

  private handleRootsError(source: TreeRootEventSource) {
    this.rootEventSubject$.next(new TreeRootEvent(true, TreeRootEventType.ERROR, source));
    this.rootEventSubject$.next(new TreeRootEvent(false, TreeRootEventType.LOADING, source));
  }

  private beforeLoadChildrenRequest(node: TreeNode<NodeData>, source: NodeEventSource): void {
    if (source === NodeEventSource.EXPAND_NODE) {
      node.toggleLoading();
      node.setLoadMoreError(false);
      node.setExpandError(false);
    } else {
      node.toggleLoading();
      node.setLoadMoreError(false);
    }
  }

  private handleLoadChildrenSuccess(
    parent: TreeNode<NodeData>,
    children: TreeNode<NodeData>[],
    isInitLoad = false
  ): void {
    parent.toggleLoading();
    if (isInitLoad) {
      parent.setChildren(children);
    } else {
      parent.appendChildren(children);
    }
  }

  private handleNodeError(node: TreeNode<NodeData>, source: NodeEventSource): void {
    if (source === NodeEventSource.EXPAND_NODE) {
      node.toggleLoading();
      node.setExpandError(true);
    } else {
      node.toggleLoading();
      node.setLoadMoreError(true);
    }
  }

  private mapToRootNodes(roots: RootNodes<NodeData>): TreeNode<NodeData>[] {
    this.totalRootsCount = roots.totalCount;
    return this.mapToNodes(roots.nodes, true);
  }

  private mapToNodes(data: NodeData[], isRoot = false): TreeNode<NodeData>[] {
    return data.map((datum) => {
      const node = this.createNodeInstance(datum);
      node.isExpandable = this.isExpandable(node);
      node.isRoot = isRoot;
      return node;
    });
  }

  private isExpandable(node: TreeNode<NodeData>): boolean {
    return this.getTotalChildrenCount(node) > 0;
  }

  private resolveHasMoreChildren(parentNode: TreeNode<NodeData>, size: number, offset: number): void {
    parentNode.hasMoreChildren = this.getTotalChildrenCount(parentNode) > size + offset;
  }

  private resolveHasMoreRoots(size: number, offset: number, source: TreeRootEventSource): void {
    const hasMoreRoots = this.totalRootsCount > size + offset;
    this.rootEventSubject$.next(new TreeRootEvent(hasMoreRoots, TreeRootEventType.HAS_MORE, source));
  }

  private setRoots(roots: TreeNode<NodeData>[]) {
    this.selectLast(roots);
    this.rootNodesSubject$.next(roots);
  }

  private appendRoots(toAppend: TreeNode<NodeData>[]) {
    const curr = this.rootNodesSubject$.getValue();
    curr.push(...toAppend);
    this.selectLast(curr);
    this.rootNodesSubject$.next(curr);
  }

  private selectLast(nodes: TreeNode<NodeData>[]): void {
    nodes.forEach((node, index) => (node.isLast = index === nodes.length - 1));
  }

  private incrementOffset(length: number): void {
    this.rootsOffset += length;
  }
}
