import { TestBed, waitForAsync } from '@angular/core/testing';
import { take } from 'rxjs/operators';
import { PaginationBaseInit } from './pagination-base-init';
import { INIT_PAGINATION_EVENT_BASE_TOKEN, PAGINATION_TYPE_TOKEN } from './pagination-token';
import { PaginationService } from './pagination.service';
import { PaginationType } from './types';
import { ICursorPaginationEvent, IPaginationEvent, IOffsetPaginationEvent } from '@sentinel/common/pagination';

describe('PaginationService', () => {
  let service: PaginationService;
  let paginationType: PaginationType | undefined = undefined;
  let initPagination: PaginationBaseInit | undefined = undefined;

  beforeEach(waitForAsync(() => {
    TestBed.configureTestingModule({
      providers: [
        PaginationService,
        { provide: PAGINATION_TYPE_TOKEN, useValue: paginationType },
        {
          provide: INIT_PAGINATION_EVENT_BASE_TOKEN,
          useValue: initPagination,
        },
      ],
    });
    service = TestBed.inject(PaginationService);
  }));

  describe('With default configuration', () => {
    it('should create', () => {
      expect(service).toBeTruthy();
    });

    it('should only emit if updated pagination is different from prev', (done) => {
      const expectedPaginations: IPaginationEvent[] = [
        { ...service.INIT_PAGINATION },
        { ...service.INIT_PAGINATION, size: 10 },
        { ...service.INIT_PAGINATION, size: 20 },
      ];
      let resultIndex = 0;
      const expectedEmissions = expectedPaginations.length;
      service.pagination$.pipe(take(expectedEmissions)).subscribe({
        next: (paginationEvent) => {
          expect(Object.assign({}, paginationEvent)).toEqual(Object.assign({}, expectedPaginations[resultIndex]));
          resultIndex++;
          if (expectedEmissions === resultIndex) {
            done();
          }
        },
        error: () => fail(),
      });
      const currPagination = service.get();
      const pagination1 = { ...currPagination, size: 10 };
      const pagination2 = { ...pagination1, size: 20 };
      service.update(pagination1);
      service.update(pagination1); // emitting the same pagination two times
      service.update(pagination2);
    });
  });

  describe('With injected offset-based pagination tokens', () => {
    const INIT_SIZE = 999;
    const INIT_SORT = 'testSort';
    const INIT_SORT_DIR = 'asc';

    beforeAll(() => {
      paginationType = 'offset';
      initPagination = new PaginationBaseInit(INIT_SIZE, INIT_SORT, INIT_SORT_DIR);
    });

    it('should instantiate with injected tokens', () => {
      const initPagination = service.get();
      const expectedPagination: IOffsetPaginationEvent = {
        size: INIT_SIZE,
        sort: INIT_SORT,
        sortDir: INIT_SORT_DIR,
        page: 0,
      };
      expect(Object.assign({}, initPagination)).toEqual(Object.assign({}, expectedPagination));
    });
  });

  describe('With injected cursor-based pagination tokens', () => {
    const INIT_SIZE = 999;
    const INIT_SORT = 'testSort';
    const INIT_SORT_DIR = 'asc';

    beforeAll(() => {
      paginationType = 'cursor';
      initPagination = new PaginationBaseInit(INIT_SIZE, INIT_SORT, INIT_SORT_DIR);
    });

    it('should instantiate with injected tokens', () => {
      const initPagination = service.get();
      const expectedPagination: ICursorPaginationEvent = {
        size: INIT_SIZE,
        sort: INIT_SORT,
        sortDir: INIT_SORT_DIR,
        cursor: null,
      };
      expect(Object.assign({}, initPagination)).toEqual(Object.assign({}, expectedPagination));
    });
  });
});
