import { Inject, Injectable, Optional } from '@angular/core';
import { Observable } from 'rxjs';
import { distinctUntilChanged, shareReplay } from 'rxjs/operators';
import { PaginationBase } from './pagination-base';
import { PaginationBaseInit } from './pagination-base-init';
import { INIT_PAGINATION_EVENT_BASE_TOKEN, PAGINATION_TYPE_TOKEN } from './pagination-token';
import { PaginationType } from './types';
import equal from 'fast-deep-equal/es6';
import {
  CursorRequestedPaginationManager,
  IPaginationEvent,
  isOffsetEvent,
  OffsetPaginationEvent,
  OffsetRequestedPaginationManager,
  RequestedPaginationManager,
  isCursorEvent,
  CursorPaginationEvent,
} from '@sentinel/common/pagination';

@Injectable()
export class PaginationService {
  readonly INIT_PAGINATION: IPaginationEvent;
  protected paginationManager: RequestedPaginationManager;
  pagination$: Observable<IPaginationEvent>;

  constructor(
    @Optional() @Inject(INIT_PAGINATION_EVENT_BASE_TOKEN) protected readonly initPaginationBase?: PaginationBaseInit,
    @Optional() @Inject(PAGINATION_TYPE_TOKEN) protected readonly paginationType?: PaginationType
  ) {
    if (!paginationType) {
      this.paginationType = 'cursor';
    }
    this.INIT_PAGINATION = this.createInitPagination(this.initPaginationBase);
    this.paginationManager = this.createManager(this.INIT_PAGINATION.size, this.INIT_PAGINATION);
    this.pagination$ = this.paginationManager.pagination$.asObservable().pipe(
      distinctUntilChanged((prev, curr) => equal(prev, curr)),
      shareReplay({ refCount: true })
    );
  }

  get(): IPaginationEvent {
    return this.paginationManager.pagination$.getValue();
  }

  update(pagination: IPaginationEvent): void {
    this.paginationManager.update(pagination);
  }

  resetPosition(): void {
    this.paginationManager.resetPosition();
  }

  resetSize(): void {
    this.paginationManager.resetSize();
  }

  protected createInitPagination(initPaginationBase?: PaginationBase): IPaginationEvent {
    const paginationBase: PaginationBase = initPaginationBase ?? { size: 10, sort: '', sortDir: '' };
    if (this.paginationType === 'offset') {
      return new OffsetPaginationEvent(0, paginationBase.size, paginationBase.sort, paginationBase.sortDir);
    } else if (this.paginationType === 'cursor') {
      return new CursorPaginationEvent(null, paginationBase.size, paginationBase.sort, paginationBase.sortDir);
    } else {
      throw new Error('PaginationType cannot be undefined');
    }
  }

  protected createManager(initSize: number, initPagination: IPaginationEvent): RequestedPaginationManager {
    if (this.paginationType === 'offset' && isOffsetEvent(initPagination)) {
      return new OffsetRequestedPaginationManager(initSize, initPagination);
    } else if (this.paginationType === 'cursor' && isCursorEvent(initPagination)) {
      return new CursorRequestedPaginationManager(initSize, initPagination);
    } else {
      throw new TypeError('Incompatible pagination type and initial pagination');
    }
  }
}
