import { TestBed, waitForAsync } from '@angular/core/testing';
import { ActivatedRoute, convertToParamMap, ParamMap, Router } from '@angular/router';
import { of } from 'rxjs';
import { skip, take } from 'rxjs/operators';
import { PaginationBaseInit } from './pagination-base-init';
import { INIT_PAGINATION_EVENT_BASE_TOKEN, PAGINATION_TYPE_TOKEN } from './pagination-token';
import { QueryParamPaginationService } from './query-param-pagination.service';
import { PaginationType } from './types';
import { ICursorPaginationEvent, IOffsetPaginationEvent } from '@sentinel/common/pagination';

describe('QueryParamPaginationService', () => {
  let paginationType: PaginationType;
  let initPagination: PaginationBaseInit;
  let mockQueryParamMap: ParamMap;
  let routerSpy: jasmine.SpyObj<Router>;
  let service: QueryParamPaginationService;

  beforeEach(waitForAsync(() => {
    routerSpy = jasmine.createSpyObj('Router', ['navigate']);

    TestBed.configureTestingModule({
      providers: [
        QueryParamPaginationService,
        { provide: PAGINATION_TYPE_TOKEN, useValue: paginationType },
        {
          provide: INIT_PAGINATION_EVENT_BASE_TOKEN,
          useValue: initPagination,
        },
        { provide: Router, useValue: routerSpy },
        {
          provide: ActivatedRoute,
          useValue: {
            queryParamMap: of(mockQueryParamMap),
          },
        },
      ],
    });
    service = TestBed.inject(QueryParamPaginationService);
  }));

  describe('With cursor-based configuration', () => {
    const initRoutePagination: ICursorPaginationEvent = {
      cursor: 'testCursor',
      size: 777,
      sort: 'paramSort',
      sortDir: 'desc',
    };

    beforeAll(() => {
      paginationType = 'cursor';
      new PaginationBaseInit(999, 'default', 'asc');
      mockQueryParamMap = convertToParamMap({
        cursor: initRoutePagination.cursor,
        size: initRoutePagination.size.toString(),
        sort: initRoutePagination.sort,
        sortDir: initRoutePagination.sortDir,
      });
    });

    it('should create', () => {
      expect(service).toBeTruthy();
    });

    it('should init from route params', (done) => {
      routerSpy.navigate.and.returnValue(Promise.resolve(true));
      service.pagination$.pipe(take(1)).subscribe({
        next: (pagination) => {
          expect(Object.assign({}, pagination)).toEqual(Object.assign({}, initRoutePagination));
          done();
        },
        error: () => fail(),
      });
    });

    it('should update route params', (done) => {
      routerSpy.navigate.and.returnValue(Promise.resolve(true));
      const nextPagination: ICursorPaginationEvent = {
        cursor: 'cursor123',
        size: 321,
        sort: 'sortByValue',
        sortDir: 'desc',
      };

      service.pagination$.pipe(skip(1), take(1)).subscribe({
        next: (pagination) => {
          // we need to test it in timeout because the service navigates in timeout with zero
          // delay to avoid bugs in router updates
          setTimeout(() => {
            expect(Object.assign({}, pagination)).toEqual(Object.assign({}, nextPagination));
            expect(routerSpy.navigate).toHaveBeenCalledTimes(2); // first time for init, second time for update
            done();
          }, 0);
        },
        error: () => fail(),
      });
      service.update(nextPagination);
    });

    it('should update route params multiple times', (done) => {
      routerSpy.navigate.and.returnValue(Promise.resolve(true));
      const firstPagination: ICursorPaginationEvent = {
        cursor: 'cursor123',
        size: 321,
        sort: 'sortByValue',
        sortDir: 'desc',
      };
      const secondPagination: ICursorPaginationEvent = {
        cursor: 'cursorXYZ',
        size: 1,
        sort: 'sortByValueXYZ',
        sortDir: 'asc',
      };

      const expectedPaginations = [firstPagination, secondPagination, firstPagination];
      const expectedResultsCount = expectedPaginations.length;
      let resultIndex = 0;

      // we skip the initial pagination
      service.pagination$.pipe(skip(1), take(expectedResultsCount)).subscribe({
        next: (pagination) => {
          // we need to test it in timeout because the service navigates in timeout with zero
          // delay to avoid bugs in router updates
          setTimeout(() => {
            expect(Object.assign({}, pagination)).toEqual(Object.assign({}, expectedPaginations[resultIndex]));
            expect(routerSpy.navigate).toHaveBeenCalledTimes(resultIndex + 2); // first time for init, second time for update
            resultIndex++;
            if (resultIndex === expectedResultsCount) {
              done();
            }
          }, 0);
        },
        error: () => fail(),
      });
      service.update(firstPagination);
      service.update(secondPagination);
      service.update(firstPagination);
    });
  });

  describe('With offset-based configuration', () => {
    const initRoutePagination: IOffsetPaginationEvent = {
      page: 789,
      size: 456,
      sort: 'paramSort',
      sortDir: 'desc',
    };

    beforeAll(() => {
      paginationType = 'offset';
      new PaginationBaseInit(999, 'defaultSort', 'asc');
      mockQueryParamMap = convertToParamMap({
        page: initRoutePagination.page.toString(),
        size: initRoutePagination.size.toString(),
        sort: initRoutePagination.sort,
        sortDir: initRoutePagination.sortDir,
      });
    });

    it('should create', () => {
      expect(service).toBeTruthy();
    });

    it('should init from route params', (done) => {
      routerSpy.navigate.and.returnValue(Promise.resolve(true));
      service.pagination$.pipe(take(1)).subscribe({
        next: (pagination) => {
          expect(Object.assign({}, pagination)).toEqual(Object.assign({}, initRoutePagination));
          done();
        },
        error: () => fail(),
      });
    });

    it('should update route params', (done) => {
      routerSpy.navigate.and.returnValue(Promise.resolve(true));
      const nextPagination: IOffsetPaginationEvent = {
        page: 777,
        size: 321,
        sort: 'sortByValue',
        sortDir: 'desc',
      };

      service.pagination$.pipe(skip(1), take(1)).subscribe({
        next: (pagination) => {
          // we need to test it in timeout because the service navigates in timeout with zero
          // delay to avoid bugs in router updates
          setTimeout(() => {
            expect(Object.assign({}, pagination)).toEqual(Object.assign({}, nextPagination));
            expect(routerSpy.navigate).toHaveBeenCalledTimes(2); // first time for init, second time for update
            done();
          }, 0);
        },
        error: () => fail(),
      });
      service.update(nextPagination);
    });

    it('should update route params multiple times', (done) => {
      routerSpy.navigate.and.returnValue(Promise.resolve(true));
      const firstPagination: IOffsetPaginationEvent = {
        page: 333,
        size: 321,
        sort: 'sortByValue',
        sortDir: 'desc',
      };
      const secondPagination: IOffsetPaginationEvent = {
        page: 555,
        size: 1,
        sort: 'sortByValueXYZ',
        sortDir: 'asc',
      };

      const expectedPaginations = [firstPagination, secondPagination, firstPagination];
      const expectedResultsCount = expectedPaginations.length;
      let resultIndex = 0;

      service.pagination$.pipe(skip(1), take(expectedResultsCount)).subscribe({
        next: (pagination) => {
          // we need to test it in timeout because the service navigates in timeout with zero
          // delay to avoid bugs in router updates
          setTimeout(() => {
            expect(Object.assign({}, pagination)).toEqual(Object.assign({}, expectedPaginations[resultIndex]));
            expect(routerSpy.navigate).toHaveBeenCalledTimes(resultIndex + 2); // first time for init, second time for update
            resultIndex++;
            if (resultIndex === expectedResultsCount) {
              done();
            }
          }, 0);
        },
        error: () => fail(),
      });
      service.update(firstPagination);
      service.update(secondPagination);
      service.update(firstPagination);
    });
  });

  describe('Cursor-based incorrect params fallback', () => {
    const INIT_SIZE = 999;
    const INIT_SORT = 'testSort';
    const INIT_SORT_DIR = 'asc';

    beforeAll(() => {
      paginationType = 'cursor';
      initPagination = new PaginationBaseInit(INIT_SIZE, INIT_SORT, INIT_SORT_DIR);
      mockQueryParamMap = convertToParamMap({
        unknownParam: 'hello-world',
        size: 'NaN',
        // sort missing,
      });
    });

    it('should create', () => {
      expect(service).toBeTruthy();
    });

    it('should fallback to default pagination on incorrect query params', (done) => {
      routerSpy.navigate.and.returnValue(Promise.resolve(true));
      const expectedPagination: ICursorPaginationEvent = {
        cursor: null,
        size: INIT_SIZE,
        sort: INIT_SORT,
        sortDir: INIT_SORT_DIR,
      };
      service.pagination$.pipe(take(1)).subscribe({
        next: (pagination) => {
          expect(Object.assign({}, pagination)).toEqual(Object.assign({}, expectedPagination));
          done();
        },
        error: () => fail(),
      });
    });
  });

  describe('Offset-based incorrect params fallback', () => {
    const INIT_SIZE = 999;
    const INIT_SORT = 'testSort';
    const INIT_SORT_DIR = 'asc';

    beforeAll(() => {
      paginationType = 'offset';
      initPagination = new PaginationBaseInit(INIT_SIZE, INIT_SORT, INIT_SORT_DIR);
      mockQueryParamMap = convertToParamMap({
        page: '11',
        size: 'NaN',
        sort: 'something',
        sortDir: 'asc',
      });
    });

    it('should create', () => {
      expect(service).toBeTruthy();
    });

    it('should fallback to default pagination on incorrect query params', (done) => {
      routerSpy.navigate.and.returnValue(Promise.resolve(true));
      const expectedPagination: IOffsetPaginationEvent = {
        page: 0,
        size: INIT_SIZE,
        sort: INIT_SORT,
        sortDir: INIT_SORT_DIR,
      };
      service.pagination$.pipe(take(1)).subscribe({
        next: (pagination) => {
          expect(Object.assign({}, pagination)).toEqual(Object.assign({}, expectedPagination));
          done();
        },
        error: () => fail(),
      });
    });
  });
});
