import { Inject, Injectable, Optional } from '@angular/core';
import { ActivatedRoute, ParamMap, Router } from '@angular/router';
import { Optional as Maybe, toQueryParams } from '@sentinel/common/utils';
import { Observable } from 'rxjs';
import { concatMap, map, shareReplay, take, tap } from 'rxjs/operators';
import { PaginationBase } from './pagination-base';
import { PaginationBaseInit } from './pagination-base-init';
import { INIT_PAGINATION_EVENT_BASE_TOKEN, PAGINATION_TYPE_TOKEN } from './pagination-token';
import { PaginationService } from './pagination.service';
import { PaginationType } from './types';
import { CursorPaginationEvent, IPaginationEvent, OffsetPaginationEvent } from '@sentinel/common/pagination';

@Injectable()
export class QueryParamPaginationService extends PaginationService {
  constructor(
    private router: Router,
    private activatedRoute: ActivatedRoute,
    @Optional()
    @Inject(INIT_PAGINATION_EVENT_BASE_TOKEN)
    protected override readonly initPaginationBase?: PaginationBaseInit,
    @Optional() @Inject(PAGINATION_TYPE_TOKEN) protected override readonly paginationType?: PaginationType
  ) {
    super(initPaginationBase, paginationType);
    if (!paginationType) {
      this.paginationType = 'cursor';
    }
    this.pagination$ = this.createPaginationObservable();
  }

  protected updateQueryParams(pagination: IPaginationEvent): void {
    setTimeout(() => {
      this.router
        .navigate([], {
          queryParams: toQueryParams<IPaginationEvent>(pagination),
          queryParamsHandling: 'merge',
          replaceUrl: true,
        })
        .catch(() => console.error('Failed to update pagination query params'));
    }, 0);
  }

  private createPaginationObservable(): Observable<IPaginationEvent> {
    return this.fromRoute().pipe(
      tap((pagination) => (this.paginationManager = this.createManager(this.INIT_PAGINATION.size, pagination))),
      concatMap(() => this.paginationManager.pagination$),
      tap((pagination) => this.updateQueryParams(pagination)),
      shareReplay({ refCount: true })
    );
  }

  private fromRoute(): Observable<IPaginationEvent> {
    return this.activatedRoute.queryParamMap.pipe(
      take(1),
      map((paramMap) => this.paginationFromParamMap(paramMap))
    );
  }

  private paginationFromParamMap(paramMap: ParamMap): IPaginationEvent {
    let pagination: Maybe<IPaginationEvent>;
    if (this.paginationType === 'offset') {
      pagination = QueryParamPaginationService.offsetPaginationFromParamMap(paramMap);
    } else if (this.paginationType === 'cursor') {
      pagination = QueryParamPaginationService.cursorPaginationFromParamMap(paramMap);
    }
    return pagination ?? this.INIT_PAGINATION;
  }

  private static offsetPaginationFromParamMap(paramMap: ParamMap): Maybe<IPaginationEvent> {
    const paginationBase = QueryParamPaginationService.paginationBaseFromParamMap(paramMap);
    const pageStr = paramMap.get('page');

    if (paginationBase && pageStr !== null) {
      const page = Number.parseInt(pageStr);
      if (Number.isInteger(page)) {
        return new OffsetPaginationEvent(page, paginationBase.size, paginationBase.sort, paginationBase.sortDir);
      }
    }
    return undefined;
  }

  private static cursorPaginationFromParamMap(paramMap: ParamMap): Maybe<IPaginationEvent> {
    const paginationBase = QueryParamPaginationService.paginationBaseFromParamMap(paramMap);
    const cursor = paramMap.get('cursor');
    if (paginationBase) {
      return new CursorPaginationEvent(cursor, paginationBase.size, paginationBase.sort, paginationBase.sortDir);
    }
    return undefined;
  }

  private static paginationBaseFromParamMap(paramMap: ParamMap): Maybe<PaginationBase> {
    const sizeStr = paramMap.get('size');
    const sort = paramMap.get('sort');
    const sortDir = paramMap.get('sortDir');
    if (sizeStr !== null) {
      const size = Number.parseInt(sizeStr);
      if (Number.isInteger(size)) {
        if (sort !== null && sortDir !== null) {
          return { size, sort, sortDir };
        } else {
          return { size, sort: '', sortDir: '' };
        }
      }
    }
    return undefined;
  }
}
