import { TestBed, waitForAsync } from '@angular/core/testing';
import { ActivatedRoute, convertToParamMap, ParamMap, Router } from '@angular/router';
import { of } from 'rxjs';
import { skip, take } from 'rxjs/operators';
import { INIT_SORT_TOKEN } from './order-token';
import { QueryParamSortService } from './query-param-sort.service';
import { SortEvent } from './sort-event';

describe('QueryParamSortService', () => {
  let service: QueryParamSortService;
  let initSortEvent: SortEvent | undefined = undefined;
  let mockQueryParamMap: ParamMap;
  let routerSpy: jasmine.SpyObj<Router>;

  beforeEach(waitForAsync(() => {
    routerSpy = jasmine.createSpyObj('Router', ['navigate']);

    TestBed.configureTestingModule({
      providers: [
        QueryParamSortService,
        {
          provide: INIT_SORT_TOKEN,
          useValue: initSortEvent,
        },
        { provide: Router, useValue: routerSpy },
        {
          provide: ActivatedRoute,
          useValue: {
            queryParamMap: of(mockQueryParamMap),
          },
        },
      ],
    });
    service = TestBed.inject(QueryParamSortService);
  }));

  describe('With default configuration', () => {
    beforeAll(() => {
      initSortEvent = undefined;
      mockQueryParamMap = convertToParamMap({});
    });

    it('should create', () => {
      expect(service).toBeTruthy();
    });

    it('should update route params multiple times', (done) => {
      routerSpy.navigate.and.returnValue(Promise.resolve(true));
      const sort1: SortEvent = {
        sort: 'sort1',
        sortDir: 'desc',
      };
      const sort2: SortEvent = {
        sort: 'sort2',
        sortDir: 'asc',
      };

      const expectedSorts = [sort1, sort2, sort1];
      const expectedResultsCount = expectedSorts.length;
      let resultIndex = 0;

      service.sort$.pipe(skip(1), take(expectedResultsCount)).subscribe({
        next: (sort) => {
          // we need to test it in timeout because the service navigates in timeout with zero
          // delay to avoid bugs in router updates
          setTimeout(() => {
            expect(Object.assign({}, sort)).toEqual(Object.assign({}, expectedSorts[resultIndex]));
            expect(routerSpy.navigate).toHaveBeenCalledTimes(resultIndex + 2); // first time for init, second time for update
            resultIndex++;
            if (resultIndex === expectedResultsCount) {
              done();
            }
          }, 0);
        },
        error: () => fail(),
      });
      service.update(sort1);
      service.update(sort2);
      service.update(sort1);
    });
  });

  describe('With init sort token', () => {
    const INIT_SORT = 'myInitSort';
    const INIT_SORT_DIR = 'asc';

    beforeAll(() => {
      initSortEvent = { sort: INIT_SORT, sortDir: INIT_SORT_DIR };
      mockQueryParamMap = convertToParamMap({
        sort: INIT_SORT,
        sortDir: INIT_SORT_DIR,
      });
    });

    it('should create', () => {
      expect(service).toBeTruthy();
    });

    it('should init from route params', (done) => {
      expect(service).toBeTruthy();

      routerSpy.navigate.and.returnValue(Promise.resolve(true));
      service.sort$.pipe(take(1)).subscribe({
        next: (sort) => {
          expect(Object.assign({}, sort)).toEqual(Object.assign({}, initSortEvent));
          done();
        },
        error: () => fail(),
      });
    });

    it('should not update route params is updated with same sort', (done) => {
      const sortEvent = { ...initSortEvent };
      routerSpy.navigate.and.returnValue(Promise.resolve(true));
      service.sort$.pipe(take(1)).subscribe({
        next: (sort) => {
          // we need to test it in timeout because the service navigates in timeout with zero
          // delay to avoid bugs in router updates
          setTimeout(() => {
            expect(Object.assign({}, sort)).toEqual(Object.assign({}, sortEvent));
            expect(routerSpy.navigate).toHaveBeenCalledTimes(1);
            done();
          }, 0);
        },
        error: () => fail(),
      });
      service.update(sortEvent);
      service.update(sortEvent);
      service.update(sortEvent);
    });
  });

  describe('Incorrect params fallback', () => {
    const INIT_SORT = 'fallbackSort';
    const INIT_SORT_DIR = 'asc';

    beforeAll(() => {
      initSortEvent = { sort: INIT_SORT, sortDir: INIT_SORT_DIR };
      mockQueryParamMap = convertToParamMap({
        sortWithTypo: 'something',
        sortDir: 'desc',
      });
    });

    it('should create', () => {
      expect(service).toBeTruthy();
    });

    it('should fallback to default sort on incorrect query params', (done) => {
      routerSpy.navigate.and.returnValue(Promise.resolve(true));
      service.sort$.pipe(take(1)).subscribe({
        next: (sort) => {
          expect(Object.assign({}, sort)).toEqual(Object.assign({}, initSortEvent));
          done();
        },
        error: () => fail(),
      });
    });
  });
});
