import { Inject, Injectable, Optional } from '@angular/core';
import { ActivatedRoute, ParamMap, Router } from '@angular/router';
import * as equal from 'fast-deep-equal/es6';
import { Observable } from 'rxjs';
import { concatMap, distinctUntilChanged, map, shareReplay, take, tap } from 'rxjs/operators';
import { INIT_SORT_TOKEN } from './order-token';
import { SortService } from './sort.service';
import { SortEvent } from './sort-event';
import { toQueryParams } from '@sentinel/common/utils';

@Injectable()
export class QueryParamSortService extends SortService {
  constructor(
    private router: Router,
    private activatedRoute: ActivatedRoute,
    @Optional() @Inject(INIT_SORT_TOKEN) protected override readonly initSort?: SortEvent
  ) {
    super(initSort);
    this.sort$ = this.createSortObservable();
  }

  protected updateQueryParams(sort: SortEvent): void {
    setTimeout(() => {
      this.router
        .navigate([], {
          queryParams: toQueryParams<SortEvent>(sort),
          queryParamsHandling: 'merge',
          replaceUrl: true,
        })
        .catch(() => console.error('Failed to update sort query params'));
    }, 0);
  }

  private createSortObservable(): Observable<SortEvent> {
    return this.fromRoute().pipe(
      take(1),
      concatMap(() => this.sortSubject$),
      distinctUntilChanged((prev, curr) => equal(prev, curr)),
      tap((sort) => this.updateQueryParams(sort)),
      shareReplay({ refCount: true })
    );
  }

  private fromRoute(): Observable<SortEvent> {
    return this.activatedRoute.queryParamMap.pipe(
      take(1),
      map((paramMap) => this.sortFromParamMap(paramMap))
    );
  }

  private sortFromParamMap(paramMap: ParamMap): SortEvent {
    const sort = paramMap.get('sort');
    const sortDirStr = paramMap.get('sortDir');
    if (sort !== null && sortDirStr !== null) {
      const sortDir = sortDirStr === 'asc' ? 'asc' : 'desc';
      return { sort, sortDir };
    }
    return this.INIT_SORT;
  }
}
