import { Inject, Injectable, Optional } from '@angular/core';
import { ActivatedRoute, ParamMap, Params, Router } from '@angular/router';
import { from, merge, Observable, of } from 'rxjs';
import { concatMap, distinctUntilChanged, mapTo, skip, take, tap } from 'rxjs/operators';
import { OrionTabService } from './orion-tab.service';
import { TABS_TOKEN } from './tabs-token';

@Injectable()
export class OrionQueryParamTabService extends OrionTabService {
  protected tabParams: Params[] = [];
  protected readonly TAB_PARAM_SELECTOR = 'tab';
  protected isInitialTabSelection = true;

  constructor(
    private router: Router,
    private activatedRoute: ActivatedRoute,
    @Optional() @Inject(TABS_TOKEN) public override readonly tabs?: string[]
  ) {
    super(tabs);
    this.initTabParams(this.tabs?.length ?? 0);
    const tabFromRoute$ = this.resolveTabIndexFromRoute();
    const tabFromSubject$ = this.selectedTabSubject$.asObservable().pipe(skip(1), distinctUntilChanged());
    this.selectedIndex$ = merge(tabFromRoute$, tabFromSubject$);
  }

  override reset(): Observable<boolean> {
    this.isInitialTabSelection = true;
    this.initTabParams(this.tabs?.length ?? 0);
    return this.select(0);
  }

  override select(nextIndex: number): Observable<boolean> {
    const prevIndex = this.selectedTabSubject$.getValue();
    return this.saveTabQueryParams(prevIndex, this.isInitialTabSelection)
      .pipe(concatMap(() => this.updateTabQueryParam(prevIndex, nextIndex)))
      .pipe(
        tap(() => {
          this.selectedTabSubject$.next(nextIndex);
          this.isInitialTabSelection = false;
        })
      );
  }

  protected resetCustomParams(prevTabIndex: number, nextParams: Params): Params {
    const prevParams = this.tabParams[prevTabIndex];
    const resetParams = this.createResetParams(prevParams, nextParams);
    return { ...nextParams, ...resetParams };
  }

  private initTabParams(tabCount: number): void {
    this.tabParams = [];
    for (let i = 0; i < tabCount; i++) {
      const tabName = this.getTabName(i);
      this.tabParams.push({ [this.TAB_PARAM_SELECTOR]: tabName });
    }
  }

  private createResetParams(prevParams: Params, nextParams: Params): Params {
    const resetParams: Params = {};
    const prevParamEntries = Object.entries(prevParams);
    const nextParamKeys = Object.keys(nextParams);
    prevParamEntries.forEach((entry) => {
      const prevParamKey = entry[0];
      if (this.isCustomParam(prevParamKey)) {
        if (nextParamKeys.includes(prevParamKey)) {
          // Router params are typed with any
          // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
          resetParams[prevParamKey] = nextParams[prevParamKey]; // override param value if previous and next params contain same keys
        } else {
          resetParams[prevParamKey] = undefined; // remove from query params by setting old param to undefined
        }
      }
    });
    return resetParams;
  }

  private isCustomParam(paramSelector: string): boolean {
    return paramSelector !== this.TAB_PARAM_SELECTOR;
  }

  private updateTabQueryParam(prevIndex: number, nextIndex: number): Observable<boolean> {
    let params = this.tabParams[nextIndex];
    if (params) {
      params = this.resetCustomParams(prevIndex, params);
    }
    return from(
      this.router.navigate([], {
        relativeTo: this.activatedRoute,
        queryParams: { ...params },
        queryParamsHandling: 'merge',
        replaceUrl: true,
      })
    );
  }

  private saveTabQueryParams(tabIndex: number, isInitial: boolean): Observable<Params> {
    if (isInitial) {
      const tabName = this.getTabName(tabIndex);
      const params = { [this.TAB_PARAM_SELECTOR]: tabName };
      this.tabParams[tabIndex] = params;
      return of(params);
    } else {
      return this.activatedRoute.queryParams.pipe(
        take(1),
        tap((queryParams) => (this.tabParams[tabIndex] = queryParams))
      );
    }
  }

  private getTabName(index: number): string {
    return this.tabs && this.tabs[index] ? this.tabs[index] : index.toString();
  }

  private resolveTabIndexFromRoute(): Observable<number> {
    return this.activatedRoute.queryParamMap.pipe(
      take(1),
      concatMap((queryParamMap) => this.tabIndexFromParamMap(queryParamMap)),
      tap((tabIndex) => this.selectedTabSubject$.next(tabIndex))
    );
  }

  private tabIndexFromParamMap(paramMap: ParamMap): Observable<number> {
    const tabParamValue = paramMap.get(this.TAB_PARAM_SELECTOR);
    const index = this.getTabIndexFromParamValue(tabParamValue);
    return this.select(index).pipe(take(1), mapTo(index)); // we need to select it as a side-ffect
  }

  private getTabIndexFromParamValue(paramValue: string | null): number {
    if (paramValue === null) {
      return 0;
    }
    if (this.tabs) {
      const index = this.tabs.findIndex((tab) => tab === paramValue);
      return index === -1 ? 0 : index;
    }
    return 0;
  }
}
