import {
  ChangeDetectionStrategy,
  Component,
  EventEmitter,
  Input,
  OnChanges,
  OnInit,
  Output,
  SimpleChanges,
} from '@angular/core';
import { Destroyable } from '@sentinel/common';
import { BehaviorSubject, combineLatest, Observable } from 'rxjs';
import { map, skip, takeUntil, tap } from 'rxjs/operators';
import { FilesService } from './files.service';

@Component({
  selector: 'orion-upload-files',
  templateUrl: './orion-upload-files.component.html',
  styleUrls: ['./orion-upload-files.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush,
  providers: [FilesService],
})
export class OrionUploadFilesComponent extends Destroyable() implements OnInit, OnChanges {
  @Input() cancellable = true;
  @Input() withActions = true;
  @Input() withCard = true;
  @Input() multiple = true;
  @Input() disabled?: boolean;
  @Output() upload: EventEmitter<File[]> = new EventEmitter<File[]>();
  @Output() cancel: EventEmitter<void> = new EventEmitter<void>();
  @Output() selectedFilesChange: EventEmitter<File[]> = new EventEmitter<File[]>();

  selectedFiles$?: Observable<File[]>;
  uploadDisabled$?: Observable<boolean>;
  disabled$: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);

  constructor(private filesService: FilesService) {
    super();
  }

  ngOnInit(): void {
    this.selectedFiles$ = this.filesService.filesToUpload$;
    const noSelectedFiles$ = this.selectedFiles$.pipe(map((files) => files.length === 0));
    this.uploadDisabled$ = combineLatest([this.disabled$, noSelectedFiles$]).pipe(
      map((conditions) => conditions[0] || conditions[1])
    );
    const selectionChangeEvent$ = this.selectedFiles$.pipe(
      takeUntil(this.destroy$),
      skip(1),
      tap((files) => this.selectedFilesChange.emit(files))
    );
    selectionChangeEvent$.subscribe();
  }

  ngOnChanges(changes: SimpleChanges): void {
    if ('disabled' in changes && this.disabled !== undefined) {
      this.disabled$.next(this.disabled);
    }
  }

  removeFile(fileIndex: number): void {
    this.filesService.removeFile(fileIndex);
  }

  selectFiles(files: File[]): void {
    if (this.multiple) {
      this.filesService.addFiles(files);
    } else if (files.length > 0) {
      this.filesService.clearFiles();
      this.filesService.addFiles([files[0]]);
    }
  }

  onUpload(): void {
    this.upload.emit(this.filesService.getCurrFiles());
  }

  onCancel(): void {
    this.filesService.clearFiles();
    this.cancel.emit();
  }
}
