import { EMPTY, of, Subject, BehaviorSubject, merge } from 'rxjs';
import { mergeMap, catchError, shareReplay, switchMap, tap, filter, map, distinctUntilChanged, take, concatMap } from 'rxjs/operators';
import { NetworkStatus } from '@apollo/client/core';
import { createApolloContextErrorTitle, createApolloContextRetryable, ApolloOperationResultResolver, QueryCursorPaginationDataResolver, CursorPaginationConverter, DmtOrderingConverter, CursorPaginationResolver } from '@csirtmu-orion/microframework/graphql';
import { PaginatedResource, CursorPaginationEvent, isCursorEvent } from '@sentinel/common/pagination';

class ApolloBaseResolver {
    constructor(getQuery) {
        this.getQuery = getQuery;
    }
    resolve(route) {
        const id = route.paramMap.get(this.getIdParamSelector());
        if (id !== null) {
            return this.getEntity(id);
        }
        else {
            return EMPTY;
        }
    }
    getEntity(id, ...args) {
        return this.getQuery.fetch(this.createVariables(id, ...args), { fetchPolicy: 'cache-first' }).pipe(mergeMap((res) => this.handleQueryResult(res)), catchError(() => EMPTY));
    }
    handleQueryResult(result) {
        const data = result === null || result === void 0 ? void 0 : result.data;
        if (data) {
            const entity = this.toFragment(data);
            if (entity) {
                return of(entity);
            }
        }
        return EMPTY;
    }
}

class ApolloResolver extends ApolloBaseResolver {
    constructor(getQuery) {
        super(getQuery);
        this.getQuery = getQuery;
    }
    getEntity(id) {
        return super.getEntity(id);
    }
}

const DEFAULT_OPTIONS = {
    fetchPolicy: 'network-only',
    pollInterval: undefined,
    retryable: true,
};
class ApolloQueryService {
    constructor(getQuery, operationName, options) {
        this.getQuery = getQuery;
        this.isLoadingSubject$ = new Subject();
        this.initialized$ = new BehaviorSubject(false);
        this.isInErrorState = false;
        this.operationName = operationName;
        this.options = options !== null && options !== void 0 ? options : DEFAULT_OPTIONS;
        this.data$ = this.initialized$.pipe(shareReplay({ refCount: true }), switchMap(() => {
            var _a, _b;
            return (_b = (_a = this.queryRef) === null || _a === void 0 ? void 0 : _a.valueChanges.pipe(tap((queryRes) => this.checkAndHandleQueryError(queryRes)), 
            // ignoring all responses not containing data
            filter((queryRes) => queryRes.data !== undefined && queryRes.networkStatus === NetworkStatus.ready), map((queryRes) => this.toData(queryRes)), catchError(() => EMPTY))) !== null && _b !== void 0 ? _b : EMPTY;
        }));
        this.isRefetching$ = this.initialized$.pipe(switchMap(() => {
            var _a, _b;
            return (_b = (_a = this.queryRef) === null || _a === void 0 ? void 0 : _a.valueChanges.pipe(map((queryRes) => (queryRes.networkStatus === NetworkStatus.refetch && queryRes.loading) ||
                queryRes.error !== undefined ||
                false), catchError(() => of(false)), distinctUntilChanged())) !== null && _b !== void 0 ? _b : EMPTY;
        }));
        this.isLoading$ = this.initialized$.pipe(switchMap(() => { var _a; return (_a = this.loadingInternal$) !== null && _a !== void 0 ? _a : EMPTY; }));
        this.hasError$ = this.initialized$.pipe(switchMap(() => { var _a; return (_a = this.hasErrorInternal$) !== null && _a !== void 0 ? _a : EMPTY; }));
    }
    recoverAfterError() {
        this.initialize();
    }
    updateVariables(...args) {
        this.variables = this.createVariables(args);
    }
    initialize() {
        this.initQueryWatch();
        this.initLoading();
        this.initHasError();
        this.initialized$.next(true);
    }
    initQueryWatch() {
        var _a, _b, _c, _d, _e, _f;
        this.queryRef = this.getQuery.watch(this.variables, {
            useInitialLoading: true,
            pollInterval: (_b = (_a = this.options) === null || _a === void 0 ? void 0 : _a.pollInterval) !== null && _b !== void 0 ? _b : DEFAULT_OPTIONS.pollInterval,
            fetchPolicy: (_c = this.options.fetchPolicy) !== null && _c !== void 0 ? _c : DEFAULT_OPTIONS.fetchPolicy,
            nextFetchPolicy: 'network-only',
            errorPolicy: 'all',
            returnPartialData: true,
            notifyOnNetworkStatusChange: true,
            context: Object.assign(Object.assign({}, createApolloContextErrorTitle(this.operationName)), createApolloContextRetryable(((_d = this.options) === null || _d === void 0 ? void 0 : _d.retryable) !== undefined ? (_e = this.options) === null || _e === void 0 ? void 0 : _e.retryable : ((_f = DEFAULT_OPTIONS.retryable) !== null && _f !== void 0 ? _f : false))),
        });
    }
    updateQuery() {
        if (this.isInErrorState) {
            this.initialize();
        }
        this.isLoadingSubject$.next(true);
        if (this.queryRef && this.variables) {
            this.queryRef.setVariables(this.variables);
        }
    }
    initLoading() {
        if (this.queryRef) {
            const ref = this.queryRef;
            this.loadingInternal$ = merge(this.isLoadingSubject$.asObservable(), ref.valueChanges.pipe(filter((queryRes) => queryRes.networkStatus !== NetworkStatus.refetch), 
            // we ignore refetch loadings before it causes bad ux.
            map((queryRes) => queryRes.loading || queryRes.error !== undefined), catchError(() => of(false)))).pipe(distinctUntilChanged());
        }
    }
    initHasError() {
        if (this.queryRef) {
            const ref = this.queryRef;
            this.hasErrorInternal$ = ref.valueChanges.pipe(map((queryRes) => this.hasError(queryRes)), catchError(() => of(true)), distinctUntilChanged(), tap((hasError) => (this.isInErrorState = hasError)));
        }
    }
    checkAndHandleQueryError(queryRes) {
        if (this.hasError(queryRes)) {
            this.isLoadingSubject$.next(false);
        }
    }
    hasError(queryRes) {
        return (queryRes.error !== undefined || queryRes.errors !== undefined || queryRes.networkStatus === NetworkStatus.error);
    }
}

class ApolloEntityBaseService extends ApolloQueryService {
    constructor(getQuery, operationName, options) {
        super(getQuery, operationName, options);
        this.getQuery = getQuery;
    }
    updateVariables(entityId, ...args) {
        this.variables = this.createVariables(entityId, ...args);
    }
    toData(queryRes) {
        return this.toFragment(queryRes.data);
    }
    updateVars(id, ...args) {
        this.updateVariables(this.resolveId(id), ...args);
        if (!this.initialized$.getValue()) {
            this.initialize();
        }
        this.updateQuery();
    }
    resolveId(id) {
        return typeof id === 'string' ? id : id.toString();
    }
}

class ApolloEntityService extends ApolloEntityBaseService {
    constructor(getQuery, operationName, options) {
        super(getQuery, operationName, options);
        this.getQuery = getQuery;
        this.initialized = false;
    }
}

class ApolloEntitiesBaseService extends ApolloQueryService {
    constructor(getQuery, operationName, options) {
        super(getQuery, operationName, options);
        this.getQuery = getQuery;
    }
    toData(queryRes) {
        return this.toFragments(queryRes.data);
    }
}

class ApolloEntitiesService extends ApolloEntitiesBaseService {
    constructor(getQuery, operationName, options) {
        super(getQuery, operationName, options);
        this.getQuery = getQuery;
        this.initialize();
    }
}

class ApolloBreadcrumbResolver extends ApolloBaseResolver {
    constructor(getQuery) {
        super(getQuery);
        this.getQuery = getQuery;
    }
    getEntity(id) {
        return super.getEntity(id);
    }
}

class ApolloEntityPatchService {
    constructor(patchMutation) {
        this.patchMutation = patchMutation;
    }
    patch(event) {
        const vars = this.createVariables(event);
        return this.patchMutation
            .mutate(vars, {
            refetchQueries: this.refetchQueriesNames,
            awaitRefetchQueries: true,
            context: createApolloContextErrorTitle('Updating entity'),
        })
            .pipe(take(1), map((result) => ApolloOperationResultResolver.resolve(result)), catchError((err) => ApolloOperationResultResolver.handleError(err)));
    }
}

class ApolloChildEntitiesService extends ApolloEntitiesBaseService {
    constructor(getQuery, operationName, options) {
        super(getQuery, operationName, options);
        this.getQuery = getQuery;
    }
    updateVariables(parentId, ...args) {
        this.variables = this.createVariables(parentId, ...args);
    }
    setParent(parentId, ...args) {
        this.updateVariables(this.resolveId(parentId), ...args);
        if (this.initialized$.getValue()) {
            this.updateQuery();
        }
        else {
            this.initialize();
        }
    }
    resolveId(id) {
        return typeof id === 'string' ? id : id.toString();
    }
}

class ApolloRoutedEntityService extends ApolloEntityBaseService {
    constructor(getQuery, activatedRoute, operationName, options) {
        super(getQuery, operationName, options);
        this.getQuery = getQuery;
        this.activatedRoute = activatedRoute;
        this.data$ = this.initVarsFromParamMap().pipe(switchMap(() => this.initialized$), shareReplay({ refCount: true }), concatMap(() => {
            var _a, _b;
            return (_b = (_a = this.queryRef) === null || _a === void 0 ? void 0 : _a.valueChanges.pipe(tap((queryRes) => this.checkAndHandleQueryError(queryRes)), // ignoring all responses not containing data
            filter((queryRes) => queryRes.data !== undefined && queryRes.networkStatus === NetworkStatus.ready), map((queryRes) => this.toData(queryRes)), catchError(() => EMPTY))) !== null && _b !== void 0 ? _b : EMPTY;
        }));
    }
    initVarsFromParamMap() {
        return this.activatedRoute.paramMap.pipe(map((data) => this.idFromParamMap(data)), map((id) => {
            if (id === null) {
                throw new Error('Failed to initialize id from paramMap');
            }
            else {
                this.updateVars(id);
            }
        }));
    }
}

class ApolloPaginatedBaseService extends ApolloQueryService {
    constructor(getQuery, operationName, options) {
        super(getQuery, operationName, options);
        this.getQuery = getQuery;
        this.lastPaginationEvent = this.createInitialPaginationEvent();
    }
    updateVariables(paginationEvent, ...args) {
        this.variables = this.createVariables(paginationEvent, args);
    }
    update(paginationEvent, ...args) {
        if (!this.initialized$.getValue()) {
            this.initialize();
        }
        this.lastPaginationEvent = paginationEvent;
        this.updateVariables(paginationEvent, args);
        this.updateQuery();
    }
    toData(result) {
        const entities = this.toFragments(result.data);
        const pagination = this.createPaginationFromQueryResult(result);
        return new PaginatedResource(entities, pagination);
    }
}

class ApolloPaginatedDmtService extends ApolloPaginatedBaseService {
    constructor(getQuery, operationName, options) {
        super(getQuery, operationName, options);
        this.getQuery = getQuery;
    }
    queryToPaginationData(query) {
        return QueryCursorPaginationDataResolver.resolve(this.getQueryResponse(query));
    }
    toFragments(query) {
        return this.getQueryResponse(query).data;
    }
    createInitialPaginationEvent() {
        return new CursorPaginationEvent(null, 0, '', '');
    }
    createPaginationVariables(paginationEvent) {
        if (isCursorEvent(paginationEvent)) {
            return CursorPaginationConverter.convert(paginationEvent);
        }
        else {
            throw TypeError('Unsupported type of pagination');
        }
    }
    createOrderVariables(sort, dir) {
        if (sort && sort !== '' && dir && dir !== '') {
            return DmtOrderingConverter.fromSingleSort(this.strToOrderable(sort), dir);
        }
        else {
            return {};
        }
    }
    createPaginationFromQueryResult(result) {
        if (isCursorEvent(this.lastPaginationEvent)) {
            const paginationData = this.queryToPaginationData(result.data);
            return CursorPaginationResolver.resolve(paginationData, this.lastPaginationEvent);
        }
        else {
            throw TypeError('Unsupported pagination');
        }
    }
}

class ApolloPaginatedFilterableBaseService extends ApolloPaginatedDmtService {
    constructor(getQuery, operationName, options) {
        super(getQuery, operationName, options);
        this.getQuery = getQuery;
    }
    update(paginationEvent, clientFilter, ...args) {
        if (!this.initialized$.getValue()) {
            this.initialize();
        }
        this.lastPaginationEvent = paginationEvent;
        this.updateVariables(paginationEvent, clientFilter, args);
        this.updateQuery();
    }
    updateVariables(paginationEvent, clientFilter, ...args) {
        this.variables = this.createVariables(paginationEvent, clientFilter, args);
    }
    createFilterVariables(clientFilter) {
        return { filter: this.filterConverter.convert(clientFilter) };
    }
}

class ApolloInterfaceChildEntitiesService extends ApolloChildEntitiesService {
    constructor(getQuery, operationName, options) {
        super(getQuery, operationName, options);
        this.getQuery = getQuery;
    }
    updateVariables(parentId, parentType) {
        super.updateVariables(parentId, parentType);
    }
    setParent(parentId, parentType) {
        super.setParent(parentId, parentType);
    }
}

class ApolloFilterableOrderableBaseService extends ApolloEntitiesBaseService {
    constructor(getQuery, operationName, options) {
        super(getQuery, operationName, options);
        this.getQuery = getQuery;
    }
    update(clientFilter, sort, ...args) {
        if (!this.initialized$.getValue()) {
            this.initialize();
        }
        this.updateVariables(clientFilter, sort, args);
        this.updateQuery();
    }
    updateVariables(clientFilter, sort, ...args) {
        this.variables = this.createVariables(clientFilter, sort, args);
    }
    createFilterVariables(clientFilter) {
        return { filter: this.filterConverter.convert(clientFilter) };
    }
    createOrderVariables(sortEvent) {
        if (sortEvent.sort && sortEvent.sort !== '' && sortEvent.sortDir) {
            return DmtOrderingConverter.fromSingleSort(this.strToOrderable(sortEvent.sort), sortEvent.sortDir);
        }
        else {
            return {};
        }
    }
}

class ApolloFilterableOrderableService extends ApolloFilterableOrderableBaseService {
    constructor(getQuery, operationName, options) {
        super(getQuery, operationName, options);
        this.getQuery = getQuery;
        this.filterConverter = this.createFilterConverter();
    }
}

class ApolloFilterableService extends ApolloEntitiesBaseService {
    constructor(getQuery, operationName, options) {
        super(getQuery, operationName, options);
        this.getQuery = getQuery;
        this.filterConverter = this.createFilterConverter();
    }
    update(clientFilter, ...args) {
        this.updateVariables(clientFilter, args);
        if (!this.initialized$.getValue()) {
            this.initialize();
        }
        this.updateQuery();
    }
    updateVariables(clientFilter, ...args) {
        this.variables = this.createVariables(clientFilter, args);
    }
    createFilterVariables(clientFilter) {
        return { filter: this.filterConverter.convert(clientFilter) };
    }
}

class ApolloPaginatedService extends ApolloPaginatedDmtService {
    constructor(getQuery, operationName, options) {
        super(getQuery, operationName, options);
        this.getQuery = getQuery;
    }
}

class ApolloPaginatedFilterableService extends ApolloPaginatedFilterableBaseService {
    constructor(getQuery, operationName, options) {
        super(getQuery, operationName, options);
        this.getQuery = getQuery;
        this.filterConverter = this.createFilterConverter();
    }
}

/**
 * Generated bundle index. Do not edit.
 */

export { ApolloBaseResolver, ApolloBreadcrumbResolver, ApolloChildEntitiesService, ApolloEntitiesBaseService, ApolloEntitiesService, ApolloEntityBaseService, ApolloEntityPatchService, ApolloEntityService, ApolloFilterableOrderableBaseService, ApolloFilterableOrderableService, ApolloFilterableService, ApolloInterfaceChildEntitiesService, ApolloPaginatedBaseService, ApolloPaginatedDmtService, ApolloPaginatedFilterableBaseService, ApolloPaginatedFilterableService, ApolloPaginatedService, ApolloQueryService, ApolloResolver, ApolloRoutedEntityService };
//# sourceMappingURL=csirtmu-orion-microframework-apollo-services.mjs.map
