import equal from 'fast-deep-equal/es6';
import { BehaviorSubject, merge, combineLatest } from 'rxjs';
import { distinctUntilChanged, tap, shareReplay, map, skip, concatMap } from 'rxjs/operators';

class FilterService {
    constructor() {
        this.EMPTY_FILTER = this.createEmptyFilter();
        this.filterLocalStateSubject$ = new BehaviorSubject(this.EMPTY_FILTER);
        this.filterSubject$ = new BehaviorSubject(this.filterLocalStateSubject$.getValue());
        this.filterLocalState$ = this.filterLocalStateSubject$.asObservable();
        this.filter$ = this.filterSubject$.asObservable().pipe(distinctUntilChanged((prev, curr) => equal(prev, curr)), tap((filter) => this.filterLocalStateSubject$.next(filter), shareReplay({ refCount: true })));
    }
    get() {
        return this.filterSubject$.getValue();
    }
    update(filter) {
        const curr = Object.assign({}, this.filterLocalStateSubject$.getValue());
        this.filterLocalStateSubject$.next(Object.assign(Object.assign({}, curr), filter));
    }
    updateAndSubmit(filter) {
        this.update(filter);
        this.submit();
    }
    submit() {
        this.filterSubject$.next(this.filterLocalStateSubject$.getValue());
    }
    clear() {
        this.filterLocalStateSubject$.next(Object.assign({}, this.EMPTY_FILTER));
        this.submit();
    }
    set(filter) {
        this.filterLocalStateSubject$.next(filter);
    }
}

class FilterChipsService {
    constructor(filterService) {
        this.filterService = filterService;
        this.chips$ = this.filterService.filter$.pipe(map((filter) => this.createChips(filter)));
    }
}

class QueryParamFilterService extends FilterService {
    constructor(router, activatedRoute) {
        super();
        this.router = router;
        this.activatedRoute = activatedRoute;
        this.queryParamsHandling = 'merge';
        this.replaceUrl = true;
        this.filter$ = this.createFilterObservable();
    }
    toQueryParams(filter) {
        const params = {};
        const entries = Object.entries(filter);
        entries.forEach((entry) => (params[entry[0]] = this.slugify(entry[1])));
        return params;
    }
    updateQueryParams(filter) {
        setTimeout(() => {
            this.router
                .navigate([], {
                queryParams: this.toQueryParams(filter),
                queryParamsHandling: this.queryParamsHandling,
                replaceUrl: this.replaceUrl,
            })
                .catch(() => {
                console.error('Failed to change filter query params');
            });
        }, 0);
    }
    slugify(value) {
        if (value === null || value === undefined) {
            return undefined;
        }
        else {
            return encodeURIComponent(value);
        }
    }
    deslugify(value) {
        if (value === null || value === undefined) {
            return undefined;
        }
        else {
            return decodeURIComponent(value);
        }
    }
    deslugifyTimestamp(value) {
        if (value === null || value === undefined) {
            return undefined;
        }
        else {
            const deslugified = this.deslugify(value);
            if (deslugified) {
                const timestamp = Number.parseInt(deslugified);
                if (!Number.isNaN(timestamp)) {
                    return new Date(timestamp);
                }
            }
            return undefined;
        }
    }
    createFilterObservable() {
        const fromRoute$ = this.fromRoute().pipe(tap((filter) => this.updateAndSubmit(filter)));
        const fromSubject$ = this.createFilterSubjectObservable();
        return merge(fromSubject$, fromRoute$).pipe(distinctUntilChanged((prev, curr) => equal(prev, curr)), tap((filter) => this.updateQueryParams(filter)), shareReplay({ refCount: true }));
    }
    createFilterSubjectObservable() {
        // skip subject init from route
        return this.filterSubject$.asObservable().pipe(shareReplay({ refCount: true }), skip(1));
    }
    fromRoute() {
        return combineLatest([
            this.activatedRoute.queryParamMap,
            this.activatedRoute.paramMap,
            this.activatedRoute.data,
        ]).pipe(concatMap((routeAsyncAttrs) => this.createFilterFromRoute(routeAsyncAttrs[0], routeAsyncAttrs[1], routeAsyncAttrs[2])));
    }
}

/**
 * Generated bundle index. Do not edit.
 */

export { FilterChipsService, FilterService, QueryParamFilterService };
//# sourceMappingURL=csirtmu-orion-microframework-filter-services.mjs.map
