import * as i0 from '@angular/core';
import { InjectionToken, Injectable, Optional, Inject } from '@angular/core';
import { distinctUntilChanged, shareReplay, tap, concatMap, take, map } from 'rxjs/operators';
import equal from 'fast-deep-equal/es6';
import { OffsetPaginationEvent, CursorPaginationEvent, isOffsetEvent, OffsetRequestedPaginationManager, isCursorEvent, CursorRequestedPaginationManager, OffsetPagination } from '@sentinel/common/pagination';
import { toQueryParams } from '@sentinel/common/utils';
import * as i1 from '@angular/router';

const INIT_PAGINATION_EVENT_BASE_TOKEN = new InjectionToken('INIT_PAGINATION_EVENT_TOKEN');
const PAGINATION_TYPE_TOKEN = new InjectionToken('PAGINATION_TYPE_TOKEN');

class PaginationBaseInit {
    constructor(size, sort, sortDir) {
        this.size = size;
        this.sort = sort;
        this.sortDir = sortDir;
    }
}

class PaginationService {
    constructor(initPaginationBase, paginationType) {
        this.initPaginationBase = initPaginationBase;
        this.paginationType = paginationType;
        if (!paginationType) {
            this.paginationType = 'cursor';
        }
        this.INIT_PAGINATION = this.createInitPagination(this.initPaginationBase);
        this.paginationManager = this.createManager(this.INIT_PAGINATION.size, this.INIT_PAGINATION);
        this.pagination$ = this.paginationManager.pagination$.asObservable().pipe(distinctUntilChanged((prev, curr) => equal(prev, curr)), shareReplay({ refCount: true }));
    }
    get() {
        return this.paginationManager.pagination$.getValue();
    }
    update(pagination) {
        this.paginationManager.update(pagination);
    }
    resetPosition() {
        this.paginationManager.resetPosition();
    }
    resetSize() {
        this.paginationManager.resetSize();
    }
    createInitPagination(initPaginationBase) {
        const paginationBase = initPaginationBase !== null && initPaginationBase !== void 0 ? initPaginationBase : { size: 10, sort: '', sortDir: '' };
        if (this.paginationType === 'offset') {
            return new OffsetPaginationEvent(0, paginationBase.size, paginationBase.sort, paginationBase.sortDir);
        }
        else if (this.paginationType === 'cursor') {
            return new CursorPaginationEvent(null, paginationBase.size, paginationBase.sort, paginationBase.sortDir);
        }
        else {
            throw new Error('PaginationType cannot be undefined');
        }
    }
    createManager(initSize, initPagination) {
        if (this.paginationType === 'offset' && isOffsetEvent(initPagination)) {
            return new OffsetRequestedPaginationManager(initSize, initPagination);
        }
        else if (this.paginationType === 'cursor' && isCursorEvent(initPagination)) {
            return new CursorRequestedPaginationManager(initSize, initPagination);
        }
        else {
            throw new TypeError('Incompatible pagination type and initial pagination');
        }
    }
}
PaginationService.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: PaginationService, deps: [{ token: INIT_PAGINATION_EVENT_BASE_TOKEN, optional: true }, { token: PAGINATION_TYPE_TOKEN, optional: true }], target: i0.ɵɵFactoryTarget.Injectable });
PaginationService.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: PaginationService });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: PaginationService, decorators: [{
            type: Injectable
        }], ctorParameters: function () {
        return [{ type: PaginationBaseInit, decorators: [{
                        type: Optional
                    }, {
                        type: Inject,
                        args: [INIT_PAGINATION_EVENT_BASE_TOKEN]
                    }] }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: Inject,
                        args: [PAGINATION_TYPE_TOKEN]
                    }] }];
    } });

class OffsetPaginationResolver {
    static resolve(queryData, paginationEvent) {
        const size = paginationEvent.size;
        return new OffsetPagination(paginationEvent.page, queryData.elementsCount, size, queryData.totalCount, Math.ceil(queryData.totalCount / size));
    }
}

class QueryParamPaginationService extends PaginationService {
    constructor(router, activatedRoute, initPaginationBase, paginationType) {
        super(initPaginationBase, paginationType);
        this.router = router;
        this.activatedRoute = activatedRoute;
        this.initPaginationBase = initPaginationBase;
        this.paginationType = paginationType;
        if (!paginationType) {
            this.paginationType = 'cursor';
        }
        this.pagination$ = this.createPaginationObservable();
    }
    updateQueryParams(pagination) {
        setTimeout(() => {
            this.router
                .navigate([], {
                queryParams: toQueryParams(pagination),
                queryParamsHandling: 'merge',
                replaceUrl: true,
            })
                .catch(() => console.error('Failed to update pagination query params'));
        }, 0);
    }
    createPaginationObservable() {
        return this.fromRoute().pipe(tap((pagination) => (this.paginationManager = this.createManager(this.INIT_PAGINATION.size, pagination))), concatMap(() => this.paginationManager.pagination$), tap((pagination) => this.updateQueryParams(pagination)), shareReplay({ refCount: true }));
    }
    fromRoute() {
        return this.activatedRoute.queryParamMap.pipe(take(1), map((paramMap) => this.paginationFromParamMap(paramMap)));
    }
    paginationFromParamMap(paramMap) {
        let pagination;
        if (this.paginationType === 'offset') {
            pagination = QueryParamPaginationService.offsetPaginationFromParamMap(paramMap);
        }
        else if (this.paginationType === 'cursor') {
            pagination = QueryParamPaginationService.cursorPaginationFromParamMap(paramMap);
        }
        return pagination !== null && pagination !== void 0 ? pagination : this.INIT_PAGINATION;
    }
    static offsetPaginationFromParamMap(paramMap) {
        const paginationBase = QueryParamPaginationService.paginationBaseFromParamMap(paramMap);
        const pageStr = paramMap.get('page');
        if (paginationBase && pageStr !== null) {
            const page = Number.parseInt(pageStr);
            if (Number.isInteger(page)) {
                return new OffsetPaginationEvent(page, paginationBase.size, paginationBase.sort, paginationBase.sortDir);
            }
        }
        return undefined;
    }
    static cursorPaginationFromParamMap(paramMap) {
        const paginationBase = QueryParamPaginationService.paginationBaseFromParamMap(paramMap);
        const cursor = paramMap.get('cursor');
        if (paginationBase) {
            return new CursorPaginationEvent(cursor, paginationBase.size, paginationBase.sort, paginationBase.sortDir);
        }
        return undefined;
    }
    static paginationBaseFromParamMap(paramMap) {
        const sizeStr = paramMap.get('size');
        const sort = paramMap.get('sort');
        const sortDir = paramMap.get('sortDir');
        if (sizeStr !== null) {
            const size = Number.parseInt(sizeStr);
            if (Number.isInteger(size)) {
                if (sort !== null && sortDir !== null) {
                    return { size, sort, sortDir };
                }
                else {
                    return { size, sort: '', sortDir: '' };
                }
            }
        }
        return undefined;
    }
}
QueryParamPaginationService.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: QueryParamPaginationService, deps: [{ token: i1.Router }, { token: i1.ActivatedRoute }, { token: INIT_PAGINATION_EVENT_BASE_TOKEN, optional: true }, { token: PAGINATION_TYPE_TOKEN, optional: true }], target: i0.ɵɵFactoryTarget.Injectable });
QueryParamPaginationService.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: QueryParamPaginationService });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: QueryParamPaginationService, decorators: [{
            type: Injectable
        }], ctorParameters: function () {
        return [{ type: i1.Router }, { type: i1.ActivatedRoute }, { type: PaginationBaseInit, decorators: [{
                        type: Optional
                    }, {
                        type: Inject,
                        args: [INIT_PAGINATION_EVENT_BASE_TOKEN]
                    }] }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: Inject,
                        args: [PAGINATION_TYPE_TOKEN]
                    }] }];
    } });

/**
 * Generated bundle index. Do not edit.
 */

export { INIT_PAGINATION_EVENT_BASE_TOKEN, OffsetPaginationResolver, PAGINATION_TYPE_TOKEN, PaginationBaseInit, PaginationService, QueryParamPaginationService };
//# sourceMappingURL=csirtmu-orion-microframework-pagination-services.mjs.map
