import { CdkVirtualScrollViewport } from '@angular/cdk/scrolling';
import * as i0 from '@angular/core';
import { EventEmitter, Directive, ViewChild, Input, Output } from '@angular/core';
import { isOffsetPagination, isCursorPagination, CursorPaginationEvent, OffsetPaginationEvent } from '@sentinel/common/pagination';
import { trackByIndex } from '@sentinel/common/utils';

/**
 * Abstract class providing the infinite scroll behaviour for SentinelPaginatedResource. Internally, it expects
 * CdkVirtualScrollViewport to be used in the component's template.
 */
class SentinelInfiniteScrollerDirective {
    constructor() {
        /**
         * Attribute name to sort by
         */
        this.sortByAttribute = 'title';
        /**
         * Sort direction
         */
        this.sortDir = 'asc';
        /**
         * Height (in pxs) of an item displayed in the viewport
         */
        this.viewPortItemHeight = 45;
        /**
         * Maximal height of the viewport
         */
        this.maxHeight = 600;
        /**
         * Function for tracking items in *ngFor. Defaults to tracking by index
         */
        this.trackByFn = trackByIndex;
        /**
         * Emits an event with info about the requested page. The parent component should get the requested data and pass it
         * through resource attribute
         */
        this.pageChange = new EventEmitter();
        /**
         * True if all resources are displayed (based on pagination and scroll state)
         */
        this.isEnd = false;
        /**
         * All items displayed (current and previous - cached). You should display all of those in your cdk virtual scroll component
         */
        this.scrollItems = [];
        this.cachedPages = [];
    }
    /**
     * Updates the state of the component, adds previously not seen resource to the cache.
     * @param changes
     */
    ngOnChanges(changes) {
        if ('resource' in changes && this.resource) {
            this.isLoading = false;
            if (!this.isAlreadyCached(this.resource.pagination)) {
                this.addPageToCache(this.resource.elements);
                this.flattenCachedPages();
            }
            else if (!this.isSameAsLastCache(this.resource)) {
                this.refreshCache(this.resource.pagination, this.resource.elements);
                this.flattenCachedPages();
            }
            this.isEnd = this.isLastPage(this.resource.pagination);
            this.calculateViewportHeight();
        }
    }
    /**
     * Reloads the page
     */
    reload() {
        this.isLoading = true;
        if (isOffsetPagination(this.resource.pagination)) {
            this.onOffsetPaginationReload(this.resource.pagination);
        }
        else if (isCursorPagination(this.resource.pagination)) {
            this.pageChange.emit(new CursorPaginationEvent(this.lastCursor, this.pageSize, this.sortByAttribute, this.sortDir));
        }
    }
    /**
     * Loads next page
     */
    nextPage(offset) {
        if (this.isEnd || this.isLoading || !this.resource) {
            return;
        }
        if (isOffsetPagination(this.resource.pagination)) {
            this.onNextOffsetPage(offset, this.resource.pagination);
        }
        else if (isCursorPagination(this.resource.pagination)) {
            this.onNextCursorPage(this.resource.pagination);
        }
    }
    isAlreadyCached(pagination) {
        if (isOffsetPagination(pagination)) {
            return pagination.page < this.cachedPages.length;
        }
        else if (isCursorPagination(pagination)) {
            return pagination.previous === this.lastCursor;
        }
        else {
            throw TypeError('Unsupported type of pagination: provided pagination is neither offset nor cursor pagination');
        }
    }
    addPageToCache(elements) {
        this.cachedPages.push(elements);
    }
    refreshCache(pagination, elements) {
        if (isOffsetPagination(pagination)) {
            this.cachedPages[pagination.page] = elements;
        }
        else if (isCursorPagination(pagination)) {
            this.cachedPages[this.lastCursor] = elements;
        }
    }
    isSameAsLastCache(output) {
        return this.cachedPages[this.cachedPages.length - 1].length === output.elements.length;
    }
    flattenCachedPages() {
        this.scrollItems = Array.prototype.concat.apply([], this.cachedPages);
    }
    calculateViewportHeight() {
        const itemsCount = this.isFirstPage(this.resource.pagination)
            ? this.resource.pagination.numberOfElements
            : this.resource.pagination.size;
        const itemsHeight = this.viewPortItemHeight * itemsCount;
        this.viewportHeight = Math.min(itemsHeight, this.maxHeight);
    }
    isFirstPage(pagination) {
        if (isOffsetPagination(pagination)) {
            return pagination.page === 0;
        }
        else if (isCursorPagination(pagination)) {
            return pagination.previous === null;
        }
        else {
            throw TypeError('Unsupported type of pagination: provided pagination is neither offset nor cursor pagination');
        }
    }
    isLastPage(pagination) {
        if (isOffsetPagination(pagination)) {
            return pagination.page === pagination.totalPages - 1;
        }
        else if (isCursorPagination(pagination)) {
            return pagination.next === null;
        }
        else {
            throw TypeError('Unsupported type of pagination: provided pagination is neither offset nor cursor pagination');
        }
    }
    onOffsetPaginationReload(pagination) {
        const page = pagination.page;
        const offsetPagination = new OffsetPaginationEvent(page, this.pageSize, this.sortByAttribute, this.sortDir);
        this.pageChange.emit(offsetPagination);
    }
    onNextOffsetPage(offset, pagination) {
        if (offset >= this.pageSize * pagination.page) {
            this.isLoading = true;
            const requestedPage = pagination.page + 1;
            const offsetPagination = new OffsetPaginationEvent(requestedPage, this.pageSize, this.sortByAttribute, this.sortDir);
            this.pageChange.emit(offsetPagination);
        }
    }
    onNextCursorPage(pagination) {
        this.isLoading = true;
        this.lastCursor = pagination.next;
        this.pageChange.emit(new CursorPaginationEvent(pagination.next, this.pageSize, this.sortByAttribute, this.sortDir));
    }
}
SentinelInfiniteScrollerDirective.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: SentinelInfiniteScrollerDirective, deps: [], target: i0.ɵɵFactoryTarget.Directive });
SentinelInfiniteScrollerDirective.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "14.0.0", version: "14.2.12", type: SentinelInfiniteScrollerDirective, inputs: { resource: "resource", sortByAttribute: "sortByAttribute", sortDir: "sortDir", hasError: "hasError", pageSize: "pageSize", isLoading: "isLoading", viewPortItemHeight: "viewPortItemHeight", maxHeight: "maxHeight", trackByFn: "trackByFn" }, outputs: { pageChange: "pageChange" }, viewQueries: [{ propertyName: "viewport", first: true, predicate: CdkVirtualScrollViewport, descendants: true }], usesOnChanges: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: SentinelInfiniteScrollerDirective, decorators: [{
            type: Directive
        }], propDecorators: { viewport: [{
                type: ViewChild,
                args: [CdkVirtualScrollViewport, { static: false }]
            }], resource: [{
                type: Input
            }], sortByAttribute: [{
                type: Input
            }], sortDir: [{
                type: Input
            }], hasError: [{
                type: Input
            }], pageSize: [{
                type: Input
            }], isLoading: [{
                type: Input
            }], viewPortItemHeight: [{
                type: Input
            }], maxHeight: [{
                type: Input
            }], trackByFn: [{
                type: Input
            }], pageChange: [{
                type: Output
            }] } });

/**
 * Public API of @sentinel/components/infinite-scroller
 */

/**
 * Generated bundle index. Do not edit.
 */

export { SentinelInfiniteScrollerDirective };
//# sourceMappingURL=sentinel-components-infinite-scroll.mjs.map
