import * as i0 from '@angular/core';
import { EventEmitter, Component, ChangeDetectionStrategy, Input, Output, ViewEncapsulation, ViewChild, HostListener, NgModule } from '@angular/core';
import * as i1 from '@angular/common';
import { CommonModule } from '@angular/common';
import * as d3 from 'd3';
import moment from 'moment-mini';

function flatAndSort(items) {
    const summary = items.flatMap((x) => x.summary);
    summary.sort((a, b) => b.value - a.value);
    return summary;
}
function range(count) {
    return count > 0 ? [...Array(count).keys()] : [];
}

function toYearItem(data) {
    return {
        date: new Date(data.year, 0, 1),
        count: data.count,
        children: Object.values(data.months).map((i) => {
            return {
                date: new Date(data.year, i.month, 1),
                count: i.count,
                summary: flatAndSort(i.items),
            };
        }),
    };
}
function toMonthItem(data) {
    return {
        date: new Date(data.year, data.month, 1),
        count: data.count,
        children: Object.values(data.items).map((i) => {
            return {
                date: new Date(i.date),
                count: i.summary.reduce((total, item) => total + item.value, 0),
                summary: i.summary,
            };
        }),
    };
}
function toMonth({ date }, pattern) {
    const now = moment();
    now.set('month', date.getMonth());
    return now.format(pattern);
}
function toMap(data) {
    const [start, end] = d3.extent(data.map((d) => new Date(d.date)));
    // Generate years.
    const years = d3.timeYears(moment(start).startOf('year'), moment(end).endOf('year'));
    // Generate months.
    const months = d3.timeMonths(moment().startOf('year'), moment().endOf('year'));
    // Prepare global template.
    const globalMap = {};
    for (const y of years) {
        globalMap[y.getFullYear()] = {
            count: 0,
            itemMaxCount: 0,
            year: y.getFullYear(),
            months: months.reduce((map, month) => {
                map[month.getMonth()] = {
                    year: y.getFullYear(),
                    month: month.getMonth(),
                    count: 0,
                    itemMaxCount: 0,
                    items: [],
                };
                return map;
            }, {}),
        };
    }
    // Process data.
    for (const record of data) {
        const aggregatedRecord = {
            ...record,
            count: record.summary.reduce((total, item) => total + item.value, 0),
        };
        const date = new Date(aggregatedRecord.date);
        const year = globalMap[date.getFullYear()];
        const month = year.months[date.getMonth()];
        month.count += aggregatedRecord.count;
        year.count += aggregatedRecord.count;
        if (aggregatedRecord.count > month.itemMaxCount) {
            month.itemMaxCount = aggregatedRecord.count;
        }
        if (month.count > year.itemMaxCount) {
            year.itemMaxCount = month.count;
        }
        month.items.push(aggregatedRecord);
    }
    return globalMap;
}

/**
 * Helper function for constructing HTML tooltip
 * @param d object
 */
function constructTooltipHtml(d) {
    let tooltipHtml = `<div><span><strong>Total items:</strong></span>${d.count}</div>`;
    if (d.date) {
        tooltipHtml += `<div>on ${moment(d.date).format('dddd, MMM Do YYYY')}</div><br>`;
    }
    tooltipHtml += '<br/>';
    // Add summary to the tooltip
    for (let i = 0, max = d.summary.length; i < max; i++) {
        if (i >= 5) {
            tooltipHtml += '...';
            break;
        }
        tooltipHtml += `<div><span><strong>${d.summary[i].name}</strong></span>${formatValue(d.summary[i].value)}</div>`;
    }
    return tooltipHtml;
}
function getTooltipX(event) {
    const tooltipWidth = 250;
    const offset = 50;
    let tooltipX = event.clientX;
    const totalWidth = event.view.innerWidth - offset;
    if (tooltipX + tooltipWidth > totalWidth) {
        tooltipX = totalWidth - tooltipWidth;
    }
    return tooltipX;
}
function formatValue(total) {
    return total === 1 ? '1 time' : `${total} times`;
}

class MonthHorizontalAxisComponent {
    constructor(element) {
        this.selectionChange = new EventEmitter();
        this.showTooltipEvent = new EventEmitter();
        this.hideTooltipEvent = new EventEmitter();
        this.duration = 250;
        this.labels = d3.select(element.nativeElement);
    }
    ngOnChanges() {
        const labels = this.labels.selectAll('.label-month').data(this.items, (d) => d.date);
        labels
            .exit()
            .transition()
            .duration(this.duration)
            .attr('x', this.horizontalScale.bandwidth())
            .attr('opacity', 0)
            .attr('font-weight', 'normal')
            .remove();
        labels
            .enter()
            .append('text')
            .attr('class', 'item label label-month')
            .attr('font-size', `${this.fontSize}px`)
            .attr('font-weight', 'normal')
            .attr('cursor', 'pointer')
            .attr('text-anchor', 'middle')
            .attr('dominant-baseline', 'central')
            .text((d) => toMonth(d, 'MMMM'))
            .attr('x', 0)
            .attr('y', 0)
            .on('mouseenter', (d) => this.showTooltip(d))
            .on('mouseout', () => this.hideTooltip())
            .on('click', (d) => {
            this.selectionChange.next({
                year: d.date.getFullYear(),
                month: d.date.getMonth(),
            });
        })
            .merge(labels)
            .transition()
            .duration(this.duration)
            .attr('font-weight', 'bold')
            .attr('x', (d) => this.horizontalScale(d.date.getMonth()) + this.horizontalScale.bandwidth() / 2);
    }
    showTooltip(item) {
        this.showTooltipEvent.emit({
            html: constructTooltipHtml({
                count: item.count,
                summary: flatAndSort(item.children),
            }),
            x: getTooltipX(d3.event),
            y: d3.event.clientY,
        });
    }
    hideTooltip() {
        this.hideTooltipEvent.emit();
    }
}
MonthHorizontalAxisComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: MonthHorizontalAxisComponent, deps: [{ token: i0.ElementRef }], target: i0.ɵɵFactoryTarget.Component });
MonthHorizontalAxisComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "14.0.0", version: "14.2.12", type: MonthHorizontalAxisComponent, selector: "g[heatmap-calendar-month-horizontal-axis]", inputs: { items: "items", horizontalScale: "horizontalScale", fontSize: "fontSize" }, outputs: { selectionChange: "selectionChange", showTooltipEvent: "showTooltipEvent", hideTooltipEvent: "hideTooltipEvent" }, usesOnChanges: true, ngImport: i0, template: '<ng-container></ng-container>', isInline: true, changeDetection: i0.ChangeDetectionStrategy.OnPush });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: MonthHorizontalAxisComponent, decorators: [{
            type: Component,
            args: [{
                    // eslint-disable-next-line @angular-eslint/component-selector
                    selector: 'g[heatmap-calendar-month-horizontal-axis]',
                    template: '<ng-container></ng-container>',
                    changeDetection: ChangeDetectionStrategy.OnPush,
                }]
        }], ctorParameters: function () { return [{ type: i0.ElementRef }]; }, propDecorators: { items: [{
                type: Input
            }], horizontalScale: [{
                type: Input
            }], fontSize: [{
                type: Input
            }], selectionChange: [{
                type: Output
            }], showTooltipEvent: [{
                type: Output
            }], hideTooltipEvent: [{
                type: Output
            }] } });

class MonthSeriesComponent {
    constructor(element) {
        this.daySelected = new EventEmitter();
        this.showTooltipEvent = new EventEmitter();
        this.hideTooltipEvent = new EventEmitter();
        this.duration = 250;
        this.numberOfWeeksScale = d3.scaleBand().paddingOuter(0.25);
        this.g = d3.select(element.nativeElement);
    }
    get details() {
        return this.items.flatMap((i) => i.children);
    }
    ngOnChanges() {
        const width = this.monthScale.bandwidth();
        this.numberOfWeeksScale.rangeRound([0, width]).domain(range(6));
        this.updateDayCircles();
    }
    updateDayCircles() {
        const circles = this.g
            .selectAll('.item-block-day')
            .data(this.details, (d) => moment(d.date).format('YYYY-MM-DD'));
        circles.exit().transition().duration(this.duration).attr('opacity', 0).remove();
        circles
            .enter()
            .append('circle')
            .attr('class', 'item item-block-day')
            .attr('transform', (d) => {
            return `translate( ${this.monthScale(d.date.getMonth())}, 0)`;
        })
            .attr('cx', (d) => this.numberOfWeeksScale(this.countWeekdayOccurrencesInMonth(d.date)))
            .attr('cy', (d) => this.weekdayScale(moment(d.date).day()))
            .attr('r', Math.min(this.weekdayScale.bandwidth() / 2, this.numberOfWeeksScale.bandwidth() / 2))
            .attr('fill', (d) => this.colorScale(d.count))
            .style('cursor', 'pointer')
            .attr('opacity', 0)
            .on('click', (d) => this.daySelected.emit(d.date))
            .on('mouseenter', (d) => this.showTooltip(d))
            .on('mouseleave', () => this.hideTooltip())
            .merge(circles)
            .transition()
            .delay(() => (Math.cos(Math.PI * Math.random()) + 1) * this.duration)
            .duration(this.duration)
            .ease(d3.easeLinear)
            .attr('transform', (d) => {
            return `translate( ${this.monthScale(d.date.getMonth())}, 0)`;
        })
            .attr('cx', (d) => this.numberOfWeeksScale(this.countWeekdayOccurrencesInMonth(d.date)))
            .attr('cy', (d) => this.weekdayScale(moment(d.date).day()))
            .attr('opacity', 1);
    }
    showTooltip(item) {
        this.showTooltipEvent.emit({
            html: constructTooltipHtml({
                count: item.count,
                summary: item.summary,
                date: item.date,
            }),
            x: getTooltipX(d3.event),
            y: d3.event.clientY,
        });
    }
    hideTooltip() {
        this.hideTooltipEvent.emit();
    }
    countWeekdayOccurrencesInMonth(date) {
        const m = moment(date);
        let week = m.week();
        // Last week is considered to be the first week of next year.
        if (m.month() === 11 && m.week() === 1) {
            week = 53;
        }
        const start = moment(date).startOf('month');
        return week - start.week();
    }
}
MonthSeriesComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: MonthSeriesComponent, deps: [{ token: i0.ElementRef }], target: i0.ɵɵFactoryTarget.Component });
MonthSeriesComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "14.0.0", version: "14.2.12", type: MonthSeriesComponent, selector: "g[heatmap-calendar-month-series]", inputs: { items: "items", weekdayScale: "weekdayScale", monthScale: "monthScale", colorScale: "colorScale" }, outputs: { daySelected: "daySelected", showTooltipEvent: "showTooltipEvent", hideTooltipEvent: "hideTooltipEvent" }, usesOnChanges: true, ngImport: i0, template: '<ng-container></ng-container>', isInline: true, changeDetection: i0.ChangeDetectionStrategy.OnPush });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: MonthSeriesComponent, decorators: [{
            type: Component,
            args: [{
                    // eslint-disable-next-line @angular-eslint/component-selector
                    selector: 'g[heatmap-calendar-month-series]',
                    template: '<ng-container></ng-container>',
                    changeDetection: ChangeDetectionStrategy.OnPush,
                }]
        }], ctorParameters: function () { return [{ type: i0.ElementRef }]; }, propDecorators: { items: [{
                type: Input
            }], weekdayScale: [{
                type: Input
            }], monthScale: [{
                type: Input
            }], colorScale: [{
                type: Input
            }], daySelected: [{
                type: Output
            }], showTooltipEvent: [{
                type: Output
            }], hideTooltipEvent: [{
                type: Output
            }] } });

class YearHorizontalAxisComponent {
    constructor(element) {
        this.selectionChange = new EventEmitter();
        this.showTooltipEvent = new EventEmitter();
        this.hideTooltipEvent = new EventEmitter();
        this.duration = 250;
        this.labels = d3.select(element.nativeElement);
    }
    ngOnChanges() {
        const labels = this.labels.selectAll('.label-year').data(this.items, (d) => d.date);
        labels.exit().transition().duration(this.duration).attr('x', 0).attr('font-weight', 'normal').remove();
        labels
            .enter()
            .append('text')
            .attr('class', 'item label label-year')
            .attr('font-size', `${this.fontSize}px`)
            .attr('font-weight', 'normal')
            .attr('cursor', 'pointer')
            .attr('text-anchor', 'middle')
            .attr('dominant-baseline', 'central')
            .text((d) => d.date.getFullYear())
            .attr('x', 0)
            .attr('y', 0)
            .on('mouseenter', (d) => this.showTooltip(d))
            .on('mouseout', () => this.hideTooltip())
            .on('click', (d) => {
            this.selectionChange.emit({ year: d.date.getFullYear() });
        })
            .merge(labels)
            .transition()
            .duration(this.duration)
            .attr('font-weight', 'bold')
            .attr('x', (d) => this.horizontalScale(d.date.getFullYear()) + this.horizontalScale.bandwidth() / 2);
    }
    showTooltip(item) {
        this.showTooltipEvent.emit({
            html: constructTooltipHtml({
                count: item.count,
                summary: flatAndSort(item.children),
            }),
            x: getTooltipX(d3.event),
            y: d3.event.clientY,
        });
    }
    hideTooltip() {
        this.hideTooltipEvent.emit();
    }
}
YearHorizontalAxisComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: YearHorizontalAxisComponent, deps: [{ token: i0.ElementRef }], target: i0.ɵɵFactoryTarget.Component });
YearHorizontalAxisComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "14.0.0", version: "14.2.12", type: YearHorizontalAxisComponent, selector: "g[heatmap-calendar-year-horizontal-axis]", inputs: { items: "items", horizontalScale: "horizontalScale", fontSize: "fontSize" }, outputs: { selectionChange: "selectionChange", showTooltipEvent: "showTooltipEvent", hideTooltipEvent: "hideTooltipEvent" }, usesOnChanges: true, ngImport: i0, template: '<ng-container></ng-container>', isInline: true, changeDetection: i0.ChangeDetectionStrategy.OnPush });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: YearHorizontalAxisComponent, decorators: [{
            type: Component,
            args: [{
                    // eslint-disable-next-line @angular-eslint/component-selector
                    selector: 'g[heatmap-calendar-year-horizontal-axis]',
                    template: '<ng-container></ng-container>',
                    changeDetection: ChangeDetectionStrategy.OnPush,
                }]
        }], ctorParameters: function () { return [{ type: i0.ElementRef }]; }, propDecorators: { items: [{
                type: Input
            }], horizontalScale: [{
                type: Input
            }], fontSize: [{
                type: Input
            }], selectionChange: [{
                type: Output
            }], showTooltipEvent: [{
                type: Output
            }], hideTooltipEvent: [{
                type: Output
            }] } });

class YearSeriesComponent {
    constructor(element) {
        this.selectionChange = new EventEmitter();
        this.showTooltipEvent = new EventEmitter();
        this.hideTooltipEvent = new EventEmitter();
        this.fontColor = '#aaa';
        this.whiteColor = '#fff';
        this.duration = 250;
        this.monthXAxis = d3.scaleBand().padding(0.1).domain(range(2));
        this.monthYAxis = d3.scaleBand().padding(0.15).domain(range(6));
        this.getYearX = (d) => this.yearScale(d.date.getFullYear());
        this.getMonthX = (d) => this.monthXAxis(Math.floor(d.date.getMonth() / 6));
        this.getMonthY = (d) => this.monthYAxis(d.date.getMonth() % 6);
        this.getCenteredMonthX = (d) => this.getMonthX(d) + this.monthXAxis.bandwidth() / 2;
        this.getCenteredMonthY = (d) => this.getMonthY(d) + this.monthYAxis.bandwidth() / 2;
        this.formatOutline = (d) => (this.isMonthSelected(d) ? `thin solid ${this.primaryColor}` : '');
        this.formatBackground = (d) => {
            return this.isMonthSelected(d) || d.count === 0 ? this.whiteColor : this.colorScale(d.count);
        };
        this.yearItems = d3.select(element.nativeElement).append('g');
    }
    get details() {
        return this.items.flatMap((i) => i.children);
    }
    ngOnChanges() {
        const width = this.yearScale.bandwidth();
        this.monthXAxis.range([0, width]);
        this.monthYAxis.range([0, this.height]);
        this.updateMonthRects();
        this.updateMonthLabels();
    }
    updateMonthRects() {
        const rects = this.yearItems
            .selectAll('.item-block-month')
            .data(this.details, (d) => moment(d.date).format('YYYY-MM-DD'));
        rects.exit().transition().duration(this.duration).attr('opacity', 0).remove();
        rects
            .enter()
            .append('rect')
            .attr('class', 'item item-block-month')
            .attr('transform', (d) => {
            return `translate(${this.getYearX(d)}, 0)`;
        })
            .attr('x', (d) => this.getMonthX(d))
            .attr('y', (d) => this.getMonthY(d))
            .attr('width', this.monthXAxis.bandwidth)
            .attr('height', this.monthYAxis.bandwidth)
            .style('cursor', 'pointer')
            .attr('opacity', 0)
            .on('click', (d) => {
            this.selectionChange.next({
                year: d.date.getFullYear(),
                month: d.date.getMonth(),
            });
        })
            .on('mouseenter', (d) => this.showTooltip(d))
            .on('mouseleave', () => this.hideTooltip())
            .merge(rects)
            .transition()
            .delay((_, i) => {
            return (this.duration * (i + 1)) / 100;
        })
            .duration(this.duration)
            .attr('transform', (d) => {
            return `translate(${this.getYearX(d)}, 0)`;
        })
            .attr('x', (d) => this.getMonthX(d))
            .attr('y', (d) => this.getMonthY(d))
            .attr('fill', this.formatBackground)
            .style('outline', this.formatOutline)
            .attr('opacity', 1);
    }
    updateMonthLabels() {
        const labels = this.yearItems
            .selectAll('.item-text-month')
            .data(this.details, (d) => moment(d.date).format('YYYY-MM-DD'));
        labels.exit().transition().duration(this.duration).attr('opacity', 0).remove();
        labels
            .enter()
            .append('text')
            .attr('class', 'item item-text-month')
            .attr('transform', (d) => {
            return `translate(${this.getYearX(d)}, 0)`;
        })
            .attr('x', (d) => this.getCenteredMonthX(d))
            .attr('y', (d) => this.getCenteredMonthY(d))
            .attr('font-size', `${this.fontSize}px`)
            .attr('font-weight', 'bold')
            .attr('text-anchor', 'middle')
            .attr('dominant-baseline', 'central')
            .attr('cursor', 'pointer')
            .text((d) => toMonth(d, 'MMM'))
            .on('click', (d) => {
            this.selectionChange.next({
                year: d.date.getFullYear(),
                month: d.date.getMonth(),
            });
        })
            .on('mouseenter', (d) => this.showTooltip(d))
            .on('mouseleave', () => this.hideTooltip())
            .merge(labels)
            .transition()
            .delay((d, i) => {
            return (this.duration * (i + 1)) / 100;
        })
            .duration(this.duration)
            .attr('transform', (d) => {
            return `translate(${this.getYearX(d)}, 0)`;
        })
            .attr('x', (d) => this.getCenteredMonthX(d))
            .attr('y', (d) => this.getCenteredMonthY(d))
            .attr('fill', (d) => this.formatFontColor(d))
            .attr('opacity', 1);
    }
    formatFontColor(detail) {
        if (this.isMonthSelected(detail)) {
            return this.primaryColor;
        }
        return detail.count <= this.colorScale.domain()[1] * 0.2 ? this.fontColor : this.whiteColor;
    }
    isMonthSelected(d) {
        if (!this.selectedItem) {
            return false;
        }
        return this.selectedItem.year === d.date.getFullYear() && d.date.getMonth() === this.selectedItem.month;
    }
    showTooltip({ count, summary }) {
        this.showTooltipEvent.emit({
            html: constructTooltipHtml({ count, summary }),
            x: getTooltipX(d3.event),
            y: d3.event.clientY,
        });
    }
    hideTooltip() {
        this.hideTooltipEvent.emit();
    }
}
YearSeriesComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: YearSeriesComponent, deps: [{ token: i0.ElementRef }], target: i0.ɵɵFactoryTarget.Component });
YearSeriesComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "14.0.0", version: "14.2.12", type: YearSeriesComponent, selector: "g[heatmap-calendar-year-series]", inputs: { items: "items", height: "height", primaryColor: "primaryColor", yearScale: "yearScale", colorScale: "colorScale", fontSize: "fontSize", selectedItem: "selectedItem" }, outputs: { selectionChange: "selectionChange", showTooltipEvent: "showTooltipEvent", hideTooltipEvent: "hideTooltipEvent" }, usesOnChanges: true, ngImport: i0, template: '<ng-container></ng-container>', isInline: true, changeDetection: i0.ChangeDetectionStrategy.OnPush });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: YearSeriesComponent, decorators: [{
            type: Component,
            args: [{
                    // eslint-disable-next-line @angular-eslint/component-selector
                    selector: 'g[heatmap-calendar-year-series]',
                    template: '<ng-container></ng-container>',
                    changeDetection: ChangeDetectionStrategy.OnPush,
                }]
        }], ctorParameters: function () { return [{ type: i0.ElementRef }]; }, propDecorators: { items: [{
                type: Input
            }], height: [{
                type: Input
            }], primaryColor: [{
                type: Input
            }], yearScale: [{
                type: Input
            }], colorScale: [{
                type: Input
            }], fontSize: [{
                type: Input
            }], selectedItem: [{
                type: Input
            }], selectionChange: [{
                type: Output
            }], showTooltipEvent: [{
                type: Output
            }], hideTooltipEvent: [{
                type: Output
            }] } });

class TooltipComponent {
    constructor(element) {
        this.duration = 250;
        this.tooltipElement = d3
            .select(element.nativeElement)
            .append('div')
            .attr('class', 'heatmap-tooltip')
            .style('opacity', 0);
    }
    ngOnChanges(changes) {
        if ('tooltipEvent' in changes) {
            this.tooltipEvent ? this.show(this.tooltipEvent) : this.hide();
        }
    }
    hide() {
        this.tooltipElement
            .transition()
            .duration(this.duration / 2)
            .ease(d3.easeLinear)
            .style('opacity', 0);
    }
    show(event) {
        this.tooltipElement
            .html(event.html)
            .style('left', `${event.x}px`)
            .style('top', `${event.y}px`)
            .transition()
            .duration(this.duration / 2)
            .ease(d3.easeLinear)
            .style('opacity', 1);
    }
}
TooltipComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: TooltipComponent, deps: [{ token: i0.ElementRef }], target: i0.ɵɵFactoryTarget.Component });
TooltipComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "14.0.0", version: "14.2.12", type: TooltipComponent, selector: "sentinel-heatmap-calendar-tooltip", inputs: { tooltipEvent: "tooltipEvent" }, usesOnChanges: true, ngImport: i0, template: '<ng-container></ng-container>', isInline: true, changeDetection: i0.ChangeDetectionStrategy.OnPush });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: TooltipComponent, decorators: [{
            type: Component,
            args: [{
                    selector: 'sentinel-heatmap-calendar-tooltip',
                    template: '<ng-container></ng-container>',
                    changeDetection: ChangeDetectionStrategy.OnPush,
                }]
        }], ctorParameters: function () { return [{ type: i0.ElementRef }]; }, propDecorators: { tooltipEvent: [{
                type: Input
            }] } });

class BackButtonComponent {
    constructor(element) {
        this.resetSelection = new EventEmitter();
        this.duration = 250;
        this.g = d3.select(element.nativeElement);
    }
    ngOnChanges(changes) {
        if ('isVisible' in changes && this.isVisible) {
            this.show();
        }
        if ('isVisible' in changes && !this.isVisible) {
            this.hide();
        }
    }
    hide() {
        if (this.circle) {
            this.circle.transition().duration(this.duration).ease(d3.easeLinear).style('r', 0).remove();
            this.circle = null;
        }
        if (this.arrow) {
            this.arrow.transition().duration(this.duration).ease(d3.easeLinear).style('opacity', 0).remove();
            this.arrow = null;
        }
    }
    show() {
        if (!this.circle) {
            this.circle = this.g.append('circle');
        }
        this.circle
            .attr('class', 'back-button-circle')
            .attr('cx', 8)
            .attr('cy', 0)
            .style('cursor', 'pointer')
            .html('&#x2190;')
            .attr('r', 0)
            .on('click', () => this.resetSelection.next())
            .transition()
            .duration(this.duration)
            .ease(d3.easeLinear)
            .attr('r', 8);
        if (!this.arrow) {
            this.arrow = this.g.append('text');
        }
        this.arrow
            .attr('class', 'back-button-arrow')
            .attr('font-size', `${16}px`)
            .attr('x', 1)
            .attr('y', 0)
            .attr('dy', 2)
            .style('cursor', 'pointer')
            .html('&#x2190;')
            .style('opacity', 0)
            .on('click', () => this.resetSelection.next())
            .transition()
            .duration(this.duration)
            .ease(d3.easeLinear)
            .style('opacity', 1);
    }
}
BackButtonComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: BackButtonComponent, deps: [{ token: i0.ElementRef }], target: i0.ɵɵFactoryTarget.Component });
BackButtonComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "14.0.0", version: "14.2.12", type: BackButtonComponent, selector: "g[heatmap-calendar-back-button]", inputs: { isVisible: "isVisible" }, outputs: { resetSelection: "resetSelection" }, usesOnChanges: true, ngImport: i0, template: '<ng-container></ng-container>', isInline: true, changeDetection: i0.ChangeDetectionStrategy.OnPush });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: BackButtonComponent, decorators: [{
            type: Component,
            args: [{
                    // eslint-disable-next-line @angular-eslint/component-selector
                    selector: 'g[heatmap-calendar-back-button]',
                    template: '<ng-container></ng-container>',
                    changeDetection: ChangeDetectionStrategy.OnPush,
                }]
        }], ctorParameters: function () { return [{ type: i0.ElementRef }]; }, propDecorators: { isVisible: [{
                type: Input
            }], resetSelection: [{
                type: Output
            }] } });

class WeekdayVerticalAxisComponent {
    constructor(element) {
        this.duration = 250;
        this.g = d3.select(element.nativeElement);
    }
    set isVisible(isVisible) {
        if (isVisible) {
            this.show();
        }
        else {
            this.clear();
        }
    }
    ngOnChanges() {
        const labels = this.g.selectAll('.label-month').data(this.weekdays, (d) => d.getDay());
        labels.exit().transition().duration(this.duration).attr('y', 0).attr('font-weight', 'normal').remove();
        labels
            .enter()
            .append('text')
            .attr('class', 'item label label-month')
            .attr('font-size', `${this.fontSize}px`)
            .attr('font-weight', 'normal')
            .attr('cursor', 'pointer')
            .attr('text-anchor', 'middle')
            .attr('dominant-baseline', 'central')
            .text((date) => moment(date).format('dd')[0])
            .attr('x', 0)
            .attr('y', 0)
            .merge(labels)
            .transition()
            .duration(this.duration)
            .attr('font-weight', 'bold')
            .attr('y', (d) => this.weekdayScale(d.getDay()));
    }
    show() {
        this.weekdays = d3.timeDays(moment().startOf('week').toDate(), moment().endOf('week').toDate());
        this.ngOnChanges();
    }
    clear() {
        this.weekdays = [];
        this.ngOnChanges();
    }
}
WeekdayVerticalAxisComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: WeekdayVerticalAxisComponent, deps: [{ token: i0.ElementRef }], target: i0.ɵɵFactoryTarget.Component });
WeekdayVerticalAxisComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "14.0.0", version: "14.2.12", type: WeekdayVerticalAxisComponent, selector: "g[heatmap-calendar-weekday-vertical-axis]", inputs: { height: "height", fontSize: "fontSize", weekdayScale: "weekdayScale", isVisible: "isVisible" }, usesOnChanges: true, ngImport: i0, template: '<ng-container></ng-container>', isInline: true, changeDetection: i0.ChangeDetectionStrategy.OnPush });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: WeekdayVerticalAxisComponent, decorators: [{
            type: Component,
            args: [{
                    // eslint-disable-next-line @angular-eslint/component-selector
                    selector: 'g[heatmap-calendar-weekday-vertical-axis]',
                    template: '<ng-container></ng-container>',
                    changeDetection: ChangeDetectionStrategy.OnPush,
                }]
        }], ctorParameters: function () { return [{ type: i0.ElementRef }]; }, propDecorators: { height: [{
                type: Input
            }], fontSize: [{
                type: Input
            }], weekdayScale: [{
                type: Input
            }], isVisible: [{
                type: Input
            }] } });

var AlignMonths;
(function (AlignMonths) {
    AlignMonths["none"] = "none";
    AlignMonths["left"] = "left";
    AlignMonths["right"] = "right";
    AlignMonths["around"] = "around";
})(AlignMonths || (AlignMonths = {}));

/* tslint:disable:no-bitwise */
function shadeColor(color, percent) {
    // strip the leading # if it's there
    color = color.replace(/^\s*#|\s*$/g, '');
    // convert 3 char codes --> 6, e.g. `E0F` --> `EE00FF`
    if (color.length === 3) {
        color = color.replace(/(.)/g, '$1$1');
    }
    const r = parseInt(color.substr(0, 2), 16);
    const g = parseInt(color.substr(2, 2), 16);
    const b = parseInt(color.substr(4, 2), 16);
    return ('#' +
        (0 | ((1 << 8) + r + ((256 - r) * percent) / 100)).toString(16).substr(1) +
        (0 | ((1 << 8) + g + ((256 - g) * percent) / 100)).toString(16).substr(1) +
        (0 | ((1 << 8) + b + ((256 - b) * percent) / 100)).toString(16).substr(1));
}

class HeatmapCalendarWrapperComponent {
    constructor() {
        this.color = '#ff4500';
        this.alignMonths = AlignMonths.around;
        this.monthsCount = 0;
        this.daySelected = new EventEmitter();
        this.margin = { left: 0, right: 0, top: 40, bottom: 0 };
        this.monthsOffset = 20;
        this.translateChart = `translate(0, ${this.margin.top})`;
        this.translateTopAxis = `translate(0, ${-this.margin.top / 2})`;
        this.translateTopAxisMonthsOffset = `translate(${this.monthsOffset}, ${-this.margin.top / 2})`;
        this.translateMonthsOffset = `translate(${this.monthsOffset}, 0)`;
        this.fontSize = 15;
        this.colorScale = d3.scaleLinear();
        this.yearScale = d3.scaleBand().paddingInner(0.05);
        this.monthScale = d3.scaleBand();
        this.weekdayScale = d3.scaleBand().paddingInner(0.2).domain(range(7));
    }
    ngOnInit() {
        this.initColorScale();
    }
    /**
     * Check if data is available
     */
    ngOnChanges(changes) {
        if ('width' in changes) {
            this.calculateInnerWidth();
        }
        if ('height' in changes) {
            this.calculateInnerHeight();
        }
        if (this.data) {
            this.map = toMap(this.data);
            this.weekdayScale.range([0, this.innerHeight]);
            this.yearScale.domain(Object.keys(this.map).map((k) => +k)).range([0, this.innerWidth]);
            this.onSelectionChanged(this.selectedItem);
        }
        if ('color' in changes) {
            this.initColorScale();
        }
    }
    onDaySelected(date) {
        this.daySelected.emit(date ? date.valueOf() : -1);
    }
    resetSelection() {
        this.onSelectionChanged(undefined);
    }
    onSelectionChanged(event) {
        this.selectedItem = event;
        this.onResetTooltipEvent();
        if (event) {
            this.mapMonths(event.month);
            if (event.month !== null && event.month !== undefined) {
                this.mapYears(event.year);
                // Push months next to year chart.
                this.monthsTransform = this.formatTransform(this.yearScale.bandwidth() + 30, 0);
            }
            else {
                this.yearItems = [];
                this.monthsTransform = this.formatTransform(0, 0);
            }
        }
        else {
            this.mapYears();
            this.monthItems = [];
        }
    }
    onTooltipEvent(event) {
        this.tooltipEvent = event;
    }
    onResetTooltipEvent() {
        this.tooltipEvent = undefined;
    }
    mapYears(yearFilter) {
        const items = [];
        let maxCount = 0;
        if (yearFilter) {
            const year = this.map[yearFilter];
            maxCount = year.itemMaxCount;
            const yearWidth = this.innerWidth / Object.keys(this.map).length;
            this.yearScale = this.yearScale.domain([yearFilter]).rangeRound([0, yearWidth]);
            items.push(toYearItem(year));
        }
        else {
            this.yearScale = this.yearScale.domain(Object.keys(this.map).map((k) => +k)).rangeRound([0, this.innerWidth]);
            for (const mapKey in this.map) {
                if (Object.prototype.hasOwnProperty.call(this.map, mapKey)) {
                    const year = this.map[mapKey];
                    maxCount = Math.max(maxCount, year.itemMaxCount);
                    items.push(toYearItem(year));
                }
            }
        }
        this.colorScale.domain([0, maxCount]);
        this.yearItems = items;
    }
    mapMonths(monthFilter) {
        if (monthFilter !== null && monthFilter !== undefined) {
            const half = Math.floor(this.monthsCount / 2);
            switch (this.alignMonths) {
                case AlignMonths.none:
                    this.align(monthFilter, 0, 0);
                    break;
                case AlignMonths.right:
                    this.align(monthFilter, 0, this.monthsCount);
                    break;
                case AlignMonths.left:
                    this.align(monthFilter, -this.monthsCount, 0);
                    break;
                case AlignMonths.around:
                    this.align(monthFilter, -half, half);
                    break;
            }
        }
        else {
            const items = [];
            const year = this.map[this.selectedItem.year];
            this.monthScale = this.monthScale.domain(range(12)).range([0, this.innerWidth - this.monthsOffset]);
            for (const mapKey in year.months) {
                if (Object.prototype.hasOwnProperty.call(year.months, mapKey)) {
                    const month = year.months[mapKey];
                    items.push(toMonthItem(month));
                }
            }
            this.monthItems = items;
            this.colorScale.domain([0, year.itemMaxCount]);
        }
    }
    align(monthFilter, start, end) {
        const domain = [];
        const items = [];
        let maxValue = 0;
        let year = this.selectedItem.year;
        for (let i = start; i <= end; i++) {
            const index = monthFilter + i;
            if (index < 0) {
                year = this.selectedItem.year - 1;
            }
            else if (index > 11) {
                year = this.selectedItem.year + 1;
            }
            const yearData = this.map[year];
            if (!yearData) {
                continue;
            }
            const monthNumber = index < 0 ? 12 + index : index % 12;
            const monthData = yearData.months[monthNumber];
            if (!monthData) {
                continue;
            }
            maxValue = Math.max(maxValue, monthData.count);
            domain.push(monthNumber);
            items.push(toMonthItem(monthData));
        }
        const monthWidth = (this.innerWidth - this.monthsOffset) / 12;
        const monthsCount = Math.min(12, domain.length);
        this.monthScale = this.monthScale.domain(domain).range([0, monthWidth * monthsCount]);
        this.colorScale.domain([0, maxValue]);
        this.monthItems = items;
    }
    formatTransform(x, y) {
        return `translate(${x} , ${y})`;
    }
    initColorScale() {
        this.colorScale.range([shadeColor(this.color, 90), this.color]);
    }
    calculateInnerWidth() {
        this.innerWidth = this.width - this.margin.left - this.margin.right;
    }
    calculateInnerHeight() {
        this.innerHeight = this.height - this.margin.top - this.margin.bottom;
    }
}
HeatmapCalendarWrapperComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: HeatmapCalendarWrapperComponent, deps: [], target: i0.ɵɵFactoryTarget.Component });
HeatmapCalendarWrapperComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "14.0.0", version: "14.2.12", type: HeatmapCalendarWrapperComponent, selector: "sentinel-heatmap-calendar-wrapper", inputs: { data: "data", color: "color", alignMonths: "alignMonths", monthsCount: "monthsCount", width: "width", height: "height" }, outputs: { daySelected: "daySelected" }, usesOnChanges: true, ngImport: i0, template: "<sentinel-heatmap-calendar-tooltip [tooltipEvent]='tooltipEvent'></sentinel-heatmap-calendar-tooltip>\n\n<svg class=\"svg\" [attr.width]=\"width\" [attr.height]=\"height\">\n  <g [attr.transform]=\"translateChart\" class=\"heat-map chart\">\n    <g [attr.transform]=\"translateTopAxis\">\n      <g heatmap-calendar-back-button [isVisible]=\"!!selectedItem\" (resetSelection)='resetSelection()'></g>\n      <g\n        heatmap-calendar-year-horizontal-axis\n        [items]=\"yearItems\"\n        [horizontalScale]=\"yearScale\"\n        [fontSize]=\"fontSize\"\n        (selectionChange)='onSelectionChanged($event)'\n        (showTooltipEvent)='onTooltipEvent($event)'\n        (hideTooltipEvent)='onResetTooltipEvent()'\n      ></g>\n    </g>\n\n    <g\n      heatmap-calendar-year-series\n      [height]=\"innerHeight\"\n      [items]=\"yearItems\"\n      [primaryColor]=\"color\"\n      [yearScale]=\"yearScale\"\n      [colorScale]=\"colorScale\"\n      [selectedItem]=\"selectedItem\"\n      (selectionChange)='onSelectionChanged($event)'\n      (showTooltipEvent)='onTooltipEvent($event)'\n      (hideTooltipEvent)='onResetTooltipEvent()'\n    ></g>\n\n    <g [attr.transform]=\"monthsTransform\">\n      <g\n        heatmap-calendar-weekday-vertical-axis\n        transform=\"translate(8, 0)\"\n        [height]=\"innerHeight\"\n        [isVisible]=\"monthItems.length !== 0\"\n        [fontSize]=\"fontSize\"\n        [weekdayScale]=\"weekdayScale\"\n      ></g>\n      <g\n        heatmap-calendar-month-horizontal-axis\n        [items]=\"monthItems\"\n        [horizontalScale]=\"monthScale\"\n        [fontSize]=\"fontSize\"\n        [attr.transform]=\"translateTopAxisMonthsOffset\"\n        (selectionChange)='onSelectionChanged($event)'\n        (showTooltipEvent)='onTooltipEvent($event)'\n        (hideTooltipEvent)='onResetTooltipEvent()'\n      ></g>\n\n      <g\n        heatmap-calendar-month-series\n        [items]=\"monthItems\"\n        [monthScale]=\"monthScale\"\n        [colorScale]=\"colorScale\"\n        [weekdayScale]=\"weekdayScale\"\n        [attr.transform]=\"translateMonthsOffset\"\n        (daySelected)='onDaySelected($event)'\n        (showTooltipEvent)='onTooltipEvent($event)'\n        (hideTooltipEvent)='onResetTooltipEvent()'\n      ></g>\n    </g>\n  </g>\n</svg>\n", styles: [".heatmap-tooltip{pointer-events:none;position:fixed;z-index:9999;width:250px;margin-left:40px;margin-top:40px;max-width:250px;overflow:hidden;padding:15px;font-size:12px;line-height:14px}.heatmap-tooltip .header strong{display:inline-block;width:250px}.heatmap-tooltip span{display:inline-block;width:75%;padding-right:10px;box-sizing:border-box}.heatmap-tooltip span,.heatmap-tooltip .header strong{white-space:nowrap;overflow:hidden;text-overflow:ellipsis}\n"], dependencies: [{ kind: "component", type: WeekdayVerticalAxisComponent, selector: "g[heatmap-calendar-weekday-vertical-axis]", inputs: ["height", "fontSize", "weekdayScale", "isVisible"] }, { kind: "component", type: BackButtonComponent, selector: "g[heatmap-calendar-back-button]", inputs: ["isVisible"], outputs: ["resetSelection"] }, { kind: "component", type: TooltipComponent, selector: "sentinel-heatmap-calendar-tooltip", inputs: ["tooltipEvent"] }, { kind: "component", type: YearSeriesComponent, selector: "g[heatmap-calendar-year-series]", inputs: ["items", "height", "primaryColor", "yearScale", "colorScale", "fontSize", "selectedItem"], outputs: ["selectionChange", "showTooltipEvent", "hideTooltipEvent"] }, { kind: "component", type: YearHorizontalAxisComponent, selector: "g[heatmap-calendar-year-horizontal-axis]", inputs: ["items", "horizontalScale", "fontSize"], outputs: ["selectionChange", "showTooltipEvent", "hideTooltipEvent"] }, { kind: "component", type: MonthSeriesComponent, selector: "g[heatmap-calendar-month-series]", inputs: ["items", "weekdayScale", "monthScale", "colorScale"], outputs: ["daySelected", "showTooltipEvent", "hideTooltipEvent"] }, { kind: "component", type: MonthHorizontalAxisComponent, selector: "g[heatmap-calendar-month-horizontal-axis]", inputs: ["items", "horizontalScale", "fontSize"], outputs: ["selectionChange", "showTooltipEvent", "hideTooltipEvent"] }], changeDetection: i0.ChangeDetectionStrategy.OnPush, encapsulation: i0.ViewEncapsulation.None });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: HeatmapCalendarWrapperComponent, decorators: [{
            type: Component,
            args: [{ selector: 'sentinel-heatmap-calendar-wrapper', changeDetection: ChangeDetectionStrategy.OnPush, encapsulation: ViewEncapsulation.None, template: "<sentinel-heatmap-calendar-tooltip [tooltipEvent]='tooltipEvent'></sentinel-heatmap-calendar-tooltip>\n\n<svg class=\"svg\" [attr.width]=\"width\" [attr.height]=\"height\">\n  <g [attr.transform]=\"translateChart\" class=\"heat-map chart\">\n    <g [attr.transform]=\"translateTopAxis\">\n      <g heatmap-calendar-back-button [isVisible]=\"!!selectedItem\" (resetSelection)='resetSelection()'></g>\n      <g\n        heatmap-calendar-year-horizontal-axis\n        [items]=\"yearItems\"\n        [horizontalScale]=\"yearScale\"\n        [fontSize]=\"fontSize\"\n        (selectionChange)='onSelectionChanged($event)'\n        (showTooltipEvent)='onTooltipEvent($event)'\n        (hideTooltipEvent)='onResetTooltipEvent()'\n      ></g>\n    </g>\n\n    <g\n      heatmap-calendar-year-series\n      [height]=\"innerHeight\"\n      [items]=\"yearItems\"\n      [primaryColor]=\"color\"\n      [yearScale]=\"yearScale\"\n      [colorScale]=\"colorScale\"\n      [selectedItem]=\"selectedItem\"\n      (selectionChange)='onSelectionChanged($event)'\n      (showTooltipEvent)='onTooltipEvent($event)'\n      (hideTooltipEvent)='onResetTooltipEvent()'\n    ></g>\n\n    <g [attr.transform]=\"monthsTransform\">\n      <g\n        heatmap-calendar-weekday-vertical-axis\n        transform=\"translate(8, 0)\"\n        [height]=\"innerHeight\"\n        [isVisible]=\"monthItems.length !== 0\"\n        [fontSize]=\"fontSize\"\n        [weekdayScale]=\"weekdayScale\"\n      ></g>\n      <g\n        heatmap-calendar-month-horizontal-axis\n        [items]=\"monthItems\"\n        [horizontalScale]=\"monthScale\"\n        [fontSize]=\"fontSize\"\n        [attr.transform]=\"translateTopAxisMonthsOffset\"\n        (selectionChange)='onSelectionChanged($event)'\n        (showTooltipEvent)='onTooltipEvent($event)'\n        (hideTooltipEvent)='onResetTooltipEvent()'\n      ></g>\n\n      <g\n        heatmap-calendar-month-series\n        [items]=\"monthItems\"\n        [monthScale]=\"monthScale\"\n        [colorScale]=\"colorScale\"\n        [weekdayScale]=\"weekdayScale\"\n        [attr.transform]=\"translateMonthsOffset\"\n        (daySelected)='onDaySelected($event)'\n        (showTooltipEvent)='onTooltipEvent($event)'\n        (hideTooltipEvent)='onResetTooltipEvent()'\n      ></g>\n    </g>\n  </g>\n</svg>\n", styles: [".heatmap-tooltip{pointer-events:none;position:fixed;z-index:9999;width:250px;margin-left:40px;margin-top:40px;max-width:250px;overflow:hidden;padding:15px;font-size:12px;line-height:14px}.heatmap-tooltip .header strong{display:inline-block;width:250px}.heatmap-tooltip span{display:inline-block;width:75%;padding-right:10px;box-sizing:border-box}.heatmap-tooltip span,.heatmap-tooltip .header strong{white-space:nowrap;overflow:hidden;text-overflow:ellipsis}\n"] }]
        }], propDecorators: { data: [{
                type: Input
            }], color: [{
                type: Input
            }], alignMonths: [{
                type: Input
            }], monthsCount: [{
                type: Input
            }], width: [{
                type: Input
            }], height: [{
                type: Input
            }], daySelected: [{
                type: Output
            }] } });

class SentinelHeatmapCalendarComponent {
    constructor(changeDetectionRef) {
        this.changeDetectionRef = changeDetectionRef;
        this.color = '#ff4500';
        this.alignMonths = AlignMonths.around;
        this.monthsCount = 0;
        this.daySelected = new EventEmitter();
        this.isSizeSetup = false;
    }
    onResize() {
        this.setSizes();
    }
    ngAfterViewInit() {
        this.setSizes();
        this.changeDetectionRef.detectChanges();
    }
    onDaySelected(daySelected) {
        this.daySelected.emit(daySelected);
    }
    setSizes() {
        this.width = this.container.nativeElement.offsetWidth;
        this.height = this.container.nativeElement.offsetHeight;
        this.isSizeSetup = true;
    }
}
SentinelHeatmapCalendarComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: SentinelHeatmapCalendarComponent, deps: [{ token: i0.ChangeDetectorRef }], target: i0.ɵɵFactoryTarget.Component });
SentinelHeatmapCalendarComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "14.0.0", version: "14.2.12", type: SentinelHeatmapCalendarComponent, selector: "sentinel-heatmap-calendar", inputs: { data: "data", color: "color", alignMonths: "alignMonths", monthsCount: "monthsCount" }, outputs: { daySelected: "daySelected" }, host: { listeners: { "window:resize": "onResize($event)" } }, viewQueries: [{ propertyName: "container", first: true, predicate: ["heatmapCalendarContainer"], descendants: true }], ngImport: i0, template: "<div class='container' #heatmapCalendarContainer>\n  <sentinel-heatmap-calendar-wrapper *ngIf=\"isSizeSetup\"\n                                     [data]=\"data\"\n                                     [monthsCount]=\"monthsCount\"\n                                     [alignMonths]=\"alignMonths\"\n                                     [color]=\"color\"\n                                     [width]=\"width\"\n                                     [height]=\"height\"\n                                     (daySelected)=\"onDaySelected($event)\">\n  </sentinel-heatmap-calendar-wrapper>\n</div>\n\n", styles: [".container{width:100%;height:100%;min-width:200px;min-height:200px}\n"], dependencies: [{ kind: "directive", type: i1.NgIf, selector: "[ngIf]", inputs: ["ngIf", "ngIfThen", "ngIfElse"] }, { kind: "component", type: HeatmapCalendarWrapperComponent, selector: "sentinel-heatmap-calendar-wrapper", inputs: ["data", "color", "alignMonths", "monthsCount", "width", "height"], outputs: ["daySelected"] }], changeDetection: i0.ChangeDetectionStrategy.OnPush });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: SentinelHeatmapCalendarComponent, decorators: [{
            type: Component,
            args: [{ selector: 'sentinel-heatmap-calendar', changeDetection: ChangeDetectionStrategy.OnPush, template: "<div class='container' #heatmapCalendarContainer>\n  <sentinel-heatmap-calendar-wrapper *ngIf=\"isSizeSetup\"\n                                     [data]=\"data\"\n                                     [monthsCount]=\"monthsCount\"\n                                     [alignMonths]=\"alignMonths\"\n                                     [color]=\"color\"\n                                     [width]=\"width\"\n                                     [height]=\"height\"\n                                     (daySelected)=\"onDaySelected($event)\">\n  </sentinel-heatmap-calendar-wrapper>\n</div>\n\n", styles: [".container{width:100%;height:100%;min-width:200px;min-height:200px}\n"] }]
        }], ctorParameters: function () { return [{ type: i0.ChangeDetectorRef }]; }, propDecorators: { data: [{
                type: Input
            }], color: [{
                type: Input
            }], alignMonths: [{
                type: Input
            }], monthsCount: [{
                type: Input
            }], daySelected: [{
                type: Output
            }], container: [{
                type: ViewChild,
                args: ['heatmapCalendarContainer', { static: false }]
            }], onResize: [{
                type: HostListener,
                args: ['window:resize', ['$event']]
            }] } });

class SentinelHeatmapCalendarModule {
}
SentinelHeatmapCalendarModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: SentinelHeatmapCalendarModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
SentinelHeatmapCalendarModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "14.0.0", version: "14.2.12", ngImport: i0, type: SentinelHeatmapCalendarModule, declarations: [SentinelHeatmapCalendarComponent,
        HeatmapCalendarWrapperComponent,
        WeekdayVerticalAxisComponent,
        BackButtonComponent,
        TooltipComponent,
        YearSeriesComponent,
        YearHorizontalAxisComponent,
        MonthSeriesComponent,
        MonthHorizontalAxisComponent], imports: [CommonModule], exports: [SentinelHeatmapCalendarComponent] });
SentinelHeatmapCalendarModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: SentinelHeatmapCalendarModule, imports: [CommonModule] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: SentinelHeatmapCalendarModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [
                        SentinelHeatmapCalendarComponent,
                        HeatmapCalendarWrapperComponent,
                        WeekdayVerticalAxisComponent,
                        BackButtonComponent,
                        TooltipComponent,
                        YearSeriesComponent,
                        YearHorizontalAxisComponent,
                        MonthSeriesComponent,
                        MonthHorizontalAxisComponent,
                    ],
                    imports: [CommonModule],
                    exports: [SentinelHeatmapCalendarComponent],
                }]
        }] });

/**
 * Generated bundle index. Do not edit.
 */

export { AlignMonths, SentinelHeatmapCalendarComponent, SentinelHeatmapCalendarModule };
//# sourceMappingURL=sentinel-components-heatmap-calendar.mjs.map
