import { BehaviorSubject, of, defer } from 'rxjs';
import * as i8 from '@angular/common';
import { CommonModule } from '@angular/common';
import * as i0 from '@angular/core';
import { NgModule, EventEmitter, Component, ChangeDetectionStrategy, Input, Output } from '@angular/core';
import { MatButtonModule } from '@angular/material/button';
import * as i5 from '@angular/material/card';
import { MatCardModule } from '@angular/material/card';
import * as i6 from '@angular/material/icon';
import { MatIconModule } from '@angular/material/icon';
import { OffsetPaginationEvent, CursorPaginationEvent } from '@sentinel/common/pagination';
import * as i4 from '@sentinel/components/controls';
import { SentinelControlItem, SentinelControlsModule } from '@sentinel/components/controls';
import { map, take, takeWhile } from 'rxjs/operators';
import * as i2 from '@sentinel/components/table';
import { DeleteAction, SentinelTable, Row, Column, SentinelTableModule } from '@sentinel/components/table';
import { SentinelBaseDirective } from '@sentinel/common';
import { trackById } from '@sentinel/common/utils';
import * as i3 from '@sentinel/components/resource-selector';
import { SentinelResourceSelectorModule } from '@sentinel/components/resource-selector';
import * as i7 from '@sentinel/components/loading';
import { SentinelLoadingModule } from '@sentinel/components/loading';

/**
 * Abstract service representing a layer between components and API to assign users to a resource.
 * Extend this class and its methods to connect the library components with concrete API.
 */
class SentinelUserAssignService {
    constructor() {
        this.hasErrorSubject$ = new BehaviorSubject(false);
        /**
         * True if error was thrown while getting assigned users, false otherwise
         */
        this.hasError$ = this.hasErrorSubject$.asObservable();
        this.isLoadingAssignedSubject = new BehaviorSubject(false);
        /**
         * True if getting data for assigned users is in progress, false otherwise
         */
        this.isLoadingAssigned$ = this.isLoadingAssignedSubject.asObservable();
        this.selectedUsersToAssignSubject$ = new BehaviorSubject([]);
        this.selectedUsersToAssign$ = this.selectedUsersToAssignSubject$.asObservable();
        this.selectedAssignedUsersSubject$ = new BehaviorSubject([]);
        this.selectedAssignedUsers$ = this.selectedAssignedUsersSubject$.asObservable();
    }
    setSelectedUsersToAssign(users) {
        this.selectedUsersToAssignSubject$.next(users);
    }
    clearSelectedUsersToAssign() {
        this.selectedUsersToAssignSubject$.next([]);
    }
    setSelectedAssignedUsers(users) {
        this.selectedAssignedUsersSubject$.next(users);
    }
    clearSelectedAssignedUsers() {
        this.selectedAssignedUsersSubject$.next([]);
    }
}

class SentinelUserAssignMaterialModule {
}
SentinelUserAssignMaterialModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: SentinelUserAssignMaterialModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
SentinelUserAssignMaterialModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "14.0.0", version: "14.2.12", ngImport: i0, type: SentinelUserAssignMaterialModule, imports: [MatCardModule, MatButtonModule, MatIconModule], exports: [MatCardModule, MatButtonModule, MatIconModule] });
SentinelUserAssignMaterialModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: SentinelUserAssignMaterialModule, imports: [MatCardModule, MatButtonModule, MatIconModule, MatCardModule, MatButtonModule, MatIconModule] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: SentinelUserAssignMaterialModule, decorators: [{
            type: NgModule,
            args: [{
                    imports: [MatCardModule, MatButtonModule, MatIconModule],
                    exports: [MatCardModule, MatButtonModule, MatIconModule],
                }]
        }] });

class UserRowDeleteAction extends DeleteAction {
    constructor(userTitle, disabled$, result$) {
        super(`Delete ${userTitle}`, disabled$, result$);
    }
}

/**
 * Default user table class for user assign component
 */
class UserTableAdapter extends SentinelTable {
    constructor(resource, assigneesTitle, resourceId, service) {
        const rows = resource.elements.map((user) => new Row(user, [
            new UserRowDeleteAction(assigneesTitle, of(false), defer(() => service.unassign(resourceId, [user]))),
        ]));
        const columns = [
            new Column('id', 'id', false),
            new Column('name', 'name', true, 'familyName'),
            new Column('login', 'login', true, 'login'),
        ];
        super(rows, columns);
        this.pagination = resource.pagination;
        this.filterable = true;
        this.selectable = true;
        this.filterLabel = 'Filter by name';
    }
}

/**
 * Reusable component for assignment of users to a resource
 */
class SentinelUserAssignComponent extends SentinelBaseDirective {
    constructor(usersService) {
        super();
        this.usersService = usersService;
        /**
         * Default pagination size of sentinel table
         */
        this.defaultPaginationSize = 5;
        this.paginationType = 'offset';
        /**
         * Event emitter triggered every time change is made.
         */
        this.hasUnsavedChanges = new EventEmitter();
        this.trackByFn = trackById;
        this.selectorResourceMapping = {
            id: 'id',
            title: 'name',
            subtitle: 'login',
            icon: 'picture',
        };
    }
    ngOnChanges(changes) {
        if ('resource' in changes && this.resource && this.resource.id !== undefined) {
            this.searchPlaceholder = `Search ${this.toAssignTitle.toLowerCase()} by surname or first name and surname`;
            this.init();
        }
    }
    /**
     * Calls service to get users available to assign
     * @param filterValue filter to apply on users
     */
    getToAssign(filterValue) {
        this.toAssign = this.usersService
            .getAvailableToAssign(this.resource.id, filterValue)
            .pipe(map((resource) => resource.elements));
    }
    /**
     * Resolves type of action and calls appropriate handler
     * @param event action event emitted from asignees table
     */
    onAssigneesTableAction(event) {
        event.action.result$.pipe(take(1)).subscribe();
    }
    onControlAction(controlItem) {
        controlItem.result$.pipe(take(1)).subscribe();
    }
    /**
     * Changes internal state of the selection of users to assign
     * @param selected new state of selection
     */
    onToAssignSelection(selected) {
        this.usersService.setSelectedUsersToAssign(selected);
    }
    onAssigneesLoadEvent(loadEvent) {
        this.usersService
            .getAssigned(this.resource.id, loadEvent.pagination, loadEvent.filter)
            .pipe(takeWhile(() => this.isAlive))
            .subscribe();
    }
    /**
     * Changes internal state of the selection of assigned users
     * @param selected new state of selection
     */
    onAssignedSelection(selected) {
        this.usersService.setSelectedAssignedUsers(selected);
    }
    init() {
        this.selectedToAssign$ = this.usersService.selectedUsersToAssign$;
        this.initTable();
        this.initAssignedControls();
        this.initToAssignControls();
        this.initUnsavedChangesEmitter();
    }
    initTable() {
        const initEvent = { pagination: this.initPagination() };
        this.assignees$ = this.usersService.assignedUsers$.pipe(map((paginatedUsers) => new UserTableAdapter(paginatedUsers, this.assigneesTitle, this.resource.id, this.usersService)));
        this.assigneesTableHasError$ = this.usersService.hasError$;
        this.isLoadingAssignees$ = this.usersService.isLoadingAssigned$;
        this.onAssigneesLoadEvent(initEvent);
    }
    initPagination() {
        if (this.paginationType === 'offset') {
            return new OffsetPaginationEvent(0, this.defaultPaginationSize, '', '');
        }
        else {
            return new CursorPaginationEvent(null, this.defaultPaginationSize, '', '');
        }
    }
    initAssignedControls() {
        this.usersService.selectedAssignedUsers$.pipe(takeWhile(() => this.isAlive)).subscribe((selection) => (this.assignedControls = [
            new SentinelControlItem('delete', selection.length > 0 ? `Delete (${selection.length})` : 'Delete', 'warn', of(selection.length <= 0), defer(() => this.usersService.unassignSelected(this.resource.id))),
        ]));
    }
    initToAssignControls() {
        const disabled$ = this.usersService.selectedUsersToAssign$.pipe(map((selection) => selection.length <= 0));
        this.toAssignControls = [
            new SentinelControlItem('add', 'Add', 'primary', disabled$, defer(() => this.usersService.assignSelected(this.resource.id))),
        ];
    }
    initUnsavedChangesEmitter() {
        this.usersService.selectedUsersToAssign$
            .pipe(takeWhile(() => this.isAlive))
            .subscribe((selection) => this.hasUnsavedChanges.emit(selection.length > 0));
    }
}
SentinelUserAssignComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: SentinelUserAssignComponent, deps: [{ token: SentinelUserAssignService }], target: i0.ɵɵFactoryTarget.Component });
SentinelUserAssignComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "14.0.0", version: "14.2.12", type: SentinelUserAssignComponent, selector: "sentinel-user-assign", inputs: { resource: "resource", resourceMatIcon: "resourceMatIcon", defaultPaginationSize: "defaultPaginationSize", paginationType: "paginationType", toAssignTitle: "toAssignTitle", assigneesTitle: "assigneesTitle" }, outputs: { hasUnsavedChanges: "hasUnsavedChanges" }, usesInheritance: true, usesOnChanges: true, ngImport: i0, template: "<sentinel-controls\n  class=\"to-assign-controls\"\n  [controls]=\"toAssignControls\"\n  (itemClicked)=\"onControlAction($event)\"\n></sentinel-controls>\n<mat-card class=\"to-assign-card\">\n  <mat-card-header>\n    <div mat-card-avatar>\n      <mat-icon color=\"primary\">person_add</mat-icon>\n    </div>\n    <mat-card-title>Add {{ toAssignTitle }}</mat-card-title>\n    <mat-card-subtitle>Add {{ toAssignTitle }} to {{ resource?.title }} </mat-card-subtitle>\n  </mat-card-header>\n  <mat-card-content>\n    <sentinel-resource-selector\n      [searchPlaceholder]=\"searchPlaceholder\"\n      [trackByFn]=\"trackByFn\"\n      [resources]=\"toAssign | async\"\n      [selected]=\"selectedToAssign$ | async\"\n      [resourceMapping]=\"selectorResourceMapping\"\n      (selectionChange)=\"onToAssignSelection($event)\"\n      (fetch)=\"getToAssign($event)\"\n    >\n    </sentinel-resource-selector>\n  </mat-card-content>\n</mat-card>\n<sentinel-controls\n  class=\"assigned-controls\"\n  [controls]=\"assignedControls\"\n  (itemClicked)=\"onControlAction($event)\"\n></sentinel-controls>\n<sentinel-loading-wrapper [isLoading]=\"isLoadingAssignees$ | async\">\n  <mat-card>\n    <mat-card-header>\n      <div mat-card-avatar>\n        <mat-icon color=\"primary\">{{ resourceMatIcon }}</mat-icon>\n      </div>\n      <mat-card-title> {{ assigneesTitle }} of {{ resource?.title }}</mat-card-title>\n    </mat-card-header>\n    <mat-card-content>\n      <sentinel-table\n        elevation=\"0\"\n        [trackByFn]=\"trackByFn\"\n        [hasError]=\"assigneesTableHasError$ | async\"\n        [data]=\"assignees$ | async\"\n        (rowAction)=\"onAssigneesTableAction($event)\"\n        (rowSelection)=\"onAssignedSelection($event)\"\n        (tableLoad)=\"onAssigneesLoadEvent($event)\"\n      >\n      </sentinel-table>\n    </mat-card-content>\n  </mat-card>\n</sentinel-loading-wrapper>\n", styles: ["sentinel-loading-wrapper{margin-top:10px}.to-assign-card{margin-bottom:40px}\n"], dependencies: [{ kind: "component", type: i2.SentinelTableComponent, selector: "sentinel-table", inputs: ["data", "isLoading", "elevation", "preselected", "hasError", "defaultSortDirection", "defaultSortName", "showMoreDefaultSize", "actionMode", "alternateRows", "titlecasedActions", "displayedActionsCount", "trackByFn"], outputs: ["rowAction", "rowSelection", "rowClick", "expandableRowStateChange", "tableLoad"] }, { kind: "component", type: i3.SentinelResourceSelectorComponent, selector: "sentinel-resource-selector", inputs: ["resources", "multiSelection", "resourceMapping", "appearance", "searchPlaceholder", "label", "required", "selected", "singleSelected", "fitIcon", "isLoading", "compactMode", "trackByFn", "identity", "cleanInputOnSelection"], outputs: ["selectionChange", "singleSelectionChange", "fetch", "validityChange"] }, { kind: "component", type: i4.SentinelControlsComponent, selector: "sentinel-controls", inputs: ["controls", "buttonStyle", "maxItemsCount", "align", "withMargin", "titlecased"], outputs: ["itemClicked"] }, { kind: "component", type: i5.MatCard, selector: "mat-card", exportAs: ["matCard"] }, { kind: "component", type: i5.MatCardHeader, selector: "mat-card-header" }, { kind: "directive", type: i5.MatCardContent, selector: "mat-card-content, [mat-card-content], [matCardContent]" }, { kind: "directive", type: i5.MatCardTitle, selector: "mat-card-title, [mat-card-title], [matCardTitle]" }, { kind: "directive", type: i5.MatCardSubtitle, selector: "mat-card-subtitle, [mat-card-subtitle], [matCardSubtitle]" }, { kind: "directive", type: i5.MatCardAvatar, selector: "[mat-card-avatar], [matCardAvatar]" }, { kind: "component", type: i6.MatIcon, selector: "mat-icon", inputs: ["color", "inline", "svgIcon", "fontSet", "fontIcon"], exportAs: ["matIcon"] }, { kind: "component", type: i7.SentinelLoadingWrapperComponent, selector: "sentinel-loading-wrapper", inputs: ["isLoading", "color"] }, { kind: "pipe", type: i8.AsyncPipe, name: "async" }], changeDetection: i0.ChangeDetectionStrategy.OnPush });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: SentinelUserAssignComponent, decorators: [{
            type: Component,
            args: [{ selector: 'sentinel-user-assign', changeDetection: ChangeDetectionStrategy.OnPush, template: "<sentinel-controls\n  class=\"to-assign-controls\"\n  [controls]=\"toAssignControls\"\n  (itemClicked)=\"onControlAction($event)\"\n></sentinel-controls>\n<mat-card class=\"to-assign-card\">\n  <mat-card-header>\n    <div mat-card-avatar>\n      <mat-icon color=\"primary\">person_add</mat-icon>\n    </div>\n    <mat-card-title>Add {{ toAssignTitle }}</mat-card-title>\n    <mat-card-subtitle>Add {{ toAssignTitle }} to {{ resource?.title }} </mat-card-subtitle>\n  </mat-card-header>\n  <mat-card-content>\n    <sentinel-resource-selector\n      [searchPlaceholder]=\"searchPlaceholder\"\n      [trackByFn]=\"trackByFn\"\n      [resources]=\"toAssign | async\"\n      [selected]=\"selectedToAssign$ | async\"\n      [resourceMapping]=\"selectorResourceMapping\"\n      (selectionChange)=\"onToAssignSelection($event)\"\n      (fetch)=\"getToAssign($event)\"\n    >\n    </sentinel-resource-selector>\n  </mat-card-content>\n</mat-card>\n<sentinel-controls\n  class=\"assigned-controls\"\n  [controls]=\"assignedControls\"\n  (itemClicked)=\"onControlAction($event)\"\n></sentinel-controls>\n<sentinel-loading-wrapper [isLoading]=\"isLoadingAssignees$ | async\">\n  <mat-card>\n    <mat-card-header>\n      <div mat-card-avatar>\n        <mat-icon color=\"primary\">{{ resourceMatIcon }}</mat-icon>\n      </div>\n      <mat-card-title> {{ assigneesTitle }} of {{ resource?.title }}</mat-card-title>\n    </mat-card-header>\n    <mat-card-content>\n      <sentinel-table\n        elevation=\"0\"\n        [trackByFn]=\"trackByFn\"\n        [hasError]=\"assigneesTableHasError$ | async\"\n        [data]=\"assignees$ | async\"\n        (rowAction)=\"onAssigneesTableAction($event)\"\n        (rowSelection)=\"onAssignedSelection($event)\"\n        (tableLoad)=\"onAssigneesLoadEvent($event)\"\n      >\n      </sentinel-table>\n    </mat-card-content>\n  </mat-card>\n</sentinel-loading-wrapper>\n", styles: ["sentinel-loading-wrapper{margin-top:10px}.to-assign-card{margin-bottom:40px}\n"] }]
        }], ctorParameters: function () { return [{ type: SentinelUserAssignService }]; }, propDecorators: { resource: [{
                type: Input
            }], resourceMatIcon: [{
                type: Input
            }], defaultPaginationSize: [{
                type: Input
            }], paginationType: [{
                type: Input
            }], toAssignTitle: [{
                type: Input
            }], assigneesTitle: [{
                type: Input
            }], hasUnsavedChanges: [{
                type: Output
            }] } });

/**
 * Main module of the library. Contains user assign component and imports resource selector module
 */
class SentinelUserAssignModule {
}
SentinelUserAssignModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: SentinelUserAssignModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
SentinelUserAssignModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "14.0.0", version: "14.2.12", ngImport: i0, type: SentinelUserAssignModule, declarations: [SentinelUserAssignComponent], imports: [CommonModule,
        SentinelTableModule,
        SentinelResourceSelectorModule,
        SentinelControlsModule,
        SentinelUserAssignMaterialModule,
        SentinelLoadingModule], exports: [SentinelUserAssignComponent] });
SentinelUserAssignModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: SentinelUserAssignModule, imports: [CommonModule,
        SentinelTableModule,
        SentinelResourceSelectorModule,
        SentinelControlsModule,
        SentinelUserAssignMaterialModule,
        SentinelLoadingModule] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "14.2.12", ngImport: i0, type: SentinelUserAssignModule, decorators: [{
            type: NgModule,
            args: [{
                    imports: [
                        CommonModule,
                        SentinelTableModule,
                        SentinelResourceSelectorModule,
                        SentinelControlsModule,
                        SentinelUserAssignMaterialModule,
                        SentinelLoadingModule,
                    ],
                    declarations: [SentinelUserAssignComponent],
                    exports: [SentinelUserAssignComponent],
                }]
        }] });

/*
 * Public API Surface of @sentinel/components/user-assign
 */

/**
 * Generated bundle index. Do not edit.
 */

export { SentinelUserAssignComponent, SentinelUserAssignModule, SentinelUserAssignService, UserTableAdapter };
//# sourceMappingURL=sentinel-components-user-assign.mjs.map
