import { ElementRef, EventEmitter, OnChanges, OnInit, QueryList, SimpleChanges, TrackByFunction } from '@angular/core';
import { UntypedFormControl } from '@angular/forms';
import { SentinelResourceSelectorMapping } from './model/sentinel-resource-selector-mapping';
import { SentinelBaseDirective } from '@sentinel/common';
import { MatCheckbox, MatCheckboxChange } from '@angular/material/checkbox';
import { MatAutocompleteTrigger } from '@angular/material/autocomplete';
import { MatOptionSelectionChange } from '@angular/material/core';
import { SentinelSelectorElementDirective } from './directives/sentinel-selector-element.directive';
import { SentinelSelectorSelectedElementDirective } from './directives/sentinel-selector-selected-element.directive';
import { MatFormFieldAppearance } from '@angular/material/form-field';
import * as i0 from "@angular/core";
/**
 * Generic reusable component for selection of generic resource with autocomplete, chips and other material components
 */
export declare class SentinelResourceSelectorComponent<T> extends SentinelBaseDirective implements OnInit, OnChanges {
    /**
     * Array of generic selectable resource to be displayed in combo box
     */
    resources: T[];
    multiSelection: boolean;
    /**
     * MUST BE PROVIDED if no custom element and chip directives are provided
     * Maps custom resource attribute names to predefined attribute names that this component uses internally
     */
    resourceMapping: SentinelResourceSelectorMapping;
    appearance: MatFormFieldAppearance;
    /**
     * Custom search bar placeholder
     */
    searchPlaceholder: string;
    label?: string;
    required: boolean;
    /**
     * Currently selected or pre-selected resources
     */
    selected: T[];
    singleSelected?: T;
    /**
     * Determines style of fitting of the icon in selector. If set to 'center', it will make the icon smaller that the bounding
     * element and fit in in its centre. If set to 'full' it will try to fit whole icon in its bounding element.
     */
    fitIcon: 'center' | 'full';
    isLoading: boolean;
    compactMode: boolean;
    /**
     * Function for tracking items in *ngFor. Defaults to tracking by index
     */
    trackByFn?: TrackByFunction<T>;
    /**
     * Function for comparing the resources. Defaults to comparison by 'id' attribute
     */
    identity?: (a: T, b: T) => boolean;
    cleanInputOnSelection: boolean;
    /**
     * Event emitter emitting all selected resource every time, change is made
     */
    selectionChange: EventEmitter<T[]>;
    singleSelectionChange: EventEmitter<T>;
    /**
     * Event emitter emitting event with search input from resource every time, input is changed
     */
    fetch: EventEmitter<string>;
    validityChange: EventEmitter<boolean>;
    elementDirective?: SentinelSelectorElementDirective;
    selectedElementDirective?: SentinelSelectorSelectedElementDirective;
    resourceControl: UntypedFormControl;
    selectedResources: T[];
    selectedResource?: T;
    selectedCheckboxes: number[];
    internalIsLoading: boolean;
    externalLoadingSet: boolean;
    valid: boolean;
    readonly DUE_TIME = 300;
    userInput: ElementRef<HTMLInputElement>;
    autocomplete: MatAutocompleteTrigger;
    selectionCheckboxes: QueryList<MatCheckbox>;
    ngOnInit(): void;
    ngOnChanges(changes: SimpleChanges): void;
    /**
     * Emits event to fetch resources based on search bar input
     * @param filter value based on which resources should be filtered
     */
    search(filter: string): void;
    searchByLastValue(): void;
    /**
     * Selects resource and emits event
     * @param event containing information if selection was done by user
     * @param resource selected resource
     */
    resourceSelected(event: MatOptionSelectionChange, resource: T): void;
    singleResourceSelected(event: MatOptionSelectionChange, resource?: T): void;
    private onResourceSelected;
    /**
     * Selects resources with checkboxes
     * @param event checkbox check event
     * @param resource selected/unselected resource
     */
    resourceSelectedByCheckbox(event: MatCheckboxChange, resource: T): void;
    /**
     * Removes selection of a resource
     * @param index index of a resource to be removed
     */
    removeSelected(index: number): void;
    /**
     * Returns true if resource is already selected, false otherwise. Compares based on ID provided in resource mapping
     * @param resource resource to compare
     */
    alreadySelected(resource: T): boolean;
    private defaultIdentity;
    private onSearch;
    private resetFilter;
    private updateValidity;
    static ɵfac: i0.ɵɵFactoryDeclaration<SentinelResourceSelectorComponent<any>, never>;
    static ɵcmp: i0.ɵɵComponentDeclaration<SentinelResourceSelectorComponent<any>, "sentinel-resource-selector", never, { "resources": "resources"; "multiSelection": "multiSelection"; "resourceMapping": "resourceMapping"; "appearance": "appearance"; "searchPlaceholder": "searchPlaceholder"; "label": "label"; "required": "required"; "selected": "selected"; "singleSelected": "singleSelected"; "fitIcon": "fitIcon"; "isLoading": "isLoading"; "compactMode": "compactMode"; "trackByFn": "trackByFn"; "identity": "identity"; "cleanInputOnSelection": "cleanInputOnSelection"; }, { "selectionChange": "selectionChange"; "singleSelectionChange": "singleSelectionChange"; "fetch": "fetch"; "validityChange": "validityChange"; }, ["elementDirective", "selectedElementDirective"], never, false>;
}
