import { AfterContentInit, EventEmitter, OnChanges, QueryList, SimpleChanges, TrackByFunction } from '@angular/core';
import { MatSort } from '@angular/material/sort';
import { MatTableDataSource } from '@angular/material/table';
import { Observable } from 'rxjs';
import { SentinelTable } from './model/sentinel-table';
import { Row } from './model/row';
import { RowAction } from './model/row-action';
import { TableActionEvent } from './model/table-action-event';
import { TableActionMode } from './model/table-action-mode';
import { SentinelRowDirective } from './directives/sentinel-cell-def.directive';
import { SentinelProjectedCells } from './model/sentinel-projected-cells';
import { SentinelBaseDirective } from '@sentinel/common';
import { TableEventService } from './services/table-event.service';
import { TableSortService } from './services/table-sort.service';
import { TableFilterService } from './services/table-filter.service';
import { TablePageService } from './services/table-page.service';
import { TableSelectionService } from './services/selection/table-selection.service';
import { ShowMoreService } from './services/show-more.service';
import { TableLoadEvent } from './model/table-load-event';
import { ExpandableRowChangeEvent } from './model/table-events/expandable-row-change-event';
import { trackByIndex } from '@sentinel/common/utils';
import * as i0 from "@angular/core";
/**
 * Presentational table component in material design. Columns, rows and its behaviour is defined by input data
 * passed to the component through data binding. Table supports displays selected attributes of a data source. Table
 * also supports Observables, so async data can be displayed in table cells. Sorting by selected columns is also supported
 * as well as filtering, pagination and row selection (checkboxes). Each table row can have unique actions (mat button icon)
 * and they can be disabled asynchronously by passing an Observable. Another important feature is row expansion, where row can
 * be expanded by clicking on arrow-like button and row detail component will roll out of the selected row. Detail components
 * are rendered dynamically and efficiently. The detail component must be created and provided by client. Current row is passed
 * to the detail component through `@Input() data` attribute, so make sure to include it in your detail component.
 */
export declare class SentinelTableComponent<T> extends SentinelBaseDirective implements OnChanges, AfterContentInit {
    filterService: TableFilterService;
    sortService: TableSortService;
    pageService: TablePageService;
    eventService: TableEventService;
    selectionService: TableSelectionService<T>;
    showMoreService: ShowMoreService;
    /**
     * Table configuration, columns and rows to be displayed
     */
    data: SentinelTable<T>;
    /**
     * True if data for table is loading, false otherwise (table displays loading bar)
     */
    isLoading: boolean;
    /**
     * Angular material elevation size of the table. Default value is 8. See more at https://material.angular.io/guide/elevation
     */
    elevation: number;
    /**
     * List of rows that should be preselected. Works only when table is configured with row selection checkboxes
     */
    preselected: T[];
    /**
     * True if there was an error when getting data for the table. Displays message with a try again button
     */
    hasError: boolean;
    /**
     * Direction of initial sort setting. Works only when table is configured with sorting
     */
    defaultSortDirection?: 'desc' | 'asc';
    /**
     * Name of initial sort column. Works only when table is configured with sorting
     */
    defaultSortName?: string;
    /**
     * By how much should the pagination be expanded after clicking "Show More" button.
     */
    showMoreDefaultSize: number;
    /**
     * Decides how should the actions be displayed. "Icon" mode displays actions as mat icon components, while "button mode"
     * displays them as a mat button components
     */
    actionMode: TableActionMode;
    /**
     * True if rows should have alternating colors, false otherwise.
     */
    alternateRows: boolean;
    /**
     * True if row actions should be titlecased
     */
    titlecasedActions: boolean;
    /**
     * Number of actions to be displayed in actions column. Actions above the number are hidden in menu
     */
    displayedActionsCount: number;
    /**
     * Function for tracking items in *ngFor. Defaults to tracking by index
     */
    trackByFn?: TrackByFunction<Row<T>>;
    /**
     * Emits event when user clicks on row action
     */
    rowAction: EventEmitter<TableActionEvent<T>>;
    /**
     * Emits event when user changes row selection
     */
    rowSelection: EventEmitter<T[]>;
    /**
     * Emits event when user click on a row
     */
    rowClick: EventEmitter<T>;
    /**
     * Emits event when user expands or collapses expandable row
     */
    expandableRowStateChange: EventEmitter<ExpandableRowChangeEvent<T>>;
    /**
     * Emits event when data refresh for the table is requested (for example change of pagination settings, sort etc)
     */
    tableLoad: EventEmitter<TableLoadEvent>;
    sort: MatSort;
    cellDirectives: QueryList<SentinelRowDirective>;
    trackByFnCols: typeof trackByIndex;
    displayedColumns: string[];
    expandedElement: T;
    isExpandable: boolean;
    dataSource: MatTableDataSource<Row<T>>;
    projectedCells: SentinelProjectedCells;
    showMoreCount: number;
    onTryAgain$: Observable<any>;
    private alreadySetup;
    private pageAlreadySetup;
    constructor(filterService: TableFilterService, sortService: TableSortService, pageService: TablePageService, eventService: TableEventService, selectionService: TableSelectionService<T>, showMoreService: ShowMoreService);
    ngOnChanges(changes: SimpleChanges): void;
    ngAfterContentInit(): void;
    /**
     * Emits event to refresh data for the table. Optionally, adds pagination data and filters
     */
    onTryAgain(): void;
    /**
     * Emits event when row action is selection
     * @param row row of the selected action
     * @param action action that was selected
     */
    onAction(row: Row<T>, action: RowAction): void;
    onRowClick(row: Row<T>): void;
    /**
     * Opens or closes detail of expandable row
     * @param row row which was toggled
     * @param index index of toggled row
     */
    toggleExpandableRow(row: Row<T>, index: number): void;
    private onDataChange;
    private createDataSource;
    private setup;
    /**
     * ORDER MATTERS!
     */
    private setupDisplayedColumns;
    private setupActions;
    private setupSort;
    private actionsColumnAlreadySetup;
    private getInitSortName;
    private createProjectedCells;
    private onFetchEvent;
    static ɵfac: i0.ɵɵFactoryDeclaration<SentinelTableComponent<any>, never>;
    static ɵcmp: i0.ɵɵComponentDeclaration<SentinelTableComponent<any>, "sentinel-table", never, { "data": "data"; "isLoading": "isLoading"; "elevation": "elevation"; "preselected": "preselected"; "hasError": "hasError"; "defaultSortDirection": "defaultSortDirection"; "defaultSortName": "defaultSortName"; "showMoreDefaultSize": "showMoreDefaultSize"; "actionMode": "actionMode"; "alternateRows": "alternateRows"; "titlecasedActions": "titlecasedActions"; "displayedActionsCount": "displayedActionsCount"; "trackByFn": "trackByFn"; }, { "rowAction": "rowAction"; "rowSelection": "rowSelection"; "rowClick": "rowClick"; "expandableRowStateChange": "expandableRowStateChange"; "tableLoad": "tableLoad"; }, ["cellDirectives"], never, false>;
}
