import { BehaviorSubject, Observable } from 'rxjs';
import { AssignableUser } from './model/assignable-user';
import { PaginatedResource, PaginationBaseEvent } from '@sentinel/common/pagination';
/**
 * Abstract service representing a layer between components and API to assign users to a resource.
 * Extend this class and its methods to connect the library components with concrete API.
 */
export declare abstract class SentinelUserAssignService {
    protected hasErrorSubject$: BehaviorSubject<boolean>;
    /**
     * True if error was thrown while getting assigned users, false otherwise
     */
    hasError$: Observable<boolean>;
    protected isLoadingAssignedSubject: BehaviorSubject<boolean>;
    /**
     * True if getting data for assigned users is in progress, false otherwise
     */
    isLoadingAssigned$: Observable<boolean>;
    /**
     * List of users already assigned to the resource
     */
    abstract assignedUsers$: Observable<PaginatedResource<AssignableUser>>;
    protected selectedUsersToAssignSubject$: BehaviorSubject<AssignableUser[]>;
    selectedUsersToAssign$: Observable<AssignableUser[]>;
    protected selectedAssignedUsersSubject$: BehaviorSubject<AssignableUser[]>;
    selectedAssignedUsers$: Observable<AssignableUser[]>;
    setSelectedUsersToAssign(users: AssignableUser[]): void;
    clearSelectedUsersToAssign(): void;
    setSelectedAssignedUsers(users: AssignableUser[]): void;
    clearSelectedAssignedUsers(): void;
    /**
     * Gets users available to assign to resource
     * @param resourceId id of a resource to assign to
     * @param filter filter to be applied on users
     */
    abstract getAvailableToAssign(resourceId: number, filter: string): Observable<PaginatedResource<AssignableUser>>;
    /**
     * Gets users already assigned to the resource
     * @contract MUST update assignedUsers observable.
     * @param resourceId id of the resource associated with requested users
     * @param pagination pagination event
     * @param filter filter to be applied on users
     */
    abstract getAssigned(resourceId: number, pagination: PaginationBaseEvent, filter: string): Observable<PaginatedResource<AssignableUser>>;
    /**
     * Assigns selected users to a resource
     * @param resourceId id of a resource to associate with users
     */
    abstract assignSelected(resourceId: number): Observable<any>;
    /**
     * Assigns users to the resource
     * @param resourceId id of the resource to assign to
     * @param users users to be assigned
     */
    abstract assign(resourceId: number, users: AssignableUser[]): Observable<any>;
    /**
     * Unassigns selected users from resource
     * @param resourceId id of resource which association should be cancelled
     */
    abstract unassignSelected(resourceId: number): Observable<any>;
    /**
     * Unassigns all users from resource (cancels the association)
     * @param resourceId id of the resource which associations should be cancelled
     * @param users users to unassign
     */
    abstract unassign(resourceId: number, users: AssignableUser[]): Observable<any>;
    /**
     * Assigns and unassigns users from/to resource at the same time
     * @param resourceId id of a resource which associations should be updated
     * @param additions users to assign to a resource
     * @param removals users to unnassign from a resource
     */
    abstract update(resourceId: number, additions: AssignableUser[], removals: AssignableUser[]): Observable<any>;
}
